Package: REDCapSync
Title: Encapsulated 'REDCap' Projects for Synchronized Data Pipelines
Version: 0.1.1
Authors@R: c(
    person("Brandon", "Rose", , "thecodingdocs@gmail.com", role = c("cre", "aut", "cph"),
           comment = c(ORCID = "0009-0009-7813-1960")),
    person("Natalie", "Goulett", role = "ctb")
  )
Description: Wraps dozens of 'REDCap' API endpoints into a standardized R6
    object.  Research Electronic Data Capture ('REDCap') is a survey and
    database web application software maintained by Vanderbilt University.
    It has a robust application programming interface (API) utilized by
    several R packages. 'REDCapSync' uses 'redcapAPI' and 'REDCapR'
    behind-the-scenes to retrieve all metadata, data, and log details for
    a project. To minimize unnecessary server calls, the interim 'REDCap'
    log is analyzed and used to only update necessary records.
    Furthermore, the user can define custom datasets that save to a
    directory. Those datasets continue to refresh when projects are
    synced.  Having a secure, standardized, API-efficient,
    project-agnostic R object for 'REDCap' projects, streamlines
    downstream use in scripts, functions, and shiny applications.
License: GPL (>= 3)
URL: https://github.com/thecodingdocs/REDCapSync,
        https://thecodingdocs.github.io/REDCapSync/
BugReports: https://github.com/thecodingdocs/REDCapSync/issues
Depends: R (>= 4.1)
Imports: checkmate, cli, dplyr, hoardr, lubridate, openxlsx2, R6,
        readxl, redcapAPI, REDCapR, skimr, stats, stringr, tools, utils
Suggests: keyring, knitr, listviewer, rmarkdown, spelling, testthat (>=
        3.3.0), withr
VignetteBuilder: knitr
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
Language: en-US
NeedsCompilation: no
Packaged: 2026-06-09 00:31:14 UTC; brandonrose
Author: Brandon Rose [cre, aut, cph] (ORCID:
    <https://orcid.org/0009-0009-7813-1960>),
  Natalie Goulett [ctb]
Maintainer: Brandon Rose <thecodingdocs@gmail.com>
Repository: CRAN
Date/Publication: 2026-06-09 14:10:02 UTC
Built: R 4.6.0; ; 2026-06-09 16:31:29 UTC; unix

# Generated by roxygen2: do not edit by hand

export(REDCapSyncDataset)
export(REDCapSyncProject)
export(cache_clear)
export(config)
export(load_project)
export(projects)
export(setup_project)
export(sync)
importFrom(R6,R6Class)
importFrom(REDCapR,redcap_delete)
importFrom(REDCapR,redcap_file_download_oneshot)
importFrom(REDCapR,redcap_read)
importFrom(REDCapR,redcap_read_oneshot)
importFrom(REDCapR,redcap_report)
importFrom(REDCapR,redcap_write)
importFrom(REDCapR,sanitize_token)
importFrom(checkmate,assert)
importFrom(checkmate,assert_character)
importFrom(checkmate,assert_choice)
importFrom(checkmate,assert_class)
importFrom(checkmate,assert_data_frame)
importFrom(checkmate,assert_date)
importFrom(checkmate,assert_directory)
importFrom(checkmate,assert_directory_exists)
importFrom(checkmate,assert_environment)
importFrom(checkmate,assert_integerish)
importFrom(checkmate,assert_list)
importFrom(checkmate,assert_logical)
importFrom(checkmate,assert_names)
importFrom(checkmate,assert_path_for_output)
importFrom(checkmate,assert_posixct)
importFrom(checkmate,assert_r6)
importFrom(checkmate,assert_set_equal)
importFrom(checkmate,assert_string)
importFrom(checkmate,assert_tibble)
importFrom(checkmate,assert_true)
importFrom(checkmate,check_directory_exists)
importFrom(checkmate,check_set_equal)
importFrom(checkmate,check_true)
importFrom(checkmate,expect_character)
importFrom(checkmate,expect_class)
importFrom(checkmate,expect_data_frame)
importFrom(checkmate,expect_directory_exists)
importFrom(checkmate,expect_file_exists)
importFrom(checkmate,expect_list)
importFrom(checkmate,expect_logical)
importFrom(checkmate,expect_r6)
importFrom(checkmate,expect_scalar_na)
importFrom(checkmate,test_character)
importFrom(checkmate,test_choice)
importFrom(checkmate,test_class)
importFrom(checkmate,test_directory_exists)
importFrom(checkmate,test_file_exists)
importFrom(checkmate,test_integerish)
importFrom(checkmate,test_logical)
importFrom(checkmate,test_scalar_na)
importFrom(cli,cli_abort)
importFrom(cli,cli_alert_danger)
importFrom(cli,cli_alert_info)
importFrom(cli,cli_alert_success)
importFrom(cli,cli_alert_warning)
importFrom(cli,cli_bullets)
importFrom(cli,cli_code)
importFrom(cli,cli_h1)
importFrom(cli,cli_h2)
importFrom(cli,cli_h3)
importFrom(cli,cli_li)
importFrom(cli,cli_progress_bar)
importFrom(cli,cli_progress_update)
importFrom(cli,cli_text)
importFrom(cli,cli_warn)
importFrom(cli,format_message)
importFrom(dplyr,as_tibble)
importFrom(dplyr,bind_rows)
importFrom(dplyr,filter)
importFrom(dplyr,first)
importFrom(dplyr,last)
importFrom(dplyr,pull)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(dplyr,starts_with)
importFrom(dplyr,tibble)
importFrom(openxlsx2,create_border)
importFrom(openxlsx2,create_cell_style)
importFrom(openxlsx2,create_fill)
importFrom(openxlsx2,create_font)
importFrom(openxlsx2,wb_color)
importFrom(openxlsx2,wb_dims)
importFrom(openxlsx2,wb_workbook)
importFrom(redcapAPI,REDCAP_METADATA_FIELDTYPE)
importFrom(redcapAPI,REDCAP_METADATA_VALIDATION_TYPE)
importFrom(redcapAPI,REDCAP_PROJECT_PURPOSE)
importFrom(redcapAPI,REDCAP_SYSTEM_FIELDS)
importFrom(redcapAPI,exportLogging)
importFrom(redcapAPI,exportVersion)
importFrom(redcapAPI,redcapConnection)
importFrom(stringr,str_extract)
importFrom(stringr,str_length)
importFrom(stringr,str_match)
importFrom(stringr,str_remove)
importFrom(stringr,str_replace_all)
importFrom(stringr,str_split_fixed)
importFrom(stringr,str_to_title)
importFrom(stringr,str_trunc)
importFrom(utils,browseURL)
importFrom(utils,head)
importFrom(utils,object.size)
importFrom(utils,write.csv)

# REDCapSync <a href="https://thecodingdocs.github.io/REDCapSync/"><img src="man/figures/logo.png" align="right" height="138" alt="REDCapSync website" /></a>

<!-- badges: start -->

[![Lifecycle:
experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/thecodingdocs/REDCapSync/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/thecodingdocs/REDCapSync/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/thecodingdocs/REDCapSync/graph/badge.svg)](https://app.codecov.io/gh/thecodingdocs/REDCapSync)
[![License: GPL
v3](https://img.shields.io/badge/License-GPLv3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0)
[![CRAN
status](https://www.r-pkg.org/badges/version/REDCapSync)](https://CRAN.R-project.org/package=REDCapSync)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/grand-total/REDCapSync)](https://CRAN.R-project.org/package=REDCapSync)
<!-- badges: end -->

Several R packages exist for using the
[REDCap](https://projectredcap.org/) Application Program Interface (API)
such as, [`redcapAPI`](https://github.com/vubiostat/redcapAPI),
[`REDCapR`](https://ouhscbbmc.github.io/REDCapR/), and
[`tidyREDCap`](https://raymondbalise.github.io/tidyREDCap/). However,
[`REDCapSync`](https://thecodingdocs.github.io/REDCapSync/) is the first
“get-everything” REDCap R package that converts REDCap projects into a
standardized, API-efficient, and project-agnostic
[R6](https://r6.r-lib.org/) object.

## What is REDCapSync?

[`REDCapSync`](https://thecodingdocs.github.io/REDCapSync/) unleashes
the full power of the REDCap API even for the basic R user. When a sync
is performed,
[`REDCapSync`](https://thecodingdocs.github.io/REDCapSync/) uses a cache
of previous saves, a user-defined directory, and the REDCap log to only
update data that changed since the last API call. Project objects can be
used for the best that R has to offer via statistics, visualization,
functions, shiny apps, and more!

The aims of [`REDCapSync`](https://thecodingdocs.github.io/REDCapSync/)
are to…

1.  Encapsulate the REDCap API into one standardized object to
    streamline use.
2.  Automate common tasks such as cleaning, deidentification and merges.
3.  Automate distribution of user-defined Excel datasets to local/cloud
    storage for one or many REDCap projects.
4.  Convert ***labelled*** REDCap data from R or Excel for upload using
    REDCap API.
5.  Power the companion shiny app
    [`RosyREDCap`](https://thecodingdocs.github.io/RosyREDCap/ "RosyREDCap R package")

By leveraging the combined strengths of R and REDCap, users can maintain
strong data pipelines that include statistics, visuals, and even shiny
applications!

## Installation

The stable *release* version can be installed from
[CRAN](https://cran.r-project.org/package=REDCapSync).

``` r
install.packages("REDCapSync")
```

### Development Version

You can install the development version from GitHub with the
[`pak`](https://pak.r-lib.org). Windows users may need to install
[RTools version
4.5](https://cran.r-project.org/bin/windows/Rtools/rtools45/rtools.html "R Getting Started")
to use pak.

``` r
# install.packages("pak")
pak::pak("thecodingdocs/REDCapSync")
```

Alternatively, you can install the development version from GitHub with
the [`remotes`](https://remotes.r-lib.org) package.

``` r
# install.packages("remotes")
remotes::install_github("thecodingdocs/REDCapSync")
```

If you have any issues, try downloading the most recent version of R at
RStudio and update all packages in RStudio. See
[thecodingdocs.com/r/getting-started](https://www.thecodingdocs.com/r/getting-started "R Getting Started").

## Getting Started!

Getting started is as simple as 1.) setting your token, 2.) setting up a
project, and 3.) running project\$sync(). See [Getting
Started](https://thecodingdocs.github.io/REDCapSync/articles/REDCapSync.html "Getting Started")
page for the basics!

``` r
# 1.) setting your token -------------------------------------------------------
Sys.setenv(REDCAPSYNC_FIRST_PROJECT = "YoUrNevErShaReToken") # put in console
# or WAY BETTER put this in your .Renviron file...
# REDCAPSYNC_FIRST_PROJECT = 'YoUrNevErShaReToken'
# Then save file, restart R session (`.rs.restartR()`) and library(REDCapSync)

# 2.) setting up a project -----------------------------------------------------
project <- setup_project(
  project_name = "FIRST_PROJECT",                       
  redcap_uri = "https://redcap.fake.edu/api/",             # same as REDCapR
  dir_path = getwd(),                            # choose appropriate folder
  sync_frequency = "daily",                          # only checks max daily 
  get_entire_log = TRUE                       # for small or medium projects
)

# install.packages("keyring") 
project$test_token()                    # will launch keyring if token fails

# 3.) running project$sync() ---------------------------------------------------
project$sync() 

project$generate_dataset("custom", envir = globalenv())        
```

For an in-depth demonstration of both REDCapSync and RosyREDCap, see
[RMed26-Demo](https://thecodingdocs.github.io/RMed26-Demo/ "RMed26-Demo").

## Framework

![](man/figures/framework.png)

## Minimum Requirements

- R (and RStudio) installed on your computer or server.
- Access to at least one REDCap project (real or test) with API Token
  privileges according to user rights.
- Ideally, you should have User Permissions to logging in order to use
  the package efficiently
- Appropriate permission to export and analyze data for projects for
  which you have a token.
- Basic R knowledge such as installing a package and running code.
- Thoughtful attention to how and where data you create is used and
  stored.

![](man/figures/redcap_priv.png)

## Contributing

If you wish to contribute to this software, use github issues and pull
requests.

### Style Guidelines

#### Functions

- Exported function names: **snake_case**
  - Examples: setup_project, load_project, sync
- Internal function names: **snake_case**
  - Examples: get_project_token, get_redcap_data
  - Exceptions: .onLoad (R standard)
    - redcapAPI functions (**dromedaryCase**)
    - utils functions: write.csv, object.size, browseURL
- Function parameters: **snake_case**
  - examples: project_name, dir_path, redcap_uri, sync_frequency
- Function variables: **snake_case**

#### Constants/Data

- Internal constants: **SCREAMING_SNAKE_CASE**
  - Examples: FAKE_TOKEN, BLANK_PROJECT_COLS, SYNC_FREQUENCY
- Internal datasets: **SCREAMING_SNAKE_CASE**
  - Examples: TEST_CLASSIC, TEST_REPEATING, TEST_CANCER
- Environment variable names: **SCREAMING_SNAKE_CASE**
  - Examples: REDCAPSYNC_TEST_CLASSIC

#### R6 Class

- R6ClassGenerator and R6 Classes: **PascalCase** with capitalized
  “REDCap” acronym
  - Example: REDCapSyncProject
- R6 public_methods: snake_case
- R6 active_bindings: snake_case
  - Exception: .internal for custom functions

#### Package Use

- project_name: SCREAMING_SNAKE_CASE (numbers allowed if not first)
- REDCapSyncProject object environment name can be “project” or
  user-defined
- Package options: **snake.case**
  - Examples: redcapsync.config.show.api.messages

## Disclaimers

- With great power comes great responsibility! The REDCap API has the
  with ability to read and write sensitive data. The API token holder is
  ultimately responsible for their activity and security.
- Always confirm that you have the appropriate permission to use and
  store data.
- REDCap is maintained by Vanderbilt University and is not responsible
  for testing, developing, or maintaining this software.
- This package is still in development and is subject to changes,
  especially pre-CRAN submission.

## Links

- The REDCapSync package is at
  [github.com/thecodingdocs/REDCapSync](https://github.com/thecodingdocs/REDCapSync "REDCapSync R package").
  See instructions above.
- The RosyREDCap package is at
  [github.com/thecodingdocs/RosyREDCap](https://github.com/thecodingdocs/RosyREDCap "RosyREDCap R package").
- A demonstration of both packages is at
  [thecodingdocs.github.io/RMed26-Demo](https://thecodingdocs.github.io/RMed26-Demo/ "RMed26-Demo").
- For more R coding visit
  [TheCodingDocs.com](https://www.thecodingdocs.com/ "TheCodingDocs.com")
- For correspondence/feedback/issues, please email
  <TheCodingDocs@gmail.com>!
- Follow us on Twitter/X
  [x.com/TheCodingDocs](https://x.com/TheCodingDocs/ "TheCodingDocs Twitter")
- Follow me on Twitter/X
  [x.com/BRoseMDMPH](https://x.com/BRoseMDMPH/ "BRoseMDMPH Twitter")

![](man/figures/cover.jpg)

[![](man/figures/TCD.png)](https://www.thecodingdocs.com)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE---------------------------------------------------------------
# #Install usethis if you don't have it.
# #install.packages("usethis")
# usethis::edit_r_environ()
# # Now save your token.... (without the comment symbol '#')
# # REDCAPSYNC_FIRST_PROJECT = "faKeTokeN"
# # Save the file and Close
# # Restart R Session (session tab)
# # .rs.restartR() # this will also restart R session for you.
# Sys.getenv("REDCAPSYNC_FIRST_PROJECT") # now should contain your token

## ----eval=FALSE---------------------------------------------------------------
# project <- load_project("TEST_CLASSIC")
# 
# project$set_keyring_token() # now enter token in pop-up
# # internally the package checks ...
# #1. Sys.getenv()
# token <- Sys.getenv(project$.internal$token_name)
# #2. followed by keyring
# token <- keyring::key_get(service = config$keyring.service(),
#                           username = project$project_name,
#                           keyring = config$keyring())
# # you could set with the following; not preferred because token in script
# # if you do this set in securely in file that no one can ever see; never github!
# keyring::key_set_with_value(service = config$keyring.service(),
#                             username = project$project_name,
#                             password = "VeryNOTsecureWayToSetYourTOKEN",
#                             keyring = config$keyring())

## ----eval=FALSE---------------------------------------------------------------
# options(redcapsync.config.keyring = "REDCapSync")
# # with this kind of method you could define multiple tokens in a private/secret
# # .R file which can be sourced to set all tokens in akeyring. Can also be
# # locked and unlocked (see keyring webstite for details)

## ----eval=FALSE---------------------------------------------------------------
# # Set your token manually
# # again having this in a script is not advised but possible
# Sys.setenv(REDCAPSYNC_FIRST_PROJECT="a_FaKe_TOkEn")
# 
# # Get your token
# Sys.getenv("REDCAPSYNC_FIRST_PROJECT")
# #>[1] "a_FaKe_TOkEn"

## ----eval=FALSE---------------------------------------------------------------
# project <- load_project("TEST_CLASSIC") # replace project name with yours
# 
# project$test_token()

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, eval=FALSE--------------------------------------------------------
# library(REDCapSync)

## ----eval=FALSE, message=FALSE------------------------------------------------
# project <- load_project("TEST_CLASSIC")$sync()
# 
# new_branching <- sample(c("Yes", "No"), size = 50, replace = TRUE)
# 
# # add 50 records
# upload_this <- data.frame(record_id = as.character(51:100),
#                           var_branching = new_branching,
#                           ecog_at_diagnosis = "0")
# 
# project$upload(upload_this) # will also trigger sync/update

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----example, eval=FALSE------------------------------------------------------
# library(REDCapSync)
# projects$print()
# # View(projects$df())                   # Vew your previously saved projects

## ----eval=FALSE---------------------------------------------------------------
# # 1.) setting your token -------------------------------------------------------
# Sys.setenv(REDCAPSYNC_FIRST_PROJECT = "YoUrNevErShaReToken") # in console
# # or WAY BETTER put this in your .Renviron file...
# # REDCAPSYNC_FIRST_PROJECT = 'YoUrNevErShaReToken'
# # Then save file, restart R session (`.rs.restartR()`) and library(REDCapSync)
# 
# # 2.) setting up a project -----------------------------------------------------
# project <- setup_project(
#   project_name = "FIRST_PROJECT",
#   redcap_uri = "https://redcap.fake.edu/api/",             # same as REDCapR
#   dir_path = getwd(),                            # choose appropriate folder
#   sync_frequency = "daily",                          # only checks max daily
#   get_entire_log = TRUE                       # for small or medium projects
# )
# 
# # install.packages("keyring")
# project$test_token()                    # will launch keyring if token fails
# 
# # 3.) running project$sync() ---------------------------------------------------
# project$sync()
# 
# project$generate_dataset("custom", envir = globalenv())

## ----eval=FALSE---------------------------------------------------------------
# #1. set in your Renviron file
# usethis::edit_r_environ()                                 #finds your file
# # another way to find same file
# rstudioapi::navigateToFile(file.path(Sys.getenv("HOME")), ".Renviron"))
# # Then you add --> `REDCAPSYNC_FIRST_PROJECT = 'YoUrNevErShaReToken'`
# # then save file and restart R session (`.rs.restartR()`) and reload package
# 
# #2. set to your token with keyring package! (make sure is installed)
# keyring::key_set(service = config$keyring.service(), # for keyring
#                  username = "FIRST_PROJECT", # is project_name
#                  keyring = config$keyring()) # now enter token in pop-up
# # ... or if have project object already
# project$set_keyring_token() # now enter token in pop-up
# 
# #3. set each time in your console or script (not recommended!) RISKY
# Sys.setenv(REDCAPSYNC_FIRST_PROJECT = "YoUrNevErShaReToken")
# 
# # Confirm
# # If it worked you will see your token when you run...
# Sys.getenv("REDCAPSYNC_FIRST_PROJECT") # for R eviron
# # or...
# keyring::key_get(service = config$keyring.service(), # for keyring
#                  username = "FIRST_PROJECT", # is project_name
#                  keyring = config$keyring())

## ----eval=FALSE---------------------------------------------------------------
# # raw REDCap -------------------------------------------------------------------
# forms <- project$metadata$forms
# fields <- project$metadata$fields
# choices <- project$metadata$choices
# log <- project$redcap$log
# users <- project$redcap$users
# form1 <- project$data$form1
# form2 <- project$data$form2
# # project$data |> list2env(globalenv()) # all forms to global
# 
# # summarized REDCap ------------------------------------------------------------
# # has added columns using log to annotate metadata and users etc. customizable
# # help(dataset)
# # dataset <- project$load_dataset("REDCapSync") # de-identified and no free text
# dataset <- project$generate_dataset("custom") # identified (see parameters)
# annotated_forms <- dataset$metadata$forms
# annotated_fields <- dataset$metadata$fields
# annotated_choices <- dataset$metadata$choices
# annotated_users <- dataset$users
# merged <- dataset$data$merged
# # dataset$data |> list2env(globalenv()) # all forms to global
# # dataset$to_envir(globalenv()) # send to R global env
# # can modify if you want before save (the actual dataset object not global)
# # dataset$save() # will save to output folder by default

## ----eval=FALSE---------------------------------------------------------------
# project$load_dataset(summary_name = "REDCapSync", envir = globalenv())

## ----eval=FALSE---------------------------------------------------------------
# project$generate_dataset(
#   dataset_name = "custom_name"
#   filter_field = "has_something"  # Raw REDCap variable name
#   filter_choices = "Yes"          # REDCap labelled choice
#   exclude_identifiers = TRUE,
#   exclude_free_text = TRUE,
#   include_metadata = TRUE,
#   envir = globalenv()
# )

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(REDCapSync)

## ----eval=TRUE----------------------------------------------------------------
TEST_CLASSIC <- load_project(project_name = "TEST_CLASSIC")
listviewer::jsonedit(TEST_CLASSIC$.internal)

## ----setup, eval=TRUE---------------------------------------------------------
project <- load_project("TEST_CLASSIC")

# projects have read-only active bindings
names(REDCapSyncProject$active)

# projects have public methods
names(REDCapSyncProject$public_methods) |> setdiff("initialize")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE---------------------------------------------------------------
# library(REDCapSync)

## ----eval=FALSE---------------------------------------------------------------
# save_dir <- tempdir() # replace with your intended folder
# 
# project <- load_project("TEST_CLASSIC")
# 
# project$sync() #optional sync
# 
# dataset <- project$load_dataset("REDCapSync")
# 
# # add quick custom variable
# # in future version you can use project$add_field for this
# dataset$data$merged$letter_b <- dataset$data$merged$var_text_letters == "b"
# 
# # send data in global environment
# dataset$to_envir(globalenv())
# 
# # preview file from temporary directory
# dataset$preview()
# 
# # save data in custom location
# dataset$save(dir_other = save_dir)
# 
# # open folder where file was saved
# browseURL(save_dir)

## ----eval=FALSE---------------------------------------------------------------
# save_dir <- tempdir() # replace with your intended folder
# 
# dataset <- projects$load("TEST_CLASSIC")$sync()$load_dataset("REDCapSync")
# # add quick custom variable
# # in future version you can use project$add_field for this
# dataset$data$merged$letter_b <- dataset$data$merged$var_text_letters == "b"
# 
# # send data in global environment
# dataset$to_envir(globalenv())
# 
# # save data in custom location
# dataset$save(dir_other = save_dir) # dir_other not needed for non-test-projects
# 
# # open folder where file was saved
# browseURL(save_dir)

## ----eval=FALSE---------------------------------------------------------------
# projects$load("TEST_CLASSIC")$
#   sync()$
#   load_dataset("REDCapSync", envir = globalenv())

## ----eval=FALSE---------------------------------------------------------------
# save_dir <- tempdir() # replace with your intended folder
# 
# #load and sync (load_project(...) is the same as projects$load(...))
# project <- load_project("TEST_LONGITUDINAL")$sync()
# 
# # see field_names
# project$metadata$fields$field_name
# 
# males_dataset <- project$generate_dataset(dataset_name = "males",
#                                           envir = globalenv(), # global env
#                                           filter_field = "birth_sex",
#                                           filter_choices = "Male")
# # if you check merged and survey it will only contain males
# unique(males_dataset$data$merged$birth_sex) # Male
# unique(males_dataset$data$survey$birth_sex) # Male
# 
# females_dataset <- project$generate_dataset(dataset_name = "females",
#                                             envir = globalenv(), # global env
#                                             filter_field = "birth_sex",
#                                             filter_choices = "Female")
# # if you check merged and survey it will only contain males
# unique(females_dataset$data$merged$birth_sex) # Female
# unique(females_dataset$data$survey$birth_sex) # Female
# 
# # if you want this saved to excel one time you can do
# females_dataset$save(dir_other = save_dir) # dir_other not needed for default
# 
# # To keep this analysis refreshing with each sync use project$add_dataset(..,)
# project$add_dataset(dataset_name = "males",
#                     filter_field = "birth_sex",
#                     filter_choices = "Male")
# # not it can be loaded by name!
# males_dataset <- project$load_dataset("males") # will not send to global
# 
# project$load_dataset("males", envir = globalenv()) # will send to global
# 
# project$save_dataset("males") # will save AND keep track in project object
# 
# project$sync() # once defined will save datasets with sync

## ----eval=FALSE---------------------------------------------------------------
# #load and sync add datasets and save
# project <- load_project("TEST_LONGITUDINAL")$
#   sync()$
#   add_dataset(dataset_name = "males",
#               filter_field = "birth_sex",
#               filter_choices = "Male")$
#   add_dataset(dataset_name = "females",
#               filter_field = "birth_sex",
#               filter_choices = "Female")
# # load males
# project$load_dataset("males", envir = globalenv()) # will send to global
# 
# # load females
# project$load_dataset("females", envir = globalenv()) # will send to global
# 
# # load all default
# project$load_dataset("REDCapSync", envir = globalenv()) # will send to global
# 
# project$load_dataset("males")$preview() # open temporary excel
# project$load_dataset("females")$preview() # open temporary excel
# 
# # can trigger saves with `project$save_datasets()` or default `project$sync()`
# project$save_datasets()
# project$sync()

#' @noRd
clean_redcap_log <- function(redcap_log, drop_exports = FALSE) {
  redcap_log <- unique(redcap_log)
  redcap_log$record_id <- NA
  redcap_log$action_type <- NA
  redcap_log <- redcap_log |>
    lapply(function(x) {
      trimws(x, whitespace = WHITESPACE)
    }) |>
    as.data.frame()
  design_test <- redcap_log$action == "Manage/Design"
  design_rows <- which(design_test)
  not_design_rows <- which(!design_test)
  # notdesign action -----
  keep <- starts_with(match = LOG_ACTION_RECORDS,
                      vars = redcap_log$action[not_design_rows])
  record_rows <- not_design_rows[keep]
  redcap_log$record_id[record_rows] <- gsub(
    paste0(
      "Update record",
      "|",
      "Delete record",
      "|",
      "Create record",
      "|",
      "[:(:]API[:):]",
      "|",
      "Auto",
      "|",
      "calculation",
      "|",
      "Lock/Unlock Record ",
      "|",
      " ",
      "|",
      "[:):]",
      "|",
      "[:(:]"
    ),
    "",
    redcap_log$action[record_rows]
  )
  redcap_log$action_type[record_rows] <- redcap_log$action[record_rows] |>
    strsplit(" ") |>
    lapply(function(x) {
      x[[1L]]
    }) |>
    unlist()
  redcap_log <- redcap_log_labeller(redcap_log = redcap_log,
                                    compare_rows = not_design_rows,
                                    action_var = "action",
                                    start_match = LOG_ACTION_EXPORTS,
                                    label = "Exports")
  redcap_log <- redcap_log_labeller(redcap_log = redcap_log,
                                    compare_rows = not_design_rows,
                                    action_var = "action",
                                    start_match = LOG_ACTION_USERS,
                                    label = "Users")
  redcap_log <- redcap_log_labeller(redcap_log = redcap_log,
                                    compare_rows = not_design_rows,
                                    action_var = "action",
                                    start_match = LOG_ACTION_NO_CHANGES,
                                    label = "No Changes")
  # design details  -------------------
  keep <- starts_with(match = LOG_DETAILS_COMMENTS,
                      vars = redcap_log$details[design_rows])
  comment_rows <- design_rows[keep]
  extracted_ids <- str_extract(string = redcap_log$details[comment_rows],
                               pattern = "(?<=Record: )[^,]+")
  redcap_log$record_id[comment_rows] <- extracted_ids
  redcap_log$action_type[comment_rows] <- "Comment"
  redcap_log <- redcap_log_labeller(redcap_log = redcap_log,
                                    compare_rows = design_rows,
                                    action_var = "details",
                                    start_match = LOG_DETAILS_EXPORTS,
                                    label = "Exports")
  redcap_log <- redcap_log_labeller(redcap_log = redcap_log,
                                    compare_rows = design_rows,
                                    action_var = "details",
                                    start_match = LOG_DETAILS_METADATA_MAJOR,
                                    label = "Metadata Change Major")
  redcap_log <- redcap_log_labeller(redcap_log = redcap_log,
                                    compare_rows = design_rows,
                                    action_var = "details",
                                    start_match = LOG_DETAILS_METADATA_MINOR,
                                    label = "Metadata Change Minor")
  redcap_log <- redcap_log_labeller(redcap_log = redcap_log,
                                    compare_rows = design_rows,
                                    action_var = "details",
                                    start_match = LOG_DETAILS_NO_CHANGES,
                                    label = "No Changes")
  redcap_log <- redcap_log_labeller(redcap_log = redcap_log,
                                    compare_rows = design_rows,
                                    action_var = "details",
                                    start_match = LOG_DETAILS_TOKENS,
                                    label = "Tokens")
  redcap_log <- redcap_log_labeller(redcap_log = redcap_log,
                                    compare_rows = design_rows,
                                    action_var = "details",
                                    start_match = LOG_DETAILS_REPOSITORY,
                                    label = "Repository")
  # end ------------
  row_index <- which(is.na(redcap_log$record) &
                       !is.na(redcap_log$record_id))
  redcap_log$record[row_index] <- redcap_log$record_id[row_index]
  row_index <- which(!is.na(redcap_log$record) &
                       is.na(redcap_log$record_id))
  redcap_log$action_type[row_index] <- "Users"
  redcap_log$record_id <- NULL
  redcap_log$username[which(redcap_log$username == "[survey respondent]")] <- NA
  cannot_label_rows <- which(is.na(redcap_log$action_type))
  if (length(cannot_label_rows) > 0L) {
    warning_about_unlabelled_log <- paste0(
      "Some log elements could not be labelled... Report to issues page \n",
      "    `x <- project$redcap$log[which(is.na(redcap_log$action_type)), ]`"
    )
    cli_alert_warning(warning_about_unlabelled_log)
  }
  if (drop_exports) {
    redcap_log <- redcap_log[which(redcap_log$action_type != "Exports"), ]
  }
  if (nrow(redcap_log) == 0L) {
    return(redcap_log)
  }
  new_order <- order(redcap_log$timestamp, decreasing = TRUE)
  redcap_log <- unique(redcap_log[new_order, ])
  redcap_log
}
#' @noRd
redcap_log_labeller <- function(redcap_log,
                                compare_rows,
                                action_var,
                                start_match,
                                label) {
  start_vars <- redcap_log[[action_var]][compare_rows]
  keep <- starts_with(match = start_match, vars = start_vars)
  redcap_log$action_type[compare_rows[keep]] <- label
  redcap_log
}
#' @noRd
get_interim_log <- function(project) {
  interim_log <- NULL
  if (!project$redcap$has_log_access) {
    cli_alert_danger(
      paste0( # fix for when access is changed
        "You do not have logging access to this REDCap project. If you ",
        "want to refresh the data use `project$sync(hard_reset = TRUE)`.",
        " Request logging access for more effecient use of API exports."
      )
    )
    return(interim_log)
  }
  interim_log <- get_redcap_log(project, log_begin_date = as.Date(
    strptime(project$redcap$log$timestamp[1L], format = "%Y-%m-%d")
  )) |> unique()
  if (nrow(interim_log) <= nrow(project$redcap$log)) {
    head_of_log <- head(project$redcap$log, n = nrow(interim_log))
  } else {
    head_of_log <- project$redcap$log
  }
  unique_log <- interim_log |>
    bind_rows(head_of_log) |>
    unique()
  # dup <- unique_log[
  #which(duplicated(
  #rbind(unique_log, head_of_log),fromLast = TRUE)[
  #seq_len(nrow(unique_log)]), ]
  log_to_clean <- bind_rows(unique_log, head_of_log)
  indices <- seq_len(nrow(unique_log))
  unique_rows <- which(!duplicated(log_to_clean, fromLast = TRUE)[indices])
  interim_log <- unique_log[unique_rows, ]
  interim_log
}
#' @noRd
analyze_log <- function(interim_log, id_col) {
  # check log interim
  #assert_log?
  log_changes <- list(
    hard_reset = FALSE,
    refresh_metadata = FALSE,
    refresh_metadata_major = FALSE,
    refresh_metadata_minor = FALSE,
    refresh_users = FALSE,
    refresh_data = FALSE,
    deleted_records = NULL,
    updated_records = NULL,
    renamed_records = NULL,
    comment_records = NULL,
    length_deleted_records = 0L,
    length_updated_records = 0L,
    length_renamed_records = 0L,
    length_comment_records = 0L
  )
  if (nrow(interim_log) > 0L) {
    # interim_log timestamp ? NULL
    maybe_metadata <- interim_log$action_type[which(is.na(interim_log$record))]
    # inclusion
    refresh_metadata_major <- "Metadata Change Major" %in% maybe_metadata
    refresh_metadata_minor <- "Metadata Change Minor" %in% maybe_metadata
    log_changes$refresh_metadata_major <- refresh_metadata_major
    log_changes$refresh_metadata_minor <- refresh_metadata_minor
    log_changes$refresh_users <- "Users" %in% interim_log$action_type
    log_changes$refresh_metadata <- refresh_metadata_major ||
      refresh_metadata_minor
    if (log_changes$refresh_metadata) {
      # account for minor changes later
      log_changes$hard_reset <- TRUE
      log_changes$refresh_data <- TRUE # this can change with more code
      log_changes$refresh_users <- TRUE
      return(log_changes)
    }
    interim_log_data <- interim_log[which(!is.na(interim_log$record)), ]
    interim_log_data <-
      interim_log_data[which(interim_log_data$action_type != "Users"), ]
    if (nrow(interim_log_data) > 0L) {
      log_list <- split(interim_log_data, interim_log_data$action_type)
      log_changes$comment_records <- unique(log_list$Comment$record)
      log_changes$deleted_records <- unique(log_list$Delete$record)
      log_changes$updated_records <- log_changes$deleted_records |>
        append(log_list$Create$record) |>
        append(log_list$Update$record) |>
        unique()
      update_list <- log_list$Update$details |>
        str_replace_all(" = '[^']*'", "") |>
        strsplit(", ")
      names(update_list) <- log_list$Update$record
      if (length(update_list) > 0L) {
        log_changes$renamed_records <-  update_list |>
          lapply(function(detail) {
            id_col %in% detail
          }) |>
          unlist() |>
          which() |>
          names() |>
          unique()
      }
      log_changes$length_deleted_records <- length(log_changes$deleted_records)
      log_changes$length_updated_records <- length(log_changes$updated_records)
      log_changes$length_renamed_records <- length(log_changes$renamed_records)
      log_changes$length_comment_records <- length(log_changes$comment_records)
      log_changes$refresh_data <- (log_changes$length_deleted_records > 0L) ||
        (log_changes$length_updated_records > 0L)
      log_changes$hard_reset <- log_changes$length_renamed_records > 0L
    }
  }
  log_changes
}
#' @noRd
log_change_messages <- function(log_changes, max_print = 8L) {
  if (log_changes$refresh_metadata) {
    cli_alert_warning("Full update triggered: Metadata was changed!")
    return(invisible())
  }
  nothing_to_do <- sum(
    log_changes$length_deleted_records,
    log_changes$length_updated_records,
    log_changes$length_renamed_records,
    log_changes$length_comment_records
  ) == 0L
  if (nothing_to_do) {
    cli_alert_success("Up to date already!")
    return(invisible())
  }
  if (log_changes$length_renamed_records > 0L) {
    cli_alert_info(paste(
      "Renamed:",
      ifelse(
        log_changes$length_renamed_records > max_print,
        paste(log_changes$length_renamed_records, "records"),
        toString(log_changes$renamed_records)
      )
    ))
    cli_alert_warning("Full update triggered: Records were renamed!")
    return(invisible())
  }
  if (log_changes$length_deleted_records > 0L) {
    cli_alert_info(paste(
      "Deleted:",
      ifelse(
        log_changes$length_deleted_records > max_print,
        paste(log_changes$length_deleted_records, "records"),
        toString(log_changes$deleted_records)
      )
    ))
  }
  if (log_changes$length_updated_records > 0L) {
    cli_alert_info(paste0(
      "Updated: ",
      ifelse(
        log_changes$length_updated_records > max_print,
        paste(log_changes$length_updated_records, "records"),
        toString(log_changes$updated_records)
      )
    ))
  }
  if (log_changes$length_comment_records > 0L) {
    cli_alert_info(paste(
      "Comments:",
      ifelse(
        log_changes$length_comment_records > max_print,
        paste(log_changes$length_comment_records, " records"),
        toString(log_changes$comment_records)
      )
    ))
  }
  invisible(NULL)
}
#' @noRd
LOG_ACTION_EXPORTS <- c("Data export", "Download uploaded ")
#' @noRd
LOG_DETAILS_EXPORTS <- c("Download ", "Export ")
#' @noRd
LOG_ACTION_USERS <- c("Add user ",
                      "Create user role",
                      "Delete user ",
                      "Edit user ",
                      "Rename user role",
                      "User assigned to role ",
                      "User removed from user role")
#' @noRd
LOG_DETAILS_COMMENTS <- c("Add field comment ",
                          "Edit field comment ",
                          "Delete field comment ")
#' @noRd
LOG_ACTION_RECORDS <- c("Create record ",
                        "Delete record ",
                        "Lock/Unlock Record ",
                        "Update record ")
#' @noRd
LOG_ACTION_NO_CHANGES <- c("Enable external module ",
                           "Disable external module ",
                           "Modify configuration for external module ",
                           "Update Response")
#' @noRd
LOG_DETAILS_NO_CHANGES <- c("Add settings for automated survey invitations",
                            "Add/edit stop actions for survey",
                            "Approve production project modifications",
                            "Automatically schedule survey invitation",
                            "Cancel draft mode",
                            "CDIS settings updated",
                            "Checked off item in project checklist",
                            "Click project bookmark",
                            "Copy report",
                            "Create custom record dashboard",
                            "Create data access group",
                            "Create project bookmark",
                            "Create project dashboard",
                            "Create report",
                            "Delete custom record dashboard",
                            "Delete data access group",
                            "Delete project bookmark",
                            "Delete project dashboard",
                            "Delete report",
                            "Disable auto variable",
                            "Disable Google reCAPTCHA",
                            "Disabled survey notification for user",
                            "Edit project bookmark",
                            "Edit project dashboard",
                            "Edit report",
                            "Edit settings for automated survey invitation",
                            "Edit settings for Form Render Skip Logic",
                            "Edit settings for survey queue",
                            "Email survey link",
                            "Enable auto variable",
                            "Enable Clinical Data Pull (CDP) module",
                            "Enable Google reCAPTCHA on public survey",
                            "Enabled survey notification",
                            "Enter draft mode",
                            "Entered Draft Preview",
                            "Execute data quality rule",
                            "Exited Draft Preview mode",
                            "Mapping configuration created",
                            "Mapping configuration deleted",
                            "Mapping configuration updated",
                            "Message added to the queue",
                            "Modify custom record dashboard",
                            "Multi-Language Management",
                            "Reject production proj",
                            "Reorder project bookmarks",
                            "Reorder report",
                            "Request approval for",
                            "Send email ",
                            "Send file ",
                            "Send request to copy project",
                            "Send request to move project",
                            "Send survey invitation",
                            "Switch DAG ",
                            "Unchecked item in project checklist",
                            "updated non-adjudicated item_count",
                            "User opted in to access")
#' @noRd
LOG_DETAILS_TOKENS <- c("Create API token",
                        "Request API token",
                        "User delete own API token",
                        "User regenerate own API token")
#' @noRd
LOG_DETAILS_REPOSITORY <- c("Create folder in File Repository",
                            "Delete file from File Repository",
                            "Delete folder from File Repository",
                            "Upload document to file repository",
                            "Upload file to File Repository")
#' @noRd
LOG_DETAILS_METADATA_MINOR <- c("Add/edit branching logic",
                                "Delete section header",
                                "Modify survey title",
                                "Move project field",
                                "Reorder data collection instruments",
                                "Reorder events",
                                "Reorder project fields",
                                "Tag new identifier fields")
#' @noRd
LOG_DETAILS_METADATA_MAJOR <- c("Copy data collection instrument",
                                "Copy project ",
                                "Create arm",
                                "Create data collection instrument",
                                "Create event",
                                "Create matrix of fields",
                                "Create project",
                                "Create project field",
                                "Delete arm",
                                "Delete data collection instrument",
                                "Delete event",
                                "Delete matrix of fields",
                                "Delete project field",
                                "Delete survey",
                                "Download instrument from Shared Library",
                                "Edit arm",
                                "Edit event",
                                "Edit project field",
                                "Erase all data",
                                "Make project customizations",
                                "Modify project settings",
                                "Modify survey info",
                                "Move project ",
                                "Perform instrument-event mappings",
                                "Rename data collection instrument",
                                "Set up repeating instruments",
                                "Set up survey",
                                "Upload data dictionary",
                                "Upload document for image")

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# clean_redcap_log (Internal)
test_that("clean_redcap_log removes duplicates", {
  redcap_log <- data.frame(
    timestamp = c("2024-01-15 10:30:00",
                  "2024-01-15 10:30:00",
                  "2024-01-15 10:35:00"),
    username = c("user1", "user1", "user2"),
    action = c("Update record 123", "Update record 123", "Create record 456"),
    details = c("Field updated", "Field updated", "New record"),
    record = c("123", "123", "456"),
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  expect_identical(nrow(result), 2L)
})
test_that("clean_redcap_log trims whitespace", {
  redcap_log <- data.frame(
    timestamp = "2024-01-15 10:30:00",
    username = "  user1  ",
    action = "  Update record 123  ",
    details = "  Field updated  ",
    record = "123",
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  expect_false(grepl("^\\s|\\s$", result$username[1L]))
  expect_false(grepl("^\\s|\\s$", result$action[1L]))
  expect_false(grepl("^\\s|\\s$", result$details[1L]))
})
test_that("clean_redcap_log identifies record actions", {
  redcap_log <- data.frame(
    timestamp = c(
      "2024-01-15 10:30:00",
      "2024-01-15 10:35:00",
      "2024-01-15 10:40:00",
      "2024-01-15 10:45:00"
    ),
    username = c("user1", "user2", "user3", "user4"),
    action = c(
      "Update record 123",
      "Delete record 456",
      "Create record 789",
      "Lock/Unlock Record 321"
    ),
    details = c("Updated", "Deleted", "Created", "Locked"),
    record = c("123", "456", "789", "321"),
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  # Check that record_id was extracted and action_type was set
  expect_all_true(!is.na(result$action_type[1L:4L]))
  expect_all_true(
    result$action_type[1L:4L] %in%
      c("Update", "Delete", "Create", "Lock/Unlock")
  )
})
test_that("clean_redcap_log handles Manage/Design actions", {
  redcap_log <- data.frame(
    timestamp = c(
      "2024-01-15 10:30:00",
      "2024-01-15 10:35:00"
    ),
    username = c("user1", "user2"),
    action = c("Manage/Design", "Manage/Design"),
    details = c(
      "Edit project field: field_name",
      "Create project field: new_field"
    ),
    record = c(NA, NA),
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  expect_true(any(result$action_type == "Metadata Change Major", na.rm = TRUE))
})
test_that("clean_redcap_log converts [survey respondent] to NA", {
  redcap_log <- data.frame(
    timestamp = c("2024-01-15 10:35:00", "2024-01-15 10:30:00"),
    username = c("[survey respondent]", "user1"),
    action = c("Update record 123", "Create record 456"),
    details = c("Survey submitted", "Created"),
    record = c("123", "456"),
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  expect_s3_class(result, "data.frame")
  expect_true(is.na(result$username[1L]))
  expect_identical(result$username[2L], "user1")
})
test_that("clean_redcap_log sets action_type column", {
  redcap_log <- data.frame(
    timestamp = "2024-01-15 10:30:00",
    username = "user1",
    action = "Update record 123",
    details = "Updated",
    record = "123",
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  expect_true("action_type" %in% colnames(result))
  expect_false(all(is.na(result$action_type)))
})
test_that("clean_redcap_log handles export actions", {
  redcap_log <- data.frame(
    timestamp = c(
      "2024-01-15 10:30:00",
      "2024-01-15 10:35:00"
    ),
    username = c("user1", "user2"),
    action = c("Data export", "Download uploaded file"),
    details = c("Exported data", "Downloaded file"),
    record = c(NA, NA),
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  expect_true(all(result$action_type %in% c("Exports", NA)))
})
test_that("clean_redcap_log handles user actions", {
  redcap_log <- data.frame(
    timestamp = c(
      "2024-01-15 10:30:00",
      "2024-01-15 10:35:00"
    ),
    username = c("user1", "user2"),
    action = c("Add user test", "Edit user admin"),
    details = c("User added", "User edited"),
    record = c(NA, NA),
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  expect_true(all(result$action_type %in% c("Users", NA)))
})
test_that("clean_redcap_log handles no-changes actions", {
  redcap_log <- data.frame(
    timestamp = "2024-01-15 10:30:00",
    username = "user1",
    action = "Enable external module test",
    details = "Module enabled",
    record = NA,
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  expect_identical(result$action_type[1L], "No Changes")
})
test_that("clean_redcap_log sorts by timestamp descending", {
  redcap_log <- data.frame(
    timestamp = c(
      "2024-01-15 10:30:00",
      "2024-01-15 10:45:00",
      "2024-01-15 10:35:00"
    ),
    username = c("user1", "user2", "user3"),
    action = c("Update record 1", "Update record 2", "Update record 3"),
    details = c("Updated", "Updated", "Updated"),
    record = c("1", "2", "3"),
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  # Should be sorted by timestamp descending
  expect_identical(result$timestamp[1L], "2024-01-15 10:45:00")
  expect_identical(result$timestamp[2L], "2024-01-15 10:35:00")
  expect_identical(result$timestamp[3L], "2024-01-15 10:30:00")
})
test_that("clean_redcap_log removes record_id column", {
  redcap_log <- data.frame(
    timestamp = "2024-01-15 10:30:00",
    username = "user1",
    action = "Update record 123",
    details = "Updated",
    record = "123",
    record_id = NA,
    stringsAsFactors = FALSE
  )
  result <- clean_redcap_log(redcap_log)
  expect_false("record_id" %in% colnames(result))
})
# redcap_log_labeller (Internal)
test_that("redcap_log_labeller works!", {
})
# analyze_log (Internal)
test_that("analyze_log works!", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  id_col <- project$metadata$id_col
  interim_log <- data.frame(
    timestamp = as.character(Sys.time()),
    username = "u1231",
    action = "Create record (API) 99",
    details = NA,
    record = "99",
    action_type = "Create",
    stringsAsFactors = FALSE
  )
  expect_identical(analyze_log(interim_log, id_col)$length_updated_records, 1L)
  interim_log <- data.frame(
    timestamp = as.character(Sys.time()),
    username = "u1231",
    action = "Create record (API) 99",
    details = NA,
    record = "99",
    action_type = c("Create", "Update"),
    stringsAsFactors = FALSE
  )
  expect_identical(analyze_log(interim_log, id_col)$length_updated_records, 1L)
  interim_log <- data.frame(
    timestamp = as.character(Sys.time()),
    username = "u1231",
    action = "Create record (API) 99",
    details = NA,
    record = "99",
    action_type = "Delete",
    stringsAsFactors = FALSE
  )
  log_changes <- analyze_log(interim_log, id_col)
  expect_identical(log_changes$length_updated_records, 1L)
  expect_identical(log_changes$length_deleted_records, 1L)
  expect_identical(log_changes$updated_records, "99")
  expect_identical(log_changes$deleted_records, "99")
  expect_true(log_changes$refresh_data)
  expect_false(log_changes$refresh_metadata)
  interim_log <- data.frame(
    timestamp = as.character(Sys.time()),
    username = "u1231",
    action = "Changes to metadata",
    details = NA,
    record = NA,
    action_type = "Metadata Change Major",
    stringsAsFactors = FALSE
  )
  expect_true(analyze_log(interim_log, id_col)$refresh_metadata)
  interim_log <- data.frame(
    timestamp = as.character(Sys.time()),
    username = "u1231",
    action = "Changes to metadata",
    details = NA,
    record = NA,
    action_type = "Metadata Change Minor",
    stringsAsFactors = FALSE
  )
  expect_true(analyze_log(interim_log, id_col)$refresh_metadata)
})
# log_change_messages (Internal)
test_that("log_change_messages works!", {
  log_changes <- list(
    hard_reset = FALSE,
    refresh_metadata = FALSE,
    refresh_metadata_major = FALSE,
    refresh_metadata_minor = FALSE,
    refresh_users = FALSE,
    refresh_data = FALSE,
    deleted_records = NULL,
    updated_records = NULL,
    renamed_records = NULL,
    comment_records = NULL,
    length_deleted_records = 0L,
    length_updated_records = 0L,
    length_renamed_records = 0L,
    length_comment_records = 0L
  )
  expect_message(log_change_messages(log_changes), "Up to date")
  log_changes$refresh_metadata <- TRUE
  expect_message(log_change_messages(log_changes), "Full update triggered")
  log_changes$refresh_metadata <- FALSE
  log_changes$renamed_records <- "1c"
  log_changes$length_renamed_records <- 1L
  expect_message(log_change_messages(log_changes), "Full update triggered")
  log_changes$renamed_records <- NULL
  log_changes$length_renamed_records <- 0L
  log_changes$length_comment_records <- 1L
  log_changes$length_deleted_records <- 1L
  log_changes$length_updated_records <- 1L
  log_changes$comment_records <- "comment_record"
  log_changes$deleted_records <- "deleted_record"
  log_changes$updated_records <- "updated_record"
  expect_message(log_change_messages(log_changes), "Deleted: deleted")
  expect_message(log_change_messages(log_changes), "Updated: updated")
  expect_message(log_change_messages(log_changes), "Comments: comment")
})
# generate_comment_table (Internal)
test_that("generate_comment_table works!", {
  redcap_log <- TEST_CLASSIC$redcap$log[10L, ]
  time_offset <- seq(from = 100000L, to = 1000000L, by = 100000L)
  redcap_log_comments <- data.frame(
    timestamp = rep(Sys.time(), 10L) - time_offset,
    username = "commenter5",
    action = "Manage/Design",
    details = c(
      generate_comments("Edit", "1", "var_multi_dropdown", "some comment"),
      generate_comments("Add", "2", "var_branching", "different comment"),
      generate_comments("Delete", "2", "var_branching", "A comment"),
      generate_comments("Add", "2", "var_branching", "A comment"),
      generate_comments("Add", "3", "var_multi_dropdown", "some comment"),
      generate_comments("Add", "4", "var_branching", "a comment"),
      generate_comments("Edit", "1", "var_multi_dropdown", NULL),
      generate_comments("Add", "5", "var_multi_dropdown", "another comment"),
      generate_comments("Add", "1", "var_branching", "some comment"),
      generate_comments("Add", "1", "var_multi_dropdown", "some comment")
    ),
    record = c("1", "1", "2", "2", "3", "4", "1", "5", "1", "1"),
    action_type = "Comment",
    stringsAsFactors = FALSE
  )
  comment_table <- generate_comment_table(redcap_log = redcap_log_comments)
  expect_contains(comment_table$comment_type, c("Add", "Edit", "Delete"))
  expect_all_false(is.na(comment_table$comment_type))
  expect_all_false(is.na(comment_table$comment_field))
  expect_all_false(is.na(comment_table$comment_details[-7L]))
  expect_scalar_na(comment_table$comment_details[7L])
  expect_data_frame(comment_table, nrows = 10L)
  comment_table <- generate_comment_table(redcap_log = redcap_log_comments,
                                          only_most_recent = TRUE)
  comment_table
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=F-------------------------------------------------------------------
# library(REDCapSync) # don't forget to load the package

## ----eval=F-------------------------------------------------------------------
# user_cache_location <- rappdirs::user_cache_dir("R")
# redcapsync_cache_location <- file.path(user_cache_location, "REDCapSync")
# redcapsync_cache_location
# # launch the folder on your computer
# utils::browseURL(redcapsync_cache_location)
# # can modify with config options
# config$cache.dir()

## ----eval=F-------------------------------------------------------------------
# # clear entire cache (all projects)
# cache_clear()
# # clear specific project
# cache_clear("OLD_PROJECT1")
# # clear specific project1
# cache_clear(c("OLD_PROJECT2", "OLD_PROJECT3"))

## ----eval=F-------------------------------------------------------------------
# print.data.frame(projects$df)

#' @noRd
list_to_excel <- function(input_list,
                          dir,
                          file_name = NULL,
                          separate = FALSE,
                          overwrite = TRUE,
                          key_cols_list = list(),
                          derived_cols_list = list(),
                          link_col_list = list(),
                          str_trunc_length = 32000L,
                          header_df_list = NULL,
                          header_style = config$openxlsx.header.style(),
                          body_style = config$openxlsx.body.style(),
                          freeze_header = TRUE,
                          pad_rows = 0L,
                          pad_cols = 0L,
                          freeze_keys = TRUE,
                          drop_empty = TRUE) {
  if (is.null(key_cols_list)) {
    key_cols_list <- list()
  }
  if (is.null(link_col_list)) {
    link_col_list <- list()
  }
  if (is.null(derived_cols_list)) {
    derived_cols_list <- list()
  }
  input_list <- process_df_list(input_list, drop_empty = drop_empty)
  list_names <- names(input_list)
  if (length(input_list) == 0L) {
    return(warning("empty input_list cannot be saved", immediate. = TRUE))
  }
  if (separate) {
    for (i in seq_along(input_list)) {
      sub_list <- input_list[i]
      file_name2 <- names(sub_list)
      if (!is.null(file_name)) {
        file_name2 <- paste0(file_name, "_", file_name2)
      }
      keep_keys <- which(names(key_cols_list) == list_names[i])
      keep_deriveds <- which(names(derived_cols_list) == list_names[i])
      key_cols_list_i <- key_cols_list[keep_keys]
      derived_cols_list_i <- derived_cols_list[keep_deriveds]
      save_wb(
        wb = list_to_wb(
          input_list = sub_list,
          key_cols_list = key_cols_list_i,
          derived_cols_list = derived_cols_list_i,
          link_col_list = link_col_list,
          str_trunc_length = str_trunc_length,
          header_df_list = header_df_list,
          header_style = header_style,
          body_style = body_style,
          freeze_header = freeze_header,
          pad_rows = pad_rows,
          pad_cols = pad_cols,
          freeze_keys = freeze_keys,
          drop_empty = drop_empty
        ),
        dir = dir,
        file_name = file_name2,
        overwrite = overwrite
      )
    }
  } else {
    save_wb(
      wb = list_to_wb(
        input_list = input_list,
        key_cols_list = key_cols_list,
        derived_cols_list = derived_cols_list,
        link_col_list = link_col_list,
        str_trunc_length = str_trunc_length,
        header_df_list = header_df_list,
        header_style = header_style,
        body_style = body_style,
        freeze_header = freeze_header,
        pad_rows = pad_rows,
        pad_cols = pad_cols,
        freeze_keys = freeze_keys,
        drop_empty = drop_empty
      ),
      dir = dir,
      file_name = file_name,
      overwrite = overwrite
    )
  }
}
#' @noRd
save_wb <- function(wb, dir, file_name, overwrite = TRUE) {
  if (!dir.exists(dir)) {
    cli_abort("dir doesn't exist: {.file {dir}}")
  }
  path <- file.path(dir, paste0(file_name, ".xlsx")) |> sanitize_path()
  openxlsx::saveWorkbook(wb = wb,
                         file = path,
                         overwrite = overwrite)
  cli_alert_success("Saved {basename(path)}: {.file {path}}")
}
#' @noRd
list_to_csv <- function(input_list,
                        dir,
                        file_name = NULL,
                        overwrite = TRUE,
                        drop_empty = TRUE) {
  input_list <- process_df_list(input_list, drop_empty = drop_empty)
  for (i in seq_along(input_list)) {
    sub_list <- input_list[i]
    file_name2 <- names(sub_list)
    if (!is.null(file_name)) {
      file_name2 <- paste0(file_name, "_", file_name2)
    }
    save_csv(
      form = sub_list[[1L]],
      dir = dir,
      file_name = file_name2,
      overwrite = overwrite
    )
  }
}
#' @noRd
save_csv <- function(form, dir, file_name, overwrite = TRUE) {
  if (!dir.exists(dir)) {
    cli_abort("dir doesn't exist: {.file {dir}}")
  }
  path <- file.path(dir, paste0(file_name, ".csv")) |> sanitize_path()
  write_it <- TRUE
  if (!overwrite && file.exists(path)) {
    write_it <- FALSE
    cli_alert_info("Already a file! {.file {path}}")
  }
  if (write_it) {
    write.csv(x = form, file = path, row.names = FALSE)
    cli_alert_success("Saved {basename(path)}: {.file {path}}")
  }
}
#' @noRd
list_to_wb <- function(input_list,
                       key_cols_list = list(),
                       derived_cols_list = list(),
                       link_col_list = list(),
                       str_trunc_length = 32000L,
                       header_df_list = NULL,
                       header_style = NULL,
                       body_style = NULL,
                       freeze_header = TRUE,
                       pad_rows = 0L,
                       pad_cols = 0L,
                       freeze_keys = TRUE,
                       drop_empty = TRUE) {
  if (is.null(key_cols_list)) {
    key_cols_list <- list()
  }
  if (is.null(link_col_list)) {
    link_col_list <- list()
  }
  if (is.null(derived_cols_list)) {
    derived_cols_list <- list()
  }
  wb <- openxlsx::createWorkbook()
  input_list <- process_df_list(input_list, drop_empty = drop_empty)
  list_names <- names(input_list)
  list_link_names <- list()
  if (length(link_col_list) > 0L) {
    if (is_named_list(link_col_list)) {
      if (!all(names(link_col_list) %in% list_names)) {
        for (list_name in list_names) {
          list_link_names[[list_name]] <- link_col_list
        }
      }
    }
  }
  list_names_rename <- rename_list_names_excel(list_names = list_names)
  for (i in seq_along(list_names)) {
    wb <- form_to_wb(
      wb = wb,
      form = input_list[[list_names[i]]],
      form_name = list_names_rename[i],
      key_cols = key_cols_list[[list_names[i]]],
      derived_cols = NULL,
      link_col_list = list_link_names[[list_names[i]]],
      str_trunc_length = str_trunc_length,
      header_df = header_df_list[[list_names[i]]],
      header_style = header_style,
      body_style = body_style,
      freeze_header = freeze_header,
      pad_rows = pad_rows,
      pad_cols = pad_cols,
      freeze_keys = freeze_keys
    )
  }
  wb
}
#' @noRd
form_to_wb <- function(form,
                       form_name,
                       wb = openxlsx::createWorkbook(),
                       key_cols = NULL,
                       derived_cols = NULL,
                       link_col_list = list(),
                       str_trunc_length = 32000L,
                       header_df = NULL,
                       header_style = NULL,
                       body_style = NULL,
                       freeze_header = TRUE,
                       pad_rows = 0L,
                       pad_cols = 0L,
                       freeze_keys = TRUE) {
  if (nchar(form_name) > 31L) {
    cli_abort("{.arg form_name} is longer than 31 char")
  }
  form[] <- lapply(form, function(col) {
    out <- col
    if (is.character(col)) {
      out <- str_trunc(col, str_trunc_length, ellipsis = "")
    }
    out
  })
  hyperlink_col <- NULL
  if (is_something(header_df)) {
    names(header_df)[match(names(form), names(header_df))]
    missing_headers <- names(form) |>  vec1_not_in_vec2(names(header_df))
    if (length(missing_headers) > 0L) {
      for (missing_header in missing_headers){
        header_df[[missing_header]] <- ""
        #fix later
      }
    }
  }
  if (freeze_keys) {
    all_cols <- colnames(form)
    if (!all(key_cols %in% all_cols)) {
      cli_abort("all key_cols must be in the forms")
    }
    freeze_key_cols <- which(all_cols %in% key_cols)
    if (length(freeze_key_cols) > 0L) {
      if (!is_consecutive_srt_1(freeze_key_cols)) {
        # warning?
        non_key_cols <- seq_len(ncol(form))
        non_key_cols <- non_key_cols[which(!non_key_cols %in% freeze_key_cols)]
        new_col_order <- c(freeze_key_cols, non_key_cols)
        if (is_something(header_df)) {
          header_df <- header_df[, new_col_order]
        }
        form <- form[, new_col_order]
      }
    }
  }
  openxlsx::addWorksheet(wb, form_name)
  start_row_header <- pad_rows + 1L
  start_row_table <- start_row_header
  start_col <- pad_cols + 1L
  if (is_something(header_df)) {
    openxlsx::writeData(
      wb,
      sheet = form_name,
      x = header_df,
      startRow = start_row_header,
      startCol = start_col,
      colNames = FALSE
    )
    start_row_table <- start_row_header + nrow(header_df)
  }
  if (length(link_col_list) > 0L) {
    has_names <- !is.null(names(link_col_list))
    for (i in seq_along(link_col_list)) {
      if (link_col_list[[i]] %in% colnames(form)) {
        class(form[[link_col_list[[i]]]]) <- "hyperlink"
      } # else warning?
      if (has_names) {
        if (names(link_col_list)[i] %in% colnames(form)) {
          hyperlink_col <- which(colnames(form) == names(link_col_list)[i])
          openxlsx::writeData(
            wb,
            sheet = form_name,
            x = form[[link_col_list[[i]]]],
            startRow = start_row_table + 1L,
            startCol = hyperlink_col + pad_cols
          )
          form[[link_col_list[[i]]]] <- NULL
        } # else warning?
      }
    }
  }
  openxlsx::writeDataTable(
    wb,
    sheet = form_name,
    x = form,
    startRow = start_row_table,
    startCol = start_col,
    tableStyle = "none"
  )
  # add derived style
  style_cols <- seq_len(ncol(form)) + pad_cols
  openxlsx::addStyle(
    wb,
    sheet = form_name,
    style = header_style,
    rows = seq(from = start_row_header, to = start_row_table),
    cols = style_cols,
    gridExpand = TRUE,
    stack = TRUE
  )
  openxlsx::addStyle(
    wb,
    sheet = form_name,
    style = body_style,
    rows = seq_len(nrow(form)) + start_row_table,
    cols = style_cols,
    gridExpand = TRUE,
    stack = TRUE
  )
  if (freeze_header || freeze_keys) {
    first_active_row <- NULL
    if (freeze_header) {
      first_active_row <- start_row_table + 1L
    }
    first_active_col <- NULL
    if (freeze_keys) {
      first_active_col <- start_col
      freeze_key_cols <- which(colnames(form) %in% key_cols)
      if (length(freeze_key_cols) > 0L) {
        if (is_consecutive_srt_1(freeze_key_cols)) {
          first_active_col <- first_active_col +
            freeze_key_cols[length(freeze_key_cols)]
        } else {
          the_warning <- "key_cols should be consecutive and start at the left."
          warning(the_warning, immediate. = TRUE)
        }
      }
      openxlsx::freezePane(wb,
                           form_name,
                           firstActiveRow = first_active_row,
                           firstActiveCol = first_active_col)
    }
  }
  wb
}
#' @noRd
rename_list_names_excel <- function(list_names) {
  list_names_rename <- str_trunc(list_names,
                                 width = 31L,
                                 side = "right",
                                 ellipsis = "")
  bad_names <- which_duplicated(list_names_rename)
  if (length(bad_names) > 0L) {
    cli_alert_danger(paste0(
      "Duplicated names when trimmed from right 31 max in Excel: ",
      toString(list_names[bad_names])
    ))
    cli_alert_info(
      paste0(
        "Use CSV or shorten the names and make sure they are unique if they",
        " are trimmed to 31 char. For now will make unique by adding number."
      )
    )
    list_names_rename <- unique_trimmed_strings(strings = list_names_rename,
                                                max_length = 31L)
  }
  list_names_rename
}
#' @noRd
excel_to_list <- function(path) {
  sheets <- openxlsx::getSheetNames(path)
  names(sheets) <- seq_along(sheets)
  clean_sheets <- clean_env_names(sheets)
  out <- list()
  if ("dataset_details" %in% sheets) {
    dataset_details <- readxl::read_xlsx(
      path,
      col_types = "text",
      sheet = which(sheets == "dataset_details")
    )
    if (all(c("paramater", "value") %in% colnames(dataset_details))) {
      the_row <- which(dataset_details$paramater == "raw_form_names")
      form_names <- dataset_details$value[the_row] |>
        strsplit(" [|] ") |>
        unlist()
      the_row <- which(dataset_details$paramater == "cols_start")
      cols_start <- as.integer(dataset_details$value[the_row])
      if (cols_start > 1L) {
        for (i in as.integer(names(sheets)[match(form_names, sheets)])) {
          suppressMessages({
            out[[i]] <- readxl::read_xlsx(path,
                                          col_types = "text",
                                          sheet = i,
                                          col_names = FALSE)
          })
          final_nrow <- nrow(out[[i]])
          if (cols_start < final_nrow) {
            true_colnames <- out[[i]][cols_start, ] |> unlist() |> unname()
            out[[i]] <- out[[i]][(cols_start + 1L):final_nrow, ]
            colnames(out[[i]]) <- true_colnames
            sheets <- sheets[which(sheets != sheets[i])]
          }
        }
      }
    }
  }
  for (i in as.integer(names(sheets))) {
    out[[i]] <- readxl::read_xlsx(path, col_types = "text", sheet = i)
  }
  names(out) <- clean_sheets
  out
}
#' @noRd
openxlsx_header_style <- function() {
  openxlsx::createStyle(fgFill = "#74DFFF",
                        halign = "center",
                        valign = "center",
                        textDecoration = "Bold",
                        fontSize = 14L,
                        fontColour = "black",
                        border = "TopBottomLeftRight")
}
#' @noRd
openxlsx_body_style <- function() {
  openxlsx::createStyle(halign = "left",
                        valign = "center",
                        fontSize = 12L)
}

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# config (Exported)
# config_get (Internal)
test_that("config_get works!", {
  # defaults
  option_list <- list(
    redcapsync.config.allow.test.names = NULL,
    redcapsync.config.show.api.messages = NULL,
    redcapsync.config.verbose = NULL,
    redcapsync.config.offline = NULL,
    redcapsync.config.cache.dir = NULL,
    redcapsync.config.keyring = NULL,
    redcapsync.config.keyring.service = NULL,
    redcapsync.config.openxlsx.header.style = NULL,
    redcapsync.config.openxlsx.body.style = NULL
  )
  envvar_list <- list(
    R_USER_CACHE_DIR = NA,
    REDCAPSYNC_CONFIG_ALLOW_TEST_NAMES = NA,
    REDCAPSYNC_CONFIG_SHOW_API_MESSAGES = NA,
    REDCAPSYNC_CONFIG_VERBOSE = NA,
    REDCAPSYNC_CONFIG_OFFLINE = NA,
    REDCAPSYNC_CONFIG_KEYRING = NA,
    REDCAPSYNC_CONFIG_KEYRING_SERVICE = NA,
    REDCAPSYNC_CONFIG_CACHE_DIR = NA,
    REDCAPSYNC_CONFIG_HEADER_STYLE = NA,
    REDCAPSYNC_CONFIG_BODY_STYLE = NA
  )
  withr::local_options(option_list)
  withr::local_envvar(envvar_list)
  value <- config_get(opt_name = "allow.test.names",
                      type = "logical",
                      default = FALSE)
  expect_false(value)
  value <- config_get(opt_name = "show.api.messages",
                      type = "logical",
                      default = FALSE)
  expect_false(value)
  value <- config_get(opt_name = "verbose",
                      type = "logical",
                      default = TRUE)
  expect_true(value)
  value <- config_get(opt_name = "offline",
                      type = "logical",
                      default = FALSE)
  expect_false(value)
  value <- config_get(opt_name = "keyring",
                      type = "character",
                      default = NULL)
  expect_null(value)
  value <- config_get(opt_name = "keyring.service",
                      type = "character",
                      default = "R-REDCapSync")
  expect_identical(value, "R-REDCapSync")
  value <- config_get(opt_name = "cache.dir",
                      type = "filepath",
                      default = cache_path_default())
  expect_identical(value, cache_path_default())
  value <- config_get(opt_name = "openxlsx.header.style",
                      type = "openxlsx_style",
                      default = openxlsx_header_style())
  expect_identical(value, openxlsx_header_style())
  value <- config_get(opt_name = "openxlsx.body.style",
                      type = "openxlsx_style",
                      default = openxlsx_body_style())
  expect_identical(value, openxlsx_body_style())
  # envvar
  tempdir_test <- sanitize_path(withr::local_tempdir())
  envvar_list <- list(
    R_USER_CACHE_DIR = NA,
    REDCAPSYNC_CONFIG_ALLOW_TEST_NAMES = "True",
    REDCAPSYNC_CONFIG_SHOW_API_MESSAGES = "true",
    REDCAPSYNC_CONFIG_VERBOSE = "FALSE",
    REDCAPSYNC_CONFIG_OFFLINE = TRUE,
    REDCAPSYNC_CONFIG_CACHE_DIR = tempdir_test,
    REDCAPSYNC_CONFIG_KEYRING = "REDCapSync",
    REDCAPSYNC_CONFIG_KEYRING_SERVICE = "R-REDCapSync-new",
    REDCAPSYNC_CONFIG_HEADER_STYLE = "ignored",
    REDCAPSYNC_CONFIG_BODY_STYLE = "ignored"
  )
  withr::local_envvar(envvar_list)
  value <- config_get(opt_name = "allow.test.names",
                      type = "logical",
                      default = FALSE)
  expect_true(value)
  value <- config_get(opt_name = "show.api.messages",
                      type = "logical",
                      default = FALSE)
  expect_true(value)
  value <- config_get(opt_name = "verbose",
                      type = "logical",
                      default = TRUE)
  expect_false(value)
  value <- config_get(opt_name = "offline",
                      type = "logical",
                      default = FALSE)
  expect_true(value)
  value <- config_get(opt_name = "cache.dir",
                      type = "filepath",
                      default = cache_path_default())
  expect_identical(value, tempdir_test)
  value <- config_get(opt_name = "keyring",
                      type = "character",
                      default = NULL)
  expect_identical(value, "REDCapSync")
  value <- config_get(opt_name = "keyring.service",
                      type = "character",
                      default = "R-REDCapSync")
  expect_identical(value, "R-REDCapSync-new")
  value <- config_get(opt_name = "openxlsx.header.style",
                      type = "openxlsx_style",
                      default = openxlsx_header_style())
  expect_identical(value, openxlsx_header_style())
  value <- config_get(opt_name = "openxlsx.body.style",
                      type = "openxlsx_style",
                      default = openxlsx_body_style())
  expect_identical(value, openxlsx_body_style())
  #options override envvar
  tempdir_test2 <- sanitize_path(withr::local_tempdir())
  option_list <- list(
    redcapsync.config.allow.test.names = FALSE,
    redcapsync.config.show.api.messages = "FALSe",
    redcapsync.config.verbose = TRUE,
    redcapsync.config.offline = "false",
    redcapsync.config.cache.dir = tempdir_test2,
    redcapsync.config.keyring = "new-keyring",
    redcapsync.config.keyring.service = "R-REDCapSync",
    redcapsync.config.openxlsx.header.style = openxlsx_body_style(), # flipped
    redcapsync.config.openxlsx.body.style = openxlsx_header_style() # flipped
  )
  withr::local_options(option_list)
  value <- config_get(opt_name = "allow.test.names",
                      type = "logical",
                      default = FALSE)
  expect_false(value)
  value <- config_get(opt_name = "show.api.messages",
                      type = "logical",
                      default = FALSE)
  expect_false(value)
  value <- config_get(opt_name = "verbose",
                      type = "logical",
                      default = TRUE)
  expect_true(value)
  value <- config_get(opt_name = "offline",
                      type = "logical",
                      default = FALSE)
  expect_false(value)
  value <- config_get(opt_name = "cache.dir",
                      type = "filepath",
                      default = cache_path_default())
  expect_identical(value, tempdir_test2)
  value <- config_get(opt_name = "keyring",
                      type = "character",
                      default = NULL)
  expect_identical(value, "new-keyring")
  value <- config_get(opt_name = "keyring.service",
                      type = "character",
                      default = "R-REDCapSync")
  expect_identical(value, "R-REDCapSync")
  value <- config_get(opt_name = "openxlsx.header.style",
                      type = "openxlsx_style",
                      default = openxlsx_header_style())
  expect_identical(value, openxlsx_body_style()) # flipped
  value <- config_get(opt_name = "openxlsx.body.style",
                      type = "openxlsx_style",
                      default = openxlsx_body_style())
  expect_identical(value, openxlsx_header_style()) # flipped
})
# config_validate (Internal)
test_that("config_validate works!", {
  value <- config_validate(
    opt_name = "allow.test.names",
    value = mtcars, # not logical
    type = "logical",
    default = TRUE
  )
  expect_true(value)
  value <- config_validate(
    opt_name = "allow.test.names",
    value = "FALSE", # character ok
    type = "logical",
    default = FALSE
  )
  expect_false(value)
  value <- config_validate(
    opt_name = "allow.test.names",
    value = TRUE,
    type = "logical",
    default = FALSE
  )
  expect_true(value)
})

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# excel_to_list (Internal)
test_that("excel_to_list works!", {
  expect_directory_exists(tempdir_file)
  test_file <- file.path(tempdir_file, "cars.xlsx")
  expect_false(file.exists(test_file))
  df_list <- list(one = mtcars, two = cars)
  list_to_excel(df_list, dir = tempdir_file, file_name = "cars")
  expect_file_exists(test_file)
  df_list_saved <- excel_to_list(test_file)
  expect_list(df_list, len = 2L)
  expect_named(df_list, names(df_list_saved))
  expect_identical(nrow(df_list$one), nrow(df_list_saved$one))
  expect_identical(nrow(df_list$two), nrow(df_list_saved$two))
  expect_error(list_to_excel(file.path(tempdir_file, "fake_cars.xlsx")))
})
# form_to_wb (Internal)
test_that("form_to_wb works!", {
})
# list_to_csv (Internal)
test_that("list_to_csv works!", {
  expect_directory_exists(tempdir_file)
  test_file <- file.path(tempdir_file, "cars.csv")
  test_file_one <- file.path(tempdir_file, "cars_one.csv")
  test_file_two <- file.path(tempdir_file, "cars_two.csv")
  expect_false(file.exists(test_file))
  df_list <- list(one = mtcars, two = cars)
  list_to_csv(df_list, dir = tempdir_file, file_name = "cars")
  expect_file_exists(test_file_one)
  expect_file_exists(test_file_two)
  one <- read.csv(test_file_one)
  two <- read.csv(test_file_two)
  expect_data_frame(one, nrows = nrow(mtcars), ncols = ncol(mtcars))
  expect_identical(colnames(one), colnames(mtcars))
  expect_data_frame(two, nrows = nrow(cars), ncols = ncol(cars))
  expect_identical(colnames(two), colnames(cars))
})
# list_to_excel (Internal)
test_that("list_to_excel works!", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  expect_directory_exists(tempdir_test)
  test_file <- file.path(tempdir_test, "cars.xlsx")
  test_file_one <- file.path(tempdir_test, "cars_one.xlsx")
  test_file_two <- file.path(tempdir_test, "cars_two.xlsx")
  expect_false(file.exists(test_file))
  df_list <- list(one = mtcars, two = cars)
  list_to_excel(df_list, dir = tempdir_test, file_name = "cars")
  expect_file_exists(test_file)
  list_to_excel(df_list,
                dir = tempdir_test,
                file_name = "cars",
                separate = TRUE)
  expect_file_exists(test_file_one)
  expect_file_exists(test_file_two)
  one <- readxl::read_excel(test_file_one)
  two <- readxl::read_excel(test_file_two)
  expect_data_frame(one, nrows = nrow(mtcars), ncols = ncol(mtcars))
  expect_identical(colnames(one), colnames(mtcars))
  expect_data_frame(two, nrows = nrow(cars), ncols = ncol(cars))
  expect_identical(colnames(two), colnames(cars))
})
# list_to_wb (Internal)
test_that("list_to_wb works!", {
})
# rename_list_names_excel (Internal)
test_that("rename_list_names_excel works!", {
  list_names <- c("data", "metadata", "fields", "forms", "choices")
  expect_identical(rename_list_names_excel(list_names), list_names)
  list_names <- c(
    "data",
    "metadata",
    "fields",
    "forms",
    "choices",
    "long_var_name_that_is_not_unique_until_the_very_end1"
  )
  list_names_match <- unique_trimmed_strings(strings = list_names,
                                             max_length = 31L)
  expect_identical(rename_list_names_excel(list_names), list_names_match)
  long_name <- "long_var_name_that_is_not_unique_until_the_very_end2"
  list_names <- append(list_names, long_name)
  expect_false(anyDuplicated(list_names) > 0L)
  expect_message({
    new_names <- rename_list_names_excel(list_names)
  }, "Duplicated names when trimmed from right")
  expect_all_true(str_length(new_names) <= 31L)
})
# save_csv (Internal)
test_that("save_csv works!", {
  expect_directory_exists(tempdir_file)
  test_file <- file.path(tempdir_file, "cars.csv")
  expect_false(file.exists(test_file))
  save_csv(
    form = mtcars,
    dir = tempdir_file,
    file_name = "cars",
    overwrite = TRUE
  )
  expect_file_exists(test_file)
  test_saved_csv <- read.csv(test_file)
  expect_data_frame(test_saved_csv,
                    nrows = nrow(mtcars),
                    ncols = ncol(mtcars))
  expect_message(
    save_csv(
      form = cars,
      dir = tempdir_file,
      file_name = "cars",
      overwrite = FALSE
    ),
    "Already a file"
  )
  test_saved_csv <- read.csv(test_file)
  expect_data_frame(test_saved_csv,
                    nrows = nrow(mtcars),
                    ncols = ncol(mtcars))
})
# save_wb (Internal)
test_that("save_wb works!", {
})

#' @title REDCapSync Dataset Object
#' @description
#' `r lifecycle::badge("experimental")`
#' [R6][R6::R6Class] object representing a standardized dataset generated from
#' a [REDCapSync] project. Use this object to inspect transformed REDCap data,
#' export datasets, and assign prepared analysis tables into the calling
#' environment.
#'
#' @details
#' A `REDCapSyncDataset` can be created ad-hoc from a project with
#' `dataset <- project$generate_dataset()`. For reusability, you can also
#' define with `project$add_dataset()` and then load with
#' `dataset <- project$load_dataset()`.
#'
#' You can add variables manually if you want them to be passed to Excel. But
#' `add_field` is a feature in development.
#'
#' ## Typical workflow
#' ```r
#' setup project from test object
#' project <- setup_project(
#'   project_name = "TEST_CLASSIC",
#'   dir_path = tempdir()
#' )
#' # Create and save a filtered dataset
#' project$add_dataset(
#'   dataset_name = "analysis_set",
#'   filter_field = "var_yesno",
#'   filter_choices = "Yes",
#'   field_names = c("ecog_at_diagnosis", "stage_at_diagnosis")
#' )
#' # generate dataset for R environment
#' dataset <- project$load_dataset("analysis_set")
#' # optional send to global environment
#' dataset$to_envir(globalenv()) # keep in mind potential name conflicts
#' # Optional modify (final save depends on what is in the dataset object)
#' dataset$data$merged$stage_2 <- dataset$data$merged$stage_at_diagnosis == "II"
#' # save to directory
#' dataset$save() # can specify `dir_other`, by default saves to output folder
#' ```
#'
#' This object is designed for users who want a stable dataset output from
#' REDCap without modifying the underlying project. This is also used
#' behind-the-scenes in the RosyREDCap shiny app.
#'
#' ## Key features
#' - Stores a project-specific dataset definition and resulting data
#' - Keeps metadata, record details, user information, and comments in sync
#' - Saves datasets in Excel or CSV formats with optional hyperlinks
#' - Exports dataset components to a user-specified environment
#'
#' @param project Project object from [setup_project()] or [load_project()].
#' @param dataset_name Character. Name of the dataset to generate or load.
#'   If the dataset already exists in the project, the existing definition is
#'   reused.
#' @param transformation_type Character. Data transformation strategy: "default"
#'   (preferred merged output), "none" (raw data structure), or
#'   "merge_non_repeating" (merge only non-repeating forms). Default is
#'   "default".
#' @param merge_form_name Character. Name used for merged non-repeating records.
#'   Default is "merged".
#' @param filter_field Character. Field used for filtering the dataset.
#' @param filter_choices Vector. Allowed values for `filter_field`.
#' @param filter_list List. Named list mapping field names to allowed values.
#'   Use instead of `filter_field`/`filter_choices` for more complex filters.
#' @param filter_strict Logical. If `TRUE`, filters are applied to every form.
#'   If `FALSE`, filters apply only to the record identifier. Default is
#'   `TRUE`.
#' @param form_names Character vector. Forms to include in the dataset.
#'   Default is `NULL` (all forms).
#' @param field_names Character vector. Variables to include in the dataset.
#'   Default is `NULL` (all fields).
#' @param exclude_identifiers Logical. Remove identifier fields. Default is
#'   `TRUE`.
#' @param exclude_free_text Logical. Remove free text fields. Default is
#'   `FALSE`.
#' @param date_handling Character. Date handling method: "none",
#'   "exclude_dates", "random_shift_by_record", "random_shift_by_project",
#'   "zero_by_record", or "zero_by_project". Default is "none".
#' @param labelled Logical. Convert values to labelled vectors if `TRUE`.
#'   Default is `TRUE`.
#' @param clean Logical. Clean the dataset by standardizing missing values and
#'   blanks. Default is `TRUE`.
#' @param drop_blanks Logical. Drop records with blank fields. Default is
#'   `FALSE`.
#' @param drop_missing_codes Logical. Convert REDCap missing codes to `NA`.
#'   Default is `FALSE`.
#' @param drop_others Character vector of additional values to remove.
#' @param include_metadata Logical. Include field metadata in the dataset.
#'   Default is `TRUE`.
#' @param include_users Logical. Include user information in the dataset.
#'   Default is `TRUE`.
#' @param include_records Logical. Include record-level details. Default is
#'   `TRUE`.
#' @param include_log Logical. Include REDCap activity log details. Default is
#'   `FALSE`.
#' @param annotate_from_log Logical. Annotate metadata and records using the
#'   change log. Default is `TRUE`.
#' @param include_comments Logical. Include REDCap comments. Default is
#'   `FALSE`.
#' @param with_links Logical. Include hyperlinks in Excel exports. Default is
#'   `TRUE`.
#' @param separate Logical. Save each form as a separate file instead of a
#'   multi-sheet workbook. Default is `FALSE`.
#' @param use_csv Logical. Write CSV files instead of Excel. Default is
#'   `FALSE`.
#' @param dir_other Character. Directory where the dataset file should be
#'   saved. Defaults to the project's output folder.
#' @param file_name Character. Base file name for saved datasets. Defaults to
#'   `<project_name>_<dataset_name>`.
#' @param envir Environment to assign exported dataset objects. Default is
#'   `NULL`.
#' @examples
#' project <- load_project("TEST_CLASSIC")
#'
#' dataset <- project$generate_dataset(
#'   dataset_name = "stage_2_patients",
#'   filter_field = "stage_at_diagnosis",
#'   filter_choices = "II",
#'   field_names = c("ecog_at_diagnosis", "stage_at_diagnosis")
#'  )
#'
#' dataset$save(dir_other = tempdir())
#'
#' @seealso
#' [project] for using the project objects
#' \code{vignette("Datasets", package = "REDCapSync")}
#' \code{vignette("RosyREDCap", package = "REDCapSync")}
#'
#' @returns
#' An R6 `REDCapSyncDataset` object containing dataset output, metadata,
#' records, user information, and optional REDCap log data.
#' @name dataset
#' @rdname dataset
#' @export
REDCapSyncDataset <- R6Class(
  "REDCapSyncDataset",
  active = list(
    #' @field project_details Read-only list of dataset details from
    #' [setup_project()].
    project_details = function(value) {
      if (!missing(value)) {
        message(
          "`project_details` is read only. To change use `setup_project()`"
        )
      }
      private$details$project_details
    },
    #' @field dataset_details Read-only list of dataset details
    dataset_details = function(value) {
      if (!missing(value)) {
        message(
          "`dataset_details` is read only. To change use `generate_dataset()`"
        )
      }
      private$details$dataset_details
    },
    #' @field redcap Read-only list of dataset details from [project]
    redcap = function(value) {
      if (!missing(value)) {
        message(
          "`redcap` is read only. To change use `generate_dataset()`"
        )
      }
      private$redcap_list
    },
    #' @field links Read-only list of dataset details from [project]
    links = function(value) {
      if (!missing(value)) {
        message(
          "`links` is read only. To change use `generate_dataset()`"
        )
      }
      private$links_list
    }
  ),
  public = list(
    #' @description
    #' The end user will not see `dataset$new()`. This is handled internally.
    #' Users should construct objects using [REDCapSyncProject].
    initialize = function(project,
                          dataset_name,
                          transformation_type = "default",
                          merge_form_name = "merged",
                          filter_field = NULL,
                          filter_choices = NULL,
                          filter_list = NULL,
                          filter_strict = TRUE,
                          field_names = NULL,
                          form_names = NULL,
                          exclude_identifiers = FALSE,
                          exclude_free_text = FALSE,
                          date_handling = "none",
                          labelled = TRUE,
                          clean = TRUE,
                          drop_blanks = FALSE,
                          drop_missing_codes = FALSE,
                          drop_others = NULL,
                          include_metadata = TRUE,
                          include_users = TRUE,
                          include_records = TRUE,
                          include_log = FALSE,
                          annotate_from_log = TRUE,
                          include_comments = FALSE) {
      assert_setup_project(project) # message about internal?
      dataset <- NULL
      if (dataset_name %in% names(project$datasets)) {
        dataset <- load_project_dataset(project = project,
                                        dataset_name = dataset_name)
      }
      if (is.null(dataset)) {
        dataset <- generate_project_dataset(
          project = project,
          dataset_name = dataset_name,
          transformation_type = transformation_type,
          merge_form_name = merge_form_name,
          filter_field = filter_field,
          filter_choices = filter_choices,
          filter_list = filter_list,
          filter_strict = filter_strict,
          field_names = field_names,
          form_names = form_names,
          exclude_identifiers = exclude_identifiers,
          exclude_free_text = exclude_free_text,
          date_handling = date_handling,
          labelled = labelled,
          clean = clean,
          drop_blanks = drop_blanks,
          drop_missing_codes = drop_missing_codes,
          drop_others = drop_others,
          include_metadata = include_metadata,
          include_users = include_users,
          include_records = include_records,
          include_log = include_log,
          annotate_from_log = annotate_from_log,
          include_comments = include_comments
        )
      }
      self$data <- dataset$data
      self$metadata <- dataset$metadata
      self$records <- dataset$records
      self$users <- dataset$users
      self$log <- dataset$log
      self$comments <- dataset$comments
      private$details$dataset_details <- dataset$dataset_details
      private$details$project_details <- dataset$project_details
      private$redcap_list <- project$redcap
      private$redcap_list$log <- NULL
      private$redcap_list$users <- NULL
      private$links_list <- project$links
      invisible(self)
    },
    #' @description Print some key dataset information
    print = function() {
      cli_h1("REDCapSyncDataset")
      cli_text("Project: {private$details$project_details$project_name}")
      cli_text("Dataset: {private$details$dataset_details$dataset_name}")
      help_cli_text()
      invisible(self)
    },
    #' @field data list of data where names are forms
    data = NULL,
    #' @field metadata list of metadata
    metadata = NULL,
    #' @field records data.frame of records with timestamps
    records = NULL,
    #' @field users data.frame of users with timestamps
    users = NULL,
    #' @field log data.frame of log
    log = NULL,
    #' @field comments data.frame of comments
    comments = NULL,
    #' @description Return flat list
    save = function(with_links = TRUE,
                    separate = FALSE,
                    use_csv = FALSE,
                    dir_other = NULL,
                    file_name = NULL) {
      data_list <- save_project_data_list(
        data_list = self,
        with_links = with_links,
        separate = separate,
        use_csv = use_csv,
        dir_other = dir_other,
        file_name = file_name
      )
      # would be nice if updated supclass project in future version
      invisible(data_list)
    },
    #' @description export dataset to envir of your choosing. Keep in mind
    #' potential name conflicts
    to_envir = function(envir = NULL) {
      data_list <- data_list_to_save(self)
      if (!is.null(envir)) {
        list2env(data_list, envir = envir)
      }
      invisible(data_list)
    }
  ),
  private = list(
    details = NULL,
    redcap_list = NULL,
    links_list = NULL
  ),
  lock_objects = FALSE,
  cloneable = FALSE
)

#' @noRd
labelled_to_raw_form <- function(form, project) {
  form <- all_character_cols(form)
  if (nrow(form) == 0L) {
    return(form)
  }
  fields <- filter_fields_from_form(form = form, project = project)
  for (i in seq_len(nrow(fields))) {
    field_row <- fields[i, ]
    field_name <- field_row$field_name
    conversion_table <- generate_choices_table(field_row, project)
    match_rows <- match(form[[field_name]], conversion_table$name)
    if (field_row$has_choices) {
      no_match <- which(is.na(match_rows) & !is.na(form[[field_name]]))
      if (length(no_match) > 0L) {
        bad_choices <- form[[field_name]][no_match] |> unique() |> toString()
        cli_alert_danger(form[[field_name]][no_match]) # confirm with real
        cli_abort("Mismatched REDCap! `{field_name}` = {bad_choices}")
      }
      form[[field_name]] <- conversion_table$code[match_rows]
    } else {
      map_row <- which(!is.na(match_rows))
      form[[field_name]][map_row] <- conversion_table$name[match_rows[map_row]]
    }
  }
  form
}
#' @noRd
raw_to_labelled_form <- function(form, project) {
  if (project$metadata$has_coding_conflicts) {
    cli_abort(
      paste0(
        "You cannot use labelled = 'TRUE' because you have a coding conflict ",
        "in your data dictionary... Try {.topic REDCapSync::setup_project} ",
        "with labelled = 'FALSE'. The conflicts are from: ",
        "{toString(project$metadata$coding_conflict_field_names)}"
      )
    )
  }
  form <- all_character_cols(form)
  if (nrow(form) == 0L) {
    return(form)
  }
  fields <- filter_fields_from_form(form = form, project = project)
  for (i in seq_len(nrow(fields))) {
    field_row <- fields[i, ]
    field_name <- field_row$field_name
    conversion_table <- generate_choices_table(field_row, project)
    match_rows <- match(form[[field_name]], conversion_table$code)
    if (field_row$has_choices) {
      no_match <- which(is.na(match_rows) & !is.na(form[[field_name]]))
      if (length(no_match) > 0L) {
        bad_choices <- form[[field_name]][no_match] |> unique() |> toString()
        cli_alert_warning("Mismatched REDCap! {field_name} = {bad_choices}")
        affected <- form[[project$metadata$id_col]][no_match] |> toString()
        cli_alert_danger("Affected records: {affected}")
      }
      form[[field_name]] <- conversion_table$name[match_rows]
    } else {
      map_row <- which(!is.na(match_rows))
      form[[field_name]][map_row] <- conversion_table$name[match_rows[map_row]]
    }
  }
  form
}
#' @noRd
generate_choices_table <- function(field_row, project) {
  missing_codes <- project$metadata$missing_codes
  use_missing_codes <- is.data.frame(project$metadata$missing_codes)
  has_choices <- field_row$has_choices
  conversion_table <- NULL
  if (has_choices) {
    choice_table <- split_choices(field_row$select_choices_or_calculations)
    conversion_table <- bind_rows(conversion_table, choice_table)
  }
  if (use_missing_codes) {
    conversion_table <- bind_rows(conversion_table, missing_codes)
  }
  conversion_table
}
#' @noRd
labelled_to_raw_project <- function(project) {
  project <- assert_blank_project(project)
  if (!project$settings$labelled) {
    cli_abort("project is already raw or coded (not labelled values)")
  }
  for (form_name in names(project$data)) {
    form <- project$data[[form_name]]
    project$data[[form_name]] <- labelled_to_raw_form(form = form,
                                                      project = project)
  }
  project$settings$labelled <- FALSE
  invisible(project)
}
#' @noRd
raw_to_labelled_project <- function(project) {
  project <- assert_blank_project(project)
  if (project$settings$labelled) {
    cli_abort("project is already labelled (not raw or coded values)")
  }
  if (project$metadata$has_coding_conflicts) {
    cli_abort("project has codebook conflict(s) and cannot be labelled!")
  }
  for (form_name in names(project$data)) {
    form <- project$data[[form_name]]
    project$data[[form_name]] <- raw_to_labelled_form(form = form,
                                                      project = project)
  }
  project$settings$labelled <- TRUE
  invisible(project)
}
#' @noRd
normalize_redcap <- function(denormalized, project, labelled) {
  forms <- project$metadata$forms
  fields <- project$metadata$fields
  events <- project$metadata$events
  event_mapping <- project$metadata$event_mapping
  form_list <- list()
  if (nrow(denormalized) == 0L) {
    return(form_list)
  }
  is_longitudinal <- project$metadata$is_longitudinal
  denormalized <- all_character_cols(denormalized)
  add_ons <- c(project$metadata$id_col,
               "arm_number",
               "event_name",
               "redcap_event_name",
               REDCAP_REPEAT_COLS)
  if (is_longitudinal) {
    denormalized$id_temp <- seq_len(nrow(denormalized))
    denormalized <- merge(
      denormalized,
      events[, c("arm_number", "event_name", "unique_event_name")],
      by.x = "redcap_event_name",
      by.y = "unique_event_name",
      sort = FALSE,
      all.x = TRUE
    )
    add_ons <- add_ons[which(add_ons %in% colnames(denormalized))]
    col_names <- c(add_ons, colnames(denormalized)) |> unique()
    new_order <- order(denormalized$id_temp)
    cols_to_keep <- col_names |>
      lapply(function(c) {
        which(colnames(denormalized) == c)
      }) |>
      unlist() |>
      as.integer()
    denormalized <- denormalized[new_order, cols_to_keep]
    denormalized$id_temp <- NULL
  }
  add_ons <- add_ons[which(add_ons %in% colnames(denormalized))]
  if (!all(project$metadata$raw_structure_cols %in% colnames(denormalized))) {
    cli_abort(
      paste0(
        "denormalized is missing one of the following... and that's weird: ",
        toString(project$metadata$raw_structure_cols)
      )
    )
  }
  form_rows <- which(forms$form_name %in% unique(fields$form_name))
  form_names <- forms$form_name[form_rows]
  has_repeating_forms <- project$metadata$has_repeating_forms
  repeating_forms <- forms$form_name[which(forms$repeating)]
  for (form_name in form_names) {
    form_field_names <- fields$field_name[which(
      fields$form_name == form_name &
        fields$field_name %in% colnames(denormalized) &
        fields$field_name != project$metadata$id_col
    )]
    if (length(form_field_names) == 0L) {
      cli_alert_danger("Possibly no access to '{form_name}'. Unable to obtain.")
    }
    # consider message for what vars are missing with vec1_not_in_vec2
    row_index <- NULL
    if (length(form_field_names) > 0L) {
      add_ons_x <- add_ons
      is_repeating_form <- form_name %in% repeating_forms
      row_index <- seq_len(nrow(denormalized))
      if (is_repeating_form) {
        if (!"redcap_repeat_instrument" %in% colnames(denormalized)) {
          cli_abort("redcap_repeat_instrument not in colnames(denormalized)")
        }
        if (is_longitudinal) {
          has_form_event <- event_mapping$form == form_name
          no_rep_form_event <- which(!event_mapping$repeating & has_form_event)
          no_rep_event <- event_mapping$unique_event_name[no_rep_form_event]
          form_no_rep_event <- denormalized$redcap_event_name %in% no_rep_event
          form_has_rep <- denormalized$redcap_repeat_instrument == form_name
          row_index <- which(form_has_rep | form_no_rep_event)
        }
        if (!is_longitudinal) {
          row_index <- which(denormalized$redcap_repeat_instrument == form_name)
        }
      }
      if (!is_repeating_form) {
        add_ons_x <- add_ons_x[which(!add_ons_x %in% REDCAP_REPEAT_COLS)]
      }
      if (!is_repeating_form && is_longitudinal && has_repeating_forms) {
        form_event_rows <- which(event_mapping$form == form_name)
        event_forms <- unique(event_mapping$unique_event_name[form_event_rows])
        has_event_name <- denormalized$redcap_event_name %in% event_forms
        has_rep_instrument <- is.na(denormalized$redcap_repeat_instrument)
        row_index <- which(has_rep_instrument & has_event_name)
      }
      if (!is_repeating_form && is_longitudinal && !has_repeating_forms) {
        form_event_rows <- which(event_mapping$form == form_name)
        event_forms <- unique(event_mapping$unique_event_name[form_event_rows])
        row_index <- which(denormalized$redcap_event_name %in% event_forms)
      }
      if (!is_repeating_form && !is_longitudinal && has_repeating_forms) {
        row_index <- which(is.na(denormalized$redcap_repeat_instrument))
      }
    }
    if (is_something(row_index)) {
      col_names <- unique(c(add_ons_x, form_field_names))
      subset <- denormalized[row_index, col_names]
      if (labelled) {
        subset <- raw_to_labelled_form(form = subset, project = project)
      }
      form_list[[form_name]] <- subset
    }
  }
  form_list <- all_character_cols_list(form_list)
  form_list
}
#' @noRd
get_min_dates <- function(data_list) {
  # assert_data_list contains data and metadata with forms and fields
  data_forms <- data_list$data
  metadata <- data_list$metadata
  fields <- metadata$fields
  id_cols <- metadata$form_key_cols |>
    unlist() |>
    unique()
  empty <- data.frame(
    record_id = character(),
    min_date = as.Date(character()),
    stringsAsFactors = FALSE
  )
  if (!is_something(data_forms)) {
    return(empty)
  }
  date_vector <- fields$field_name[which(fields$field_type_r == "date")]
  all_dates <- list()
  # Loop through each form in the list
  for (form in data_forms) {
    if (id_cols[1L] %in% names(form)) {
      # Find the date fields that exist in this form
      existing_fields <- intersect(names(form), date_vector)
      if (length(existing_fields) > 0L) {
        df_subset <- form[, c(id_cols[1L], existing_fields), drop = FALSE]
        df_long <- stats::reshape(
          df_subset,
          varying = existing_fields,
          v.names = "date",
          times = existing_fields,
          timevar = "field",
          direction = "long"
        )
        df_long$date <- as.Date(df_long$date, format = "%Y-%m-%d")
        all_dates <- c(all_dates, list(df_long))
      }
    }
  }
  if (length(all_dates) == 0L) {
    return(empty)
  }
  combined <- do.call(rbind, all_dates)
  combined <- combined[!is.na(combined$date), ]
  min_dates <- stats::aggregate(date ~ record_id, data = combined, FUN = min)
  min_dates
}
#' @noRd
get_project_url <- function(project,
                            link_type = "home",
                            open_browser = TRUE) {
  assert_choice(link_type, LINK_TYPES)
  the_link <- project$links[[paste0("redcap_", link_type)]]
  if (open_browser) {
    browseURL(the_link)
  }
  invisible(the_link)
}
#' @noRd
get_record_url <- function(project,
                           record = NULL,
                           page = NULL,
                           instance = NULL,
                           open_browser = TRUE) {
  # FIX
  link <- paste0(
    project$links$redcap_base,
    "redcap_v",
    project$redcap$version,
    "/DataEntry/record_home.php?pid=",
    project$redcap$project_id
  )
  id_col <- project$metadata$id_col
  if (!is.null(record)) {
    if (!record %in% project$record_summary[[id_col]]) {
      cli_abort("`{record}` is not one of the records inside project")
    }
    if ("arm_number" %in% colnames(project$record_summary)) {
      arm_row <- which(project$record_summary[[id_col]] == record)
      link <- paste0(link,
                     "&arm=",
                     project$record_summary$arm_number[arm_row])
    }
    link <- paste0(link, "&id=", record)
  }
  if (!is.null(page)) {
    link <- gsub("record_home", "index", link)
    if (!page %in% project$metadata$forms$form_name) {
      cli_abort(
        paste0(
          "{.arg page} has to be one of the instrument names: ",
          toString(project$metadata$forms$form_name)
        )
      )
    }
    link <- paste0(link, "&page=", page)
    if (!is.null(instance)) {
      rep_rows <- which(project$metadata$forms$repeating)
      repeating_form_names <- project$metadata$forms$form_name[rep_rows]
      if (!page %in% repeating_form_names) {
        cli_abort(
          paste0(
            "If you provide an instance, it has to be a repeating instrument: ",
            toString(repeating_form_names)
          )
        )
      }
      link <- paste0(link, "&instance=", instance)
    }
  }
  if (open_browser) {
    browseURL(link)
  }
  invisible(link)
}
#' @noRd
get_key_col_list <- function(data_list) {
  forms <- data_list$metadata$forms
  if (!is_something(forms)) {
    cli_abort("Empty --> `project$metadata$forms`")
  }
  out_list <- seq_len(nrow(forms)) |> lapply(function(i) {
    out <- data_list$metadata$id_col
    if (data_list$metadata$is_longitudinal) {
      out <- append(out, "redcap_event_name")
    }
    if (forms$repeating[i]) {
      out <- append(out, "redcap_repeat_instrument")
      out <- append(out, "redcap_repeat_instance")
    }
    out
  })
  names(out_list) <- forms$form_name
  out_list
}
#' @noRd
get_all_field_names <- function(data_list) {
  data_list$data |>
    lapply(colnames) |>
    unlist() |>
    unique()
}
#' @noRd
get_identifier_fields <- function(data_list,
                                  get_type = "deidentified",
                                  invert = FALSE) {
  assert_choice(get_type, choices = GET_TYPE)
  # assert data list
  # what to do when record_id is marked as identifier? add psuedo
  # handle dates seprately if shifted
  id_cols <- data_list$metadata$form_key_cols |> unlist() |> unique()
  fields <- data_list$metadata$fields
  fields$validation_type <- fields$text_validation_type_or_show_slider_number
  not_id <- !fields$field_name %in% id_cols
  is_identifier <- fields$identifier == "y"
  is_likely_identifier <- fields$validation_type %in% REDCAP_MAYBE_IDS_STRICT
  more_strict <- fields$validation_type %in% REDCAP_MAYBE_IDS_SUPER_STRICT
  is_date <- fields$validation_type
  if (get_type == "identified") {
    keep_rows <- NULL
  }
  if (get_type == "deidentified") {
    keep_rows <- which(is_identifier)
  }
  if (get_type == "deidentified_strict") {
    keep_rows <- which(is_identifier | is_likely_identifier)
  }
  if (get_type == "deidentified_super_strict") {
    is_notes <- fields$field_type == "notes"
    is_text <- fields$field_type == "text"
    has_validation <- !is.na(fields$validation_type)
    is_free_text <- is_notes | (is_text & !has_validation) & not_id
    keep_rows <- which(is_identifier |
                         is_likely_identifier |
                         is_free_text |
                         more_strict)
    #account for drops
  }
  id_fields <- fields$field_name[keep_rows]
  if (invert) {
    id_fields <- setdiff(fields$field_name, id_fields)
  }
  id_fields
}
#' @noRd
field_to_form_names <- function(project,
                                field_names,
                                transform = FALSE,
                                strict = FALSE) {
  metadata <- project$metadata
  if (transform) {
    metadata <- project$transformation$metadata
  }
  fields <- metadata$fields
  form_key_cols <- metadata$form_key_cols |>
    unlist() |>
    unique()
  field_names_keys <- field_names[which(field_names %in% form_key_cols)]
  form_names_keys <- field_names_keys |>
    lapply(function(field_name) {
      metadata$form_key_cols |>
        names() |>
        lapply(function(form_name) {
          if (!field_name %in% metadata$form_key_cols[[form_name]]) {
            return(NULL)
          }
          form_name
        }) |>
        unlist()
    }) |>
    unlist() |>
    as.character() |>
    unique()
  field_names_not_keys <-
    field_names[which(!field_names %in% form_key_cols)] |> unique()
  form_names_not_keys <-
    fields$form_name[match(field_names_not_keys, fields$field_name)] |>
    drop_nas() |>
    unique()
  form_names <- form_names_not_keys
  if (!strict) {
    form_names <- form_names |>
      append(form_names_keys) |>
      unique()
  }
  form_names
}
#' @noRd
construct_header_list <- function(data_list,
                                  md_elements = c("form_name",
                                                  "field_type",
                                                  "field_label")) {
  fields <- data_list$metadata$fields
  if (anyDuplicated(fields$field_name) > 0L)
    cli_abort("duplicate names not allowed in fields")
  data_field_list <- lapply(data_list$data, colnames)
  header_df_list <- data_field_list |> lapply(function(field_names) {
    x <- field_names |>
      lapply(function(field_name) {
        x_row <- which(fields$field_name == field_name)
        if (length(x_row) > 0L) {
          return(as.character(fields[md_elements][x_row, ]))
        }
        rep("", length(md_elements))
      }) |>
      as.data.frame()
    colnames(x) <- field_names
    x <- x[which(apply(x, 1L, function(x_row) {
      any(nzchar(x_row)) #consider use NA?
    })), ]
    x
  }) |>
    process_df_list(silent = TRUE)
  header_df_list
}
#' @noRd
field_names_metadata <- function(project, field_names, col_names) {
  fields <- project$metadata$fields
  bad_field_names <- field_names[which(
    !field_names %in% c(
      project$metadata$fields$field_name,
      project$metadata$raw_structure_cols,
      "arm_number",
      "event_name"
    )
  )]
  if (length(bad_field_names) > 0L)
    cli_abort(
      paste0(
        "All column names in your form must match items in your metadata, ",
        "`project$metadata$fields$field_name`... ",
        toString(bad_field_names)
      )
    )
  fields <- fields[which(fields$field_name %in% field_names), ]
  if (!missing(col_names)) {
    if (is_something(col_names))
      fields <- fields[[col_names]]
  }
  fields
}
#' @noRd
filter_fields_from_form <- function(form, project) {
  forms <- field_to_form_names(project = project,
                               field_names = colnames(form),
                               strict = TRUE)
  repeating_form_rows <- which(project$metadata$forms$repeating)
  repeating_forms <- project$metadata$forms$form_name[repeating_form_rows]
  if (any(forms %in% repeating_forms) && length(forms) > 1L) {
    cli_abort(
      paste0(
        "All column names in your form must match only one form in your",
        "metadata, `project$metadata$forms$form_name`, unless they are",
        " all non-repeating"
      )
    )
  }
  fields <- project |> field_names_metadata(field_names = colnames(form))
  fields <- fields[which(fields$field_type != "descriptive"), ]
  fields$has_choices <- !is.na(fields$select_choices_or_calculations)
  rows_no_choices <- which(fields$field_type %in% c("calc", "text", "slider"))
  fields$has_choices[rows_no_choices] <- FALSE
  fields
}
#' @noRd
extract_values_from_form_list <- function(form_list, col_name) {
  names(form_list) |>
    lapply(function(form_name) {
      form_list[[form_name]][[col_name]]
    }) |>
    unlist() |>
    drop_nas() |>
    unique()
}
#' @noRd
extract_project_records <- function(data_list) {
  all_records <- NULL
  id_col <- data_list$metadata$id_col
  if (is_something(data_list$data)) {
    record_id_col <- extract_values_from_form_list(
      form_list = data_list$data,
      col_name = id_col
    )
    all_records <- data.frame(
      record_id_col = record_id_col,
      last_api_call = NA,
      was_saved = FALSE,
      stringsAsFactors = FALSE
    )
    rownames(all_records) <- NULL
    id_col_match <- which(colnames(all_records) == "record_id_col")
    colnames(all_records)[id_col_match] <- id_col
  }
  all_records
}
#' @noRd
LINK_TYPES <- c(
  "base",
  "home",
  "record_home",
  "records_dashboard",
  "api",
  "api_playground",
  "codebook",
  "user_rights",
  "setup",
  "logging",
  "designer",
  "dictionary",
  "data_quality",
  "identifiers"
)
REDCAP_REPEAT_COLS <- c("redcap_repeat_instrument", "redcap_repeat_instance")

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# labelled_to_raw_project (Internal)
test_that("labelled_to_raw_project and raw_to_labelled_project works!", {
  project <- mock_test_project()$.internal
  # ensure project is marked as labelled and has labelled values
  expect_true(project$settings$labelled)
  # sanity check: labelled value present in example form
  expect_all_true(project$data$other$var_yesno %in% c("Yes", "No"))
  # convert and capture returned project
  expect_no_error({
    project_converted <- labelled_to_raw_project(project)
  })
  # internals updated
  expect_false(project_converted$settings$labelled)
  # values converted from labelled ("Yes"/"No") to raw codes ("1"/"0")
  expect_all_true(project_converted$data$other$var_yesno %in% c("0", "1"))
  expect_no_error({
    project_again <- raw_to_labelled_project(project_converted)
  })
  expect_all_true(project_again$data$other$var_yesno %in% c("Yes", "No"))
})
# labelled_to_raw_form (Internal)
test_that("labelled_to_raw_form and raw_to_labelled_form works!", {
  project <- mock_test_project()$.internal
  project_summary <- load_project_dataset(project, "REDCapSync") |>
    data_list_to_save()
  merged <- all_character_cols(project_summary$merged)
  var_yesno_labelled <- merged$var_yesno
  values <- unique(merged$var_yesno)
  expect_vector(values, size = 2L)
  expect_contains(values, c("Yes", "No"))
  merged <- labelled_to_raw_form(merged, project)
  var_yesno_coded <- merged$var_yesno
  values <- unique(merged$var_yesno)
  expect_vector(values, size = 2L)
  expect_contains(values, c("1", "0"))
  var_yesno_labelled_check <- (var_yesno_labelled == "Yes") |>
    as.integer() |>
    as.character()
  expect_identical(var_yesno_labelled_check, var_yesno_coded)
  merged <- raw_to_labelled_form(merged, project)
  var_yesno_labelled_again <- merged$var_yesno
  values <- unique(merged$var_yesno)
  expect_vector(values, size = 2L)
  expect_contains(values, c("Yes", "No"))
  expect_identical(var_yesno_labelled, var_yesno_labelled_again)
  expect_in("Unknown", project$metadata$missing_codes$name)
  labelled <- select(merged, record_id, var_branching)
  labelled$var_branching[3L] <- "Unknown"
  labelled$var_branching[5L] <- "Not applicable"
  raw <- labelled_to_raw_form(labelled, project)
  expect_identical(raw$var_branching[3L], "UNK")
  expect_identical(raw$var_branching[5L], "NA")
  labelled <- raw_to_labelled_form(raw, project)
  expect_identical(labelled$var_branching[3L], "Unknown")
  expect_identical(labelled$var_branching[5L], "Not applicable")
  mismatch <- select(merged, record_id, var_branching)
  mismatch$var_branching[3L] <- "Random Thing"
  error_message <- "Mismatched REDCap"
  expect_error(labelled_to_raw_form(mismatch, project), error_message)
  raw_mismatch <- raw
  raw_mismatch$var_branching[3L] <- "Random Thing"
  expect_message(raw_to_labelled_form(raw_mismatch, project), error_message)
  project$metadata$has_coding_conflicts <- TRUE
  error_text <- "you have a coding conflict"
  expect_error(raw_to_labelled_form(raw_mismatch, project), error_text)
})
# normalize_redcap (Internal)
test_that("normalize_redcap works with classic project", {
  project <- mock_test_project()$.internal
  call_list <- mock_test_calls()
  denormalized <- call_list$data
  result <- normalize_redcap(denormalized, project, labelled = TRUE)
  expect_type(result, type = "list")
  form_names <- names(result)
  expected_forms <- project$metadata$forms$form_name
  expect_all_true(expected_forms %in% form_names)
  expect_all_true(unlist(lapply(result, is.data.frame)))
  id_col <- project$metadata$id_col
  fields <- project$metadata$fields
  fields <- fields[which(fields$field_name != id_col), ]
  text_fields <- fields$field_name[which(fields$form_name == "text")]
  expect_true(id_col %in% colnames(result$text))
  expect_all_true(text_fields %in% colnames(result$text))
  rows_x <- which(fields$form_name == "other" &
                    (!fields$field_type %in% FIELD_TYPES_NOT_IN_DATA))
  other_fields <- fields$field_name[rows_x]
  expect_all_true(other_fields %in% colnames(result$other))
  expect_all_false(text_fields %in% colnames(result$other))
  expect_all_false(other_fields %in% colnames(result$text))
})
test_that("normalize_redcap works with longitudinal project", {
  project_name <- "TEST_REDCAPR_LONGITUDINAL"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  denormalized <- call_list$data
  result <- normalize_redcap(denormalized, project, labelled = TRUE)
  expect_true(project$metadata$is_longitudinal)
  expect_type(result, type = "list")
  form_names <- names(result)
  expected_forms <- project$metadata$forms$form_name
  expect_all_true(expected_forms %in% form_names)
  expect_all_true(unlist(lapply(result, is.data.frame)))
  id_col <- project$metadata$id_col
  fields <- project$metadata$fields
  expect_true(id_col %in% colnames(result$demographics))
  field_rows <- which(fields$form_name == "demographics" &
                        (!fields$field_type %in% FIELD_TYPES_NOT_IN_DATA))
  fields_to_check <- fields$field_name[field_rows]
  expect_all_true(fields_to_check %in% colnames(result$demographics))
  fields <- fields[which(fields$field_name != id_col), ]
  field_rows <-  which(fields$form_name == "demographics" &
                         (!fields$field_type %in% FIELD_TYPES_NOT_IN_DATA))
  fields_to_check <- fields$field_name[field_rows]
  expect_all_true(fields_to_check %in% colnames(result$demographics))
  expect_all_false(fields_to_check %in% colnames(result$other))
  expect_all_false(fields_to_check %in% colnames(result$text))
})
test_that("normalize_redcap works with repeating project", {
  project_name <- "TEST_REPEATING"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  denormalized <- call_list$data
  result <- normalize_redcap(denormalized, project, labelled = TRUE)
  expect_false(project$metadata$is_longitudinal)
  expect_type(result, type = "list")
  form_names <- names(result)
  expected_forms <- project$metadata$forms$form_name
  expect_all_true(expected_forms %in% form_names)
  expect_all_true(unlist(lapply(result, is.data.frame)))
  id_col <- project$metadata$id_col
  fields <- project$metadata$fields
  expect_true(id_col %in% colnames(result$form_1))
  field_rows <- which(fields$form_name == "form_1" &
                        (!fields$field_type %in% FIELD_TYPES_NOT_IN_DATA))
  fields_to_check <- fields$field_name[field_rows]
  expect_all_true(fields_to_check %in% colnames(result$form_1))
  extra_cols <- c("record_id",
                  "redcap_repeat_instrument",
                  "redcap_repeat_instance")
  field_rows <-  which(fields$form_name == "repeating" &
                         (!fields$field_type %in% FIELD_TYPES_NOT_IN_DATA))
  fields_to_check <- unique(append(extra_cols, fields$field_name[field_rows]))
  expect_all_true(fields_to_check %in% colnames(result$repeating))
  field_rows <-  which(fields$form_name == "repeating_2" &
                         (!fields$field_type %in% FIELD_TYPES_NOT_IN_DATA))
  fields_to_check <- unique(append(extra_cols, fields$field_name[field_rows]))
  expect_all_true(fields_to_check %in% colnames(result$repeating_2))
})
# raw_to_labelled_project (Internal)
test_that("raw_to_labelled_project works!", {
})
# raw_to_labelled_form (Internal)
test_that("raw_to_labelled_form works!", {
})
# construct_header_list (Internal)
test_that("construct_header_list works!", {
})
# extract_project_records (Internal)
test_that("extract_project_records works!", {
})
# extract_values_from_form_list (Internal)
test_that("extract_values_from_form_list works!", {
})
# field_names_metadata (Internal)
test_that("field_names_metadata works!", {
})
# field_to_form_names (Internal)
test_that("field_to_form_names works!", {
})
# filter_fields_from_form (Internal)
test_that("filter_fields_from_form works!", {
  project <- mock_test_project("TEST_REPEATING")$.internal
  form <- bind_rows(project$data$repeating, project$data$repeating_2)
  expect_error(filter_fields_from_form(form, project))
  expect_error(filter_fields_from_form(form, project))
  form <- project$data$form_1
  expect_no_error(filter_fields_from_form(form, project))
  fields <- filter_fields_from_form(form, project)
  expect_data_frame(fields, nrows = ncol(form))
})
# generate_choices_table (Internal)
test_that("generate_choices_table works!", {
})
# get_all_field_names (Internal)
test_that("get_all_field_names works!", {
  project <- mock_test_project()$.internal
  field_names <- get_all_field_names(project)
  expect_all_true(field_names %in% project$metadata$fields$field_name)
})
# get_identifier_fields (Internal)
test_that("get_identifier_fields works!", {
  # minimal metadata/fields to exercise deidentified / strict / super_strict
  field_names <- c("record_id", "email", "cell", "dob", "other", "notes", "int")
  valid_types <- c(NA, "email", "phone", "date_mdy", NA, NA, "integer")
  fields <- data.frame(
    field_name = field_names,
    identifier = c("y", NA, NA, "y", NA, NA, NA),
    field_type = c("text", "text", "text", "text", "text", "notes", "text"),
    text_validation_type_or_show_slider_number = valid_types,
    stringsAsFactors = FALSE
  )
  data_list <- list(metadata = list(fields = fields))
  data_list$metadata$form_key_cols <- "record_id"
  # deidentified: only explicit identifier flagged "y"
  out_deid <- get_identifier_fields(data_list, get_type = "deidentified")
  expect_identical(out_deid, c("record_id", "dob"))
  # deidentified_strict: includes fields with validation types
  # in strict list (email, phone)
  out_strict <- get_identifier_fields(data_list = data_list,
                                      get_type = "deidentified_strict")
  expect_identical(out_strict, c("record_id", "email", "cell", "dob"))
  # deidentified_super_strict: includes additional validation  (dates, etc.)
  out_super <- get_identifier_fields(data_list = data_list,
                                     get_type = "deidentified_super_strict")
  expect_identical(out_super, setdiff(field_names, "int"))
  # invert = TRUE should return the complement set
  out_inv <- get_identifier_fields(data_list = data_list,
                                   get_type = "deidentified",
                                   invert = TRUE)
  expect_identical(out_inv, c("email", "cell", "other", "notes", "int"))
})
# get_key_col_list (Internal)
test_that("get_key_col_list works!", {
})
# get_min_dates (Internal)
test_that("get_min_dates works!", {
  # construct minimal data_list with date fields across forms
  data_list <- list(
    data = list(
      form1 = data.frame(
        record_id = c("1", "2"),
        date1 = c("2020-01-01", "2020-01-05"),
        stringsAsFactors = FALSE
      ),
      form2 = data.frame(
        record_id = c("2", "3"),
        date2 = c("2020-02-01", "2020-01-15"),
        stringsAsFactors = FALSE
      )
    ),
    metadata = list(
      fields = data.frame(
        field_name = c("record_id", "date1", "date2"),
        field_type_r = c(NA, "date", "date"),
        stringsAsFactors = FALSE
      ),
      form_key_cols = list(record_id = "record_id")
    )
  )
  out <- get_min_dates(data_list)
  # basic structure checks
  expect_data_frame(out, nrows = 3L, ncols = 2L)
  expect_true(all(c("record_id", "date") %in% colnames(out)))
  # values: min date per record across forms
  one <- as.character(out$date[match("1", out$record_id)])
  two <- as.character(out$date[match("2", out$record_id)])
  three <- as.character(out$date[match("3", out$record_id)])
  expect_identical(one, "2020-01-01")
  expect_identical(two, "2020-01-05")
  expect_identical(three, "2020-01-15")
})
# get_project_url (Internal)
test_that("get_project_url works!", {
  project <- mock_test_project()$.internal
  project <- update_project_links(project)
  e <- new.env(parent = emptyenv())
  # get_project_url
  e$url <- NULL
  local_mocked_bindings(
    browseURL = function(url) {
      e$url <- url
    }
  )
  for (link_type in LINK_TYPES) {
    e$url <- NULL
    get_project_url(project, link_type = link_type, open_browser = TRUE)
    expect_identical(e$url, project$links[[paste0("redcap_", link_type)]])
    out <- get_project_url(project, link_type = link_type, open_browser = FALSE)
    expect_identical(out, project$links[[paste0("redcap_", link_type)]])
  }
})
# get_record_url (Internal)
test_that("get_record_url works!", {
  project <- mock_test_project()$.internal
  project$links$redcap_base <- "https://fakeredcap.com/"
  expected_link <- paste0(
    project$links$redcap_base,
    "redcap_v",
    project$redcap$version,
    "/DataEntry/record_home.php?pid=",
    project$redcap$project_id
  )
  e <- new.env(parent = emptyenv())
  # get_project_url
  e$url <- NULL
  local_mocked_bindings(
    browseURL = function(url) {
      e$url <- url
    }
  )
  expect_identical(get_record_url(project, open_browser = FALSE), expected_link)
  expect_null(e$url) # does not call url
  e$url <- NULL
  get_record_url(project, open_browser = TRUE)
  expect_identical(e$url, expected_link)
  e$url <- NULL
  expect_error(get_record_url(project, record = "59", open_browser = FALSE),
               regexp = "is not one of the records")
  expect_identical(
    get_record_url(project, record = "1", open_browser = FALSE),
    paste0(expected_link, "&id=1")
  )
  expect_identical(
    get_record_url(project, page = "text", open_browser = FALSE),
    paste0(
      gsub("record_home", "index", expected_link, fixed = TRUE),
      "&page=text"
    )
  )
  expect_error(get_record_url(project, page = "text_2", open_browser = FALSE))
})

#' @title Clear your cached projects
#' @description
#' Finds the location of the cache established by \code{\link[hoardr]{hoard}}
#' and deletes stored project information (not data)! If you provide
#' `project_names`, it will remove only those projects from the cache. If you
#' want to truly delete the project files, you must do so at the project
#' directory you set up.
#' @details
#' The cache only stores information like project_name, token_name, directory
#' location, and other details from [setup_project()].The default location of
#' the cache location is defined by using R_USER_CACHE_DIR if set. Otherwise,
#' it follows platform conventions via [hoardr], saving a file
#' "R/REDCapSync/projects.rds". No direct project data is stored in the cache.
#' Notably, tokens and data are not stored here. The key variables stored in
#' the cache are...
#' * `project_name` - unique identifier for REDCapSync package
#' * `redcap_uri` - server location
#' * `token_name` - where to find token environment with [Sys.getenv()]
#' * `dir_path` - where to saved project and associated files locally
#' * `project_id` - obtained from API call and "locks-in" the connection
#' * `redcap_version` - obtained from API call and affects links
#' * `last_sync` and `sync_frequency` - informs REDCap sync of when to update
#' * other variables from project info and some internal package mechanics
#' @inheritParams sync
#' @examples
#' \dontrun{
#' cache_clear("OLD_PROJECT")
#' cache_clear() # every project
#' }
#'
#' @seealso
#' \code{vignette("Cache", package = "REDCapSync")}
#' [setup_project] for initializing projects
#'
#' @returns Message of outcome and invisible NULL.
#' @export
cache_clear <- function(project_names = NULL) {
  cache <- get_cache()
  if (!is.null(project_names)) {
    for (project_name in project_names) {
      cache_remove_project(project_name)
    }
    return(invisible())
  }
  if (cache$exists("projects.rds")$exists) {
    cache$delete("projects.rds")
  }
  cli_alert_warning("You must delete files manually from the directory.")
  cli_alert_success("REDCapSync cache cleared: {.path {cache_path()}}")
  invisible()
}
#' @noRd
cache_remove_project <- function(project_name) {
  assert_env_name(project_name, max.chars = 31L, all_caps = TRUE)
  projects <- get_projects()
  is_in_cache <- project_name %in% projects$project_name
  if (!is_in_cache) {
    cli_alert_warning("'{project_name}' is not in your cache. Nothing to do.")
    return(invisible())
  }
  projects <- projects[which(projects$project_name != project_name), ]
  save_projects_to_cache(projects, silent = TRUE)
  success_message <- paste0("'{project_name}' removed from cache, but files",
                            " may still be located in the directory.")
  cli_alert_success(success_message)
  invisible()
}
#' @noRd
get_cache_default <- function() {
  cache <- hoardr::hoard()
  cache$cache_path_set(path = "REDCapSync", type = "user_cache_dir")
  cache$mkdir()
  cache
}
#' @noRd
get_cache <- function() {
  cache <- hoardr::hoard()
  cache$cache_path_set(full_path = config$cache.dir())
  cache$mkdir()
  cache
}
#' @noRd
cache_path_default <- function() {
  cache <- get_cache_default()
  path <- sanitize_path(cache$cache_path_get())
  path
}
#' @noRd
cache_path <- function() {
  cache <- get_cache()
  path <- sanitize_path(cache$cache_path_get())
  path
}
#' @noRd
sweep_dirs_for_cache <- function(project_names = NULL) {
  # for if others are using the same object
  projects <- get_projects()
  if (nrow(projects) == 0L) {
    return(invisible())
  }
  project_list <- split(projects, projects$project_name)
  had_change <- FALSE
  all_project_names <- names(project_list)
  if (is.null(project_names)) {
    project_names <- all_project_names
  }
  project_names <- project_names[which(project_names %in% all_project_names)]
  updated_projects <- NULL
  for (project_name in project_names) {
    from_cache <- project_list[[project_name]]
    expected_path <- get_project_path(
      project_name = project_name,
      dir_path = from_cache$dir_path,
      type = "details"
    )
    from_cache <- try_else_null(assert_project_details(from_cache, nrows = 1L))
    to_cache <- NULL
    if (file.exists(expected_path)) {
      to_cache <- try_else_null({
        x <- suppressWarnings({
          readRDS(expected_path)
        })
        assert_project_details(x, nrows = 1L)
        x
      })
    }
    if (is.null(from_cache) || is.null(to_cache)) {
      loaded_cache <- try_else_null({
        load_project(project_name = project_name)$.internal |>
          extract_project_details()
      })
      if (is.null(loaded_cache)) {
        danger_message <- paste0("Unable to load ",
                                 project_name,
                                 ". Removed! Retry `setup_project(...)`")
        cli_alert_danger(danger_message)
        project_list[[project_name]] <- NULL
        had_change <- TRUE
      } else {
        project_list[[project_name]] <- loaded_cache
      }
    }
    rownames(from_cache) <- NULL
    rownames(to_cache) <- NULL
    if (!is.null(to_cache) && !identical(from_cache, to_cache)) {
      if (!is.null(from_cache)) {
        to_cache$dir_path <- from_cache$dir_path
      } # could cause issue?
      project_list[[project_name]] <- to_cache
      updated_projects <-  append(updated_projects, project_name)
      had_change <- TRUE
    }
  }
  if (had_change) {
    updating_message <- paste0("Updated cache! This can happen when using",
                               " multiple computers or with version changes.")
    cli_alert_info(updating_message)
    save_projects_to_cache(bind_rows(project_list), silent = FALSE)
  }
}

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
test_that("local_envvar seen by tests, exists, but empty at first", {
  withr::local_envvar(REDCAPSYNC_CONFIG_CACHE_DIR = tempdir_file)
  expect_true(nzchar(Sys.getenv("R_USER_CACHE_DIR")))
  expect_directory_exists(Sys.getenv("R_USER_CACHE_DIR"))
  expect_directory_exists(Sys.getenv("REDCAPSYNC_CONFIG_CACHE_DIR"))
  testing_path2 <- Sys.getenv("REDCAPSYNC_CONFIG_CACHE_DIR")
  # real cache path projects file path testing path projects.rds
  real_cache_path_projects2 <- file.path(testing_path2, "projects.rds")
  # expect false test directory exists testing path
  expect_directory_exists(testing_path2)
  # expect false test file exists real cache path projects
  expect_false(test_file_exists(real_cache_path_projects2))
})
# cache_clear (Exported)
test_that("cache_clear works!", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  fake_cache <- get_cache()
  test_file <- file.path(fake_cache$cache_path_get(), "projects.rds")
  fake_project_details <- BLANK_PROJECT_DETAILS  |>
    bind_rows(data.frame(project_name = "FAKE1", stringsAsFactors = FALSE)) |>
    bind_rows(data.frame(project_name = "FAKE2", stringsAsFactors = FALSE)) |>
    bind_rows(data.frame(project_name = "FAKE3", stringsAsFactors = FALSE)) |>
    bind_rows(data.frame(project_name = "FAKE4", stringsAsFactors = FALSE))
  saveRDS(fake_project_details, test_file)
  expect_file_exists(test_file)
  projects <- get_projects()
  expect_data_frame(projects, nrows = 4L)
  expect_message(cache_clear("FAKE2"), "removed from cache")
  expect_message(cache_clear("FAKE2"), "not in your cache")
  projects <- get_projects()
  expect_data_frame(projects, nrows = 3L)
  expect_false("FAKE2" %in% projects$project_name)
  expect_message(cache_clear(c("FAKE1", "FAKE4")), "removed from cache")
  expect_message(cache_clear(c("FAKE1", "FAKE4")), "not in your cache")
  projects <- get_projects()
  expect_data_frame(projects, nrows = 1L)
  expect_identical(projects$project_name, "FAKE3")
  expect_file_exists(test_file)
  expect_message(cache_clear(), "REDCapSync cache cleared")
  expect_message({
    projects <- get_projects()
  }, "No cached projects")
  expect_false(file.exists(test_file))
})
# cache_path (Internal)
test_that("cache_path works CONFIG overrides R_USER!", {
  user_cache <- cache_path()
  expect_true(nzchar(Sys.getenv("R_USER_CACHE_DIR")))
  expect_directory_exists(user_cache)
  expect_identical(user_cache, config$cache.dir())
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(REDCAPSYNC_CONFIG_CACHE_DIR = tempdir_test)
  defined_cache <- cache_path()
  expect_identical(defined_cache, tempdir_test)
  expect_identical(defined_cache, config$cache.dir())
  withr::local_envvar(R_USER_CACHE_DIR = NA)
  expect_identical(Sys.getenv("R_USER_CACHE_DIR"), "")
  fake_other_cache <- hoardr::hoard()
  fake_other_cache$cache_path_set(path = "REDCapSync", type = "user_cache_dir")
  expected_user_path <- sanitize_path(fake_other_cache$cache_path_get())
  fake_other_cache$mkdir()
  expect_directory_exists(cache_path())
  expect_directory_exists(expected_user_path)
  expect_identical(cache_path(), tempdir_test)
  withr::local_envvar(REDCAPSYNC_CONFIG_CACHE_DIR = NA)
  expect_identical(Sys.getenv("REDCAPSYNC_CONFIG_CACHE_DIR"), "")
  expect_identical(cache_path(), expected_user_path)
})
# cache_remove_project (Internal)
test_that("cache_remove_project works!", {
  project <- mock_test_project()$.internal
  projects <- get_projects()
  project_details <- extract_project_details(project)
  expect_message(cache_remove_project(project_name = "TEST_OTHER"),
                 "Nothing to do")
  expect_false("TEST_other" %in% projects$project_name)
  expect_false("TEST_CLASSIC" %in% projects$project_name)
  add_project_details_to_cache(project_details)
  projects <- get_projects()
  expect_true("TEST_CLASSIC" %in% projects$project_name)
  expect_message(cache_remove_project(project_name = "TEST_CLASSIC"),
                 "removed from cache")
  projects <- get_projects()
  expect_false("TEST_CLASSIC" %in% projects$project_name)
})
# get_cache (Internal)
test_that("get_cache works and deletes!", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  test_cache <- get_cache()
  test_cache_path <- test_cache$cache_path_get()
  expect_directory_exists(test_cache_path)
  test_file <- file.path(test_cache_path, "projects.rds")
  expect_false(file.exists(test_file))
  file.create(test_file)
  expect_file_exists(test_file)
  test_cache$delete_all()
  expect_false(file.exists(test_file))
})
# sweep_dirs_for_cache (Internal)
test_that("sweep_dirs_for_cache updates cache when project files exist", {
  project <- mock_test_project()$.internal
  project_details <- extract_project_details(project)
  add_project_details_to_cache(project_details)
  # Create project files on disk
  project_path <- get_project_path(
    project_name = project$project_name,
    dir_path = project$dir_path
  )
  saveRDS(project, file = project_path)
  details_path <- get_project_path(
    project_name = project$project_name,
    dir_path = project$dir_path,
    type = "details"
  )
  saveRDS(project_details, file = details_path)
  # Run sweep
  expect_no_error(sweep_dirs_for_cache())
  # Cache should still contain the project
  projects <- get_projects()
  expect_true(project$project_name %in% projects$project_name)
})
test_that("sweep_dirs_for_cache handles empty cache", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  projects_before <- get_projects()
  expect_identical(nrow(projects_before), 0L)
  # Sweep should handle empty cache
  expect_no_error(sweep_dirs_for_cache())
  projects_after <- get_projects()
  expect_identical(nrow(projects_after), 0L)
})
test_that("sweep_dirs_for_cache handles non-existent project_names", {
  project <- mock_test_project()$.internal
  project_details <- extract_project_details(project)
  add_project_details_to_cache(project_details)
  # Try to sweep with non-existent project names
  expect_no_error(sweep_dirs_for_cache(
    project_names = "NONEXISTENT_PROJECT"
  ))
  # Original project should still be in cache
  projects <- get_projects()
  expect_true(project$project_name %in% projects$project_name)
})
test_that("sweep_dirs_for_cache compares cached and disk project details", {
  project <- mock_test_project()$.internal
  project_details <- extract_project_details(project)
  add_project_details_to_cache(project_details)
  # Create project files with updated details
  project_path <- get_project_path(
    project_name = project$project_name,
    dir_path = project$dir_path
  )
  # Modify the project details
  project$redcap$version <- "13.0.0"
  updated_project_details <- extract_project_details(project)
  details_path <- get_project_path(
    project_name = project$project_name,
    dir_path = project$dir_path,
    type = "details"
  )
  saveRDS(updated_project_details, file = details_path)
  # Save full project
  saveRDS(project, file = project_path)
  # Run sweep
  expect_no_error(sweep_dirs_for_cache())
  # Cache should be updated with new version
  projects <- get_projects()
  project_row <- which(projects$project_name == project$project_name)
  cached_project <- projects[project_row, ]
  expect_identical(cached_project$version, "13.0.0")
})

#' @noRd
save_project_dataset <- function(project, dataset_name) {
  id_col <- project$metadata$id_col
  dataset_details <- project$datasets[[dataset_name]]
  data_list <- load_project_dataset(project = project,
                                    dataset_name = dataset_name)
  records <- extract_project_records(data_list)[[id_col]]
  # consider also saving as Rdata to lazy load
  data_list <- save_project_data_list(data_list,
                                      with_links = dataset_details$with_links,
                                      separate = dataset_details$separate,
                                      use_csv = dataset_details$use_csv,
                                      dir_other = dataset_details$dir_other,
                                      file_name = dataset_details$file_name)
  dataset_details <- data_list$dataset_details # account potential for conflicts
  last_save_time <- dataset_details$last_save_time
  final_form_tab_names <- dataset_details$final_form_tab_names
  project$datasets[[dataset_name]]$n_records <- dataset_details$n_records
  project$datasets[[dataset_name]]$last_save_time <- last_save_time
  project$datasets[[dataset_name]]$final_form_tab_names <- final_form_tab_names
  row_new <- which(project$record_summary[[id_col]] %in% records)
  project$record_summary[[dataset_name]][row_new] <- TRUE
  project$record_summary$was_saved[row_new] <- TRUE
  project$internals$last_dataset_save <- now_time()
  invisible(project)
}
#' @noRd
save_project_data_list <- function(data_list,
                                   with_links = TRUE,
                                   separate = FALSE,
                                   use_csv = FALSE,
                                   dir_other = NULL,
                                   file_name = NULL) {
  id_col <- data_list$metadata$id_col
  dataset_details <- data_list$dataset_details
  dataset_name <- dataset_details$dataset_name
  file_extension <- ifelse(use_csv, ".csv", ".xlsx")
  if (is.null(dir_other)) {
    dir_path <- data_list$project_details$dir_path
    dir_other <- file.path(dir_path, "output")
  }
  if (is.null(file_name)) {
    project_name <- data_list$project_details$project_name
    file_name <- paste0(project_name, "_", dataset_name)
  }
  file_path <- file.path(dir_other, paste0(file_name, file_extension))
  dataset_details$with_links <- with_links
  dataset_details$separate <- separate
  dataset_details$use_csv <- use_csv
  dataset_details$dir_other <- dir_other
  dataset_details$file_name <- file_name
  dataset_details$file_path <- file_path
  # add headers-------
  form_names <- names(data_list$data)
  header_df_list <- construct_header_list(data_list)
  key_cols_list <- get_key_col_list(data_list)
  cols_start <- 4L
  if (dataset_name == "REDCapSync_raw") {
    cols_start <- 1L
    header_df_list <- NULL
  }
  # track form names
  link_col_list <- list()
  if (dataset_details$with_links) {
    if (length(data_list$data) > 0L) {
      data_list$data <- data_list$data |>
        lapply(function(form) {
          add_redcap_links(form, data_list)
        })
      if (dataset_details$include_records) {
        data_list$records <- add_redcap_links(form = data_list$records,
                                              project = data_list)
      }
      #check for conflicting name
      link_col_list <- list("redcap_link")
      names(link_col_list) <- id_col
    }
  }
  data_list <- data_list_to_save(data_list)
  # save -----
  last_save_time <- now_time()
  final_form_tab_names <- rename_list_names_excel(list_names = names(data_list))
  names(final_form_tab_names) <- names(data_list)
  dataset_details$last_save_time <- last_save_time
  dataset_details$final_form_tab_names <- final_form_tab_names
  dataset_details$raw_form_names <- form_names
  retained_details <- dataset_details
  dataset_details$cols_start <- cols_start
  value <- dataset_details |>
    lapply(function(x_row) {
      paste0(x_row, collapse = " | ")
    }) |>
    unlist() |>
    unname()
  data_list$dataset_details <- data.frame(
    paramater = names(dataset_details),
    value = value
  )
  if (dataset_details$use_csv) {
    list_to_csv(
      input_list = data_list,
      dir = dataset_details$dir_other,
      file_name = dataset_details$file_name,
      overwrite = TRUE
    ) # account for links with CSV like new column
  } else {
    list_to_excel(
      input_list = data_list,
      dir = dataset_details$dir_other,
      separate = dataset_details$separate,
      link_col_list = link_col_list,
      key_cols_list = key_cols_list,
      file_name = dataset_details$file_name,
      header_df_list = header_df_list,
      overwrite = TRUE
    )
  }
  data_list$dataset_details <- retained_details
  invisible(data_list)
}
#' @noRd
save_project_datasets <- function(project, hard_reset = FALSE) {
  assert_setup_project(project)
  if (is_something(project$data)) {
    dataset_names <- check_datasets(project)
    if (hard_reset) {
      dataset_names <- names(project$datasets)
    }
    if (is_something(dataset_names)) {
      for (dataset_name in dataset_names) {
        warning_message <- "Failed to save dataset `{dataset_name}`."
        project <- tryCatch(expr = save_project_dataset(project, dataset_name),
                            error = function(e) {
                              cli_alert_warning(warning_message)
                              invisible(project)
                            })
      }
    }
  }
  invisible(project)
}
#' @noRd
generate_project_dataset <- function(project,
                                     dataset_name,
                                     transformation_type = "default",
                                     merge_form_name = "merged",
                                     filter_field = NULL,
                                     filter_choices = NULL,
                                     filter_list = NULL,
                                     filter_strict = TRUE,
                                     field_names = NULL,
                                     form_names = NULL,
                                     exclude_identifiers = FALSE,
                                     exclude_free_text = FALSE,
                                     date_handling = "none",
                                     labelled = TRUE,
                                     clean = TRUE,
                                     drop_blanks = FALSE,
                                     drop_missing_codes = FALSE,
                                     drop_others = NULL,
                                     include_metadata = TRUE,
                                     include_users = TRUE,
                                     include_records = TRUE,
                                     include_log = FALSE,
                                     annotate_from_log = TRUE,
                                     include_comments = FALSE) {
  assert_setup_project(project)
  id_col <- project$metadata$id_col
  if (missing(dataset_name)) {
    cli_abort("You must name generated datasets with 'dataset_name'")
  }
  assert_env_name(dataset_name, max.chars = 31L)
  assert_env_name(merge_form_name, max.chars = 31L)
  assert_choice(date_handling, choices = DATE_HANDLING_CHOICES)
  # add more asserts
  # function to do asserts here
  assert_choice(transformation_type, TRANFORMATION_TYPES)
  data_list <- NULL
  # add new fields and calculate them
  if (labelled != project$settings$labelled) {
    if (project$settings$labelled) {
      project <- labelled_to_raw_project(project)
    } else {
      project <- raw_to_labelled_project(project)
    }
  }
  data_list$metadata <- project$metadata
  data_list$data <- project$data
  data_list$redcap <- project$redcap
  data_list$links <- project$links
  record_sum <- project$record_summary
  data_list <- metadata_add_default_cols(data_list)
  data_list$data <- filter_data_list(
    data_list = data_list,
    field_names = field_names,
    form_names = form_names,
    filter_field = filter_field,
    filter_choices = filter_choices,
    filter_list = filter_list,
    filter_strict = filter_strict
  )
  #cache or store these to make it faster?
  if (transformation_type == "default") {
    data_list <- merge_non_repeating(
      data_list = data_list,
      merge_form_name = merge_form_name,
      merge_to_rep = TRUE
    )
  }
  if (transformation_type == "merge_non_repeating") {
    data_list <- merge_non_repeating(
      data_list = data_list,
      merge_form_name = merge_form_name,
      merge_to_rep = FALSE
    )
  }
  data_list$data <- deidentify_data_list(
    data_list = data_list,
    exclude_identifiers = exclude_identifiers,
    exclude_free_text = exclude_free_text,
    date_handling = date_handling
  )
  if (clean) {
    #include warning for if missing codes will prevent uploads
    if (is_something(data_list$metadata$missing_codes)) {
      if (drop_missing_codes) {
        if (labelled) {
          exclude_these <- data_list$metadata$missing_codes$name
        } else {
          exclude_these <- data_list$metadata$missing_codes$code
        }
        drop_others <- drop_others |>
          append(exclude_these) |>
          unique() |>
          drop_nas()
      }
    }
    data_list$data <- clean_data_list(
      data_list = data_list,
      labelled = labelled,
      drop_blanks = drop_blanks,
      drop_others = drop_others
    )
  }
  if (include_metadata) {
    if (is_something(data_list$metadata)) {
      data_list$metadata$forms <- annotate_forms(
        data_list = data_list,
        annotate = TRUE,
        drop_blanks = drop_blanks || !is.null(form_names)
      )
      data_list$metadata$fields <- annotate_fields(
        data_list = data_list,
        annotate = TRUE,
        drop_blanks = drop_blanks || !is.null(field_names)
      )
      data_list$metadata$choices <- annotate_choices(
        data_list = data_list,
        annotate = TRUE,
        drop_blanks = drop_blanks || !is.null(field_names)
      )
    }
  }
  records <- extract_project_records(data_list)[[1L]]
  record_sum <- record_sum[which(record_sum[[id_col]] %in% records), ]
  rownames(record_sum) <- NULL
  data_list$records <- record_sum
  if (data_list$redcap$has_log_access) {
    data_list$log <- extract_log(
      data_list = data_list,
      records = records,
      drop_details = project$settings$log_drop_details
    )
    if (include_comments) {
      data_list$comments <- generate_comment_table(redcap_log = data_list$log,
                                                   only_most_recent = TRUE)
    }
  } else {
    cli_alert_warning("You don't have log access so that data can't be used.")
    include_log <- FALSE
    include_comments <- FALSE
    annotate_from_log <- FALSE
  }
  if (include_records) {
    if (!is.null(records)) {
      data_list$records <- annotate_records(data_list = data_list,
                                            annotate = annotate_from_log)
    }
  }
  if (include_users) {
    if (data_list$redcap$has_user_access) {
      data_list$users <- annotate_users(
        data_list = data_list,
        records = records,
        annotate = annotate_from_log,
        drop_blanks = drop_blanks
      )
    } else {
      cli_alert_warning("You don't have user access that can't be included.")
      include_users <- FALSE
    }
  }
  data_list$redcap <- NULL
  if (is.null(filter_list)) {
    if (!is.null(filter_choices) && !is.null(filter_field)) {
      filter_list <- list(filter_choices)
      names(filter_list) <- filter_field
    } else {
      # warning
    }
  }
  # add headers-------
  form_names <- names(data_list$data)
  n_records <- length(records)
  data_list$dataset_details <- list(
    dataset_name = dataset_name,
    transformation_type = transformation_type,
    merge_form_name = merge_form_name,
    filter_list = filter_list,
    filter_strict = filter_strict,
    field_names = field_names,
    form_names = form_names,
    exclude_identifiers = exclude_identifiers,
    exclude_free_text = exclude_free_text,
    date_handling = date_handling,
    labelled = labelled,
    clean = clean,
    drop_blanks = drop_blanks,
    drop_missing_codes = drop_missing_codes,
    drop_others = drop_others,
    include_metadata = include_metadata,
    include_records = include_records,
    include_users = include_users,
    include_log = include_log,
    annotate_from_log = annotate_from_log,
    include_comments = include_comments,
    with_links = NULL,
    separate = NULL,
    use_csv = NULL,
    dir_other = NULL,
    file_name = NULL,
    file_path = NULL,
    n_records = n_records,
    last_save_time = NULL,
    final_form_tab_names = NULL
  )
  data_list$project_details <- list(
    project_name = project$project_name,
    dir_path = project$dir_path,
    n_records = nrow(project$record_summary),
    help = project$links$help,
    is_test = project$internals$is_test
  )
  invisible(data_list)
}
#' @noRd
load_project_dataset <- function(project, dataset_name) {
  assert_setup_project(project)
  assert_env_name(dataset_name, max.chars = 31L)
  if (!dataset_name %in% names(project$datasets)) {
    cli_abort("{dataset_name} is not included in the current project datasets")
  } # can save to R_objects for faster retrieval
  dataset_details <- project$datasets[[dataset_name]]#warning for other params?
  data_list <- generate_project_dataset( # filter for adding new params
    project = project,
    dataset_name = dataset_name,
    transformation_type = dataset_details$transformation_type, # verificaiton?
    merge_form_name = dataset_details$merge_form_name,
    filter_list = dataset_details$filter_list,
    filter_strict = dataset_details$filter_strict,
    field_names = dataset_details$field_names,
    form_names = dataset_details$form_names,
    exclude_identifiers = dataset_details$exclude_identifiers,
    exclude_free_text = dataset_details$exclude_free_text,
    date_handling = dataset_details$date_handling,
    labelled = dataset_details$labelled,
    clean = dataset_details$clean,
    drop_blanks = dataset_details$drop_blanks,
    drop_missing_codes = dataset_details$drop_missing_codes,
    drop_others = dataset_details$drop_others,
    include_metadata = dataset_details$include_metadata,
    include_records = dataset_details$include_records,
    include_users = dataset_details$include_users,
    include_log = dataset_details$include_log,
    annotate_from_log = dataset_details$annotate_from_log,
    include_comments = dataset_details$include_comments
  )
  invisible(data_list)
}
#' @noRd
add_project_dataset <- function(project,
                                dataset_name,
                                transformation_type = "default",
                                merge_form_name = "merged",
                                filter_field = NULL,
                                filter_choices = NULL,
                                filter_list = NULL,
                                filter_strict = TRUE,
                                field_names = NULL,
                                form_names = NULL,
                                exclude_identifiers = FALSE,
                                exclude_free_text = FALSE,
                                date_handling = "none",
                                labelled = TRUE,
                                clean = TRUE,
                                drop_blanks = FALSE,
                                drop_missing_codes = FALSE,
                                drop_others = NULL,
                                include_metadata = TRUE,
                                include_records = TRUE,
                                include_users = TRUE,
                                include_log = FALSE,
                                annotate_from_log = TRUE,
                                include_comments = TRUE,
                                with_links = TRUE,
                                separate = FALSE,
                                use_csv = FALSE,
                                dir_other = NULL,
                                file_name = NULL,
                                hard_reset = FALSE) {
  # sync_frequency ... project$settings$sync_frequency
  assert_env_name(dataset_name, max.chars = 31L)
  if (is.null(dir_other)) {
    dir_other <- file.path(project$dir_path, "output")
  }# else warn will not work on different computer
  if (is.null(file_name)) {
    file_name <- paste0(project$project_name, "_", dataset_name)
  }
  if (is.null(filter_list)) {
    if (!is.null(filter_choices) && !is.null(filter_field)) {
      filter_list <- list(filter_choices)
      names(filter_list) <- filter_field
    } else {
      # warning
    }
  }
  file_extension <- ifelse(use_csv, ".csv", ".xlsx")
  dataset_list_new <- list(
    dataset_name = dataset_name,
    transformation_type = transformation_type,
    merge_form_name = merge_form_name,
    filter_list = filter_list,
    filter_strict = filter_strict,
    field_names = field_names,
    form_names = form_names,
    exclude_identifiers = exclude_identifiers,
    exclude_free_text = exclude_free_text,
    date_handling = date_handling,
    labelled = labelled,
    clean = clean,
    drop_blanks = drop_blanks,
    drop_missing_codes = drop_missing_codes,
    drop_others = drop_others,
    include_metadata = include_metadata,
    include_records = include_records,
    include_users = include_users,
    include_log = include_log,
    annotate_from_log = annotate_from_log,
    include_comments = include_comments,
    with_links = with_links,
    separate = separate,
    use_csv = use_csv,
    dir_other = dir_other,
    file_name = file_name,
    file_path = file.path(dir_other, paste0(file_name, file_extension)),
    n_records = NULL,
    last_save_time = NULL,
    final_form_tab_names = NULL
  )
  dataset_list_old <- project$datasets[[dataset_name]]
  if (!is.null(dataset_list_old) && !hard_reset) {
    important_vars <- names(dataset_list_new) |>
      vec1_not_in_vec2(NOT_IMPORTANT_DATASET_NAMES)
    are_identical <- identical(dataset_list_new[important_vars],
                               dataset_list_old[important_vars])
    if (are_identical) {
      # optional message?
      return(invisible(project))
    }
  }
  project$datasets[[dataset_name]] <- dataset_list_new
  project$record_summary[[dataset_name]] <- FALSE
  invisible(project)
}
#' @noRd
deidentify_data_list <- function(data_list,
                                 exclude_identifiers = TRUE,
                                 exclude_free_text = FALSE,
                                 exclude_additional = NULL,
                                 date_handling = "none") {
  # assert_data_list contains data and metadata with forms and fields
  assert_choice(date_handling, choices = DATE_HANDLING_CHOICES)
  assert_logical(exclude_identifiers)
  assert_logical(exclude_free_text)
  data_forms <- data_list$data
  metadata <- data_list$metadata
  fields <- metadata$fields
  exclusions <- exclude_additional
  if (exclude_identifiers) {
    is_marked_id <- fields$identifier == "y"
    validation_type <- fields$text_validation_type_or_show_slider_number
    is_likely_id <- validation_type %in% REDCAP_MAYBE_IDS_STRICT
    initial_identifiers <- fields$field_name[which(is_marked_id | is_likely_id)]
    if (length(initial_identifiers) == 0L) {
      warning(
        "You have no identifiers marked in ",
        "`project$metadata$fields$identifier`. ",
        "You can set it in REDCap Project Setup and update ",
        "project OR define your idenitifiers in this functions ",
        "`identifiers` argument.",
        call. = TRUE,
        immediate. = TRUE
      )
    }
    exclusions <- exclusions |>
      append(initial_identifiers) |>
      unique()
  }
  bad_identifiers <- exclusions[which(!exclusions %in% fields$field_name)]
  if (length(bad_identifiers) > 0L) {
    cli_abort(
      paste0(
        "There is a bad identifier. See {.code ",
        "project$metadata$fields$field_name}: {toString(bad_identifiers)}"
      )
    )
  }
  id_cols <- metadata$form_key_cols |> unlist() |> unique()
  if (is_something(id_cols) && any(id_cols %in% exclusions)) {
    cli_abort("ID cols not allowed to be excluded for now... ")
  }
  if (exclude_free_text) {
    # get_identifier_fields here
    # drop free text only if there is no validation
    # make function for that ?external
    is_text_field <- fields$field_type == "text"
    is_note_field <- fields$field_type == "notes"
    not_id <- !fields$field_name %in% id_cols
    has_no_valid <- is.na(fields$text_validation_type_or_show_slider_number)
    # ? & fields$in_original_redcap
    is_free_text <- is_text_field & has_no_valid & not_id
    free_text_rows <- which(is_note_field | is_free_text)
    free_text_fields <- fields$field_name[free_text_rows]
    exclusions <- exclusions |>
      append(free_text_fields) |>
      unique()
  }
  if (is_something(data_forms)) {
    date_vector <- fields$field_name[which(fields$field_type_r == "date")]
    date_list <- Map(
      f = function(x, col_names) {
        date_vector[which(date_vector %in% col_names)]
      },
      names(data_forms),
      lapply(data_forms, colnames)
    )
    if (date_handling != "none") {
      if (date_handling == "exclude_dates") {
        exclusions <- exclusions |>
          append(date_vector) |>
          unique()
      }
      if (date_handling %in% c("random_shift_by_record",
                               "random_shift_by_project",
                               "zero_by_record",
                               "zero_by_project")) {
        number <- 90L # can set in options
        shift_range <- setdiff(-number:number, 0L)
        min_dates <- get_min_dates(data_list)
        if (date_handling == "random_shift_by_record") {
          min_dates$shift_amount <- sample(shift_range,
                                           size = nrow(min_dates),
                                           replace = TRUE)
        }
        if (date_handling == "random_shift_by_project") {
          min_dates$shift_amount <-
            sample(shift_range, size = 1L, replace = TRUE)
        }
        if (date_handling == "zero_by_record") {
          # should you edit fields to now be field_type_r integer?
          min_dates$shift_amount <- min_dates$date
        }
        if (date_handling == "zero_by_project") {
          # should you edit fields to now be field_type_r integer?
          min_dates$shift_amount <- min(min_dates$date)
        }
        for (form_name in names(date_list)) {
          field_record <- data_forms[[form_name]][[id_cols[1L]]]
          match_date_diff <- match(field_record, min_dates$record_id)
          difference <- min_dates$shift_amount[match_date_diff]
          for (field_name in date_list[[form_name]]) {
            field <- data_forms[[form_name]][[field_name]]
            data_forms[[form_name]][[field_name]] <-
              as.character(as.Date(field) - difference)
          }
        }
      }
      # if you have dates you already mutated no need to drop anymore
    }
    drop_list <- Map(function(x, col_names) {
      exclusions[which(exclusions %in% col_names)]
    }, names(data_forms), lapply(data_forms, colnames))
    drop_list <- drop_list[unlist(lapply(drop_list, length)) > 0L]
    for (form_name in names(drop_list)) {
      for (field_name in drop_list[[form_name]]) {
        data_forms[[form_name]][[field_name]] <- NULL
      }
    }
  }
  invisible(data_forms)
}
#' @noRd
filter_data_list <- function(data_list,
                             field_names = NULL,
                             form_names = NULL,
                             filter_field = NULL,
                             filter_choices = NULL,
                             filter_list = NULL,
                             filter_strict = TRUE) {
  if (is.null(field_names)) {
    field_names <- get_all_field_names(data_list)
  }
  # add assert/warning for accurate field_names
  if (is.null(form_names)) {
    form_names <- data_list$metadata$forms$form_name
  }
  the_rows <- which(!field_names %in% data_list$metadata$raw_structure_cols)
  field_names_minus <- field_names[the_rows]
  if (length(field_names_minus) > 0L) {
    form_names_minus <- field_to_form_names(project = data_list,
                                            field_names = field_names_minus,
                                            strict = TRUE)
    form_names <- vec1_in_vec2(form_names, form_names_minus)
  }
  # missing_filter ?
  out_list <- list()
  if (is.null(filter_list)) {
    if (!is.null(filter_field) && !is.null(filter_choices)) {
      filter_list <- list(filter_choices)
      names(filter_list) <- filter_field
    }
  } else if (!is.null(filter_field) || !is.null(filter_choices)) {
    warning_message <- "use `filter_list` or `filter_field` & `filter_choices`"
    cli_alert_warning(warning_message)
  }
  filter_field_names <- NULL
  if (!is.null(filter_list)) {
    filter_field_names <- filter_list |>
      names() |>
      drop_if("")
    # should be unique
    filter_form <- field_to_form_names(project = data_list,
                                       field_names = filter_field_names)
    if (length(filter_field_names) == 1L) {
      if (filter_field_names == data_list$metadata$id_col) {
        filter_form <- data_list$metadata$forms$form_name[1L]
        # RISKY? id_position like REDCapR, add to setup
      }
    }
    # should be length 1
    if (length(filter_form) > 1L) {
      cli_abort("You can only filter_list by multiple columns part of one form")
    }
    form_key_cols <- data_list$metadata$form_key_cols |>
      unlist() |>
      unique()
    is_key <- all(filter_field_names %in% form_key_cols)
  }
  # can use this to have repeats capture non-rep events
  for (form_name in form_names) {
    form <- data_list$data[[form_name]]
    if (is_something(form)) {
      row_index <- seq_len(nrow(form))
      if (!is.null(filter_list)) {
        row_logic <- NULL
        for (filter_field_name in filter_field_names) {
          filter_field_final <- filter_field_name
          filter_choices_final <- filter_list[[filter_field_name]]
          if (!filter_strict) {
            if (!is_key) {
              # need to account for instances
              if (form_name != filter_form) {
                filter_field_final <- data_list$metadata$id_col
                filtered_form <- data_list$data[[filter_form]]
                name_filtered <- filtered_form[[filter_field_name]]
                final_filtered <- filtered_form[[filter_field_final]]
                the_rows <- which(name_filtered %in% filter_choices_final)
                filter_choices_final <- unique(final_filtered[the_rows])
              }
            }
          }
          index_test <- rep(FALSE, nrow(form))
          if (filter_field_final %in% colnames(data_list$data[[form_name]])) {
            index_test <- data_list$data[[form_name]][[filter_field_final]] %in%
              filter_choices_final
          }
          if (is.null(row_logic)) {
            row_logic <- index_test
          }
          field_index <- which(names(filter_list) == filter_field_name)
          op_index <- (field_index - 1L)
          if (op_index <= length(filter_list)) {
            if (field_index != 1L) {
              is_and <- filter_list[[op_index]] == "and"
              if (is_and) {
                row_logic <- row_logic & index_test
              } else {
                row_logic <- row_logic | index_test
              }
            }
          }
        }
        if (is.null(row_logic))
          row_logic <- NA
        row_index <- which(row_logic)
      }
      field_names_adj <- c(field_names, filter_field_names)
      col_names <- colnames(form)[which(colnames(form) %in% field_names_adj)]
      if (length(row_index) > 0L && length(col_names) > 0L) {
        col_names <- colnames(form)[which(colnames(form) %in% unique(
          c(data_list$metadata$form_key_cols[[form_name]], field_names_adj)
        ))]
        out_list[[form_name]] <- form[row_index, col_names, drop = FALSE]
      }
    }
  }
  invisible(out_list)
}
#' @noRd
clean_data_list <- function(data_list,
                            labelled,
                            drop_blanks = TRUE,
                            drop_others = NULL) {
  # assert data list
  data_forms <- data_list$data
  metadata <- data_list$metadata
  for (form_name in names(data_forms)) {
    data_forms[[form_name]] <- clean_form(
      form = data_forms[[form_name]],
      fields = metadata$fields,
      labelled = labelled,
      drop_blanks = drop_blanks,
      drop_others = drop_others
    )
  }
  invisible(data_forms)
}
#' @noRd
merge_non_repeating <- function(data_list,
                                merge_form_name,
                                merge_to_rep = FALSE) {
  forms_transformed <- data_list$metadata$forms
  form_colnames <- colnames(data_list$metadata$forms)
  is_longitudinal <- "repeating_via_events" %in% form_colnames
  if (is_longitudinal) {
    new_order <- order(forms_transformed$repeating_via_events)
    forms_transformed <- forms_transformed[new_order, ]
  }
  new_order <- order(forms_transformed$repeating)
  forms_transformed <- forms_transformed[new_order, ]
  forms_transformed$form_name_remap <- forms_transformed$form_name
  forms_transformed$form_label_remap <- forms_transformed$form_label
  row_check <- !forms_transformed$repeating
  if (is_longitudinal) {
    row_check <- row_check & !forms_transformed$repeating_via_events
  }
  forms_transformed$form_name_remap[which(row_check)] <- merge_form_name
  merge_form_name_label <- merge_form_name # can captialize here
  if (merge_form_name %in% forms_transformed$form_name) {
    label_match <- which(forms_transformed$form_name == merge_form_name)
    merge_form_name_label <- forms_transformed$form_label[label_match]
  }
  non_rep_form_names <- forms_transformed$form_name[which(row_check)]
  forms_transformed$form_label_remap[which(row_check)] <- merge_form_name_label
  forms_transformed_original <- forms_transformed
  cols_to_keep <- c(
    "form_name_remap",
    "form_label_remap",
    "repeating",
    "repeating_via_events",
    "key_cols",
    "key_names",
    ""
  )
  included_cols <- which(cols_to_keep %in% colnames(forms_transformed))
  cols_to_keep <- cols_to_keep[included_cols]
  forms_transformed <- unique(forms_transformed[, cols_to_keep])
  name_cols <- which(colnames(forms_transformed) == "form_name_remap")
  colnames(forms_transformed)[name_cols] <- "form_name"
  label_cols <- which(colnames(forms_transformed) == "form_label_remap")
  colnames(forms_transformed)[label_cols] <- "form_label"
  forms_transformed$original_form_name <- forms_transformed$form_name |>
    lapply(function(form_name) {
      keep <- which(forms_transformed_original$form_name_remap == form_name)
      paste0(forms_transformed_original$form_name[keep], collapse = " | ")
    }) |>
    unlist() |>
    as.character()
  data_list$metadata$forms <- forms_transformed
  fields <- data_list$metadata$fields
  fields$original_form_name <- fields$form_name
  new_match <- match(fields$form_name, forms_transformed_original$form_name)
  fields$form_name <- forms_transformed_original$form_name_remap[new_match]
  new_order <- order(match(fields$form_name, forms_transformed$form_name))
  fields <- fields[new_order, ]
  # new function RosyUtils
  first <- 1L:which(colnames(fields) == "form_name")
  move <- which(colnames(fields) == "original_form_name")
  last <- which(colnames(fields) != "original_form_name")[-first]
  fields <- fields[, c(first, move, last)]
  data_list$metadata$fields <- fields
  data_list$metadata$choices <- fields_to_choices(fields)
  data_list$metadata$form_key_cols <- get_key_col_list(data_list = data_list)
  if (is_something(data_list$data)) {
    merge_form <- NULL
    keep_rows <- non_rep_form_names |>
      lapply(function(non_rep_form_name) {
        is_something(data_list$data[[non_rep_form_name]])
      }) |>
      unlist() |>
      which()
    non_rep_form_names <- non_rep_form_names[keep_rows]
    i <- 0L
    if (is_something(non_rep_form_names)) { # message here?
      for (non_rep_form_name in non_rep_form_names) {
        if (non_rep_form_name == non_rep_form_names[[1L]]) {
          merge_form <- data_list$data[[non_rep_form_name]]
        } else {
          to_be_merged <- data_list$data[[non_rep_form_name]]
          to_be_merged$arm_number <- NULL
          merge_form <- merge(
            x = merge_form,
            y = to_be_merged,
            by = data_list$metadata$id_col,
            all = TRUE,
            sort = FALSE,
            suffixes = c("", paste0("_merged_", i))
          )
        }
        data_list$data[[non_rep_form_name]] <- NULL
        i <- i + 1L
      }
      data_list$data[[merge_form_name]] <- merge_form
      other_forms <- setdiff(names(data_list$data), merge_form_name)
      if (merge_to_rep) {
        for (other_form in other_forms) {
          data_list$data[[other_form]] <- merge(
            x = data_list$data[[other_form]],
            y = merge_form,
            by = data_list$metadata$id_col,
            all.x = TRUE,
            sort = FALSE,
            suffixes = c("", "_merged")
          )
        }
      }
      data_list$data <- data_list$data[c(merge_form_name, other_forms)]
    }
  }
  data_list
}
#' @noRd
fields_to_choices <- function(fields) {
  fields <- fields[which(fields$field_type %in% REDCAP_FACTOR_FIELDS), ]
  fields <- fields[which(!is.na(fields$select_choices_or_calculations)), ]
  choices <- NULL
  for (i in seq_len(nrow(fields))) {
    field_name <- fields$field_name[i]
    form_name <- fields$form_name[i]
    field_label <- fields$field_label[i]
    field_type <- fields$field_type[i]
    selections <- split_choices(fields$select_choices_or_calculations[i])
    choices <- choices |> bind_rows(
      data.frame(
        form_name = form_name,
        field_name = field_name,
        field_type = field_type,
        field_label = ifelse(!is.na(field_label), field_label, field_name),
        code = selections$code,
        name = selections$name
      )
    )
  }
  choices$label <- paste(choices$form_name,
                         "-",
                         choices$field_label,
                         "-",
                         choices$name)
  rownames(choices) <- NULL
  choices
}
#' @noRd
add_labels_to_checkbox <- function(fields) {
  row_index <- which(fields$field_type == "checkbox_choice")
  x <- fields$field_name[row_index] |>
    strsplit("___") |>
    lapply(function(x) {
      x[[1L]]
    }) |>
    unlist()
  y <- fields$field_label[row_index]
  z <- paste0(fields$field_label[match(x, fields$field_name)], " - ", y)
  fields$field_label[row_index] <- z
  fields
}
#' @noRd
annotate_fields <- function(data_list,
                            annotate = TRUE,
                            drop_blanks = FALSE) {
  fields <- data_list$metadata$fields # [,colnames(data_list$metadata$fields)]
  if (drop_blanks) {
    keep_rows <- which(fields$field_name %in% get_all_field_names(data_list))
    fields <- fields[keep_rows, ]
    #fix transform dropping missing fields
  }
  if (nrow(fields) > 0L && is_something(data_list$data)) {
    if (annotate) {
      skimmed <- NULL
      for (form_name in drop_nas(unique(fields$form_name))) {
        col_names <- fields$field_name[which(fields$form_name == form_name)]
        form <- data_list$data[[form_name]]
        col_names <- col_names[which(col_names %in% colnames(form))]
        skimmed <- bind_rows(skimmed, skimr::skim(form[, col_names]))
      }
      field_names <- fields$field_name
      fields <-  merge(x = fields,
                       y = skimmed,
                       by.x = "field_name",
                       by.y = "skim_variable",
                       all = TRUE)
      fields <- field_names |>
        lapply(function(x) {
          fields[which(fields$field_name == x), ]
        }) |>
        bind_rows()
    }
  }
  fields
}
#' @noRd
annotate_forms <- function(data_list,
                           annotate = TRUE,
                           drop_blanks = FALSE) {
  forms <- data_list$metadata$forms
  if (drop_blanks) {
    forms <- forms[which(forms$form_name %in% names(data_list$data)), ]
  }
  if (nrow(forms) > 0L) {
    # add metadata info like n fields
    if (annotate) {
      var_list <- forms$form_name |> lapply(function(form_name) {
        paste0(form_name, "_complete")
      })
      if ("original_form_name" %in% colnames(forms)) {
        var_list <- forms$original_form_name |>
          strsplit(" [|] ") |>
          lapply(function(form_name) {
            paste0(form_name, "_complete")
          })
      }
      names(var_list) <- forms$form_name
      for (status in c("Incomplete", "Unverified", "Complete")) {
        forms[[tolower(status)]] <- forms$form_name |>
          lapply(function(form_name) {
            var_list[[form_name]] |>
              lapply(function(var_name) {
                (data_list$data[[form_name]][[var_name]] == status) |>
                  which() |>
                  length()
              }) |>
              unlist() |>
              paste0(collapse = " | ")
          }) |>
          unlist()
      }
    }
  }
  forms
}
#' @noRd
annotate_choices <- function(data_list,
                             annotate = TRUE,
                             drop_blanks = FALSE) {
  choices <- data_list$metadata$choices
  if (drop_blanks) {
    keep_rows <- which(choices$field_name %in% get_all_field_names(data_list))
    choices <- choices[keep_rows, ]
  }
  #used to have more code
  if (annotate) {
    choices$n <- seq_len(nrow(choices)) |>
      lapply(function(i) {
        form <- data_list$data[[choices$form_name[i]]]
        if (is.null(form)) {
          return(0L)
        }
        if (nrow(form) == 0L) {
          return(0L)
        }
        the_col <- choices$field_name[i]
        if (!the_col %in% colnames(form)) {
          return(0L)
        }
        sum(form[, the_col] == choices$name[i], na.rm = TRUE)
      }) |>
      unlist()
    choices$n_total <- seq_len(nrow(choices)) |>
      lapply(function(i) {
        form <- data_list$data[[choices$form_name[i]]]
        if (is.null(form)) {
          return(0L)
        }
        if (nrow(form) == 0L) {
          return(0L)
        }
        the_col <- choices$field_name[i]
        if (!the_col %in% colnames(form)) {
          return(0L)
        }
        sum(!is.na(form[, the_col]), na.rm = TRUE)
      }) |>
      unlist()
    choices$perc <- round((choices$n / choices$n_total), 4L)
    choices$perc_text <- (choices$perc * 100L) |>
      round(1L) |>
      paste0("%")
  }
  choices
}
#' @noRd
annotate_records <- function(data_list, annotate = TRUE) {
  id_col <- data_list$metadata$id_col
  records <- extract_project_records(data_list)[[1L]]
  the_rows <- which(data_list$records[[id_col]] %in% records)
  all_records <- data_list$records[the_rows, ]
  redcap_log <- extract_log(data_list = data_list,
                            records = records,
                            drop_details = TRUE)
  # convert to date level?
  if (!is_something(all_records) || !is_something(redcap_log)) {
    return(all_records)
  }
  if (annotate) {
    if (is_something(redcap_log)) {
      redcap_log <- redcap_log[, c("timestamp", "username", "record")]
      has_users <- is_something(data_list$redcap$users)
      if (has_users) {
        users <- data_list$redcap$users
        users <- users[, c("username", "firstname", "lastname", "email")]
        redcap_log <- merge(redcap_log, users, by = "username", all.x = TRUE)
      }
      cool_list <- split(redcap_log, redcap_log$record)
      if (length(cool_list) > 1L) {
        cool_df <- seq_along(cool_list) |> lapply(function(i) {
          x <- cool_list[[i]]
          the_last <- first(x)
          the_first <- last(x)
          out_df <- data.frame(
            record = the_first$record,
            first_timestamp = the_first$timestamp,
            last_timestamp = the_last$timestamp,
            unique_users = length_unique(x$username)
          )
          if (has_users) {
            out_df$last_username <- the_last$username
            out_df$last_user <- paste(the_last$firstname, the_last$lastname)
          }
          out_df
        }) |>
          bind_rows()
        colnames(cool_df)[1L] <- id_col
        all_records <- merge(all_records, cool_df, by = id_col, all.x = TRUE)
      }
    }
  }
  invisible(all_records)
}
#' @noRd
clean_form <- function(form,
                       fields,
                       labelled,
                       drop_blanks = TRUE,
                       drop_others = NULL) {
  name_code <- ifelse(labelled, 2L, 1L)
  for (field_name in colnames(form)) {
    if (field_name %in% fields$field_name) {
      x_row <- which(fields$field_name == field_name)
      x_units <- NULL
      if (!is.na(fields$units[x_row])) {
        x_units <- fields$units[x_row]
      }
      x_class <- fields$field_type_r[x_row][[1L]]
      label <- ifelse(is.na(fields$field_label[x_row]),
                      field_name,
                      fields$field_label[x_row])[[1L]]
      x_levels <- NULL
      if (!is.na(x_class)) {
        if (is_something(drop_others)) {
          erase_rows <- which(form[[field_name]] %in% drop_others)
          form[[field_name]][erase_rows] <- NA
        }
        if (x_class == "factor") {
          select_choices <- fields$select_choices_or_calculations[x_row]
          if (!is.na(select_choices)) {
            x_levels <- split_choices(select_choices)[[name_code]]
          } else {
            x_levels <- unique(form[[field_name]]) |> drop_nas()
          }
          if (anyDuplicated(x_levels) > 0L) {
            duplicate_levels <- x_levels |>
              duplicated() |>
              which()
            warning(
              "You have a variable (",
              field_name,
              ") with dupplicate names (",
              toString(x_levels[duplicate_levels]),
              "). This is not great but for this proccess they will be merged",
              "and treated as identical responses."
            )
            x_levels <- unique(x_levels)
          }
          if (drop_blanks) {
            x_levels <-
              x_levels[which(x_levels %in% unique(form[[field_name]]))]
          }
          if (is_something(drop_others)) {
            x_levels <- x_levels[which(!x_levels %in% drop_others)]
          }
        }
        if (x_class == "integer") {
          #warning about missing codes?
          form[[field_name]] <- as.integer(form[[field_name]])
        }
        if (x_class == "numeric") {
          #warning about missing codes?
          form[[field_name]] <- as.numeric(form[[field_name]])
        }
        form
      }
      form[[field_name]] <- clean_column_for_table(
        field = form[[field_name]],
        class = x_class,
        label = label,
        units = x_units,
        levels = x_levels
      )
    }
  }
  form
}
#' @noRd
clean_column_for_table <- function(field,
                                   class = NULL,
                                   label = NULL,
                                   units = NULL,
                                   levels = NULL) {
  if (is_something(class)) {
    if (class == "integer") {
      field <- as.integer(field)
    }
    if (class == "factor") {
      field <- factor(x = field, levels = levels, ordered = TRUE)
    }
    if (class == "numeric") {
      field <- as.numeric(field)
    }
  }
  if (is_something(label)) {
    attr(field, "label") <- label
  }
  if (is_something(units)) {
    attr(field, "units") <- units
  }
  field
}
#' @noRd
NOT_IMPORTANT_DATASET_NAMES <- c("n_records",
                                 "last_save_time",
                                 "final_form_tab_names")
#' @noRd
data_list_to_save <- function(data_list) {
  to_save_list <- NULL
  for (form_name in names(data_list$data)) {
    to_save_list[[form_name]] <- data_list$data[[form_name]]
  }
  if (data_list$dataset_details$include_metadata) {
    metadata_form_names <- c("forms", "fields", "choices", "missing_codes")
    metadata_names <- data_list$metadata[metadata_form_names] |>
      process_df_list(silent = TRUE) |>
      names()
    metadata_names_alt <- metadata_names
    if (any(metadata_names %in% names(data_list$data))) {
      metadata_names_alt <- paste0("redcap_", metadata_names)
      #check for conflicts
    }
    for (i in seq_along(metadata_names)) {
      to_save_list[[metadata_names_alt[i]]] <-
        data_list$metadata[[metadata_names[i]]]
    }
  }
  if (data_list$dataset_details$include_users) {
    users_name <- "users"
    users_name_alt <- users_name
    if (any(users_name %in% names(data_list$data))) {
      users_name_alt <- paste0("redcap_", users_name)
      #check for conflicts
    }
    if ("users" %in% names(data_list)) {
      to_save_list[[users_name_alt]] <- data_list$users
    }
  }
  if (data_list$dataset_details$include_records) {
    records_name <- "records"
    records_name_alt <- records_name
    if (any(records_name %in% names(data_list$data))) {
      records_name_alt <- paste0("redcap_", records_name)
      #check for conflicts
    }
    if ("records" %in% names(data_list)) {
      to_save_list[[records_name_alt]] <- data_list$records
    }
  }
  if (data_list$dataset_details$include_log) {
    log_name <- "log"
    log_name_alt <- log_name
    if (any(log_name %in% names(data_list$data))) {
      log_name_alt <- paste0("redcap_", log_name)
      #check for conflicts
    }
    if ("log" %in% names(data_list)) {
      to_save_list[[log_name_alt]] <- data_list$log
    }
  }
  if (data_list$dataset_details$include_comments) {
    comments_name <- "comments"
    comments_name_alt <- comments_name
    if (any(comments_name %in% names(data_list$data))) {
      comments_name_alt <- paste0("redcap_", comments_name)
      #check for conflicts
    }
    if ("comments" %in% names(data_list)) {
      to_save_list[[comments_name_alt]] <- data_list$comments
    }
  }
  to_save_list
}
#' @noRd
TRANFORMATION_TYPES <- c("default", "none", "merge_non_repeating")
#' @noRd
metadata_add_default_cols <- function(data_list) {
  fields <- data_list$metadata$fields
  fields <- fields[which(fields$field_type != "descriptive"), ]
  fields <- add_labels_to_checkbox(fields)
  fields <- fields[which(fields$field_type != "checkbox"), ]
  fields$field_label[which(is.na(fields$field_label))] <-
    fields$field_name[which(is.na(fields$field_label))]
  fields <- unique(fields$form_name) |>
    lapply(function(x) {
      fields[which(fields$form_name == x), ]
    }) |>
    bind_rows()
  fields$field_type_r <- field_types_to_r(fields)
  fields$in_original_redcap <- TRUE
  fields$original_form_name <- fields$form_name
  if (!"units" %in% colnames(fields)) {
    fields$units <- NA
  }
  if (!"field_label_short" %in% colnames(fields)) {
    fields$field_label_short <- fields$field_label
  }
  data_list$metadata$fields <- fields
  data_list
}
#' @noRd
field_types_to_r <- function(fields) {
  field_types_r <- rep("character", nrow(fields))
  field_types <- fields$field_type
  field_validations <- fields$text_validation_type_or_show_slider_number
  field_types_r[which(field_types %in% REDCAP_FACTOR_FIELDS)] <- "factor"
  is_text <- field_types == "text"
  is_date <- is_text & field_validations %in% REDCAP_TEXT_DATE_FIELDS
  is_datetime <- is_text & field_validations %in% REDCAP_TEXT_DATETIME_FIELDS
  is_int <- is_text & field_validations %in% REDCAP_INTEGER_FIELDS
  is_num <- is_text & field_validations %in% REDCAP_NUMERIC_FIELDS
  field_types_r[which(is_date)] <- "date"
  field_types_r[which(is_datetime)] <- "datetime"
  field_types_r[which(is_int)] <- "integer"
  field_types_r[which(is_num)] <- "numeric"
  field_types_r
}
#' @noRd
REDCAP_FACTOR_FIELDS <- c("radio",
                          "yesno",
                          "dropdown",
                          "checkbox_choice",
                          "truefalse")
#' @noRd
REDCAP_LOGICAL_FIELDS <- c("yesno", "checkbox_choice", "truefalse") # tbd
#' @noRd
REDCAP_TEXT_DATE_FIELDS <- c("date_mdy", "date_ymd", "date_dmy")
#' @noRd
REDCAP_TEXT_DATETIME_FIELDS <- c("datetime_dmy", "datetime_seconds_ymd")
#' @noRd
REDCAP_INTEGER_FIELDS <- "integer"
#' @noRd
REDCAP_NUMERIC_FIELDS <- "number"
#' @noRd
REDCAP_MAYBE_IDS_STRICT <- c("email", "phone", "vmrn", "zipcode")
#' @noRd
REDCAP_MAYBE_IDS_SUPER_STRICT <- c("date_dmy",
                                   "date_mdy",
                                   "date_ymd",
                                   "datetime_dmy",
                                   "datetime_mdy",
                                   "datetime_seconds_dmy",
                                   "datetime_seconds_mdy",
                                   "datetime_seconds_ymd",
                                   "datetime_ymd",
                                   "email",
                                   "phone",
                                   "vmrn",
                                   "zipcode")
#' @noRd
FIELD_TYPES_NOT_IN_DATA <- c("descriptive", "checkbox")
#' @noRd
clear_project_datasets <- function(project, dataset_names = NULL) {
  assert_setup_project(project)
  all_dataset_names <- names(project$datasets)
  if (is.null(dataset_names)) {
    dataset_names <- all_dataset_names
  }
  for (dataset_name in dataset_names) {
    project$datasets[[dataset_name]] <- NULL
    project$record_summary[[dataset_name]] <- NULL
  }
  cli_alert_success("Cleared project datasets!")
  invisible(project)
}
#' @noRd
extract_log <- function(data_list, records, drop_details = FALSE) {
  redcap_log <- data_list$redcap$log
  redcap_log <- redcap_log[which(!is.na(redcap_log$username)), ]
  redcap_log <- redcap_log[which(!is.na(redcap_log$record)), ]
  if (!missing(records)) {
    if (!is.null(records)) {
      redcap_log <- redcap_log[which(redcap_log$record %in% records), ]
    }
  }
  if (drop_details) {
    redcap_log$details <- NULL
  }
  redcap_log
}
#' @noRd
annotate_users <- function(data_list,
                           records = NULL,
                           annotate = TRUE,
                           drop_blanks = FALSE) {
  if (is.null(records)) {
    id_col <- data_list$metadata$id_col
    records <- data_list$records[[id_col]]
  }
  redcap_log <- extract_log(data_list, records, drop_details = TRUE)
  # role_label not inculded now
  users_df <- data_list$redcap$users |>
    select(c("username", "email", "firstname", "lastname"))
  user_groups <- NULL
  if (is_something(redcap_log)) {
    user_groups <- split(redcap_log, redcap_log$username)
  }
  names_in_log <- names(user_groups)
  if (!is_something(user_groups) || length(names_in_log) == 0L) {
    return(users_df)
  }
  if (annotate) {
    only_in_log <- vec1_not_in_vec2(names_in_log, users_df$username)
    if (length(only_in_log) > 0L) {
      users_df <- users_df |>
        bind_rows(data.frame(username = only_in_log))
    }
    users_df$first_timestamp <- NA
    users_df$last_timestamp <- NA
    users_df$unique_records_n <- NA
    for (user_group_name in names_in_log) {
      the_row <- match(user_group_name, users_df$username)
      group <- user_groups[[user_group_name]]
      users_df$first_timestamp[the_row] <- last(group$timestamp)
      users_df$last_timestamp[the_row] <- group$timestamp[[1L]]
      users_df$unique_records_n[the_row] <- length_unique(group$record)
    }
  }
  if (drop_blanks) {
    keep_rows <- which(users_df$username %in% names_in_log)
    users_df <- users_df[keep_rows, ]
    keep_rows <- drop_nas(match(users_df$username, names_in_log))
    user_groups <- user_groups[keep_rows]
  }
  users_df
}
#' @noRd
get_dataset_records <- function(project, dataset_name) {
  id_col <- project$metadata$id_col
  if (missing(dataset_name)) {
    return(project$record_summary[[id_col]])
  }
  dataset_details <- project$datasets[[dataset_name]]
  data_list <- generate_project_dataset(
    project = project,
    dataset_name = dataset_name,
    transformation_type = dataset_details$transformation_type,
    filter_list = dataset_details$filter_list,
    filter_strict = FALSE,
    form_names = dataset_details$form_names,
    field_names = project$metadata$id_col,
    exclude_identifiers = FALSE,
    exclude_free_text = FALSE,
    clean = FALSE,
    include_metadata = FALSE,
    include_records = FALSE,
    include_users = FALSE,
    include_log = FALSE,
    annotate_from_log = FALSE,
    include_comments = FALSE
  )
  data_list$records[[id_col]]
}
#' @noRd
dataset_records_due <- function(project, dataset_name) {
  dataset_names <- names(project$datasets)
  if (!dataset_name %in% dataset_names) {
    cli_alert_danger("{dataset_name} not included in current dataset_names!")
    return(FALSE)
  }
  dataset_details <- project$datasets[[dataset_name]]
  id_col <- project$metadata$id_col
  if (is.null(dataset_details$last_save_time)) {
    return(TRUE)
  }
  if (!file.exists(dataset_details$file_path) && !dataset_details$separate) {
    # can't do this for separate = TRUE unless more code is written
    return(TRUE)
  }
  old_rec_rows <- which(project$record_summary[[dataset_name]])
  old_records <- project$record_summary[[id_col]][old_rec_rows]
  records <- get_dataset_records(project = project, dataset_name = dataset_name)
  if (!identical(records, old_records)) {
    return(TRUE)
  }
  record_rows <- which(project$record_summary[[id_col]] %in% records)
  record_cols <- c(id_col, "last_api_call", "was_saved", dataset_name)
  relevant_records <- project$record_summary[record_rows, record_cols]
  if (!all(relevant_records$was_saved)) {
    return(TRUE)
  }
  if (any(relevant_records$last_api_call > dataset_details$last_save_time)) {
    return(TRUE)
  }
  FALSE
}
#' @noRd
check_datasets <- function(project, dataset_names) {
  if (missing(dataset_names)) {
    dataset_names <- names(project$datasets)
  }
  needs_refresh <- NULL
  if (is.null(dataset_names)) {
    cli_alert_info("No datasets. Use `project$add_dataset()`!")
  }
  # need_to_check <- any(project$record_summary$last_api_call >
  # dataset_details$last_save_time)
  for (dataset_name in dataset_names) {
    test_dataset <- dataset_records_due(project = project,
                                        dataset_name = dataset_name)
    if (test_dataset) {
      needs_refresh <- append(needs_refresh, dataset_name)
    }
  }
  if (is.null(needs_refresh)) {
    cli_alert_success("Refresh of datasets not needed!")
  }
  needs_refresh
}
#' @noRd
add_default_datasets <- function(project,
                                 exclude_identifiers = TRUE,
                                 exclude_free_text = TRUE,
                                 date_handling = "none",
                                 use_csv = FALSE) {
  with_links <- FALSE
  if (is_something(project$data)) {
    with_links <- nrow(project$record_summary) <= 3000L
  }
  assert_logical(exclude_identifiers)
  assert_logical(exclude_free_text)
  dataset_name <- "REDCapSync_raw"
  project <- add_project_dataset(
    project = project,
    dataset_name = dataset_name,
    transformation_type = "none",
    filter_list = NULL,
    exclude_identifiers = exclude_identifiers,
    exclude_free_text = exclude_free_text,
    date_handling = "none",
    labelled = FALSE,
    clean = FALSE,
    drop_blanks = FALSE,
    drop_others = NULL,
    include_metadata = TRUE,
    include_records = FALSE,
    include_users = TRUE,
    include_log = FALSE,
    annotate_from_log = FALSE,
    include_comments = TRUE,
    with_links = with_links,
    separate = TRUE,
    use_csv = use_csv,
    dir_other = file.path(project$dir_path, "REDCap", project$project_name),
    file_name = project$project_name
  )
  dataset_name <- "REDCapSync"
  project <- add_project_dataset(
    project = project,
    dataset_name = dataset_name,
    transformation_type = "default",
    filter_list = NULL,
    exclude_identifiers = exclude_identifiers,
    exclude_free_text = exclude_free_text,
    date_handling = "none",
    labelled = TRUE,
    clean = TRUE,
    drop_blanks = FALSE,
    drop_others = NULL,
    include_metadata = TRUE,
    include_records = TRUE,
    include_users = TRUE,
    include_log = FALSE,
    annotate_from_log = TRUE,
    include_comments = TRUE,
    with_links = with_links,
    separate = FALSE,
    use_csv = use_csv,
    dir_other = file.path(project$dir_path, "output"),
    file_name = paste0(project$project_name, "_", dataset_name)
  )
  invisible(project)
}
#' @noRd
DATE_HANDLING_CHOICES <- c(
  "none",
  "exclude_dates",
  "random_shift_by_record",
  "random_shift_by_project",
  "zero_by_record",
  "zero_by_project"
)
#' @noRd
read_dataset_from_file <- function(project, dataset_name, file_path) {
  if (!missing(dataset_name)) {
    if (!missing(file_path)) {
      cli_alert_warning("`file_path` only needed if dataset_name not provided.")
      cli_alert_info("Using `file_path` from `dataset_name`...")
    }
    if (!dataset_name %in% names(project$dataset)) {
      cli_abort("`dataset_name` is not one of `project$dataset`")
    }
    file_path <- project$dataset[[dataset_name]]$file_path
  } else {
    # add data_updates check
    if (!endsWith(file_path, ".xlsx")) {
      cli_abort("File type must be '.xlsx' --> {.file {file_path}}")
    }
    if (!file.exists(file_path)) {
      cli_abort("Path does not exist --> {.file {file_path}}")
    }
  }
  data_list <- excel_to_list(file_path)
  the_row <- which(data_list$dataset_details$paramater == "raw_form_names")
  form_names <- data_list$dataset_details$value[the_row] |>
    strsplit(" [|] ") |>
    unlist()
  drop_names <- names(data_list) |> setdiff(form_names)
  # use dataset_details to drop non-uploadcompatibles and warn
  for (drop_sheet in drop_names) {
    data_list[[drop_sheet]] <- NULL
  }
  if (length(data_list) == 0L) {
    message("nothing to return")
    return(NULL)
  }
  #check data not already there
  data_list
}

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# add_default_datasets (Internal)
test_that("add_default_datasets works!", {
  project <- mock_test_project()$.internal
  project <- clear_project_datasets(project)
  datasets <- c("REDCapSync", "REDCapSync_raw")
  expect_all_false(datasets %in% names(project$datasets))
  expect_all_false(datasets %in% names(project$record_summary))
  expect_no_error({
    project <- add_default_datasets(
      project = project,
      exclude_identifiers = TRUE,
      exclude_free_text = TRUE,
      date_handling = "none"
    )
  })
  expect_all_true(datasets %in% names(project$datasets))
  expect_all_true(datasets %in% names(project$record_summary))
  raw <- project$datasets$REDCapSync_raw
  main <- project$datasets$REDCapSync
  # check expected settings for raw datasets
  expect_identical(raw$transformation_type, "none")
  expect_false(raw$include_records)
  expect_true(raw$separate)
  # check expected settings for main datasets
  expect_identical(main$transformation_type, "default")
  expect_true(main$include_records)
  expect_false(main$separate)
})
# add_labels_to_checkbox (Internal)
test_that("add_labels_to_checkbox works!", {
})
# add_project_dataset (Internal)
test_that("add_project_dataset works!", {
  project <- mock_test_project()$.internal
  dataset_name <- "MY_SUMMARY_TEST"
  project <- add_project_dataset(
    project = project,
    dataset_name = dataset_name,
    transformation_type = "default",
    drop_blanks = FALSE,
    drop_missing_codes = TRUE,
    include_metadata = TRUE,
    include_records = FALSE,
    include_users = TRUE,
    include_log = FALSE,
    annotate_from_log = FALSE,
    with_links = FALSE,
    separate = FALSE,
    use_csv = FALSE
  )
  expect_true(dataset_name %in% names(project$datasets))
  datasets <- project$datasets[[dataset_name]]
  expect_identical(datasets$dataset_name, dataset_name)
  expect_identical(datasets$transformation_type, "default")
  expect_true(datasets$include_metadata)
  expect_false(datasets$include_records)
  expect_true(datasets$include_users)
  expect_false(datasets$with_links)
  expect_false(datasets$separate)
  expect_type(datasets$file_name, type = "character")
})
# annotate_choices (Internal)
test_that("annotate_choices works!", {
})
# annotate_fields (Internal)
test_that("annotate_fields works!", {
})
# annotate_forms (Internal)
test_that("annotate_forms works!", {
})
# annotate_records (Internal)
test_that("annotate_records works!", {
})
# annotate_users (Internal)
test_that("annotate_users works!", {
  data_list <- mock_test_project()$.internal
  data_list$redcap$users <- data_list$redcap$users |>
    bind_rows(data.frame(username = "person_not_in_log",
                         stringsAsFactors = FALSE))
  data_list$redcap$log <- data_list$redcap$log |>
    bind_rows(data.frame(username = "person_not_in_users",
                         record = "40",
                         timestamp = as.character(Sys.time()),
                         stringsAsFactors = FALSE))
  users_default <- annotate_users(data_list)
  users_drop_blank <- annotate_users(data_list, drop_blanks = TRUE)
  expect_data_frame(users_default, nrows = 3L)
  expect_data_frame(users_drop_blank, nrows = 2L)
  all_users <- c("u1230", "person_not_in_log", "person_not_in_users")
  expect_in(users_default$username, all_users)
  log_users <- c("u1230", "person_not_in_users")
  expect_in(users_drop_blank$username, all_users)
})
# check_datasets (Internal)
test_that("check_datasets works!", {
})
# clean_column_for_table (Internal)
test_that("clean_column_for_table works!", {
})
# clean_data_list (Internal)
test_that("clean_data_list works!", {
})
# clean_form (Internal)
test_that("clean_form works!", {
  project <- mock_test_project()$.internal
  form <- project$data$text
  project <- metadata_add_default_cols(project)
  fields <- project$metadata$fields
  form <- clean_form(form = form, labelled = TRUE, fields = fields)
  expect_identical(attr(form$var_text_only, "label"), "Text Only")
  expect_identical(attr(form$var_birth_date, "label"), "Birth Date")
  row_to_change <- which(fields$field_name == "var_multi_radio")
  new_var <- "1, 1 | 2, 2 | 3, 3 | 4, 4 | 5, 5 | 6, 5"
  fields$select_choices_or_calculations[row_to_change] <- new_var
  form <- project$data$other
  expect_warning(clean_form(
    form = form,
    labelled = TRUE,
    fields = fields
  ),
  "dupplicate names")
})
# clear_project_datasets (Internal)
test_that("clear_project_datasets works!", {
  project <- mock_test_project()$.internal
  datasets <- c("REDCapSync", "REDCapSync_raw")
  expect_all_true(datasets %in% names(project$datasets))
  expect_all_true(datasets %in% names(project$record_summary))
  project <- clear_project_datasets(project)
  expect_all_false(datasets %in% names(project$datasets))
  expect_all_false(datasets %in% names(project$record_summary))
})
# data_list_to_save (Internal)
test_that("data_list_to_save works!", {
})
# deidentify_data_list (Internal)
test_that("deidentify_data_list works!", {
  project <- mock_test_project()$.internal
  data_list <- merge_non_repeating(TEST_CLASSIC, "merged")
  data_list$metadata$fields$field_type_r <- NA
  data_list$metadata$fields$in_original_redcap <- NA
  id_cols <- data_list$metadata$form_key_cols |>
    unlist() |>
    unique()
  fields <- data_list$metadata$fields
  is_identifier <- fields$identifier == "y"
  fields$validation_type <- fields$text_validation_type_or_show_slider_number
  is_likely_identifier <- fields$validation_type %in% REDCAP_MAYBE_IDS_STRICT
  identifier_rows <- which(is_identifier | is_likely_identifier)
  initial_identifiers <- fields$field_name[identifier_rows]
  is_notes <- fields$field_type == "notes"
  is_text <- fields$field_type == "text"
  has_validation <- is.na(fields$text_validation_type_or_show_slider_number)
  not_id <- !fields$field_name %in% id_cols
  is_free_text <- is_notes | (is_text & has_validation) & not_id
  free_text_fields <- fields$field_name[which(is_free_text)]
  expect_all_true(initial_identifiers %in% colnames(data_list$data$merged))
  no_ids <- deidentify_data_list(data_list = data_list,
                                 exclude_identifiers = TRUE,
                                 exclude_free_text = FALSE)
  expect_all_false(initial_identifiers %in% colnames(no_ids$data$merged))
  expect_all_true(free_text_fields %in% colnames(data_list$data$merged))
  no_free_text <- deidentify_data_list(data_list = data_list,
                                       exclude_identifiers = FALSE,
                                       exclude_free_text = TRUE)
  expect_all_false(initial_identifiers %in% colnames(no_free_text$data$merged))
  keep_rows <-
    which(!data_list$metadata$fields$field_name %in% initial_identifiers)
  data_list$metadata$fields <- data_list$metadata$fields[keep_rows, ]
  keep_cols <- which(!colnames(data_list$data$merged) %in% initial_identifiers)
  data_list$data$merged <- data_list$data$merged[, keep_cols]
  expect_warning(
    deidentify_data_list(
      data_list = data_list,
      exclude_identifiers = TRUE,
      exclude_free_text = FALSE
    ),
    "You have no identifiers marked"
  )
})
test_that("deidentify_data_list works2", {
  project <- mock_test_project()$.internal
  data_list <- merge_non_repeating(TEST_CLASSIC, "merged")
  data_list <- metadata_add_default_cols(data_list)
  fields <- data_list$metadata$fields
  identifiers <- fields$field_name[which(fields$identifier == "y")]
  merged <- data_list$data$merged
  expected_cols <- c("var_text_date_dmy",
                     "var_text_date_mdy",
                     "var_text_date_ymd")
  expect_in(colnames(merged), fields$field_name)
  expect_error(deidentify_data_list(data_list = data_list, date_handling = "1"))
  # 'none'
  merged_none <- deidentify_data_list(
    data_list = data_list,
    exclude_identifiers = TRUE,
    date_handling = "none"
  )$merged
  expect_all_false(identifiers %in% colnames(merged_none))
  # 'exclude_dates'
  excluded_dates <- deidentify_data_list(data_list = data_list,
                                         date_handling = "exclude_dates")$merged
  expect_all_false(expected_cols %in% colnames(excluded_dates))
  # 'random_shift_by_record'
  random_shift <- deidentify_data_list(data_list = data_list,
                                       exclude_identifiers = FALSE,
                                       date_handling = "random_shift_by_record")
  random_shift <- random_shift$merged
  expect_all_true(expected_cols %in% colnames(random_shift))
  expect_all_false(random_shift$var_text_date_dmy == merged$var_text_date_dmy)
  time_check1 <- as.Date(merged$var_text_date_dmy) -
    as.Date(merged$var_birth_date)
  time_check2 <- as.Date(random_shift$var_text_date_dmy) -
    as.Date(random_shift$var_birth_date)
  expect_all_true(time_check1 == time_check2) #math is same
  as.Date(merged$var_birth_date) - as.Date(random_shift$var_birth_date)
  # 'random_shift_by_project'
  # merged_random_shift_by_project <- deidentify_data_list(
  #   data_list = data_list,date_handling = "random_shift_by_project")
  # # 'zero_by_record'
  # merged_zero_by_record <- deidentify_data_list(
  #   data_list = data_list,date_handling = "zero_by_record")
  # # 'zero_by_project'
  # merged_zero_by_project <- deidentify_data_list(
  #   data_list = data_list,date_handling = "zero_by_project")
})
# field_types_to_r (Internal)
test_that("field_types_to_r works!", {
})
# fields_to_choices (Internal)
test_that("fields_to_choices works!", {
  project <- mock_test_project()$.internal
  fields <- project$metadata$fields
  fields <- fields[0L, ] |> dplyr::bind_rows(
    data.frame(
      field_name = c("sex", "yes_no1", "yes_no2", "race", "location"),
      form_name = "form_one",
      field_type = c("radio", "yesno", "yesno", "dropdown", "checkbox"),
      select_choices_or_calculations = c(
        "0, Female | 1, Male",
        # 2
        "0, No | 1, Yes",
        # 2
        "0,No | 1,Yes",
        # 2
        paste0(
          "0, American Indian/Alaska Native",
          " | ",
          "1, Asian",
          " | ",
          "2, Native Hawaiian or Other Pacific Islander",
          " | ",
          "3, Black or African American",
          " | ",
          "4, White",
          " | ",
          "5, More Than One Race",
          " | ",
          "6, Unknown / Not Reported"
        ),
        # 7
        "1, Lung | 2, Liver | 3, Bone | 4, Lymph node | 5, Other | 6, Brain" #6
      ),
      stringsAsFactors = FALSE
    )
  )
  fields <- add_field_elements(fields)
  sum_of_choices <- 2L + 2L + 2L + 7L + 6L * (2L) + 3L
  choices <- fields_to_choices(fields)
  expect_data_frame(choices, nrows = sum_of_choices)
  expect_false(anyNA(choices$code))
  expect_false(anyNA(choices$name))
  expect_false(anyNA(choices$field_name))
  expect_false(anyNA(choices$form_name))
  should_be_unique <- paste0(choices$field_name, "_", choices$code)
  #check warning for same name different code here?
  expect_false(anyDuplicated(should_be_unique) > 0L)
})
# filter_data_list (Internal)
test_that("filter_data_list works!", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  new_data <- filter_data_list(
    data_list = project,
    filter_field = "var_branching",
    filter_choices = "Yes"
  )
  expect_gt(nrow(project$data$other), nrow(new_data$other))
  expect_all_true(new_data$other$var_branching == "Yes")
})
# generate_project_dataset (Internal)
# load_project_dataset (Internal)
test_that("generate_project_dataset works!", {
  project <- mock_test_project()$.internal
  expect_error(load_project_dataset(project = project,
                                    dataset_name = "non_existing"),
               regexp = "not included in the current project datasets")
  project_summary <- load_project_dataset(project = project,
                                          dataset_name = "REDCapSync")
  expect_true(is_df_list(project_summary))
  project_summary <- generate_project_dataset(project = project,
                                              dataset_name = "custom",
                                              include_metadata = TRUE,
                                              include_log = TRUE)
  expect_contains(names(project_summary), "metadata")
  expect_contains(names(project_summary$metadata), "forms")
  expect_contains(names(project_summary$metadata), "fields")
  expect_contains(names(project_summary$metadata), "choices")
  project_summary <- generate_project_dataset(project = project,
                                              dataset_name = "custom",
                                              include_metadata = FALSE,
                                              include_log = TRUE)
  project_summary <- data_list_to_save(project_summary)
  expect_false("metadata" %in% names(project_summary))
  project_summary <- generate_project_dataset(project = project,
                                              dataset_name = "custom",
                                              exclude_identifiers = FALSE)
  project_summary <- data_list_to_save(project_summary)
  fields <- project$metadata$fields
  fields$field_name[which(fields$identifier == "y")]
  colnames(project_summary$merged)
  should_be_missing <- project_summary$merged$var_text_datetime_ymd_hm[1L]
  expect_identical(should_be_missing, "Unknown")
  project_summary <- generate_project_dataset(project = project,
                                              dataset_name = "custom",
                                              drop_missing_codes = TRUE)
  project_summary <- data_list_to_save(project_summary)
  should_not_be_missing <- project_summary$merged$var_text_datetime_ymd_hm[1L]
  expect_scalar_na(should_not_be_missing)
})
# extract_log (Internal)
test_that("extract_log works!", {
  project <- mock_test_project()$.internal
  log <- project$redcap$log
  records <- unique(log$record)[1L]
  log_subset <- extract_log(project, records)
  subset_record_id <- unique(log_subset$record)
  expect_in(subset_record_id, records)
})
# get_dataset_records (Internal)
test_that("get_dataset_records works!", {
  project <- mock_test_project()$.internal
  project <- project |>
    add_project_dataset(
      dataset_name = "test_branching_yes",
      filter_field = "var_branching",
      filter_choices = "Yes"
    ) |>
    add_project_dataset(
      dataset_name = "test_branching_no",
      filter_field = "var_branching",
      filter_choices = "No"
    )
  other <- project$data$other
  expect_true("test_branching_yes" %in% names(project$datasets))
  expect_true("test_branching_no" %in% names(project$datasets))
  record_ids_yes <- other$record_id[which(other$var_branching == "Yes")]
  record_ids_no <- other$record_id[which(other$var_branching == "No")]
  get_sum_records_yes <- get_dataset_records(project, "test_branching_yes")
  get_sum_records_no <- get_dataset_records(project, "test_branching_no")
  expect_identical(record_ids_yes, get_sum_records_yes)
  expect_identical(record_ids_no, get_sum_records_no)
})
# merge_non_repeating (Internal)
test_that("merge_non_repeating works!", {
  project <- mock_test_project()$.internal
  expect_contains(names(project$data), project$metadata$forms$form_name)
  id_col <- project$metadata$id_col
  text_field_names <- colnames(project$data$text) |> setdiff(id_col)
  other_field_names <- colnames(project$data$other) |> setdiff(id_col)
  data_field_names <- colnames(project$data$cancer) |> setdiff(id_col)
  text_field_names2 <- project$metadata$fields |>
    get_match(
      match_field = "form_name",
      match_text = "text",
      return_field = "field_name"
    ) |>
    setdiff(id_col)
  other_field_names2 <- project$metadata$fields |>
    get_match(
      match_field = "form_name",
      match_text = "other",
      return_field = "field_name"
    ) |>
    setdiff(id_col)
  data_field_names2 <- project$metadata$fields |>
    get_match(
      match_field = "form_name",
      match_text = "data",
      return_field = "field_name"
    ) |>
    setdiff(id_col)
  merge_form_name <- "merged_form"
  merged <- merge_non_repeating(
    data_list = project,
    merge_form_name = merge_form_name,
    merge_to_rep = TRUE
  ) #NA for classic)
  expect_identical(merge_form_name, names(merged$data))
  a <- nrow(merged$data[[merge_form_name]])
  b <- nrow(project$data$text)
  c <- nrow(project$data$other)
  d <- nrow(project$data$cancer)
  expect_identical(a, b)
  expect_identical(a, c)
  expect_identical(a, d)
  all_cols <- c(id_col, text_field_names, other_field_names, data_field_names)
  expected_col_names <- unique(all_cols)
  expected_col_length <- length_unique(expected_col_names)
  expect_identical(ncol(merged$data[[merge_form_name]]), expected_col_length)
  expect_named(merged$data[[merge_form_name]], expected_col_names)
})
# metadata_add_default_cols (Internal)
test_that("metadata_add_default_cols works!", {
})
# save_project_dataset (Internal)
test_that("save_project_dataset works!", {
  project <- mock_test_project()$.internal
  dataset_name <- "SAVE_SUMMARY_TEST"
  project <- add_project_dataset(
    project = project,
    dataset_name = dataset_name,
    transformation_type = "default",
    include_metadata = TRUE,
    include_records = TRUE,
    include_users = TRUE,
    include_log = FALSE,
    annotate_from_log = FALSE,
    with_links = FALSE,
    separate = FALSE,
    use_csv = FALSE
  )
  # save and capture returned project
  expect_no_error({
    project <- save_project_dataset(project, dataset_name)
  })
  # file created
  file_path <- project$datasets[[dataset_name]]$file_path
  expect_file_exists(file_path)
  data_list <- read_dataset_from_file(project = project,
                                      dataset_name = dataset_name)
  # datasets metadata updated
  expect_false(is.null(project$datasets[[dataset_name]]$n_records))
  expect_false(is.null(project$datasets[[dataset_name]]$last_save_time))
  what_was_saved <- excel_to_list(file_path)
  expect_identical(what_was_saved$merged, data_list$merged)
  expect_all_true(
    c(
      "merged",
      "forms",
      "fields",
      "choices",
      "missing_codes",
      "users",
      "records",
      "dataset_details"
    ) %in% names(what_was_saved)
  )
  # csv
  dataset_name <- "SAVE_SUMMARY_TEST_CSV"
  project <- add_project_dataset(
    project = project,
    dataset_name = dataset_name,
    transformation_type = "default",
    include_metadata = TRUE,
    include_records = TRUE,
    include_users = TRUE,
    include_log = FALSE,
    annotate_from_log = FALSE,
    with_links = FALSE,
    separate = FALSE,
    use_csv = TRUE
  )
  # save and capture returned project
  expect_no_error({
    project <- save_project_dataset(project, dataset_name)
  })
  # file created
  file_path <- file.path(project$dir_path,
                         "output",
                         paste0("TEST_CLASSIC_SAVE_SUMMARY_TEST_CSV",
                                c("_merged.csv",
                                  "_forms.csv",
                                  "_dataset_details.csv")))
  expect_file_exists(file_path)
})
# save_project_datasets (Internal)
test_that("save_project_datasets works!", {
  project <- mock_test_project()$.internal
  expect_directory_exists(project$dir_path)
  # ensure default datasets present
  project <- clear_project_datasets(project)
  project <- add_default_datasets(
    project = project,
    exclude_identifiers = TRUE,
    exclude_free_text = TRUE,
    date_handling = "none"
  )
  expect_no_error({
    project_saved <- save_project_datasets(project, hard_reset = TRUE)
  })
  expect_false(is.null(project_saved$internals$last_dataset_save))
  expect_s3_class(project_saved$internals$last_dataset_save, "POSIXt")
  expect_true("REDCapSync" %in% names(project_saved$datasets))
  expect_file_exists(project_saved$datasets$REDCapSync$file_path)
})
# dataset_records_due (Internal)
test_that("dataset_records_due works!", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  project <- mock_test_project()$.internal |> clear_project_datasets()
  expect_false(dataset_records_due(project, "REDCapSync"))
  expect_error(save_project_dataset(project, "REDCapSync"))
  project <- add_default_datasets(project)
  expect_contains(names(project$datasets), "REDCapSync")
  expect_true(dataset_records_due(project, "REDCapSync"))
  expect_no_error({
    project <- save_project_dataset(project, "REDCapSync")
  })
  expect_false(dataset_records_due(project, "REDCapSync"))
  project$record_summary$REDCapSync[3L] <- NA
  expect_true(dataset_records_due(project, "REDCapSync"))
})

#' @title Configuration
#'
#' @description
#' `r lifecycle::badge("experimental")`
#' Internal configuration helpers used to retrieve package configuration
#' values from options or environment variables.
#'
#' Configuration is resolved in the following order:
#' 1. `getOption("redcapsync.config.option.name")`
#' 2. `Sys.getenv("REDCAPSYNC_CONFIG_OPTION_NAME")` # skipped for functions
#' 3. Default if unable to find and validate from above.
#'
#' @details
#' The config function is operational but only some methods presently
#' affect internal code. Most users will not ever need to modify default config.
#' This is included to improve future versions of the package.
#' * Working configs: `allow.test.names`, `cache.dir`, `keyring`,
#' `keyring.service`, `openxlsx.header.style`, `openxlsx.body.style`.
#' * Partial coverage: `offline`.
#' * Placeholder: `verbose`, `show.api.messages`
#'
#' ## allow.test.names
#' Logical for [setup_project()] allowing `project_name` starting with TEST_.
#' Default is `FALSE`. This is provided for developers and testing environments.
#' ```
#' # check current value package is using...
#' config$allow.test.names()
#' # set with options (which will be prioritized over envvar)
#' getOption("redcapsync.config.allow.test.names") # get
#' options(redcapsync.config.allow.test.names = FALSE) # set
#'
#' # set with ennvar (which will be prioritized when options not defined)
#' Sys.getenv("REDCAPSYNC_CONFIG_ALLOW_TEST_NAMES")  # get
#' Sys.setenv(REDCAPSYNC_CONFIG_ALLOW_TEST_NAMES = FALSE) # or set in .Renviron
#' ```
#'
#' ## show.api.messages
#' Logical for showing display API messages from REDCapR and redcapAPI.
#' Default is `FALSE`.
#'
#' ```
#' # check current value package is using...
#' config$show.api.messages()
#' # set with options (which will be prioritized over envvar)
#' getOption("redcapsync.config.show.api.messages") # get
#' options(redcapsync.config.show.api.messages = FALSE) # set
#'
#' # set with ennvar (which will be prioritized when options not defined)
#' Sys.getenv("REDCAPSYNC_CONFIG_SHOW_API_MESSAGES") # get
#' Sys.setenv(REDCAPSYNC_CONFIG_SHOW_API_MESSAGES = FALSE) # or set in .Renviron
#' ```
#'
#' ## verbose
#' Logical for showing display API messages from REDCapR and redcapAPI.
#' Default is `FALSE`.
#'
#' ```
#' # check current value package is using...
#' config$verbose()
#' # set with options (which will be prioritized over envvar)
#' getOption("redcapsync.config.verbose") # get
#' options(redcapsync.config.verbose = FALSE) # set
#'
#' # set with ennvar (which will be prioritized when options not defined)
#' Sys.getenv("REDCAPSYNC_CONFIG_VERBOSE") # get
#' Sys.setenv(REDCAPSYNC_CONFIG_VERBOSE = FALSE) # or set in .Renviron
#' ```
#'
#' ## offline
#' Logical for offline, which if TRUE will block any API calls.
#' Default is `FALSE`.
#'
#' ```
#' # check current value package is using...
#' config$offline()
#' # set with options (which will be prioritized over envvar)
#' getOption("redcapsync.config.offline") # get
#' options(redcapsync.config.offline = FALSE) # set
#'
#' # set with ennvar (which will be prioritized when options not defined)
#' Sys.getenv("REDCAPSYNC_CONFIG_OFFLINE") # get
#' Sys.setenv(REDCAPSYNC_CONFIG_OFFLINE = FALSE) # or set in .Renviron
#' ```
#'
#' ## cache.dir
#' Character file path overriding the default cache directory. Default follow
#' system standards via rappdir, hoardr, or R_USER_CACHE_DIR
#'
#' ```
#' # check current value package is using...
#' config$cache.dir()
#' # set with options (which will be prioritized over envvar)
#' getOption("redcapsync.config.cache.dir") # get
#' options(redcapsync.config.cache.dir = "file/path/to/keep/cache") # set
#'
#' # set with ennvar (which will be prioritized when options not defined)
#' Sys.getenv("REDCAPSYNC_CONFIG_CACHE_DIR") # get
#' Sys.setenv(REDCAPSYNC_CONFIG_CACHE_DIR = "file/path/to/keep/cache") # set
#' ```
#'
#' ## keyring
#' Character keyring name (parameter from \code{\link[keyring]{keyring}}
#' package). Default is NULL, which is at the system level. For locking use a
#' keyring like "REDCapSync"
#'
#' ```
#' # check current value package is using...
#' config$keyring()
#' # set with options (which will be prioritized over envvar)
#' getOption("redcapsync.config.keyring") # get
#' options(redcapsync.config.keyring = "REDCapSync") # set
#'
#' # set with ennvar (which will be prioritized when options not defined)
#' Sys.getenv("REDCAPSYNC_CONFIG_KEYRING") # get
#' Sys.setenv(REDCAPSYNC_CONFIG_KEYRING = "REDCapSync") # set in session
#' ```
#'
#' ## keyring.service
#' Character keyring service name (parameter from \code{\link[keyring]{keyring}}
#' package). Default is "R-REDCapSync".
#'
#' ```
#' # check current value package is using...
#' config$keyring.service()
#' # set with options (which will be prioritized over envvar)
#' getOption("redcapsync.config.keyring.service") # get
#' options(redcapsync.config.keyring.service = "REDCapSync") # set
#'
#' # set with ennvar (which will be prioritized when options not defined)
#' Sys.getenv("REDCAPSYNC_CONFIG_KEYRING_SERVICE") # get
#' Sys.setenv(REDCAPSYNC_CONFIG_KEYRING_SERVICE = "REDCapSync") # set in session
#' ```
#'
#' ## openxlsx.header.style
#' Excel sheet header using [openxlsx::createStyle()]
#'
#' ```
#' # set with options (which will be prioritized over envvar)
#' getOption("redcapsync.config.openxlsx.header.style") # get
#' new_style <- openxlsx::createStyle(fontSize = 12L)
#' options(redcapsync.config.openxlsx.header.style = new_style)
#'# does not look up envvar due to it's function type
#' ```
#'
#' ## openxlsx.body.style
#' Excel sheet body using [openxlsx::createStyle()]
#'
#' ```
#' # set with options (which will be prioritized over envvar)
#' getOption("redcapsync.config.openxlsx.body.style") # get
#' new_style <- openxlsx::createStyle(fontSize = 12L)
#' options(redcapsync.config.openxlsx.body.style = new_style)
#'# does not look up envvar due to it's function type
#' ```
#'
#' # Option Names (searched first)
#'
#' ```
#' option_list <- list(
#'   redcapsync.config.allow.test.names = NULL,
#'   redcapsync.config.show.api.messages = NULL,
#'   redcapsync.config.verbose = NULL,
#'   redcapsync.config.offline = NULL,
#'   redcapsync.config.cache.dir = NULL,
#'   redcapsync.config.keyring = NULL,
#'   redcapsync.config.keyring.service = NULL,
#'   redcapsync.config.openxlsx.header.style = NULL,
#'   redcapsync.config.openxlsx.body.style = NULL
#' )
#' ```
#'
#' # Environment Variable Names (searched second)
#' ```
#' envvar_list <- list(
#'   REDCAPSYNC_CONFIG_ALLOW_TEST_NAMES = NA,
#'   REDCAPSYNC_CONFIG_SHOW_API_MESSAGES = NA,
#'   REDCAPSYNC_CONFIG_VERBOSE = NA,
#'   REDCAPSYNC_CONFIG_OFFLINE = NA,
#'   R_USER_CACHE_DIR = NA, # affects your entire cache for any package
#'   REDCAPSYNC_CONFIG_CACHE_DIR = NA, # affects only REDCapSync Cache
#'   REDCAPSYNC_CONFIG_KEYRING = NA,
#'   REDCAPSYNC_CONFIG_KEYRING_SERVICE = NA
#' )
#' ```
#'
#' @seealso
#' \code{vignette("Projects", package = "REDCapSync")}
#' [setup_project] for initializing projects
#'
#' @examples
#'
#' # disable with environment variable
#' Sys.setenv(REDCAPSYNC_CONFIG_OFFLINE = FALSE)
#'
#' config$offline()
#'
#' # change to offline
#' options(redcapsync.config.offline = TRUE)
#'
#' config$offline()
#' @returns list of functions that returns config values
#' @export
config <- list(
  allow.test.names = function(default = FALSE) {
    config_get(opt_name = "allow.test.names",
               type = "logical",
               default = default)
  },
  show.api.messages = function(default = FALSE) {
    config_get(opt_name = "show.api.messages",
               type = "logical",
               default = default)
  },
  verbose = function(default = FALSE) {
    config_get(opt_name = "verbose",
               type = "logical",
               default = default)
  },
  offline = function(default = FALSE) {
    config_get(opt_name = "offline",
               type = "logical",
               default = default)
  },
  cache.dir = function(default = cache_path_default()) {
    config_get(opt_name = "cache.dir",
               type = "filepath",
               default = default)
  },
  keyring = function(default = NULL) {
    config_get(opt_name = "keyring",
               type = "character",
               default = default)
  },
  keyring.service = function(default = "R-REDCapSync") {
    config_get(opt_name = "keyring.service",
               type = "character",
               default = default)
  },
  openxlsx.header.style = function(default = openxlsx_header_style()) {
    config_get(opt_name = "openxlsx.header.style",
               type = "openxlsx_style",
               default = default)
  },
  openxlsx.body.style = function(default = openxlsx_body_style()) {
    config_get(opt_name = "openxlsx.body.style",
               type = "openxlsx_style",
               default = default)
  }
)
#' @noRd
config_get <- function(opt_name, type, default) {
  assert_choice(opt_name, names(config))
  assert_choice(type, OPTION_TYPES)
  # check options first
  opt_key <- paste("REDCapSync", "config", opt_name, sep = ".") |> tolower()
  opt_val <- getOption(opt_key)
  if (!is.null(opt_val)) {
    return(config_validate(opt_name, opt_val, type, default))
  }
  # check envvar second
  env_name <- chartr(".", "_", opt_name)
  env_key <- paste("REDCapSync", "config", env_name, sep = "_") |> toupper()
  env_val <- Sys.getenv(env_key, unset = NA)
  if (!is.na(env_val) && nzchar(env_val) && type != "openxlsx_style") {
    return(config_validate(opt_name, env_val, type, default))
  }
  default
}
#' @noRd
config_validate <- function(opt_name, value, type, default) {
  assert_choice(opt_name, names(config))
  assert_choice(type, OPTION_TYPES)
  # add warning messages
  if (identical(type, "logical")) {
    if (test_character(value, len = 1L)) {
      if (toupper(value) %in% c("TRUE", "FALSE")) {
        value <- value |> toupper() |> as.logical()
      }
    }
    if (test_logical(value, len = 1L)) {
      return(value)
    }
  }
  if (identical(type, "openxlsx_style")) {
    if (test_openxlsx_style(value)) {
      return(value)
    }
  }
  if (identical(type, "filepath")) {
    if (test_directory_exists(value)) {
      # meesage about not existing?
      return(value)
    }
  }
  if (identical(type, "character")) {
    if (test_character(value, min.chars = 1L, any.missing = FALSE, len = 1L)) {
      if (toupper(value) != "NULL") {
        return(value)
      }
    }
  }
  default
}
OPTION_TYPES <- c("logical", "filepath", "character", "openxlsx_style")

#' @noRd
now_time <- function() {
  as.POSIXct(Sys.time(), tz = Sys.timezone())
}
#' @noRd
process_df_list <- function(list,
                            drop_empty = TRUE,
                            silent = FALSE) {
  if (is_something(list)) {
    if (!is_df_list(list))
      cli_abort("list must be ...... a list :)")
    if (drop_empty) {
      is_a_df_with_rows <- list |>
        lapply(function(x) {
          is_df <- is.data.frame(x)
          out <- FALSE
          if (is_df) {
            out <- nrow(x) > 0L
          }
          out
        }) |>
        unlist()
      keeps <- which(is_a_df_with_rows)
      drops <- which(!is_a_df_with_rows)
      if (length(drops) > 0L) {
        if (!silent) {
          msg <- "Dropping non-data.frames and empties: {toString(names(drops)}"
          cli_alert_info(msg)
        }
      }
      list <- list[keeps]
    }
    if (length(list) > 0L) {
      if (!is_named_df_list(list)) {
        names(list) <- paste0(seq_along(list))
      }
    }
  }
  list
}
#' @noRd
is_something <- function(thing, row_length = 0L) {
  if (is.null(thing)) {
    return(FALSE)
  }
  if (is.data.frame(thing)) {
    return(nrow(thing) > row_length)
  }
  the_length <- length(thing)
  if (the_length == 0L) {
    return(FALSE)
  }
  if (the_length > 1L || is.list(thing)) {
    return(TRUE)
  }
  if (is.na(thing)) {
    return(FALSE)
  }
  if (is.character(thing)) {
    return(nzchar(thing))
  }
  TRUE
}
#' @noRd
sanitize_path <- function(path) {
  sanitized <- gsub("\\\\", "/", path)
  sanitized <- normalizePath(sanitized, winslash = "/", mustWork = FALSE)
  sanitized
}
#' @noRd
all_character_cols <- function(form) {
  as.data.frame(lapply(form, as.character))
}
#' @noRd
all_character_cols_list <- function(list) {
  lapply(list, all_character_cols)
}
#' @noRd
vec1_in_vec2 <- function(vec1, vec2) {
  vec1[which(vec1 %in% vec2)]
}
#' @noRd
vec1_not_in_vec2 <- function(vec1, vec2) {
  vec1[which(!vec1 %in% vec2)]
}
#' @noRd
length_unique <- function(x) {
  length(unique(x))
}
#' @noRd
drop_nas <- function(x) {
  if (length(x) == 0L) {
    return(x)
  }
  x[!unlist(lapply(x, is.na))]
}
#' @noRd
is_named_df_list <- function(x, strict = FALSE) {
  is_named_list(x) && is_df_list(x, strict = strict)
}
#' @noRd
is_named_list <- function(x, recursive = FALSE) {
  if (!is.list(x)) {
    return(FALSE)
  }
  if (is.null(names(x))) {
    return(FALSE)
  }
  named_all <- TRUE
  if (recursive) {
    for (n in names(x)) {
      element <- x[[n]]
      if (is.list(element)) {
        named_all <- named_all && is_named_list(element)
      }
    }
  }
  named_all
}
#' @noRd
trim_string <- function(string, max_length) {
  substr(string, 1L, max_length)
}
#' @noRd
unique_trimmed_strings <- function(strings, max_length) {
  trimmed_strings <- lapply(strings, trim_string, max_length = max_length) |>
    unlist()
  # Initialize a vector to store unique strings
  unique_strings <- character(length(trimmed_strings))
  # Initialize a counter to keep track of occurrences
  counts <- integer(length(trimmed_strings))
  for (i in seq_along(trimmed_strings)) {
    base_string <- trimmed_strings[i]
    new_string <- base_string
    counter <- 1L
    # Keep adjusting the string until it's unique
    while (new_string %in% unique_strings) {
      new_string <- paste0(
        str_trunc(
          base_string,
          width = max_length - str_length(counter),
          side = "right",
          ellipsis = ""
        ),
        counter
      )
      counter <- counter + 1L
    }
    unique_strings[i] <- new_string
    counts[i] <- counter
  }
  unique_strings
}
#' @noRd
which_duplicated <- function(x) {
  which(duplicated(x))
}
#' @noRd
is_consecutive_srt_1 <- function(vec) {
  if (vec[1L] != 1L) {
    return(FALSE)
  }
  if (length(vec) > 1L) {
    for (i in 2L:length(vec)) {
      if (vec[i] != vec[i - 1L] + 1L) {
        return(FALSE)
      }
    }
  }
  TRUE
}
#' @noRd
remove_html_tags <- function(text_vector) {
  html_pattern <- "<[^>]+>"
  cleaned_vector <- gsub(html_pattern, "", text_vector)
  cleaned_vector
}
#' @noRd
object_size <- function(x) {
  format(object.size(x), units = "auto")
}
#' @noRd
drop_if <- function(x, drops) {
  x[which(!x %in% drops)]
}
#' @noRd
clean_env_names <- function(env_names,
                            silent = FALSE,
                            lowercase = TRUE) {
  cleaned_names <- character(length(env_names))
  for (i in seq_along(env_names)) {
    name <- env_names[i]
    is_valid <- is_env_name(name, silent = TRUE)
    if (is_valid) {
      cleaned_name <- name
    }
    if (!is_valid) {
      if (!silent) {
        message("Invalid environment name: '", name)
      }
      cleaned_name <- gsub("__", "_", gsub(" ", "_", gsub("-", "", name)))
    }
    if (lowercase) {
      cleaned_name <- tolower(cleaned_name)
    }
    if (cleaned_name %in% cleaned_names) {
      if (!silent) {
        message("Non-unique environment name: '",
                name,
                "', added numbers...")
      }
      cleaned_name <- cleaned_name |>
        paste0("_", max(length(which(cleaned_name %in% cleaned_names))) + 1L)
    }
    cleaned_names[i] <- cleaned_name
  }
  cleaned_names
}
#' @noRd
is_df_list <- function(x, strict = FALSE) {
  if (!is.list(x)) {
    return(FALSE)
  }
  if (length(x) == 0L) {
    return(FALSE)
  }
  if (is_nested_list(x)) {
    return(FALSE)
  }
  out <- unlist(lapply(x, is.data.frame))
  if (strict) {
    return(all(out))
  }
  any(out)
}
#' @noRd
check_match <- function(vec_list) {
  sorted_vecs <- lapply(vec_list, sort)
  all(unlist(lapply(sorted_vecs[-1L], function(x) {
    identical(sorted_vecs[[1L]], x)
  })))
}
#' @noRd
is_env_name <- function(env_name, silent = FALSE) {
  result <- tryCatch({
    if (is.null(env_name)) {
      cli_abort("env_name is NULL")
    }
    if (is.na(env_name)) {
      cli_abort("env_name is NA")
    }
    if (!nzchar(env_name)) {
      cli_abort("Short name cannot be empty.")
    }
    if (grepl("^\\d", env_name)) {
      cli_abort("Short name cannot start with a number.")
    }
    if (grepl("[^A-Za-z0-9_]", env_name)) {
      cli_abort("Short name can only have letters, numbers, and underscores.")
    }
    return(TRUE)
  }, error = function(e) {
    if (!silent) {
      cli_alert_danger(e$message)
    }
    FALSE
  })
  result
}
#' @noRd
is_nested_list <- function(x) {
  if (!is.list(x)) {
    return(FALSE)
  }
  if (is.data.frame(x)) {
    return(FALSE)
  }
  outcome <- length(x) == 0L
  for (i in seq_along(x)) {
    outcome <- outcome || is_nested_list(x[[i]])
  }
  outcome
}
#' @noRd
add_redcap_links <- function(form, project) {
  # add instance links
  if (project$metadata$id_col %in% colnames(form)) {
    rows_x <- which(
      project$metadata$raw_structure_cols %in% colnames(form) &
        project$metadata$raw_structure_cols != project$metadata$id_col
    )
    form_structure_cols <- project$metadata$raw_structure_cols[rows_x]
    link_head <- project$links$redcap_record_home
    link_tail <- paste0("&id=", form[[project$metadata$id_col]])
    if ("redcap_repeat_instrument" %in% form_structure_cols) {
      link_head <- project$links$redcap_record_subpage
      link_tail <-
        paste0(link_tail, "&page=", form[["redcap_repeat_instrument"]])
    }
    if ("redcap_repeat_instance" %in% form_structure_cols) {
      link_head <- project$links$redcap_record_subpage
      link_tail <-
        paste0(link_tail, "&instance=", form[["redcap_repeat_instance"]])
    }
    form$redcap_link <- paste0(link_head, link_tail)
    if ("arm_number" %in% colnames(form)) {
      form$redcap_link <-
        paste0(form$redcap_link, "&arm=", form[["arm_number"]])
    }
  }
  form
}
#' @noRd
split_choices <- function(x) {
  result <- x
  # added this to account for redcap metadata output if not a number
  result <- gsub("\n", " | ", result)
  result <- result |>
    strsplit("[|]") |>
    unlist() |>
    str_split_fixed(",", 2L)
  check_length <- length(result[, 1L])
  choices_data <- data.frame(
    code = trimws(result[, 1L]),
    name = trimws(result[, 2L]),
    stringsAsFactors = FALSE
  )
  rownames(choices_data) <- NULL
  if (anyNA(choices_data$code)) {
    cli_abort("split choice error: {x}")
  }
  if (anyNA(choices_data$name)) {
    cli_abort("split choice error: {x}")
  }
  if (nrow(choices_data) != check_length) {
    cli_abort("split choice error: {x}")
  }
  if (!all(nzchar(choices_data$name))) {
    cli_abort("split choice error: {x}")
  }
  choices_data
}
#' @noRd
get_match <- function(x, return_field, match_field, match_text) {
  x[[return_field]][which(x[[match_field]] %in% match_text)]
}
#' @noRd
file_ext_alias <- function(x) {
  pos <- regexpr("\\.([[:alnum:]]+)$", x)
  ifelse(pos > -1L, substring(x, pos + 1L), "")
}
#' @noRd
try_else_null <- function(expr) {
  tryCatch(
    expr = expr,
    error = function(e) {
      NULL
    }
  )
}
#' @noRd
na_if_null <- function(x) {
  ifelse(is.null(x), NA, x)
}
#' @noRd
null_if_na <- function(x) {
  if (is.na(x)) {
    return(NULL)
  }
  x
}
#' @noRd
quotes_if_na <- function(x) {
  ifelse(is.na(x), "", x)
}

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# is_consecutive_srt_1 (Internal)
test_that("is_consecutive_srt_1 works!", {
  expect_true(is_consecutive_srt_1(1L))
  expect_true(is_consecutive_srt_1(1L:4L))
  expect_true(is_consecutive_srt_1(c(1L, 2L, 3L)))
  expect_false(is_consecutive_srt_1(c(2L, 3L)))    # doesn't start at 1
  expect_false(is_consecutive_srt_1(c(1L, 3L)))    # gap present
  expect_false(is_consecutive_srt_1(c(1L, 2L, 4L))) # non-consecutive later
})
# is_df_list (Internal)
test_that("is_df_list works!", {
  # non-list inputs
  expect_false(is_df_list(NULL))
  expect_false(is_df_list(1L))
  expect_false(is_df_list(data.frame()))
  # empty list -> FALSE
  expect_false(is_df_list(list()))
  # list with no data.frames -> FALSE
  expect_false(is_df_list(list(a = 1L, b = "x")))
  # pure list of data.frames -> TRUE
  df1 <- mtcars
  df2 <- iris
  expect_true(is_df_list(list(df1, df2)))
  expect_true(is_df_list(list(one = df1, two = df2)))
  # mixed list -> TRUE when strict = FALSE, FALSE when strict = TRUE
  mixed <- list(df1, list())
  expect_false(is_df_list(mixed, strict = TRUE))
  mixed2 <- list(df1, 1L)
  expect_true(is_df_list(mixed2))
  expect_false(is_df_list(mixed2, strict = TRUE))
})
# is_env_name (Internal)
test_that("is_env_name works!", {
  # valid names
  expect_true(is_env_name("one"))
  expect_true(is_env_name("A"))
  expect_true(is_env_name("a1_b2"))
  # invalid inputs
  expect_false(is_env_name(NULL))
  expect_false(is_env_name(""))
  expect_false(is_env_name("1start"))    # starts with digit
  expect_false(is_env_name("bad-char!")) # invalid character
  # messaging behavior
  expect_message(is_env_name("1start", silent = FALSE),
                 "Short name cannot start with a number")
  expect_silent(is_env_name("1start", silent = TRUE))
})
# is_named_df_list (Internal)
test_that("is_named_df_list works!", {
})
# is_named_list (Internal)
test_that("is_named_list works!", {
})
# is_nested_list (Internal)
test_that("is_nested_list works!", {
  expect_false(is_nested_list(NULL))
  expect_false(is_nested_list(1L))
  expect_false(is_nested_list("a"))
  expect_false(is_nested_list(data.frame()))
  expect_true(is_nested_list(list()))
  expect_false(is_nested_list(list(a = 1L, b = "x", c = TRUE)))
  expect_true(is_nested_list(list(a = list(), b = 1L)))
})
# is_something (Internal)
test_that("is_something works!", {
  expect_true(is_something("a"))
  expect_true(is_something(list(one = NA)))
  expect_true(is_something(mtcars))
  expect_true(is_something(1L))
  expect_false(is_something(NA))
  expect_false(is_something(NULL))
  expect_false(is_something(list()))
  expect_false(is_something(data.frame()))
  expect_false(is_something(tibble()))
})
# length_unique (Internal)
test_that("length_unique works!", {
  expect_identical(length_unique(c("a", "a", "b", "b", "c", "c", "c")), 3L)
  expect_identical(length_unique(c("a", "a", "b", "b", "c", "d")), 4L)
})
# now_time (Internal)
test_that("now_time works!", {
  expect_class(now_time(), "POSIXct")
  expect_class(now_time(), "POSIXt")
})
# object_size (Internal)
test_that("object_size works!", {
})
# process_df_list (Internal)
test_that("process_df_list works!", {
})
# remove_html_tags (Internal)
test_that("remove_html_tags works!", {
  # basic tag removal
  expect_identical(remove_html_tags("<b>bold</b>"), "bold")
  expect_identical(remove_html_tags("no tags"), "no tags")
  # vectorized input
  expect_identical(
    remove_html_tags(c("<p>one</p>", "<div>two</div>")),
    c("one", "two")
  )
  # tags with attributes and surrounding text
  expect_identical(remove_html_tags("<a href='x'>link</a> and text"),
                   "link and text")
  # nested tags
  expect_identical(remove_html_tags("<div><span>nested</span></div>"), "nested")
  # empty string preserved, NA preserved
  expect_identical(remove_html_tags(""), "")
  expect_true(is.na(remove_html_tags(NA_character_)))
  # tags with newlines preserved inside text
  expect_identical(remove_html_tags("<p>line\nbreak</p>"), "line\nbreak")
})
# sanitize_path (Internal)
test_that("sanitize_path works!", {
})
# split_choices (Internal)
test_that("split_choices works!", {
  choices <- split_choices("1, one | 2, two")
  expect_data_frame(choices, nrows = 2L, ncols = 2L)
  expect_false(anyNA(choices$code))
  expect_false(anyNA(choices$name))
  expect_identical(choices$code[[1L]], "1")
  expect_identical(choices$code[[2L]], "2")
  expect_identical(choices$name[[1L]], "one")
  expect_identical(choices$name[[2L]], "two")
  expect_error(split_choices("1. one | 2, two"))
  choices <- split_choices("1, one even with , second comma | 2, two")
  expect_data_frame(choices, nrows = 2L, ncols = 2L)
  expect_false(anyNA(choices$code))
  expect_false(anyNA(choices$name))
})
test_that("split_choices works if space missing!", {
  choices <- split_choices("1,one | 2,two")
  expect_data_frame(choices, nrows = 2L, ncols = 2L)
  expect_false(anyNA(choices$code))
  expect_false(anyNA(choices$name))
})
test_that("split_choices works if split space missing!", {
  choices <- split_choices("1, one|2, two")
  expect_data_frame(choices, nrows = 2L, ncols = 2L)
  expect_false(anyNA(choices$code))
  expect_false(anyNA(choices$name))
})
# trim_string (Internal)
test_that("trim_string works!", {
  expect_identical(trim_string("one", 5L), "one")
  expect_identical(trim_string("one123", 5L), "one12")
  expect_identical(trim_string("one123", 3L), "one")
})
# try_else_null (Internal)
test_that("try_else_null works!", {
})
# unique_trimmed_strings (Internal)
test_that("unique_trimmed_strings works!", {
  expect_error(unique_trimmed_strings())
  expect_identical(unique_trimmed_strings("one", 5L), "one")
  test_vector <- c("one_version_1", "one_version_2", "one_version_3")
  expect_identical(unique_trimmed_strings(test_vector, 6L),
                   c("one_ve", "one_v1", "one_v2"))
  expect_identical(unique_trimmed_strings(test_vector, 20L), test_vector)
})
# vec1_in_vec2 (Internal)
# vec1_not_in_vec2 (Internal)
test_that("vec1_in_vec2 and vec1_not_in_vec2 works!", {
  vec1 <- c("a", "b", "c")
  vec2 <- c("b", "d", "c")
  out <- vec1_in_vec2(vec1, vec2)
  expect_contains(out, "b")
  expect_contains(out, "c")
  expect_false("a" %in% out)
  out <- vec1_not_in_vec2(vec1, vec2)
  expect_contains(out, "a")
  expect_false("b" %in% out)
  expect_false("c" %in% out)
})
# which_duplicated (Internal)
test_that("which_duplicated works!", {
})
# add_redcap_links (Internal)
test_that("add_redcap_links works!", {
  project <- mock_test_project()$.internal
  form <- data.frame(
    record_id = c("1", "2"),
    var_text = c("A", "B"),
    stringsAsFactors = FALSE
  )
  # Should add redcap_link column
  out <- add_redcap_links(form, project)
  expect_true("redcap_link" %in% colnames(out))
  expect_length(out$redcap_link, 2L)
  # Links should contain record_id and project id
  expect_all_true(grepl("id=1|id=2", out$redcap_link))
  project_id <- project$redcap$project_id
  compare <- paste0("_home\\.php\\?pid=", project_id, "&id=", form$record_id)
  expect_true(grepl(compare[1L], out$redcap_link[1L]))
  expect_true(grepl(compare[2L], out$redcap_link[2L]))
  # Should not modify other columns
  expect_identical(out$var_text, form$var_text)
})
# all_character_cols (Internal)
test_that("all_character_cols works!", {
  form <- mtcars
  form$name <- rownames(form)
  form$cyl_more_then_4 <- form$cyl > 4L
  col_types <- form |> lapply(class) |> unlist() |> unique()
  expect_vector(col_types, size = 3L)
  expect_identical(col_types, c("numeric", "character", "logical"))
  form <- all_character_cols(form)
  col_types <- form |> lapply(class) |> unlist() |> unique()
  expect_vector(col_types, size = 1L)
  expect_identical(col_types, "character")
})
# all_character_cols_list (Internal)
test_that("all_character_cols_list works!", {
  form <- mtcars
  form$name <- rownames(form)
  form$cyl_more_then_4 <- form$cyl > 4L
  form_list <- list(one = form, two = airquality)
  col_types1 <- form_list$one |> lapply(class) |> unlist() |> unique()
  col_types2 <- form_list$two |> lapply(class) |> unlist() |> unique()
  expect_contains(col_types1, "numeric")
  expect_contains(col_types2, "numeric")
  form_list <- all_character_cols_list(form_list)
  col_types1 <- form_list$one |> lapply(class) |> unlist() |> unique()
  col_types2 <- form_list$two |> lapply(class) |> unlist() |> unique()
  expect_identical(col_types1, "character")
  expect_identical(col_types2, "character")
})
# check_match (Internal)
test_that("check_match works!", {
  # numeric vectors in different orders -> match
  expect_true(check_match(list(
    c(1L, 2L, 3L),
    c(3L, 2L, 1L),
    c(2L, 1L, 3L)
  )))
  # character vectors in different orders -> match
  expect_true(check_match(list(
    c("a", "b", "c"),
    c("c", "b", "a")
  )))
  # differing contents -> no match
  expect_false(check_match(list(
    c("a", "b"),
    c("a", "c")
  )))
  # different lengths -> no match
  expect_false(check_match(list(
    c(1L, 2L),
    c(1L, 2L, 3L)
  )))
  # empty vectors match
  expect_true(check_match(list(
    integer(0L),
    integer(0L)
  )))
})
# clean_env_names (Internal)
test_that("clean_env_names works!", {
  # valid names preserved
  expect_identical(clean_env_names(c("one", "two")), c("one", "two"))
  # invalid characters are cleaned and lowercased
  expect_identical(
    clean_env_names(c("My Name", "Another-Name")),
    c("my_name", "anothername")
  )
  # duplicates produce unique cleaned names
  out <- clean_env_names(c("Dup Name", "Dup-Name", "dup_name"))
  expect_length(unique(out), length(out))
  # all results should be valid environment names
  expect_true(all(vapply(out, is_env_name, logical(1L))))
  # empty input returns empty character vector
  expect_identical(clean_env_names(character(0L)), character(0L))
})
# clean_for_cli (Internal)
test_that("clean_for_cli works!", {
})
# cli_alert_wrap (Internal)
test_that("cli_alert_wrap works!", {
})
# drop_if (Internal)
test_that("drop_if works!", {
  abc <- c("a", "b", "c")
  expect_identical(drop_if(abc, "b"), c("a", "c"))
  expect_identical(drop_if(c(1L, 2L, 3L, 4L), c(2L, 4L)), c(1L, 3L))
  expect_identical(drop_if(c("a", "b"), character(0L)), c("a", "b"))
  expect_identical(drop_if(c("a", "b"), c("a", "b")), character(0L))
  expect_identical(drop_if(character(0L), "a"), character(0L))
  expect_identical(drop_if(factor(abc), "b"), factor(c("a", "c"), levels = abc))
})
# drop_nas (Internal)
test_that("drop_nas works!", {
})
# file_ext_alias (Internal)
test_that("file_ext_alias works!", {
})
# get_match (Internal)
test_that("get_match works!", {
})
# na_if_null (Internal)
test_that("na_if_null works!", {
})

#' @noRd
repair_projects <- function(projects) {
  if (test_project_details(projects)) {
    return(projects)
  }
  if (!is.data.frame(projects)) {
    return(NULL)
  }
  if (nrow(projects) == 0L) {
    return(BLANK_PROJECT_DETAILS)
  }
  must_have <- c("project_name",
                 "redcap_uri",
                 "token_name",
                 "project_id",
                 "dir_path")
  if (!all(must_have %in% colnames(projects))) {
    return(NULL)
  }
  for (project_name in projects$project_name) {
    project_details <- projects[which(projects$project_name == project_name), ]
    project_details <- repair_project_details(project_details)
    if (is.null(project_details)) {
      cli_alert_danger("Cannot extract project details from '{project_name}'")
    } else {
      projects <- projects[which(projects$project_name != project_name), ] |>
        bind_rows(project_details)
    }
  }
  projects
}
#' @noRd
repair_project_details <- function(project_details) {
  project_name <- project_details$project_name
  dir_path <- project_details$dir_path
  if (is.na(dir_path)) {
    return(NULL)
  }
  project_path <- get_project_path(project_name = project_name,
                                   dir_path = dir_path)
  if (!file.exists(project_path)) {
    return(NULL)
  }
  project_details <- try_else_null({
    suppressWarnings({
      project <- readRDS(file = project_path)
      project$dir_path <- dir_path
      extract_project_details(project)
    })
  })
  project_details
}
#' @noRd
repair_setup_project <- function(project) {
  if (!is_named_list(project)) {
    return(NULL)
  }
  must_have <- c("project_name",
                 "dir_path",
                 "redcap",
                 "metadata",
                 "data",
                 "internals")
  missing_important <-
    !test_env_name(project$project_name) ||
    !test_character(project$dir_path, len = 1L) ||
    !test_character(project$links$redcap_uri,
                    len = 1L,
                    min.chars = 4L)
  if (missing_important || !all(must_have %in% names(project))) {
    return(NULL)
  }
  if (!test_setup_project(project)) {
    project$project_name <- toupper(project$project_name)
    project$token_name <- project$token_name |>
      test_env_name(max.chars = 50L, all_caps = TRUE) |>
      ifelse(project$token_name,
             paste0("REDCAPSYNC_", project$project_name))
    # redcap
    project$redcap$timezone <-
      project$redcap$timezone |>
      test_choice(OlsonNames()) |>
      ifelse(project$redcap$timezone, Sys.timezone())
    # settings
    project$settings$sync_frequency <-
      project$settings$sync_frequency |>
      test_choice(choices = SYNC_FREQUENCY) |>
      ifelse(project$settings$sync_frequency, "daily")
    project$settings$labelled <-
      project$settings$labelled |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$labelled, TRUE)
    project$settings$hard_reset <-
      project$settings$hard_reset |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$hard_reset, FALSE)
    project$settings$get_type <-
      project$settings$get_type |>
      test_choice(choices = GET_TYPE) |>
      ifelse(project$settings$get_type, "identified")
    records_na <- test_scalar_na(project$settings$records)
    records_char <- test_unique_character(project$settings$records)
    project$settings$records <- (records_na || records_char) |>
      ifelse(project$settings$records, NA)
    fields_na <- test_scalar_na(project$settings$fields)
    fields_char <- test_unique_character(project$settings$fields)
    project$settings$fields <- (fields_na || fields_char) |>
      ifelse(project$settings$fields, NA)
    forms_na <- test_scalar_na(project$settings$forms)
    forms_char <- test_unique_character(project$settings$forms)
    project$settings$forms <- (forms_na || forms_char) |>
      ifelse(project$settings$forms, NA)
    events_na <- test_scalar_na(project$settings$events)
    events_char <- test_unique_character(project$settings$events)
    project$settings$events <- (events_na || events_char) |>
      ifelse(project$settings$events, NA)
    filter_logic_na <- test_scalar_na(project$settings$filter_logic)
    filter_logic_char <- test_unique_character(project$settings$filter_logic)
    project$settings$filter_logic <- (filter_logic_na || filter_logic_char) |>
      ifelse(project$settings$filter_logic, NA)
    project$settings$id_position <-
      project$settings$id_position |>
      test_integerish(len = 1L, lower = 1L, any.missing = FALSE) |>
      ifelse(project$settings$id_position, 1L)
    project$settings$get_users <-
      project$settings$get_users |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$get_users, TRUE)
    project$settings$get_data <-
      project$settings$get_data |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$get_data, TRUE)
    project$settings$batch_size_download <-
      project$settings$batch_size_download |>
      test_integerish(len = 1L, lower = 1L, any.missing = FALSE) |>
      ifelse(project$settings$batch_size_download, 2000L)
    project$settings$batch_size_upload <-
      project$settings$batch_size_upload |>
      test_integerish(len = 1L, lower = 1L, any.missing = FALSE) |>
      ifelse(project$settings$batch_size_upload, 500L)
    project$settings$get_entire_log <-
      project$settings$get_entire_log |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$get_entire_log, FALSE)
    project$settings$log_days <-
      project$settings$log_days |>
      test_integerish(len = 1L, lower = 1L, any.missing = FALSE) |>
      ifelse(project$settings$log_days, 10L)
    project$settings$log_drop_details <-
      project$settings$log_drop_details |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$log_drop_details, FALSE)
    project$settings$log_drop_exports <-
      project$settings$log_drop_exports |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$log_drop_exports, FALSE)
    project$settings$get_files <-
      project$settings$get_files |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$get_files, FALSE)
    project$settings$get_file_repository <-
      project$settings$get_file_repository |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$get_file_repository, FALSE)
    project$settings$original_file_names <-
      project$settings$original_file_names |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$original_file_names, FALSE)
    project$settings$add_default_datasets <-
      project$settings$add_default_datasets |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$settings$add_default_datasets, TRUE)
    # internals
    project$internals$was_updated <-
      project$internals$was_updated |>
      test_logical(len = 1L, any.missing = FALSE) |>
      ifelse(project$internals$was_updated, FALSE)
  }
  project
}

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# repair_project_details (Internal)
test_that("repair_project_details works with valid project file", {
  project <- mock_test_project()$.internal
  # Save project to disk
  project_path <- get_project_path(
    project_name = project$project_name,
    dir_path = project$dir_path
  )
  saveRDS(project, file = project_path)
  # Create project_details with matching project_name and dir_path
  project_details <- data.frame(
    project_name = project$project_name,
    dir_path = project$dir_path,
    redcap_uri = project$links$redcap_uri,
    stringsAsFactors = FALSE
  )
  result <- repair_project_details(project_details)
  expect_s3_class(result, "data.frame")
  expect_identical(result$project_name, project$project_name)
  expect_identical(result$dir_path, project$dir_path)
})
test_that("repair_project_details returns NULL when project file missing", {
  project_details <- data.frame(
    project_name = "NONEXISTENT",
    dir_path = file.path("nonexistent", "path"),
    stringsAsFactors = FALSE
  )
  result <- repair_project_details(project_details)
  expect_null(result)
})
test_that("repair_project_details returns NULL when dir_path is NA", {
  project_details <- data.frame(
    project_name = "TEST",
    dir_path = NA_character_,
    stringsAsFactors = FALSE
  )
  result <- repair_project_details(project_details)
  expect_null(result)
})
test_that("repair_project_details handles corrupted project file", {
  project <- mock_test_project()$.internal
  project_details_df <- data.frame(
    project_name = "CORRUPT_PROJECT",
    dir_path = project$dir_path,
    stringsAsFactors = FALSE
  )
  # Create corrupted project file
  project_path <- get_project_path(
    project_name = "CORRUPT_PROJECT",
    dir_path = project$dir_path
  )
  saveRDS("not a valid project object", file = project_path)
  result <- repair_project_details(project_details_df)
  expect_null(result)
})
test_that("repair_project_details extracts correct project details", {
  project <- mock_test_project()$.internal
  # Save project to disk
  project_path <- get_project_path(
    project_name = project$project_name,
    dir_path = project$dir_path
  )
  saveRDS(project, file = project_path)
  project_details <- data.frame(
    project_name = project$project_name,
    dir_path = project$dir_path,
    stringsAsFactors = FALSE
  )
  result <- repair_project_details(project_details)
  # Verify extracted details match original project
  expect_identical(result$project_name, project$project_name)
  expect_identical(result$project_id, project$redcap$project_id)
  expect_identical(result$token_name, project$token_name)
  expect_identical(result$redcap_uri, project$links$redcap_uri)
})
# repair_projects (Internal)
test_that("repair_projects works with valid project details", {
  project <- mock_test_project()$.internal
  project_path <- get_project_path(
    project_name = project$project_name,
    dir_path = project$dir_path
  )
  saveRDS(project, file = project_path)
  # Create valid project details
  project_details <- extract_project_details(project)
  projects <- project_details
  # Test repair_projects with valid data
  result <- repair_projects(projects)
  expect_s3_class(result, "data.frame")
  expect_true(test_project_details(result))
  expect_identical(nrow(result), 1L)
  expect_identical(result$project_name, project$project_name)
})
test_that("repair_projects returns blank when given empty data frame", {
  projects <- data.frame()
  result <- repair_projects(projects)
  expect_s3_class(result, "data.frame")
  expect_identical(nrow(result), 0L)
})
test_that("repair_projects returns NULL for non-data.frame input", {
  result <- repair_projects("not a data frame")
  expect_null(result)
})
test_that("repair_projects returns NULL when missing required columns", {
  projects <- data.frame(
    project_name = "TEST",
    stringsAsFactors = FALSE
  )
  result <- repair_projects(projects)
  expect_null(result)
})
test_that("repair_projects handles missing project files", {
  projects <- data.frame(
    project_name = "NONEXISTENT",
    redcap_uri = "https://redcap.fake.edu/api/",
    token_name = "REDCAPSYNC_NONEXISTENT",
    project_id = "9999",
    dir_path = file.path("nonexistent", "path"),
    stringsAsFactors = FALSE
  )
  expect_message(
    repair_projects(projects),
    "Cannot extract project details"
  )
})
# repair_setup_project (Internal)
test_that("repair_setup_project works with valid project", {
  project <- mock_test_project()$.internal
  result <- repair_setup_project(project)
  expect_list(result)
  expect_true(test_setup_project(result))
  expect_identical(result$project_name, project$project_name)
})
test_that("repair_setup_project returns NULL for non-list input", {
  result <- repair_setup_project("not a list")
  expect_null(result)
})
test_that("repair_setup_project returns NULL when missing required fields", {
  project <- list(
    project_name = "TEST",
    dir_path = file.path("test", "path")
    # missing: redcap, metadata, data, internals
  )
  result <- repair_setup_project(project)
  expect_null(result)
})
test_that("repair_setup_project returns NULL for invalid project_name", {
  project <- mock_test_project()$.internal
  project$project_name <- "invalid name with spaces"
  result <- repair_setup_project(project)
  expect_null(result)
})
test_that("repair_setup_project repairs invalid internals", {
  project <- mock_test_project()$.internal
  # Set invalid internals
  project$settings$sync_frequency <- "invalid"
  project$settings$labelled <- NA
  project$redcap$timezone <- "Invalid_Zone"
  result <- repair_setup_project(project)
  expect_true(test_setup_project(result))
  # Check that defaults were applied
  expect_identical(result$settings$sync_frequency, "daily")
  expect_true(result$settings$labelled)
  expect_identical(result$redcap$timezone, Sys.timezone())
})
test_that("repair_setup_project repairs invalid batch sizes", {
  project <- mock_test_project()$.internal
  # Set invalid batch sizes
  project$settings$batch_size_download <- 0L
  project$settings$batch_size_upload <- NA
  result <- repair_setup_project(project)
  expect_true(test_setup_project(result))
  # Check that defaults were applied
  expect_identical(result$settings$batch_size_download, 2000L)
  expect_identical(result$settings$batch_size_upload, 500L)
})
test_that("repair_setup_project repairs invalid logical fields", {
  project <- mock_test_project()$.internal
  # Set invalid logical fields
  project$settings$get_files <- NA
  project$settings$get_file_repository <- "yes"
  project$settings$get_entire_log <- 1L
  result <- repair_setup_project(project)
  expect_true(test_setup_project(result))
  # Check that defaults were applied
  expect_false(result$settings$get_files)
  expect_false(result$settings$get_file_repository)
  expect_false(result$settings$get_entire_log)
})
test_that("repair_setup_project repairs invalid get_type", {
  project <- mock_test_project()$.internal
  project$settings$get_type <- "invalid_type"
  result <- repair_setup_project(project)
  expect_true(test_setup_project(result))
  expect_identical(result$settings$get_type, "identified")
})
test_that("repair_setup_project handles character vector fields", {
  project <- mock_test_project()$.internal
  # Set valid character vectors
  project$settings$records <- c("1", "2", "3")
  project$settings$fields <- c("field1", "field2")
  project$settings$forms <- c("form1", "form2")
  result <- repair_setup_project(project)
  expect_true(test_setup_project(result))
  expect_identical(result$settings$records, c("1", "2", "3"))
  expect_identical(result$settings$fields, c("field1", "field2"))
  expect_identical(result$settings$forms, c("form1", "form2"))
})
test_that("repair_setup_project repairs invalid character vectors", {
  project <- mock_test_project()$.internal
  # Set invalid character vectors (wrong type or length)
  project$settings$records <- 123L
  project$settings$fields <- TRUE
  project$settings$forms <- NA
  result <- repair_setup_project(project)
  expect_true(test_setup_project(result))
  # Check that defaults (NA) were applied
  expect_true(is.na(result$settings$records))
  expect_true(is.na(result$settings$fields))
  expect_true(is.na(result$settings$forms))
})
test_that("repair_setup_project repairs invalid log_days", {
  project <- mock_test_project()$.internal
  project$settings$log_days <- 0L
  result <- repair_setup_project(project)
  expect_true(test_setup_project(result))
  expect_identical(result$settings$log_days, 10L)
})

#' @rdname setup-load
#' @title Setup or Load REDCapSync Project
#' @description
#' Setup or load a REDCapSync [project] object.
#' Prepares a new REDCapSync project by recording the REDCap URI,
#' token name, sync settings, and optional data selection preferences.
#' @details
#' Unless `hard_reset = TRUE`, it will first attempt to load an
#' existing project from cache or the supplied `dir_path` before creating a
#' new project object.
#' If settings differ from the loaded project, they will be used or it will
#' trigger a hard reset with a warning. `setup_project()` itself does not
#' perform a full REDCap sync. It configures the project so that subsequent sync
#' actions may retrieve data, users, files, and logs according to the project
#' settings.
#'
#' @param project_name Character scalar. Unique uppercase project name with no
#' spaces or symbols.
#' @param dir_path Optional character scalar. Directory where REDCap project
#' files are stored. If omitted, the project is only available in the current R
#' session and cannot be persisted to disk.
#' @param redcap_uri Character scalar. Base URL of the REDCap API endpoint.
#' @param token_name Optional character scalar. Environment variable name used
#' to store the REDCap token. Defaults to `REDCAPSYNC_<project_name>`.
#' @param sync_frequency Character scalar. Allowed values are "always",
#' "hourly", "daily", "weekly", "monthly", "once", and "never". This
#' setting controls how often future syncs are considered due.
#' @param labelled Logical scalar. If `TRUE`, data will be preserved as labelled
#' values. Default is `TRUE`.
#' @param get_type Character scalar. REDCap API export type. One of
#' "identified", "deidentified", "deidentified_strict", or
#' "deidentified_super_strict". Default is "identified".
#' @param records Optional character vector of record IDs to request.
#' @param fields Optional character vector of field names to request.
#' @param forms Optional character vector of form names to request.
#' @param events Optional character vector of event names to request.
#' @param filter_logic Optional character scalar. REDCap filter logic used to
#' limit returned records or record-events.
#' @param id_position Integer scalar of the variable that uniquely identifies
#' the subject (typically record_id). This defaults to the first variable in the
#' data dictionary.
#' @param get_users Logical scalar. If `TRUE`, the project will be configured to
#' retrieve REDCap users during sync.
#' @param get_data Logical scalar. If `TRUE`, the project will be configured to
#' retrieve REDCap data during sync.
#' @param batch_size_download Integer scalar. Number of records to request per
#' download batch. Default is `1000`.
#' @param batch_size_upload Integer scalar. Number of records to process per
#' upload batch. Default is `500`.
#' @param get_entire_log Logical scalar. If `TRUE`, REDCap activity logs are
#' retrieved in full. Default is `FALSE`.
#' @param log_days Integer scalar. Number of days of log history to consider
#' when a new project is being set up or a hard reset occurs. Default is `10`.
#' @param log_drop_details Logical scalar. If `TRUE`, the log details are
#' excluded.
#' @param log_drop_exports Logical scalar. If `TRUE`, the log export events are
#' excluded.
#' @param get_files Logical scalar. If `TRUE`, file attachments are configured
#' to be retrieved from REDCap. Default is `FALSE`.
#' @param get_file_repository Logical scalar. If `TRUE`, the file repository is
#' configured to be retrieved from REDCap. Default is `FALSE`.
#' @param original_file_names Logical scalar. If `TRUE`, retrieved files keep
#' their original names. Default is `FALSE`.
#' @param add_default_datasets Logical scalar. If `TRUE`, dataset templates are
#' added to a new project. Default is `TRUE`.
#' @param timezone Optional character scalar. Time zone used for REDCap data
#' timestamps. Defaults to `Sys.timezone()`.
#' @param hard_reset Logical scalar. If `TRUE`, any existing project with the
#' same `project_name` is ignored and a fresh project object is initialized.
#' Default is `FALSE`.
#'
#' @seealso
#' \code{vignette("REDCapSync", package = "REDCapSync")}
#' \code{vignette("Tokens", package = "REDCapSync")}
#' [projects] for shortcuts of cached setup projects
#' [project] for using the project objects
#' \code{vignette("Cache", package = "REDCapSync")}
#'
#' @examples
#' # Initialize the project object with the REDCap API token and URL
#' save_folder <- tempdir() # replace with real folder
#' project <- setup_project(
#'   project_name = "FIRST_PROJECT",
#'   dir_path = save_folder,
#'   redcap_uri = "https://redcap.yourinstitution.edu/api/"
#' )
#'
#' # object can be named whatever you choose to assign
#' # TEST projects can be loaded in addition to real projects
#' project_test <- load_project("TEST_CLASSIC")
#'
#' @returns R6 project object with [REDCapSyncProject] class.
#' @export
setup_project <- function(project_name,
                          dir_path,
                          redcap_uri,
                          token_name = paste0("REDCAPSYNC_", project_name),
                          sync_frequency = "daily",
                          labelled = TRUE,
                          get_type = "identified",
                          records = NA,
                          fields = NA,
                          forms = NA,
                          events = NA,
                          filter_logic = NA,
                          id_position = 1L,
                          get_users = TRUE,
                          get_data = TRUE,
                          batch_size_download = 1000L,
                          batch_size_upload = 500L,
                          get_entire_log = FALSE,
                          log_days = 10L,
                          log_drop_details = FALSE,
                          log_drop_exports = FALSE,
                          get_files = FALSE,
                          get_file_repository = FALSE,
                          original_file_names = FALSE,
                          add_default_datasets = TRUE,
                          timezone = Sys.timezone(),
                          hard_reset = FALSE) {
  missing_dir_path <- missing(dir_path)
  if (missing_dir_path) {
    missing_dir_message <- paste("If you don't supply a directory, REDCapSync",
                                 "will only run in R session. The package is",
                                 "meant to be used with a directory.")
  }
  assert_project_name(project_name, allow_test_names = TRUE)
  assert_env_name(token_name, max.chars = 50L, all_caps = TRUE)
  assert_choice(sync_frequency, choices = SYNC_FREQUENCY)
  assert_logical(labelled, len = 1L, any.missing = FALSE)
  assert_logical(hard_reset, len = 1L, any.missing = FALSE)
  assert_choice(get_type, choices = GET_TYPE)
  assert(
    test_scalar_na(records) ||
      test_character(
        records, # add exist warning
        min.chars = 1L,
        unique = TRUE,
        min.len = 1L,
        any.missing = FALSE
      )
  )
  assert(
    test_scalar_na(fields) ||
      test_character(
        fields, # add exist warning
        min.chars = 1L,
        unique = TRUE,
        min.len = 1L,
        any.missing = FALSE
      )
  )
  assert(
    test_scalar_na(forms) ||
      test_character(
        forms, # add exist warning
        min.chars = 1L,
        unique = TRUE,
        min.len = 1L,
        any.missing = FALSE
      )
  )
  assert(
    test_scalar_na(events) ||
      test_character(
        events, # add exist warning
        min.chars = 1L,
        unique = TRUE,
        min.len = 1L,
        any.missing = FALSE
      )
  )
  assert(
    test_scalar_na(filter_logic) ||
      test_character(
        filter_logic, # add exist warning
        min.chars = 1L,
        unique = TRUE,
        min.len = 1L,
        any.missing = FALSE
      )
  )
  assert_integerish(id_position, len = 1L, lower = 1L, any.missing = FALSE)
  assert_logical(get_data, len = 1L, any.missing = FALSE)
  assert_integerish(
    batch_size_download,
    len = 1L,
    lower = 1L,
    any.missing = FALSE
  )
  assert_integerish(
    batch_size_upload,
    len = 1L,
    lower = 1L,
    any.missing = FALSE
  )
  assert_logical(get_entire_log, len = 1L, any.missing = FALSE)
  assert_integerish(log_days, len = 1L, lower = 1L, any.missing = FALSE)
  assert_choice(timezone, OlsonNames())
  assert_logical(get_files, len = 1L, any.missing = FALSE)
  assert_logical(get_file_repository, len = 1L, any.missing = FALSE)
  assert_logical(original_file_names, len = 1L, any.missing = FALSE)
  assert_logical(add_default_datasets, len = 1L, any.missing = FALSE)
  # assert redcap_uri
  original_details <- NULL
  if (!hard_reset) {
    projects <- get_projects()
    in_proj_cache <- project_name %in% projects$project_name
    was_loaded <- FALSE
    if (!config$allow.test.names()) {
      if (project_name %in% TEST_PROJECT_NAMES) {
        project <- load_test_project(project_name = project_name,
                                     dir_path = dir_path)$.internal
        was_loaded <- TRUE
      }
    }
    if (!was_loaded && in_proj_cache) {
      project <- try_else_null({
        suppressWarnings({ # will ignore supplied dir_path ? add compare
          load_project(project_name = project_name)$.internal
        })
      })
      was_loaded <- !is.null(project)
    }
    if (!was_loaded && !missing_dir_path) {
      project <- try_else_null({
        suppressWarnings({ # will ignore supplied dir_path ? add compare
          load_project_from_dir(project_name = project_name,
                                dir_path = dir_path,
                                validate = TRUE)
        })
      })
      was_loaded <- !is.null(project)
    }
    # attempt to load existing project unless hard_reset
    if (!was_loaded) {
      project <- BLANK_PROJECT
      cli_alert_warning("Setup blank project. Unable to find, load, or repair.")
    }
    if (was_loaded) {
      # compare current setting to previous settings...
      original_details <- extract_project_details(project)
      if (missing(redcap_uri)) {
        redcap_uri <- original_details$redcap_uri
      }
      if (original_details$redcap_uri != redcap_uri) {
        stop_message <- paste0("There is an existing project at your chosen",
                               " directory with same `project_name` but a",
                               " different `redcap_uri`. You can use",
                               " setup_project(..., hard_reset = TRUE) to",
                               " override.")
        cli_abort(stop_message)
      }
      if (in_proj_cache) {
        project_row <- which(projects$project_name == project_name)
        cache_details <- projects[project_row, ]
        if (original_details$project_id != cache_details$project_id) {
          stop_message <- paste0("There is an existing project at your chosen",
                                 " directory with same `project_name` but a",
                                 " different `project_id`. You can use",
                                 " setup_project(..., hard_reset = TRUE) to",
                                 " override.")
          cli_abort(stop_message)
        }
      }
      params <- list(
        id_position = id_position,
        labelled = labelled,
        get_type = get_type,
        records = records,
        fields = fields,
        forms = forms,
        events = events,
        filter_logic = filter_logic
      )
      for (setting_name in names(params)) {
        if (!hard_reset) {
          if (!is.null(project$settings[[setting_name]])) {
            original <- project$settings[[setting_name]]
            value <- params[[setting_name]]
            if (!identical(original, value)) {
              hard_reset <- TRUE
              cli_alert_warning(
                paste(
                  "The loaded project was `{setting_name} = {original}` and",
                  "youchose `{setting_name} = {toString(value)}`. Therefore, a",
                  "full update was triggered."
                )
              )
            }
          }
        }
      }
    }
  }
  if (hard_reset) {
    # load blank if hard_reset = TRUE
    assert_project_name(project_name)
    project <- BLANK_PROJECT
    cli_alert_info("Setup blank project because `hard_reset = TRUE`")
  }
  if (missing_dir_path) {
    if (!is_something(project$dir_path)) {
      cli_alert_warning(missing_dir_message)
    }
  }
  project$project_name <- project_name
  if (!missing_dir_path) {
    # will also ask user if provided dir is new or different
    # (will load from original but start using new dir)
    project$dir_path <- set_dir(dir_path, silent = TRUE)
    dir.create(
      path = file.path(project$dir_path, "REDCap", project_name),
      showWarnings = FALSE
    )
  }
  project$token_name <- token_name
  project$redcap$timezone <- timezone
  project$settings$sync_frequency <- sync_frequency
  project$settings$labelled <- labelled
  project$settings$get_type <- get_type
  project$settings$records <- records
  project$settings$fields <- fields
  project$settings$forms <- forms
  project$settings$events <- events
  project$settings$filter_logic <- filter_logic
  project$settings$id_position <- id_position
  project$settings$get_data <- get_data
  project$settings$batch_size_download <- batch_size_download
  project$settings$batch_size_upload <- batch_size_upload
  project$settings$get_entire_log <- get_entire_log
  project$settings$log_days <- log_days
  project$settings$get_files <- get_files
  project$settings$get_file_repository <- get_file_repository
  project$settings$original_file_names <- original_file_names
  project$settings$add_default_datasets <- add_default_datasets
  #test these
  project$links$redcap_uri <- redcap_uri # add test, should end in / or add it
  project$links$redcap_base <- redcap_uri |> dirname() |> paste0("/")
  project$internals$version <- utils::packageVersion("REDCapSync")
  project$internals$is_blank <- FALSE
  project$internals$hard_reset <- hard_reset
  project$data <- all_character_cols_list(project$data)
  if (!is_valid_redcap_token(get_project_token(project))) {
    cli_alert_warning("No valid token in session: Sys.getenv('{token_name}')")
  }
  project <- assert_setup_project(project)
  # final_details ? compare if original_details not NULL
  invisible(REDCapSyncProject$new(project))
}
#' @noRd
SYNC_FREQUENCY <- c("always",
                    "hourly",
                    "daily",
                    "weekly",
                    "monthly",
                    "once",
                    "never")
#' @noRd
GET_TYPE <- c("identified",
              "deidentified",
              "deidentified_strict",
              "deidentified_super_strict")
#' @noRd
get_project_path <- function(project_name,
                             dir_path,
                             type = "",
                             check_dir = FALSE) {
  assert_env_name(project_name, max.chars = 31L, all_caps = TRUE)
  if (check_dir) {
    assert_dir(dir_path)
  }
  assert_choice(type, PROJECT_FILE_TYPES)
  file_name <- paste0(project_name, "_REDCapSync")
  if (type != "") {
    file_name <- paste0(file_name, "_", type)
  }
  file_name <- paste0(file_name, ".RData")
  file_path <- file.path(dir_path, "R_objects", file_name)
  sanitize_path(file_path)
}
#' @noRd
get_project_path2 <- function(project,
                              type = "",
                              check_dir = FALSE) {
  assert_setup_project(project)
  project_name <- project$project_name
  dir_path <- project$dir_path
  get_project_path(
    project_name = project_name,
    dir_path = dir_path,
    type = type,
    check_dir = check_dir
  )
}
#' @noRd
PROJECT_FILE_TYPES <- c("",
                        "details",
                        "metadata",
                        "data",
                        "data_default",
                        "data_flat",
                        "data_merge_non_repeating")
#' @noRd
PROJECT_PATH_SUFFIX <- "_REDCapSync.RData"
#' @rdname setup-load
#' @export
load_project <- function(project_name) {
  assert_project_name(project_name, allow_test_names = TRUE)
  projects <- get_projects()
  has_projects <- nrow(projects) > 0L
  project_in_cache <- project_name %in% projects$project_name
  project_is_test_project <- project_name %in% TEST_PROJECT_NAMES
  if (!has_projects && !project_is_test_project) {
    end_message <- paste0("No projects in cache. Use `setup_project(...) ",
                          " and `project$sync()` or try an offline test",
                          " project. \n\nTEST projects: ",
                          toString(TEST_PROJECT_NAMES))
    cli_abort(end_message)
  }
  project <- NULL
  if (project_in_cache) {
    project_row <- which(projects$project_name == project_name)
    dir_path <- sanitize_path(projects$dir_path[project_row])
    project <- load_project_from_dir(project_name = project_name,
                                     dir_path = dir_path,
                                     validate = TRUE)
  }
  if (!project_in_cache && !project_is_test_project) {
    end_message <- paste0("No project called {project_name} in cache.",
                          " Use `setup_project(...) and `project$sync()`?",
                          " or try an offline test. \n\nYour projects: ",
                          toString(projects$project_name),
                          "; \n\nTEST projects: ",
                          toString(TEST_PROJECT_NAMES))
    cli_abort(end_message)
  }
  if (is.null(project) && project_is_test_project) {
    project <- load_test_project(project_name)$.internal
  }
  invisible(REDCapSyncProject$new(project))
}
#' @noRd
load_project_from_dir <- function(project_name, dir_path, validate = TRUE) {
  assert_env_name(project_name, max.chars = 31L, all_caps = TRUE)
  assert_dir(dir_path, silent = TRUE)
  project_path <- get_project_path(project_name = project_name,
                                   dir_path = dir_path)
  assert_project_path(project_path)
  if (!file.exists(project_path)) {
    cli_abort(
      paste0(
        "No file at path {.file {project_path}}'",
        "'. Did you use {.fn REDCapSync::setup_project} and `project$sync()`?"
      )
    )
  }
  project <- readRDS(file = project_path)
  if (!validate) {
    return(project)
  }
  project <- repair_setup_project(project)
  if (is.null(project)) {
    abort_message <- paste0("Failed to load/repair project with proper",
                            " validation. This can happen with version",
                            " changes. You can still try to load the object",
                            " with `readRDS(\"{project_path}\")`. You should",
                            " re-run `project <- setup_project(...)`",
                            " and `project$sync()`.")
    cli_abort(abort_message)
  }
  project <- assert_setup_project(project)
  #   check if in cache already and relation!
  # SAVE
  loaded_dir <- sanitize_path(project$dir_path)
  if (!identical(dir_path, loaded_dir)) {
    cli_alert_warning(
      paste0("loaded dir_path did not match your cached dir_path. This should ",
             "only happen with cloud or shared directories.")
    )
  }
  project$dir_path <- dir_path
  project_details_path <- get_project_path(project_name = project_name,
                                           dir_path = dir_path,
                                           type = "details")
  if (file.exists(project_details_path)) { # will update timestamps
    saved_project_details <- readRDS(project_details_path)
    if (saved_project_details$project_name == project$project_name) {
      project$internals$last_sync <- saved_project_details$last_sync
      project$internals$last_directory_save <-
        saved_project_details$last_directory_save
    }
  }
  project |> extract_project_details() |> add_project_details_to_cache()
  the_message <- paste0("Loaded {project$project_name}: {.path {dir_path}}")
  if (due_for_sync(project$project_name)) {
    the_message <- paste0(the_message,
                          "\n... Due for sync. Run `project$sync()` to update.")
  }
  cli_alert_success(the_message)
  project
}
#' @noRd
load_test_project <- function(project_name, dir_path = NULL) {
  if (missing(project_name)) {
    cli_alert_info(paste(
      "TEST project choices:",
      toString(TEST_PROJECT_NAMES)
    ))
    project_name <- "TEST_CLASSIC"
    cli_alert_info("Defaulting to {project_name}")
  }
  assert_choice(project_name, TEST_PROJECT_NAMES)
  .test_projects <- list(
    TEST_CLASSIC = TEST_CLASSIC,
    TEST_REPEATING = TEST_REPEATING,
    TEST_LONGITUDINAL = TEST_LONGITUDINAL,
    TEST_MULTIARM = TEST_MULTIARM,
    TEST_EDGE = TEST_EDGE,
    TEST_DATA = TEST_DATA,
    TEST_CANCER = TEST_CANCER,
    TEST_REDCAPR_SIMPLE = TEST_REDCAPR_SIMPLE,
    TEST_REDCAPR_LONGITUDINAL = TEST_REDCAPR_LONGITUDINAL,
    TEST_REDCAPR_CLIN_TRIAL = TEST_REDCAPR_CLIN_TRIAL
  )
  project <- .test_projects[[project_name]]
  project <- repair_setup_project(project)
  assert_setup_project(project)
  if (!is.null(dir_path)) {
    project$dir_path <- set_dir(dir_path)
    dir.create(
      path = file.path(project$dir_path, "REDCap", project_name),
      showWarnings = FALSE
    )
    project <- add_default_datasets(project)
  } else {
    project$dir_path <- NA
  }
  project$internals$is_test <- TRUE
  cli_alert_success("Loaded TEST project {project$project_name}!")
  cli_alert_warning("Does not actually communicate with any REDCap API")
  invisible(REDCapSyncProject$new(project))
}
#' @noRd
save_project <- function(project, silent = FALSE) {
  assert_setup_project(project)
  if (!project$internals$ever_connected) {
    cli_alert_danger(
      paste0(
        "Did not save ",
        project$project_name,
        " because there has never been a REDCap connection! You must use ",
        "`setup_project(...)` and `project$sync()`"
      )
    )
    return(invisible(project))
  }
  project$internals$last_directory_save <- now_time()
  save_path <- get_project_path2(project = project)
  saveRDS(object = project, file = save_path) # add error check
  project_details <- extract_project_details(project)
  add_project_details_to_cache(project_details)
  saveRDS(
    object = project_details,
    file = get_project_path2(project, type = "details")
  ) # add error check
  if (!silent) {
    cli_alert_success("Saved {project$project_name}! {.file {save_path}}")
  }
  invisible(project)
}
#' @noRd
TEST_REDCAPR_NAMES <- c(
  "TEST_REDCAPR_SIMPLE",
  "TEST_REDCAPR_LONGITUDINAL",
  "TEST_REDCAPR_CLIN_TRIAL"
)
#' @noRd
TEST_PROJECT_NAMES <- c(
  "TEST_CLASSIC",
  "TEST_REPEATING",
  "TEST_LONGITUDINAL",
  "TEST_MULTIARM",
  "TEST_EDGE",
  "TEST_DATA",
  "TEST_CANCER",
  TEST_REDCAPR_NAMES
)
#' @noRd
BLANK_PROJECT <- list(
  project_name = NULL,
  dir_path = NULL,
  token_name = NULL,
  redcap = list(
    project_id = NULL,
    project_title = NULL,
    version = NULL,
    project_info = NULL,
    log = NULL,
    comments = NULL,
    users = NULL,
    current_user = NULL,
    timezone = NULL,
    has_log_access = NULL,
    has_dag_access = NULL,
    has_user_access = NULL,
    has_file_repository_access = NULL
  ),
  metadata = list(
    forms = NULL,
    fields = NULL,
    choices = NULL,
    id_col = NULL,
    form_key_cols = NULL,
    arms = NULL,
    events = NULL,
    event_mapping = NULL,
    missing_codes = NULL,
    raw_structure_cols = NULL,
    is_longitudinal = NULL,
    has_arms = NULL,
    has_multiple_arms = NULL,
    has_arms_that_matter = NULL,
    has_repeating_forms_or_events = NULL,
    has_repeating_forms = NULL,
    has_repeating_events = NULL
  ),
  data = NULL,
  record_summary = NULL,
  transformation = list(
    forms = NULL,
    fields = NULL,
    field_functions = NULL,
    data_updates = NULL
  ),
  datasets = list(),
  jobs = list(),
  links = list(
    redcap_uri = NULL,
    redcap_base = NULL,
    redcap_home = NULL,
    redcap_record_home = NULL,
    redcap_record_subpage = NULL,
    redcap_records_dashboard = NULL,
    redcap_api = NULL,
    redcap_api_playground = NULL,
    redcap_setup = NULL,
    redcap_user_rights = NULL,
    redcap_logging = NULL,
    redcap_designer = NULL,
    redcap_codebook = NULL,
    redcap_dictionary = NULL,
    redcap_data_quality = NULL,
    redcap_identifiers = NULL,
    cran = "",
    help = PKGDOWN_LINK,
    github = GITHUB_LINK,
    thecodingdocs = "https://www.thecodingdocs.com/"
  ),
  settings = list(
    sync_frequency = "daily",
    labelled = TRUE,
    get_type = "identified",
    records = NA,
    fields = NA,
    forms = NA,
    events = NA,
    filter_logic = NA,
    id_position = NULL,
    get_users = TRUE,
    get_data = TRUE,
    batch_size_download = 1000L,
    batch_size_upload = 500L,
    get_entire_log = FALSE,
    log_days = 10L,
    log_drop_details = FALSE,
    log_drop_exports = FALSE,
    get_files = FALSE,
    get_file_repository = FALSE,
    original_file_names = FALSE,
    add_default_datasets = TRUE
  ),
  internals = list(
    version = NULL,
    last_test_connection_attempt = NULL,
    last_test_connection_timezone = NULL,
    last_test_connection_outcome = NULL,
    last_metadata_update = NULL,
    last_metadata_dir_save = NULL,
    last_full_update = NULL,
    last_data_update = NULL,
    last_data_dir_save = NULL,
    last_data_transformation = NULL,
    last_dataset_save = NULL,
    last_quality_check = NULL,
    last_clean = NULL,
    last_directory_save = NULL,
    last_sync = NULL,
    project_type = "redcap",
    is_blank = TRUE,
    is_test = FALSE,
    ever_connected = FALSE,
    is_clean = FALSE,
    hard_reset = FALSE,
    was_updated = FALSE
  )
)
#' @noRd
set_dir <- function(dir_path, silent = FALSE) {
  dir_path <- clean_dir_path(dir_path)
  if (!file.exists(dir_path)) {
    cli_abort("Path not found. Use absolute path or choose existing directory.")
  }
  for (folder in DIR_FOLDERS) {
    if (!file.exists(file.path(dir_path, folder))) {
      dir.create(file.path(dir_path, folder), showWarnings = FALSE)
    }
  }
  assert_dir(dir_path, silent = silent)
}
#' @noRd
DIR_FOLDERS <- c("R_objects", "output", "scripts", "input", "REDCap")
#' @noRd
clean_dir_path <- function(dir_path) {
  if (!is.character(dir_path))
    cli_abort("dir must be a character string")
  dir_path <- dir_path |>
    trimws(whitespace = WHITESPACE) |>
    sanitize_path()
  dir_path
}
#' @noRd
WHITESPACE <- "[\\h\\v]"

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# load_project (Exported)
test_that("load_project works!", {
  expect_error(load_project("NO_PROJECTS_YET"))
  a <- mock_test_project("TEST_CLASSIC")$.internal |> extract_project_details()
  b <- mock_test_project("TEST_DATA")$.internal |> extract_project_details()
  c <- mock_test_project("TEST_LONGITUDINAL")$.internal |>
    extract_project_details()
  projects <- a |> bind_rows(b) |> bind_rows(c)
  local_mocked_bindings(
    get_projects = function(...) projects
  )
  expect_error(load_project("NO_PROJECTS_NAMED_THIS")) # no cache
  expect_error(load_project("TEST_CLASSIC")) # no file
})
# load_test_project (Exported)
test_that("load_test_project works!", {
})
# setup_project (Exported)
test_that("setup_project creates a valid project object and valid directory", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  withr::local_options(c(redcapsync.config.allow.test.names = TRUE))
  expect_error(assert_dir(dir_path = tempdir_test))
  project_name <- "TEST_CLASSIC"
  redcap_uri <- "https://redcap.miami.edu/api/"
  # test_short_names
  expect_error(assert_env_name("A project"))
  expect_error(assert_env_name("project$]"))
  expect_error(assert_env_name("1"))
  expect_error(assert_env_name(1L))
  expect_error(assert_env_name(another_name))
  expect_no_error(assert_env_name("expected_name"))
  expect_no_error(assert_env_name("expected_name2"))
  expect_no_error(assert_env_name("EXPECTED_NAME"))
  # test_redcap_base
  expect_error(assert_web_link("https://redcap.blah"))
  expect_no_error(assert_web_link("https://redcap.miami.edu/"))
  expect_no_error(assert_web_link("https://redcap.miami.edu"))
  expect_no_error(assert_web_link("https://redcap.edu"))
  # test db
  expect_no_error(assert_blank_project(BLANK_PROJECT))
  expect_error(assert_setup_project(BLANK_PROJECT))
  expect_error(assert_blank_project(BLANK_PROJECT[[-1L]]))
  expect_error(assert_blank_project(1L))
  expect_error(assert_blank_project(data.frame()))
  expect_error(assert_dir(tempdir_test))
  # Run setup_project
  project <- setup_project(
    project_name = project_name,
    dir_path = tempdir_test,
    redcap_uri = redcap_uri,
    hard_reset = TRUE
  )$.internal # change to R6 later
  expect_no_error(assert_dir(dir_path = tempdir_test))
  expect_no_error(assert_blank_project(project = project))
  check_dir <- assert_dir(project$dir_path)
  expect_identical(tempdir_test, check_dir)
  expect_type(project, type = "list")
  expect_named(project)
  expect_true("project_name" %in% names(project))
  expect_true("dir_path" %in% names(project))
  expect_false(project$internals$is_blank)
  expect_false(project$internals$is_test)
  expect_directory_exists(project$dir_path)
  expect_identical(project$project_name, project_name)
  test_dir_files <- list.files(tempdir_test)
  expect_true(all(DIR_FOLDERS %in% test_dir_files))
  project$dir_path <- file.path(tempdir_test, "another_fake_folder") |>
    sanitize_path()
  expect_error(assert_dir(project$dir_path))
})
test_that("setup_project checks existing dir", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  withr::local_options(c(redcapsync.config.allow.test.names = TRUE))
  expect_error(assert_dir(dir_path = tempdir_test))
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  save_project(project)
  redcap_uri <- "https://redcap.miami.edu/api/"
  # Run setup_project
  expect_error(
    setup_project(
      project_name = project_name,
      dir_path = tempdir_test,
      redcap_uri = redcap_uri,
      hard_reset = FALSE
    )$.internal # change to R6 later
  )
  redcap_uri <- project$links$redcap_uri
  project_loaded <- setup_project(
    project_name = project_name,
    dir_path = tempdir_test,
    redcap_uri = redcap_uri,
    hard_reset = FALSE
  )$.internal # change to R6 later
  expect_identical(project_loaded$project_name, project$project_name)
  expect_identical(project_loaded$redcap$project_id, project$redcap$project_id)
  expect_message({
    project_loaded <- setup_project(
      project_name = project_name,
      dir_path = tempdir_test,
      redcap_uri = redcap_uri,
      labelled = FALSE,
      hard_reset = FALSE
    )$.internal
  }, "The loaded project was")
  expect_null(project_loaded$internals$last_data_update)
})
# clean_dir_path (Internal)
test_that("clean_dir_path works!", {
})
# get_project_path (Internal)
test_that("get_project_path works!", {
})
# get_project_path2 (Internal)
test_that("get_project_path2 works!", {
})
# load_project_from_dir (Internal)
test_that("load_project_from_dir works!", {
})
# save_project (Internal)
test_that("save_project works!", {
  tempdir_file <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  redcap_uri <- "https://redcap.miami.edu/api/"
  project$links$redcap_uri <- redcap_uri
  expected_save_location <- file.path(project$dir_path,
                                      "R_objects",
                                      paste0(project_name, "_REDCapSync.RData"))
  expect_false(file.exists(expected_save_location))
  project$internals$ever_connected <- TRUE
  fake_time <- now_time()
  project$internals$last_sync <- fake_time
  project$internals$last_directory_save <- fake_time
  project$internals$last_data_update <- fake_time
  project$internals$last_metadata_update <- fake_time
  project$redcap$timezone <- Sys.timezone()
  expect_no_error({
    project <- save_project(project)
  })
  expect_file_exists(expected_save_location)
  # check cached proj
  projects <- get_projects()
  expect_identical(nrow(get_projects()), 1L)
  expect_identical(projects$project_name, project_name)
  expect_identical(projects$redcap_uri, redcap_uri)
  # expect identical projects dir_path tempdir_file
  # loading tests wont load unknown project
  expect_error(load_project("a_project"))
  # loads what we saved
  # change to R6 later
  project2 <- load_project(project_name = project_name)$.internal
  expect_identical(project, project2)
})
# set_dir (Internal)
test_that("set_dir handles existing directory correctly", {
  dir_path <- file.path(tempdir_file, "existing_dir")
  dir.create(dir_path)
  expect_message(set_dir(dir_path), "Directory is Valid!")
  expect_directory_exists(dir_path)
  expect_true(all(DIR_FOLDERS %in% list.files(dir_path)))
})
test_that("set_dir throws an error for invalid directory path", {
  expect_error(set_dir(123L), "dir must be a character string")
})
test_that("set_dir creates missing internal directories", {
  dir_path <- file.path(tempdir_file, "partial_dir")
  dir.create(dir_path)
  dir.create(file.path(dir_path, "R_objects"))
  expect_message(set_dir(dir_path), "Directory is Valid!")
  expect_directory_exists(dir_path)
  expect_true(all(DIR_FOLDERS %in% list.files(dir_path)))
})
test_that("set_dir stops if user chooses not to create directory", {
  dir_path <- file.path(tempdir_file, "no_create_dir")
  # Mock user input to not create the directory
  expect_error(set_dir(dir_path), "Path not found. Use absolute path")
  expect_false(file.exists(dir_path))
})
test_that("set_dir validates the directory structure", {
  dir_path <- file.path(tempdir_file, "valid_dir")
  dir.create(dir_path)
  for (folder in DIR_FOLDERS) {
    dir.create(file.path(dir_path, folder))
  }
  expect_message(set_dir(dir_path), "Directory is Valid!")
  expect_directory_exists(dir_path)
  expect_true(all(DIR_FOLDERS %in% list.files(dir_path)))
})

#' @title Manage REDCapSync projects
#' @description
#' `projects` is the central interface for managing REDCap projects in
#' **REDCapSync**. It provides a convenient, persistent registry of projects
#' that can be accessed across R sessions.
#'
#' Each time a project is set up or synced, basic metadata is stored locally so
#' that projects can be reloaded and updated without reconfiguration.
#'
#' @details
#' `projects` is implemented as a singleton list object and serves as a single
#' entry point to the REDCapSync workflow. All project-level operations—such as
#' setup, loading, syncing, and removal—are accessed through this object.
#'
#' A key advantage of this design is support for **method chaining**, even
#' without loading the namespace. Because many project methods return project
#' objects, you can write concise, readable workflows that operate in sequence:
#'
#' \preformatted{
#' # load, sync, and generate dataset to global
#' REDCapSync::projects$load("TEST_CLASSIC")$
#'   sync()$
#'   generate_dataset(envir = globalenv(),
#'                    filter_field = "var_branching",
#'                    filter_choices = "Yes")
#' }
#' @section Methods:
#'
#' \describe{
#'
#' \item{`print()`}{
#' Display a summary of all registered projects.
#'
#' - Shows total number of projects
#' - Lists project names
#' - Indicates how many are due for sync
#'
#' Returns the project data frame invisibly.
#' }
#'
#' \item{`any()`}{
#' Check whether any projects are registered.
#'
#' Returns a logical scalar.
#' }
#'
#' \item{`n()`}{
#' Return the number of registered projects.
#'
#' Returns an integer.
#' }
#'
#' \item{`names()`}{
#' Return the names of registered projects.
#'
#' Returns a character vector.
#' }
#'
#' \item{`test_names()`}{
#' Return the names of test projects.
#'
#' Returns a character vector.
#' }
#'
#' \item{`df()`}{
#' Retrieve the project registry as a data frame.
#'
#' Returns a data frame containing stored project metadata.
#' }
#'
#' \item{`load(project_name)`}{
#' Load a project by name. This is a wrapper for [load_project].
#'
#' \strong{Arguments}
#' \itemize{
#'   \item `project_name`: Character. Name of the project.
#' }
#'
#' Returns a `REDCapSyncProject` object.
#' }
#'
#' }
#' @seealso
#' \code{vignette("REDCapSync", package = "REDCapSync")}
#' \code{vignette("Cache", package = "REDCapSync")}
#' \code{vignette("RosyREDCap", package = "REDCapSync")}
#' [setup_project] for initializing projects
#' [project] for using the project objects
#'
#' @examples
#' project_df <- projects$df()
#' projects$print()
#' projects$any()
#' projects$n()
#' projects$names()
#' project <- projects$load("TEST_CLASSIC")
#' @format A named list of functions for managing REDCapSync projects.
#' @export
projects <- list(
  print = function() {
    project_df <- get_projects()
    n_projects <- nrow(project_df)
    cli_h1("REDCapSync")
    cli_h2("Your Projects (n = {n_projects})")
    if (n_projects > 0L) {
      project_df$due_for_sync <- project_df$project_name |>
        lapply(due_for_sync) |>
        unlist()
      number_due <- project_df$due_for_sync |> which() |> length()
      cli_text("{toString(project_df$project_name)}")
      cli_alert_info("{number_due} due for sync!")
    }
    cli_h2("Test Projects (n = {length(TEST_PROJECT_NAMES)})")
    cli_text("{toString(TEST_PROJECT_NAMES)}")
    help_cli_text()
    invisible(project_df)
  },
  any = function() {
    nrow(get_projects()) > 0L
  },
  n = function() {
    nrow(get_projects())
  },
  names = function() {
    get_projects()$project_name
  },
  test_names = function() {
    TEST_PROJECT_NAMES
  },
  df = function() {
    get_projects()
  },
  load = function(project_name) {
    load_project(project_name)
  }
)
#' @noRd
get_projects <- function() {
  does_exist <- FALSE
  if (file.exists(cache_path())) {
    does_exist <- cache_path() |> file.path("projects.rds") |> file.exists()
    if (!does_exist) {
      cli_alert_warning("No cached projects... use `setup_project(...)`")
    }
  }
  does_exist
  is_ok <- FALSE
  if (does_exist) {
    projects <- cache_path() |> file.path("projects.rds") |> readRDS()
    # add try catch?
    is_ok <- test_project_details(projects)
    if (!is_ok) {
      projects <- repair_projects(projects)
      is_ok <- !is.null(projects)
    }
    if (!is_ok) {
      warning_message <- paste0("You have projects cached, but due to a ",
                                "version change or other issue, it has to be ",
                                "cleared. Use `setup_project(...)`")
      cli_alert_warning(warning_message)
      cache_clear()
    }
  }
  if (!is_ok) {
    # message here?
    return(BLANK_PROJECT_DETAILS)
  }
  # message here? would have to silence internally
  projects # invisible?
}
#' @noRd
BLANK_PROJECT_COLS <- c(
  "project_name",
  "dir_path",
  "sync_frequency",
  "last_sync",
  "timezone",
  "last_directory_save",
  "last_metadata_update",
  "last_data_update",
  "version",
  "token_name",
  "project_id",
  "project_title",
  "labelled",
  "id_col",
  "is_longitudinal",
  "has_repeating_forms_or_events",
  "has_multiple_arms",
  "object_size",
  # "file_size",
  "n_records",
  "redcap_uri",
  # "redcap_base",
  "redcap_home",
  # "redcap_api_playground",
  # "log_days",
  "get_files",
  "get_file_repository",
  # "original_file_names",
  "get_entire_log",
  "get_data",
  "get_type"
)
#' @noRd
BLANK_PROJECT_DETAILS <- data.frame(
  project_name = character(0L),
  dir_path = character(0L),
  sync_frequency = character(0L),
  last_sync = character(0L) |> as.POSIXct(tz = Sys.timezone()),
  timezone = character(0L),
  last_directory_save = character(0L) |> as.POSIXct(tz = Sys.timezone()),
  last_metadata_update = character(0L) |> as.POSIXct(tz = Sys.timezone()),
  last_data_update = character(0L) |> as.POSIXct(tz = Sys.timezone()),
  version = character(0L),
  token_name = character(0L),
  project_id = character(0L),
  project_title = character(0L),
  labelled = logical(0L),
  id_col = character(0L),
  is_longitudinal = logical(0L),
  has_repeating_forms_or_events = logical(0L),
  has_multiple_arms = logical(0L),
  object_size = character(0L),
  n_records = integer(0L),
  redcap_uri = character(0L),
  redcap_home = character(0L),
  get_files = logical(0L),
  get_file_repository = logical(0L),
  get_entire_log = logical(0L),
  get_data = logical(0L),
  get_type = character(0L),
  stringsAsFactors = FALSE
)
#' @noRd
save_projects_to_cache <- function(projects, silent = TRUE) {
  assert_project_details(projects)
  new_order <- order(projects$project_name)
  projects <- projects[new_order, ]
  saveRDS(projects, file = cache_path() |> file.path("projects.rds"))
  pkg_name <- "REDCapSync"
  if (!silent) {
    cli_alert_success(
      paste0(
        pkg_name,
        " saved ",
        nrow(projects),
        " project locations to the cache...",
        toString(projects$project_name)
      ) # "   Token: ",projects$token_name,collapse = "\n"))
    )
  }
}
#' @noRd
extract_project_details <- function(project) {
  assert_setup_project(project)
  project_details <- matrix(
    data = NA,
    ncol = length(BLANK_PROJECT_COLS),
    nrow = 1L,
    dimnames = list(NULL, BLANK_PROJECT_COLS)
  ) |> as.data.frame()
  # top -----
  project_details$project_name <- project$project_name
  project_details$dir_path <- na_if_null(project$dir_path)
  # settings -------
  project_details$sync_frequency <- project$settings$sync_frequency
  project_details$get_files <- project$settings$get_files
  project_details$get_file_repository <- project$settings$get_file_repository
  project_details$get_entire_log <- project$settings$get_entire_log
  project_details$get_data <- project$settings$get_data
  project_details$labelled <- project$settings$labelled
  project_details$get_type <- project$settings$get_type
  # redcap --------
  project_details$version <- na_if_null(project$redcap$version)
  project_details$token_name <- na_if_null(project$token_name)
  project_details$project_id <- na_if_null(project$redcap$project_id)
  project_details$project_title <- na_if_null(project$redcap$project_title)
  project_details$id_col <- na_if_null(project$metadata$id_col)
  project_details$is_longitudinal <-
    na_if_null(project$metadata$is_longitudinal)
  project_details$has_repeating_forms_or_events <-
    na_if_null(project$metadata$has_repeating_forms_or_events)
  project_details$has_multiple_arms <-
    na_if_null(project$metadata$has_multiple_arms)
  project_details$n_records <- project$record_summary |>
    nrow() |>
    na_if_null() |>
    as.integer()
  project_details$redcap_uri <- project$links$redcap_uri
  project_details$redcap_home <- project$links$redcap_home |>
    na_if_null() |>
    as.character()
  project_details$timezone <- project$redcap$timezone |>
    na_if_null() |>
    as.character()
  # saving ----
  project_details$last_sync <- project$internals$last_sync |>
    na_if_null() |>
    as.POSIXct(tz = Sys.timezone())
  project_details$last_directory_save <-
    project$internals$last_directory_save |>
    na_if_null() |>
    as.POSIXct(tz = Sys.timezone())
  project_details$last_metadata_update <-
    project$internals$last_metadata_update |>
    na_if_null() |>
    as.POSIXct(tz = Sys.timezone())
  project_details$last_data_update <- project$internals$last_data_update |>
    na_if_null() |>
    as.POSIXct(tz = Sys.timezone())
  project_details$object_size <- object_size(project) |> as.character()
  project_details
}
#' @noRd
add_project_details_to_cache <- function(project_details) {
  assert_project_details(project_details, nrows = 1L)
  projects <- get_projects()
  proj_row_diff <- which(projects$project_name != project_details$project_name)
  projects <- projects[proj_row_diff, ]
  bad_row <- NULL
  if (!is.na(project_details$project_id)) {
    bad_row <- which(
      projects$project_id == project_details$project_id &
        dirname(projects$redcap_uri) == dirname(project_details$redcap_uri)
    )
  }
  if (length(bad_row) > 0L) {
    cli_abort(
      paste0(
        "You are trying to save from a project [{project_details$project_name}",
        " PID {projects$project_id[bad_row]}] that you have already setup ",
        "[{projects$project_name[bad_row]} PID {project_details$project_id}] ",
        "You can load the old project or run ",
        "`cache_clear(\"{projects$project_name[bad_row]}\")`"
      )
    )
  }
  projects <- bind_rows(projects, project_details)
  save_projects_to_cache(projects)
}
#' @noRd
help_cli_text <- function() {
  cli_h2("Help")
  cli_text("Pkgdown: {.url {PKGDOWN_LINK}}")
  cli_text("Github: {.url {GITHUB_LINK}}")
  cli_text("Datasets: {.vignette REDCapSync::Datasets}")
  cli_text("Tokens: {.vignette REDCapSync::Tokens}")
}
#' @noRd
PKGDOWN_LINK <- "https://thecodingdocs.github.io/REDCapSync/"
#' @noRd
GITHUB_LINK <- "https://github.com/thecodingdocs/REDCapSync/"

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# get_projects (Exported)
test_that("get_projects works!", {
  df <- get_projects()
  expect_s3_class(df, "data.frame")
  expect_true(all(colnames(df) %in% BLANK_PROJECT_COLS))
  df <- BLANK_PROJECT_DETAILS
  expect_s3_class(df, "data.frame")
  expect_true(all(colnames(df) %in% BLANK_PROJECT_COLS))
  # can test after adding project
})
# add_project_details_to_cache (Internal)
test_that("add_project_details_to_cache works!", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  # start empty
  projects <- get_projects()
  expect_data_frame(projects, nrows = 0L)
  # add project details to cache
  project <- mock_test_project()$.internal
  project_details <- extract_project_details(project)
  expect_no_error(add_project_details_to_cache(project_details))
  projects <- get_projects()
  expect_data_frame(projects, nrows = 1L)
  expect_true(project_details$project_name %in% projects$project_name)
  # conflict: same project_id & same base redcap_uri
  #but different project_name -> error
  project_details_conflict <- project_details
  project_details_conflict$project_name <- "TEST_OTHER"
  expect_error(add_project_details_to_cache(project_details_conflict))
})
# extract_project_details (Internal)
test_that("extract_project_details works!", {
  project <- mock_test_project()$.internal
  project_details <- extract_project_details(project)
  expect_data_frame(project_details,
                    ncols = ncol(BLANK_PROJECT_DETAILS),
                    nrows = 1L)
  expect_identical(colnames(project_details),
                   colnames(BLANK_PROJECT_DETAILS))
})
# save_projects_to_cache (Internal)
test_that("save_projects_to_cache works!", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  cache_location <- cache_path()
  projects <- get_projects()
  expect_data_frame(projects, nrows = 0L)
  project <- mock_test_project()$.internal
  #no connection should not save to cache
  project_details <- extract_project_details(project)
  projects <- get_projects()
  expect_data_frame(projects, nrows = 0L)
  project$internals$ever_connected <- TRUE
  save_project(project = project)
  projects <- get_projects()
  #connection should save to cache
  expect_data_frame(projects, nrows = 1L)
  expect_identical(project_details$project_name, projects$project_name)
  expect_data_frame(project_details, nrows = 1L)
  cache_clear()
  expect_false(file.exists(file.path(cache_location, "projects.rds")))
  projects <- get_projects()
  expect_data_frame(projects, nrows = 0L)
  expect_message(save_projects_to_cache(project_details, silent = FALSE),
                 "saved")
  expect_file_exists(file.path(cache_location, "projects.rds"))
  expect_no_message(save_projects_to_cache(project_details, silent = TRUE))
})

#' @title REDCapSync Project Object
#' @description
#' [R6][R6::R6Class] project object for managing REDCap data access and
#' synchronization. The project object is your main interface to REDCap data
#' through the [REDCapSync] package. It stores your REDCap configuration, data,
#' metadata, and provides methods to sync, transform, and export your data.
#'
#' @details
#' ## Workflow
#'
#' **Initialize** a project with [setup_project()]:
#' ```r
#' project <- setup_project(
#'   project_name = "FIRST_PROJECT",
#'   redcap_uri = "https://redcap.yourinstitution.edu/api/",
#'   dir_path = "~/redcap_projects"
#' )
#' ```
#'
#' **Synchronize** REDCap data into your project object using the REDCap API
#' and log-based change detection:
#' ```r
#' project$sync()
#' ```
#'
#' **Load** previously saved REDCap project object using cache:
#' ```r
#' project <- load_project("FIRST_PROJECT)
#' projects$names() # should see "FIRST_PROJECT"
#' ```
#'
#' **Access** REDCap data, metadata, and users via read-only fields:
#' ```r
#' project$data           # Named list of REDCap forms/instruments
#' project$metadata       # REDCap field definitions, forms, and choices
#' project$redcap$users   # REDCap user information
#' project$redcap$log     # REDCap log information
#'
#' # you can bypass read-only if needed
#' users <- project$redcap$users
#' ```
#'
#' **Transform and export** your data into clean, analysis-ready datasets to be
#' used in R and/or Excel:
#' ```r
#' project$add_dataset("analysis_data", ...) # adds to project for re-use
#' project$save_datasets() # can save to Excel but is also part of future syncs
#' ```
#'
#' **Upload** corrected or new data back to REDCap:
#' ```r
#' project$upload(updated_data)
#' ```
#'
#' ## Design Features
#'
#' The project object uses **log-based sync**: Since REDCap maintains a detailed
#' change log, `project$sync()` only retrieves and updates records that have
#' changed since the last sync. This dramatically reduces API calls and improves
#' performance for large projects.
#'
#' All methods use **method chaining**: Methods invisibly return the project
#' object (self), allowing fluent code:
#' ```r
#' dataset <- load_project("TEST_CLASSIC")$sync()$load_dataset("REDCapSync")
#' ```
#'
#' ## Read-Only Fields
#'
#' - `project$project_name`: Project identifier (set with [setup_project])
#' - `project$dir_path`: Persistent storage directory (set with [setup_project])
#' - `project$data`: Named list of synchronized REDCap forms
#' - `project$metadata`: REDCap field metadata (forms, fields, choices)
#' - `project$redcap`: REDCap project info, users, and activity log
#' - `project$.internal`: Internal project object (for advanced use)
#'
#' ## TEST Projects
#'
#' All TEST projects start with "TEST_" and by default the average user cannot
#' create a project that starts with "TEST_". They are produced from actual
#' server REDCap projects but never contained real data and were scrubbed of
#' any "real" user/log data. They are subject to change in future versions as
#' the package matures. Keep in mind there are many combinations of REDCap
#' structures to account for.
#'
#' TEST projects are meant to be used for demonstration and testing
#' purposes. In general, try to use actual REDCap project(s) to explore the
#' package.
#'
#' This is how to load a project with or without specifying a directory.
#' ```r
#' # return the list of available test projects
#' REDCapSync::projects$test_names()
#'
#' project <- load_project("TEST_CLASSIC")
#'
#' YOUR_DIRECTORY <- getwd()
#' project <- setup_project("TEST_CLASSIC", dir_path = YOUR_DIRECTORY)
#' # now can test object functions (does not use API)
#' project$sync() # will also save datasets
#' ```
#'
#' The currently available TEST projects are as follows:
#' - `TEST_CLASSIC`: Classic project (nothing repeats). Has every data type.
#' Contains data for survival analysis testing. Contains comments.
#' - `TEST_REPEATING`: Non-longitudinal project with repeating instruments.
#' - `TEST_LONGITUDINAL`: Longitudinal project (with events).
#' - `TEST_MULTIARM`: Multi-arm and longitudinal project.
#' - `TEST_EDGE`: Edge-case project. Meant to test things that may be uncommon
#' but should be accounted for as package matures.
#' - `TEST_DATA`: Placeholder project meant to contain more rich test dataset.
#' - `TEST_CANCER`: Placeholder project meant to contain more rich test dataset.
#' - `TEST_REDCAPR_SIMPLE`: Borrowed with permission from [REDCapR].
#' - `TEST_REDCAPR_LONGITUDINAL`: Borrowed with permission from [REDCapR].
#' - `TEST_REDCAPR_CLIN_TRIAL`: Borrowed with permission from [REDCapR].
#'
#' @param dataset_name Character. Name of the dataset configuration to create,
#' load, or reference.
#' @param transformation_type Character. How to transform data: "default"
#' (merge non-repeating then add to repeating), "none" (no transformation), or
#' "merge_non_repeating" (merge non-repeating only). Default is "default".
#' @param merge_form_name Character. Name for the merged non-repeating form.
#' Default is "merged".
#' @param filter_field Character. Field name to filter dataset on.
#' @param filter_choices Vector. Allowed values for `filter_field`.
#' @param filter_list List. Named list where names are field names and values
#' are allowed value sets. Alternative to `filter_field`/`filter_choices`.
#' @param filter_strict Logical. If `TRUE`, filter all forms. If `FALSE`, apply
#' filter only to records (ID column). Default is `TRUE`.
#' @param form_names Character vector. Forms to include. Default `NULL` includes
#' all forms.
#' @param field_names Character vector. Fields to include. Default `NULL`
#' includes all fields.
#' @param exclude_identifiers Logical. Exclude identifier fields. Default is
#' `TRUE`.
#' @param exclude_free_text Logical. Exclude free-text fields (for
#' deidentification). Default is `FALSE`.
#' @param date_handling Character. Date handling strategy: "none",
#' "exclude_dates", "random_shift_by_record", "random_shift_by_project",
#' "zero_by_record", or "zero_by_project". Default is "none".
#' @param labelled Logical. Convert to labelled data if `TRUE`. Default is
#' `TRUE`.
#' @param clean Logical. Clean data (standardize missing values). Default is
#' `TRUE`.
#' @param drop_blanks Logical. Drop records with blank fields. Default is
#' `TRUE`.
#' @param drop_missing_codes Logical. Convert REDCap missing codes to `NA`.
#' Default is `FALSE`.
#' @param drop_others Character vector of additional values to drop.
#' @param include_metadata Logical. Include field metadata in dataset. Default
#' is `TRUE`.
#' @param include_users Logical. Include user information. Default is `TRUE`.
#' @param include_records Logical. Include record-level information. Default is
#' `TRUE`.
#' @param include_log Logical. Include REDCap activity log. Default is `FALSE`.
#' @param annotate_from_log Logical. Annotate data using the REDCap log. Default
#' is `TRUE`.
#' @param include_comments Logical. Include field comments. Default is `TRUE`.
#' @param hard_reset Logical. Overwrite existing dataset files. Default is
#' `FALSE`.
#' @param with_links Logical. Include hyperlinks in Excel exports. Default is
#' `FALSE`.
#' @param separate Logical. Separate each form into distinct files (vs.
#' multi-tab Excel). Default is `FALSE`.
#' @param use_csv Logical. Use CSV format instead of Excel. Default is `FALSE`.
#' @param dir_other Character. Output directory (default is project's `output`
#' folder).
#' @param file_name Character. Dataset file name (default is
#' `<project_name>_<dataset_name>`).
#' @param envir Environment to assign dataset objects. Default is `NULL`.
#' @param form Character. REDCap form/instrument name, e.g., "survey_one".
#' @param link_type Character. REDCap link type: "base", "home", "record_home",
#' "records_dashboard", "api", "api_playground", "codebook", "user_rights",
#' "setup", "logging", "designer", "dictionary", "data_quality", or
#' "identifiers".
#' @param open_browser Logical. Open link in browser. Default is `TRUE`.
#' @param record Character. Record ID.
#' @param page Character. REDCap form/instrument name.
#' @param instance Character. Repeating instance number.
#' @param save_to_dir Logical. Save updated project object to directory. Default
#' is `TRUE`.
#' @param save_datasets Logical. Save dataset outputs during sync. Default is
#' `TRUE`.
#' @param hard_check Logical. Force REDCap API check regardless of
#' `sync_frequency`. Default is `FALSE`.
#' @param type Character. Metadata type to retrieve: "fields", "forms", or
#' "choices".
#' @param dataset_names Character vector. Dataset names to remove/operate on.
#' Default is `NULL`.
#' @param to_be_uploaded Data frame in raw coded format ready to upload to
#' REDCap.
#' @param keyring_if_error Logical. Launch pop-up for keyring if token is not
#' valid or fails to call REDCap. Max 3 attempts. Default is `TRUE`.
#'
#' @examples
#' # Load a test project
#' projects$test_names() # available test projects
#' project <- setup_project("TEST_CLASSIC", dir_path = tempdir())
#'
#' # Sync data from REDCap
#' project$sync()
#'
#' # Access data and metadata
#' head(project$data$text)
#' project$metadata$fields[1:5, ]
#'
#' forms <- project$metadata$forms # unchanged metadata
#' fields <- project$metadata$fields # unchanged metadata
#' choices <- project$metadata$choices # unchanged metadata
#'
#' users <- project$redcap$users # unchanged users
#' log <- project$redcap$log # unchanged log
#'
#' project$test_token()
#'
#' dataset <- project$load_dataset("REDCapSync")
#'
#' @seealso
#' \code{vignette("REDCapSync", package = "REDCapSync")}
#' [projects] for shortcuts of cached setup projects
#' \code{vignette("Tokens", package = "REDCapSync")}
#' [setup_project] for initializing projects
#' [dataset] for using the dataset objects
#' \code{vignette("Datasets", package = "REDCapSync")}
#' \code{vignette("Uploads", package = "REDCapSync")}
#'
#' @returns
#' An R6 `REDCapSyncProject` class generator for internal use. Users interact
#' with instances created by [setup_project] or [load_project].
#' @name project
#' @rdname project
#' @export
REDCapSyncProject <- R6Class(
  "REDCapSyncProject",
  #' @description
  #' Active binding are read-only
  active = list(
    #' @field project_name Read-only character string of project_name as
    #' assigned with [setup_project].
    project_name = function(value) {
      if (!missing(value)) {
        message(
          "`project_name` is read only. To change `setup_project()`"
        )
      }
      private$project$project_name
    },
    #' @field dir_path Read-only directory path assigned with [setup_project].
    dir_path = function(value) {
      if (!missing(value)) {
        message(
          "`dir_path` is read only. To change `setup_project()`"
        )
      }
      private$project$dir_path
    },
    #' @field data Read-only named list where each name is an instrument name.
    #' See public methods for [REDCapSyncProject].
    data = function(value) {
      if (!missing(value)) {
        cli_alert_danger(
          paste0(
            "`data` is read only. To change REDCap data either use",
            "`project$upload()` or work with an output by assigning the data",
            "with `form_to_edit <-project$data$<form_name>`. Alternatively",
            "use the output from `project$generate_dataset()`"
          )
        )
      }
      private$project$data
    },
    #' @field metadata Read-only named list with REDCap metadata. See
    #' public methods for [REDCapSyncProject].
    metadata = function(value) {
      if (!missing(value)) {
        cli_alert_danger(
          paste(
            "`metadata` is read only. To change, do so on the",
            "REDCap website, or use the REDCap API. If you just want to work",
            "with the object more in R, reassign the object like,",
            "`fields <- project$metadata$fields`. Alternatively use the output",
            "from `project$generate_dataset()`. See REDCap Designer",
            "{.url {private$project$links$redcap_designer}}."
          )
        )
      }
      private$project$metadata
    },
    #' @field redcap Read-only named list with REDCap information including
    #' users and log.
    redcap = function(value) {
      if (!missing(value)) {
        cli_alert_danger(
          paste(
            "`redcap` is read only. It is generated from communication with",
            "REDCap. If you just want to work with the object more in R,",
            "reassign the object like, `redcap_log <- project$redcap$log`.",
            "Alternatively, use output from `project$generate_dataset()` See",
            "REDCap Designer {.url {private$project$links$redcap_designer}}."
          )
        )
      }
      private$project$redcap
    },
    #' @field .internal Read-only internal project object for custom workflows
    .internal  = function(value) {
      if (!missing(value)) {
        cli_alert_danger(
          "`.internal` is read only. Use public `REDCapSyncProject` methods"
        )
      }
      private$project
    }
  ),
  public = list(
    #' @description
    #' The end user will not see `project$new()`. This is handled internally.
    #' Users should construct objects using [setup_project()]. The remain
    #' methods will be accessible to any user.
    #' @param project a list object meant to be stored internally within R6
    initialize = function(project) {
      assert_setup_project(project)
      private$project <- project
      invisible(self)
    },
    #' @description Print some key project information
    print = function() {
      cli_h1("REDCapSyncProject")
      cli_text("Name: {private$project$project_name}")
      cli_text("Title: {private$project$redcap$project_info$project_title}")
      cli_text("PID: {private$project$redcap$project_id}")
      cli_text("Token Name: {private$project$token_name}")
      cli_text("Sync Frequency: {private$project$settings$sync_frequency}")
      cli_text("Last Update: {private$project$internals$last_data_update}")
      cli_text("REDCap: {.url {private$project$links$redcap_home}}")
      cli_text("Directory: {.file {private$project$dir_path}}")
      help_cli_text()
      invisible(self)
    },
    #' @description
    #' Updates the REDCap data for (`project` object) by checking REDCap log for
    #' changes. Sync is performed according to the `sync_frequency` set in
    #' [setup_project()] by default. Use `hard_check` to force a check, or `
    #' hard_reset` to force a complete refresh. As a default, this object will
    #' be saved to your directory when necessary.
    sync = function(save_datasets = TRUE,
                    save_to_dir = TRUE,
                    hard_check = FALSE,
                    hard_reset = FALSE) {
      if (private$project$internals$is_test) {
        cli_alert_info("TEST projects do not communicate with the API")
        if (save_datasets && save_to_dir) {
          private$project <- save_project_datasets(project = private$project,
                                                   hard_reset = hard_reset)
        }
        return(invisible(self))
      }
      private$project <- sync_project(project = private$project,
                                      save_datasets = save_datasets,
                                      save_to_dir = save_to_dir,
                                      hard_check = hard_check,
                                      hard_reset = hard_reset)
      invisible(self)
    },
    #' @description  Add a new dataset entry
    add_dataset = function(dataset_name,
                           transformation_type = "default",
                           merge_form_name = "merged",
                           filter_field = NULL,
                           filter_choices = NULL,
                           filter_list = NULL,
                           filter_strict = TRUE,
                           field_names = NULL,
                           form_names = NULL,
                           exclude_identifiers = FALSE,
                           exclude_free_text = FALSE,
                           date_handling = "none",
                           labelled = TRUE,
                           clean = TRUE,
                           drop_blanks = FALSE,
                           drop_missing_codes = FALSE,
                           drop_others = NULL,
                           include_metadata = TRUE,
                           include_records = TRUE,
                           include_users = TRUE,
                           include_log = FALSE,
                           annotate_from_log = TRUE,
                           include_comments = TRUE,
                           with_links = TRUE,
                           separate = FALSE,
                           use_csv = FALSE,
                           dir_other = NULL,
                           file_name = NULL,
                           hard_reset = FALSE) {
      private$project <- add_project_dataset(
        project = private$project,
        dataset_name = dataset_name,
        transformation_type = transformation_type,
        merge_form_name = merge_form_name,
        filter_field = filter_field,
        filter_choices = filter_choices,
        filter_list = filter_list,
        filter_strict = filter_strict,
        field_names = field_names,
        form_names = form_names,
        exclude_identifiers = exclude_identifiers,
        exclude_free_text = exclude_free_text,
        date_handling = date_handling,
        labelled = labelled,
        clean = clean,
        drop_blanks = drop_blanks,
        drop_missing_codes = drop_missing_codes,
        drop_others = drop_others,
        include_metadata = include_metadata,
        include_records = include_records,
        include_users = include_users,
        include_log = include_log,
        annotate_from_log = annotate_from_log,
        include_comments = include_comments,
        with_links = with_links,
        separate = separate,
        use_csv = use_csv,
        dir_other = dir_other,
        file_name = file_name,
        hard_reset = hard_reset
      )
      invisible(self)
    },
    #' @description  Load dataset if previously defined with `add_dataset`.
    load_dataset = function(dataset_name,
                            envir = NULL) {
      dataset_names <- names(private$project$datasets)
      assert_choice(dataset_name, dataset_names, null.ok = FALSE)
      assert_environment(envir, null.ok = TRUE)
      dataset <- REDCapSyncDataset$new(project = private$project,
                                       dataset_name = dataset_name)
      dataset$to_envir(envir = envir)
      invisible(dataset)
    },
    #' @description  Clear all or specified datasets from the `project` object.
    remove_datasets = function(dataset_names = NULL) {
      private$project <- clear_project_datasets(
        project = private$project,
        dataset_names = dataset_names
      )
      invisible(self)
    },
    #' @description  Generate [dataset] object. This is usually handled
    #' internally for some default behavior but is provided here for ad-hoc
    #' custom datasets.
    generate_dataset = function(dataset_name,
                                envir = NULL,
                                transformation_type = "default",
                                merge_form_name = "merged",
                                filter_field = NULL,
                                filter_choices = NULL,
                                filter_list = NULL,
                                filter_strict = TRUE,
                                field_names = NULL,
                                form_names = NULL,
                                exclude_identifiers = FALSE,
                                exclude_free_text = FALSE,
                                date_handling = "none",
                                labelled = TRUE,
                                clean = TRUE,
                                drop_blanks = FALSE,
                                drop_missing_codes = FALSE,
                                drop_others = NULL,
                                include_metadata = TRUE,
                                include_records = TRUE,
                                include_users = TRUE,
                                include_log = FALSE,
                                annotate_from_log = TRUE,
                                include_comments = FALSE) {
      assert_environment(envir, null.ok = TRUE)
      provided_dataset_name <- !missing(dataset_name)
      if (provided_dataset_name) {
        if (dataset_name %in% names(private$project$datasets)) {
          cli_alert_warning("{dataset_name} is already a defined dataset")
          cli_alert_info("It will be loaded... other paramers ignored")
        }
      } else {
        dataset_name <- "custom"
        i <- 0L
        while (dataset_name %in% names(private$project$datasets)) {
          i <- i + 1L
          dataset_name <- paste0("custom", i)
        }
        cli_alert_warning("`dataset_name` not provided: Using '{dataset_name}'")
      }
      dataset <- REDCapSyncDataset$new(
        project = private$project,
        dataset_name = dataset_name,
        transformation_type = transformation_type,
        merge_form_name = merge_form_name,
        filter_field = filter_field,
        filter_choices = filter_choices,
        filter_list = filter_list,
        filter_strict = filter_strict,
        field_names = field_names,
        form_names = form_names,
        exclude_identifiers = exclude_identifiers,
        exclude_free_text = exclude_free_text,
        date_handling = date_handling,
        labelled = labelled,
        clean = clean,
        drop_blanks = drop_blanks,
        drop_missing_codes = drop_missing_codes,
        drop_others = drop_others,
        include_metadata = include_metadata,
        include_users = include_users,
        include_records = include_records,
        include_log = include_log,
        annotate_from_log = annotate_from_log,
        include_comments = include_comments
      )
      dataset$to_envir(envir = envir)
      invisible(dataset)
    },
    #' @description saves datasets to Excel via setting provided to
    #' `add_dataset`.
    save_datasets = function(hard_reset = FALSE) {
      first_stamp <- private$project$internals$last_dataset_save
      private$project <- save_project_datasets(project = private$project,
                                               hard_reset = hard_reset)
      second_stamp <- private$project$internals$last_dataset_save
      was_updated <- !identical(first_stamp, second_stamp)
      if (was_updated) {
        #consider separating just saving last in details
        private$project <- save_project(private$project)
      }
      invisible(self)
    },
    #' @description saves dataset to Excel via setting provided to
    #' `add_dataset`.
    save_dataset = function(dataset_name) {
      dataset_names <- names(private$project$datasets)
      assert_choice(dataset_name, dataset_names, null.ok = FALSE)
      private$project <- save_project_dataset(project = private$project,
                                              dataset_name = dataset_name)
      invisible(self)
    },
    #' @description  Save project object to directory chosen with
    #' [setup_project].
    save = function() {
      if (private$project$internals$is_test) {
        cli_alert_info("TEST projects do not save to directories!")
        return(invisible(self))
      }
      private$project <- save_project(private$project)
      invisible(self)
    },
    #' @description  Set keyring token. See vignette and config for detail.
    set_keyring_token = function() {
      if (private$project$internals$is_test) {
        cli_alert_info("TEST projects do not communicate with the API")
        return(invisible(self))
      }
      set_project_keyring_token(private$project)
      invisible(self)
    },
    #' @description test connection via communication with API.
    test_token = function(keyring_if_error = TRUE) {
      if (private$project$internals$is_test) {
        cli_alert_info("TEST projects do not communicate with the API")
        return(invisible(self))
      }
      had_success <- FALSE
      attempt <- 1L
      is_interactive <- interactive()
      max_attempts <- ifelse(is_interactive, 3L, 1L)
      while (!any(had_success || attempt > max_attempts)) {
        token <- get_project_token(project = private$project, silent = FALSE)
        private$project <- test_project_token(project = private$project)
        had_success <- private$project$internals$last_test_connection_outcome
        if (keyring_if_error && is_interactive && !had_success) {
          set_project_keyring_token(private$project)
        }
      }
      if (!had_success) {
        cli_alert_danger("Unable to connect!")
      }
      invisible(self)
    },
    #' @description opens links in browser.
    url_launch = function(link_type = "home",
                          open_browser = TRUE) {
      if (private$project$internals$is_test && open_browser) {
        cli_alert_info("TEST projects do not link to the web!")
        return(invisible(self))
      }
      the_link <- get_project_url(private$project,
                                  link_type = link_type,
                                  open_browser = open_browser)
      if (!open_browser) {
        return(the_link)
      }
      invisible(self)
    },
    #' @description opens record links in browser.
    url_record_launch = function(record = NULL,
                                 page = NULL,
                                 instance = NULL,
                                 open_browser = TRUE) {
      if (private$project$internals$is_test && open_browser) {
        cli_alert_info("TEST projects do not link to the web!")
        return(invisible(self))
      }
      the_link <- get_record_url(private$project,
                                 record = record,
                                 page = page,
                                 instance = instance,
                                 open_browser = open_browser)
      if (!open_browser) {
        return(the_link)
      }
      invisible(self)
    },
    #' @description
    #' This will only overwrite and new data. It will not directly delete any
    #' data. Because this is a function that can mess up your data, use it
    #' very carefully. Remember all changes are saved in the REDCap log if
    #' there's an issue. Missing rows and columns are allowed!
    upload = function(to_be_uploaded) {
      if (private$project$internals$is_test) {
        cli_alert_info("TEST projects do not communicate with the API")
        return(invisible(self))
      }
      # add detect labelled vs raw?
      if (!is_named_df_list(to_be_uploaded, strict = TRUE)) {
        if (!is.data.frame(to_be_uploaded)) {
          cli_abort(
            paste0(
              "`to_be_uploaded` must be a date.frame or named list of ",
              "data.frames!"
            )
          )
        }
        if (!is_something(to_be_uploaded)) {
          cli_alert_warning("Nothing to upload...")
          return(invisible(self))
        }
        to_be_uploaded <- list(upload = to_be_uploaded)
      }
      id_col <- private$project$metadata$id_col
      refresh_records <- extract_values_from_form_list(to_be_uploaded, id_col)
      is_labelled <- private$project$settings$labelled
      for (upload_name in names(to_be_uploaded)) {
        upload_this <- to_be_uploaded[[upload_name]]
        # add comparison check but need to test...
        if (is_labelled) {
          upload_this <- labelled_to_raw_form(form = upload_this,
                                              project = private$project)
        }
        upload_form_to_redcap(
          to_be_uploaded = upload_this,
          project = private$project,
          batch_size = private$project$settings$batch_size_upload
        )
      }
      Sys.sleep(3L)
      private$project <- sync_project_check(
        project = private$project,
        hard_reset = FALSE,
        records = refresh_records
      )
      invisible(invisible(self))
    }
  ),
  private = list(
    project = NULL
  ),
  cloneable = FALSE
)

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
test_that("REDCapSyncDataset object works!", {
  project <- mock_test_project()
  dataset <- project$generate_dataset("custom")
  expect_message(dataset$print(), "Dataset: custom")
  expected_dataset_dir <- file.path(dataset$project_details$dir_path, "output")
  file_name <- "TEST_CLASSIC_custom.xlsx"
  expect_message(dataset$save(), "Saved TEST_CLASSIC_custom.xlsx")
  expect_file_exists(file.path(expected_dataset_dir, file_name))
})

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
test_that("REDCapSyncProject object works!", {
  project_r6 <- REDCapSyncProject$new(mock_test_project()$.internal)
  expect_r6_class(project_r6, "REDCapSyncProject")
  expect_r6(
    project_r6,
    classes = "REDCapSyncProject",
    private = names(REDCapSyncProject$private_fields),
    public = c(names(REDCapSyncProject$public_methods)),
    cloneable = FALSE
  )
})
test_that("REDCapSyncProject active_bindings are read-only", {
  project_r6 <- mock_test_project()
  names(REDCapSyncProject$active)
  for (active_name in names(REDCapSyncProject$active)){
    original <- project_r6[[active_name]]
    expect_in(active_name, names(REDCapSyncProject$active))
    expect_message({
      project_r6[[active_name]] <- "fake data"
    }, "is read only")
    expect_identical(original, project_r6[[active_name]])
  }
})
test_that("REDCapSyncProject with test projects!", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  project <- mock_test_project()$.internal
  dir_path <- project$dir_path
  proj_path <- file.path(dir_path, "R_objects", "TEST_CLASSIC_REDCapSync.RData")
  project_r6 <- REDCapSyncProject$new(project)
  expect_message(project_r6$test_token(), "TEST projects do not")
  expect_message(project_r6$url_launch(), "TEST projects do not")
  expect_message(project_r6$upload(), "TEST projects do not")
  expect_false(test_file_exists(proj_path))
  expect_message(project_r6$save(), "TEST projects do not save to directories")
  expect_false(test_file_exists(proj_path))
  set_dir(project$dir_path)
  project$internals$is_test <- FALSE
  project_r6 <- REDCapSyncProject$new(project)
  project_r6$save()
  expect_file_exists(proj_path)
  project_r6$remove_datasets()
  project_r6$add_dataset(
    dataset_name = "fake"
  )
  project_r6$save_datasets()
  expect_file_exists(file.path(dir_path, "output", "TEST_CLASSIC_fake.xlsx"))
  project_r6$add_dataset(
    dataset_name = "fake2"
  )
  project_r6$save_dataset(
    dataset_name = "fake2"
  )
  expect_file_exists(file.path(dir_path, "output", "TEST_CLASSIC_fake2.xlsx"))
  expected_link <- "https://redcap.fake.edu/redcap_v16.1.4/index.php?pid=12341"
  expect_identical(project_r6$url_launch(open_browser = FALSE), expected_link)
})
# REDCapSyncProject$add_dataset (Exported)
# REDCapSyncProject$remove_datasets (Exported)
test_that("REDCapSyncProject$add_dataset and remove_datasets works!", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  project_r6 <- mock_test_project()
  dataset_names <- names(project_r6$.internal$datasets)
  expect_false("new_summary" %in% dataset_names)
  project_r6$add_dataset(
    "new_summary",
    filter_field = "var_yesno",
    filter_choices = "Yes",
    dir_other = tempdir_test
  )
  dataset_names <- names(project_r6$.internal$datasets)
  expect_in("new_summary", dataset_names)
  dataset <- project_r6$generate_dataset("new_summary")
  expect_r6_class(dataset, "REDCapSyncDataset")
  expect_r6(
    dataset,
    classes = "REDCapSyncDataset",
    private = names(REDCapSyncDataset$private_fields),
    public = c(names(REDCapSyncDataset$public_methods)),
    cloneable = FALSE
  )
  project_r6$remove_datasets("REDCapSync")
  dataset_names <- names(project_r6$.internal$datasets)
  expect_false("REDCapSync" %in% dataset_names)
  expect_in("REDCapSync_raw", dataset_names)
  expect_message(project_r6$remove_datasets(), "Cleared project datasets")
  dataset_names <- names(project_r6$.internal$datasets)
  expect_length(dataset_names, 0L)
  dataset <- project_r6$generate_dataset(filter_field = "var_branching",
                                         filter_choices = "Yes")
  expect_all_true(dataset$data$merged$var_branching == "Yes")
})
# REDCapSyncProject$generate_dataset (Exported)
test_that("REDCapSyncProject$generate_dataset works!", {
})
# REDCapSyncProject$info (Exported)
test_that("REDCapSyncProject$print works!", {
  project_r6 <- mock_test_project()
  expect_message(project_r6$print(), "Name: TEST_CLASSIC")
  expect_message(project_r6$print(), "Token Name: REDCAPSYNC_TEST_CLASSIC")
  expect_message(project_r6$print(), "PID: 12341")
})
# REDCapSyncProject$initialize (Exported)
test_that("REDCapSyncProject$initialize works!", {
})
test_that("REDCapSyncProject$remove_fields works!", {
})
test_that("REDCapSyncProject$remove_datasets works!", {
})
# REDCapSyncProject$save (Exported)
test_that("REDCapSyncProject$save works!", {
})
# REDCapSyncProject$save_dataset (Exported)
test_that("REDCapSyncProject$save_dataset works!", {
})
# REDCapSyncProject$save_datasets (Exported)
test_that("REDCapSyncProject$save_datasets works!", {
})
# REDCapSyncProject$sync (Exported)
test_that("REDCapSyncProject$sync works!", {
})
# REDCapSyncProject$test_token (Exported)
test_that("REDCapSyncProject$test_token works!", {
})
# REDCapSyncProject$upload (Exported)
test_that("REDCapSyncProject$upload works!", {
  project <- mock_test_project()$.internal
  project$internals$is_test <- FALSE # override test block for tests
  project_r6 <- REDCapSyncProject$new(project)
  local_mocked_bindings(
    upload_form_to_redcap = function(...) message("Would have uploaded!"),
    sync_project_check = function(...) {
      message("Would have synced!")
      project
    }
  )
  local_mocked_bindings(
    Sys.sleep = function(...) {
      NULL
    },
    .package = "base"
  )
  expect_error(project_r6$upload(mtcars), "names in your form must match")
  to_be_uploaded <- head(project_r6$data$text, 2L)
  to_be_uploaded$var_text_only <- c("New text", "Add text")
  to_be_uploaded$text_complete <- "Complete"
  expect_no_error(project_r6$upload(to_be_uploaded))
})
# REDCapSyncProject$url_launch (Exported)
test_that("REDCapSyncProject$url_launch works!", {
})

#' @noRd
get_project_token <- function(project, silent = TRUE) {
  assert_setup_project(project)
  project_name <- project$project_name
  token_source <- "envvar"
  token <- get_project_envvar_token(project)
  if (is.null(token)) {
    token_source <- "keyring"
    token <- get_project_keyring_token(project)
  }
  valid <- is_valid_redcap_token(token = token, silent = silent)
  if (!silent) {
    if (valid) {
      cli_alert_success("Valid token for {project_name} from {token_source}!")
    } else {
      token_help_message(project)
    }
  }
  invisible(token)
}
#' @noRd
test_project_token <- function(project) {
  assert_setup_project(project)
  token <- get_project_token(project = project)
  project$internals$last_test_connection_attempt <- now_time()
  if (!is_valid_redcap_token(token)) {
    project$internals$last_test_connection_outcome <- FALSE
    return(invisible(project))
  }
  rcon <- redcapConnection(url = project$links$redcap_uri, token = token)
  redcap_version <- tryCatch(
    expr = exportVersion(rcon = rcon),
    error = function(e) {
      NULL
    }
  )
  # add timezone
  version_error <- is.null(redcap_version)
  project$internals$last_test_connection_outcome <- !version_error
  if (version_error) {
    cli_alert_danger("Your REDCap API token check failed. Check privileges.")
    return(invisible(project))
  }
  cli_alert_success("Connected to REDCap! {.url {project$links$redcap_home}}")
  version_changed <- FALSE
  if (!is.null(project$redcap$version)) {
    version_changed <- !identical(project$redcap$version, redcap_version)
  }
  project$redcap$version <- redcap_version
  if (project$internals$ever_connected) {
    project_info <- rcon$projectInformation()
    if (!is.null(project$redcap$project_id)) { # in case tested before sync
      project_id_changed <- !identical(project$redcap$project_id,
                                       as.character(project_info$project_id))
      if (project_id_changed) {
        cli_abort("The REDCap project ID for {project$project_name} changed!")
        #trigger hard_rest or message
      } # move this?
    }
  }
  project$internals$ever_connected <- TRUE
  if (version_changed || is.null(project$links$redcap_home)) {
    project <- update_project_links(project)
  }
  invisible(project)
}
#' @noRd
update_project_links <- function(project) {
  redcap_base <- project$links$redcap_base
  redcap_version <- project$redcap$version
  link_head <- paste0(redcap_base, "redcap_v", redcap_version)
  link_tail <- paste0("?pid=", project$redcap$project_id)
  home <- paste0("/index.php", link_tail)
  record_home <-  paste0("/DataEntry/record_home.php", link_tail)
  # record_subpage <- "/DataEntry/index.php", link_tail)
  dashboard <-  paste0("/DataEntry/record_status_dashboard.php", link_tail)
  api <-  paste0("/API/project_api.php", link_tail)
  api_playground <-  paste0("/API/playground.php", link_tail)
  setup <-  paste0("/ProjectSetup/index.php", link_tail)
  user_rights <-  paste0("/UserRights/index.php", link_tail)
  logging <-  paste0("/Logging/index.php", link_tail)
  designer <-  paste0("/Design/online_designer.php", link_tail)
  codebook <-  paste0("/Design/data_dictionary_codebook.php", link_tail)
  dictionary <-  paste0("/Design/data_dictionary_upload.php", link_tail)
  data_quality <- paste0("/DataQuality/index.php", link_tail)
  identifiers <- paste0(home, "&route=IdentifierCheckController:index")
  project$links$redcap_home <- paste0(link_head, home)
  project$links$redcap_record_home <- paste0(link_head, record_home)
  project$links$redcap_records_dashboard <- paste0(link_head, dashboard)
  project$links$redcap_api <- paste0(link_head, api)
  project$links$redcap_api_playground <- paste0(link_head, api_playground)
  project$links$redcap_setup <- paste0(link_head, setup)
  project$links$redcap_user_rights <- paste0(link_head, user_rights)
  project$links$redcap_logging <- paste0(link_head, logging)
  project$links$redcap_designer <- paste0(link_head, designer)
  project$links$redcap_codebook <- paste0(link_head, codebook)
  project$links$redcap_dictionary <- paste0(link_head, dictionary)
  project$links$redcap_data_quality <- paste0(link_head, data_quality)
  project$links$redcap_identifiers <- paste0(link_head, identifiers)
  invisible(project)
}
#' @noRd
is_valid_redcap_token <- function(token, silent = TRUE) {
  start_text <- "The token "
  token_text <- NULL
  end_text <- " is not a valid 32-character hexadecimal value."
  trimmed_token <- trimws(token, whitespace = WHITESPACE)
  if (is.null(token)) { # obsolete
    token_text <- "is `NULL`,"
    if (!silent) {
      cli_alert_danger(paste0(start_text, token_text, end_text))
    }
    return(FALSE)
  }
  if (is.na(token)) { # obsolete
    token_text <- "is `NA`,"
    if (!silent) {
      cli_alert_danger(paste0(start_text, token_text, end_text))
    }
    return(FALSE)
  }
  if (!nzchar(token)) { # obsolete
    token_text <- "`` (empty),"
    if (!silent) {
      cli_alert_danger(paste0(start_text, token_text, end_text))
    }
    return(FALSE)
  }
  if (token != trimmed_token) {
    token_text <- "contains whitespace (extra lines) and is therefore"
    if (!silent) {
      cli_alert_danger(paste0(start_text, token_text, end_text))
    }
    return(FALSE)
  }
  if (!is_hexadecimal(token, length = 32L)) {
    if (!silent) {
      cli_alert_danger(paste0(start_text, token_text, end_text))
    }
    return(FALSE)
  }
  TRUE
}
#' @noRd
is_hexadecimal <- function(string, length = NULL) {
  if (!is_something(string)) {
    return(FALSE)
  }
  pattern <- if (is.null(length)) {
    "^[0-9A-Fa-f]+$" # Any length
  } else {
    paste0("^[0-9A-Fa-f]{", length, "}$") # Exact length
  }
  grepl(pattern, string)
}
#' @noRd
get_project_keyring_token <- function(project) {
  if (!is_valid_keyring()) {
    return(NULL)
  }
  token <- try_else_null({
    key_get(service = config$keyring.service(),
            username = project$project_name,
            keyring = config$keyring())
  })
  token
}
#' @noRd
get_project_envvar_token <- function(project) {
  token <- Sys.getenv(project$token_name)
  if (!is_something(token)) {
    return(NULL)
  }
  token
}
#' @noRd
set_project_keyring_token <- function(project) {
  if (!is_valid_keyring()) {
    return(NULL)
  }
  project_name <- project$project_name
  key_set(service = config$keyring.service(),
          username = project$project_name,
          keyring = config$keyring(),
          prompt = paste0(project_name, " REDCap API token: "))
}
#' @noRd
has_keyring_token <- function(project_names = NULL) {
  projects <- get_projects()
  if (nrow(projects) == 0L) {
    return(NULL)
  }
  if (is.null(project_names)) {
    project_names <- projects$project_name
  }
  project_rows <- which(projects$project_name %in% project_names)
  if (length(project_rows) == 0L) {
    return(NULL)
  }
  projects <- projects[project_rows, ]
  if (!is_valid_keyring()) {
    return(NULL)
  }
  keys <- key_list(service = config$keyring.service(),
                   keyring = config$keyring())
  if (nrow(keys) == 0L) {
    return(NULL)
  }
  project_names %in% keys$username
}
#' @noRd
is_valid_keyring <- function(service = config$keyring.service(),
                             keyring = config$keyring()) {
  if (!has_keyring_pkg()) {
    cli_alert_info("`keyring` package can be used to store tokens...")
    cli_code("install.packages(\"keyring\")")
    return(FALSE)
  }
  if (!is.null(keyring)) {
    if (!has_keyring_support()) {
      cli_alert_info("Your system does not support keyring. Set to NULL.")
      cli_code("keyring::has_keyring_support()")
      return(FALSE)
    }
    # consider what will happen in ubuntu
    if (keyring_is_locked(keyring = keyring)) {
      warning_message <- "keyring = NULL (the system keyring) is locked"
      keyring_code_start <- paste0("keyring::keyring_unlock(")
      keyring_code_end <- ") # enter system password"
      if (!is.null(keyring)) {
        warning_message <- "keyring = \"{keyring}\" is locked"
        keyring_code_end <- paste0("keyring = \"", keyring, "\")")
      }
      cli_alert_warning(warning_message)
      cli_alert_info("it can be unlocked with following code ...")
      cli_code(paste0(keyring_code_start, keyring_code_end))
      return(FALSE)
    }
    if (!keyring %in% keyring_list()$keyring) {
      cli_alert_warning("keyring = \"{keyring}\" does not exist!")
      cli_alert_info("it can be created with following code ...")
      cli_code(paste0("keyring::keyring_create(\"", keyring, "\")"))
      return(FALSE)
    }
  }
  TRUE
}
#' @noRd
key_get <- function(service,
                    username = NULL,
                    keyring = NULL) {
  keyring::key_get(service = service,
                   username = username,
                   keyring = keyring)
}
#' @noRd
key_set <- function(service,
                    username = NULL,
                    keyring = NULL,
                    prompt = "Password: ") {
  keyring::key_set(service = service,
                   username = username,
                   keyring = keyring,
                   prompt = prompt)
}
#' @noRd
key_list <- function(service,
                     keyring = NULL) {
  keyring::key_list(service = service,
                    keyring = keyring)
}
#' @noRd
has_keyring_pkg <- function() {
  requireNamespace("keyring", quietly = TRUE)
}
#' @noRd
has_keyring_support <- function() {
  keyring::has_keyring_support()
}
#' @noRd
keyring_is_locked <- function(keyring = NULL) {
  keyring::keyring_is_locked(keyring = keyring)
}
#' @noRd
keyring_list <- function() {
  keyring::keyring_list()
}
#' @noRd
has_envvar_token <- function(project_names = NULL) {
  projects <- get_projects()
  if (nrow(projects) == 0L) {
    return(NULL)
  }
  if (is.null(project_names)) {
    project_names <- projects$project_name
  }
  projects <- projects[which(projects$project_name %in% project_names), ]
  if (nrow(projects) == 0L) {
    return(NULL)
  }
  if (has_keyring_pkg()) {
    if (!is_valid_keyring()) {
      return(NULL)
    }
  }
  t_names <- projects$token_name[match(project_names, projects$project_name)]
  t_names %in% names(Sys.getenv())
}
#' @noRd
token_check <- function() {
  projects <- get_projects()
  if (nrow(projects) == 0L) {
    return(NULL)
  }
  projects$has_envvar_token <- FALSE
  ennvar_check <- has_envvar_token(projects$project_name)
  if (!is.null(ennvar_check)) {
    projects$has_envvar_token <- ennvar_check
  }
  projects$has_keyring_token <- FALSE
  keyring_check <- has_keyring_token(projects$project_name)
  if (!is.null(keyring_check)) {
    projects$has_keyring_token <- keyring_check
  }
  projects
}
#' @noRd
token_help_message <- function(project) {
  assert_setup_project(project)
  user_renviron_path <- Sys.getenv("R_ENVIRON_USER", unset = NA_character_)
  if (is.na(user_renviron_path) || !nzchar(user_renviron_path)) {
    user_renviron_path <- file.path(Sys.getenv("HOME"), ".Renviron")
  }
  user_renviron_path <- normalizePath(path = user_renviron_path,
                                      winslash = "/",
                                      mustWork = FALSE)
  cli_alert_info("You can set REDCap tokens the following ways ...")
  cli_alert_info("See the {.vignette REDCapSync::Tokens} vignette.")
  cli_alert_info("1. environment variables (preferred)")
  cli_li("User R environ file --> {.file {user_renviron_path}}") # windows?
  cli_li("Try `edit_r_environ()` from `usethis` package.")
  cli_li(paste0("Add `{project$token_name} = '",
                FAKE_TOKEN,
                "'` to your user .Renviron file... then save the file ",
                "and restart R under session tab ... The way to tell it worked",
                " is to run `Sys.getenv('{project$token_name}')`"))
  cli_alert_info("2. keyring with `project$set_keyring_token()`")
  cli_alert_info(paste0("3. console with `Sys.setenv('{project$token_name}' ",
                        "= '{FAKE_TOKEN}')`"))
  cli_alert_info("x. scripts (not recommended)")
  if (is_something(project$links$redcap_api)) {
    # wont work if version changed
    cli_alert_info("Check {.url {project$links$redcap_api}}")
  }
}
#' @noRd
TOKEN_PREFIX <- "REDCAPSYNC_"
#' @noRd
FAKE_TOKEN <- "YoUrNevErShaReToKeNfRoMREDCapWebsiTe"

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# get_project_token (Internal)
test_that("get_project_token works!", {
  project <- mock_test_project()$.internal
  token_name <- project$token_name
  token <- generate_hex(32L)
  withr::with_envvar(c(REDCAPSYNC_TEST_CLASSIC = token), {
    expect_identical(get_project_token(project), token)
    expect_no_error(get_project_token(project))
    expect_message(get_project_token(project = project, silent = FALSE),
                   "Valid token for TEST_CLASSIC from envvar")
  })
  token <- generate_hex(2L)
  withr::with_envvar(c(REDCAPSYNC_TEST_CLASSIC = token), {
    expect_identical(get_project_token(project), token)
    expect_message(get_project_token(project = project, silent = FALSE),
                   "is not a valid 32-character hexadecimal value")
  })
  withr::with_envvar(c(REDCAPSYNC_TEST_CLASSIC = "BAD"), {
    expect_identical(get_project_token(project), "BAD")
    expect_message(get_project_token(project = project, silent = FALSE),
                   "is not a valid 32-character hexadecimal value")
  })
})
# is_hexadecimal (Internal)
test_that("is_hexadecimal works!", {
  x <- generate_hex(length = 32L)
  expect_true(is_hexadecimal(x))
  expect_true(is_hexadecimal(x, length = 32L))
  expect_false(is_hexadecimal(x, length = 31L))
  x <- generate_hex(length = 30L)
  expect_true(is_hexadecimal(x, length = 30L))
  expect_false(is_hexadecimal(x, length = 31L))
  expect_true(is_hexadecimal("C234"))
  expect_true(is_hexadecimal("C"))
  expect_false(is_hexadecimal("C*"))
  expect_true(is_hexadecimal("abd3"))
  expect_true(is_hexadecimal("123"))
  expect_false(is_hexadecimal(NA))
  expect_false(is_hexadecimal(""))
  expect_error(is_hexadecimal())
})
# is_valid_redcap_token (Internal)
test_that("is_valid_redcap_token works!", {
  expect_true(is_valid_redcap_token(generate_hex(32L)))
  expect_false(is_valid_redcap_token(NA))
  expect_false(is_valid_redcap_token(NULL))
  expect_false(is_valid_redcap_token(""))
  expect_false(is_valid_redcap_token(generate_hex(31L)))
  expect_false(is_valid_redcap_token(generate_hex(33L)))
  expect_false(is_valid_redcap_token(paste0(" ", generate_hex(31L))))
  expect_false(is_valid_redcap_token(paste0("J", generate_hex(31L))))
  expect_false(is_valid_redcap_token(paste0("_", generate_hex(31L))))
})
# test_project_token (Internal)
test_that("test_project_token works!", {
  # check with API. move to API section?
})
test_that("test_project_token works when exportVersion returns version", {
  project <- mock_test_project()$.internal
  withr::local_envvar(REDCAPSYNC_TEST_CLASSIC = generate_hex(length = 32L))
  # exportVersion to simulate success
  local_mocked_bindings(
    redcapConnection = function(...) {
      list(
        projectInformation = function() {
          list(project_id = project$redcap$project_id)
        }
      )
    },
    exportVersion = function(...) "12.1.1"
  )
  expect_message(test_project_token(project), "Connected to REDCap")
  out <- test_project_token(project)
  expect_true(out$internals$last_test_connection_outcome)
  expect_false(is.null(out$internals$last_test_connection_attempt))
  local_mocked_bindings(
    redcapConnection = function(...) {
      list(
        projectInformation = function() {
          list(project_id = "5678")
        }
      )
    },
    exportVersion = function(...) "12.1.1"
  )
  regexp <- "The REDCap project ID for TEST_CLASSIC changed"
  expect_error(test_project_token(project), regexp = regexp)
})
test_that("test_project_token marks failure when exportVersion returns NULL", {
  project <- mock_test_project()$.internal
  withr::local_envvar(REDCAPSYNC_TEST_CLASSIC = generate_hex(length = 32L))
  local_mocked_bindings(
    redcapConnection = function(...) list(),
    exportVersion = function(...) NULL
  )
  out <- test_project_token(project)
  expect_false(out$internals$last_test_connection_outcome)
  expect_false(is.null(out$internals$last_test_connection_attempt))
})
# update_project_links (Internal)
test_that("update_project_links works!", {
  project <- mock_test_project()$.internal
  expect_false(is.null(project$links$redcap_uri))
  expect_false(is.null(project$links$redcap_base))
  link_vector <- paste0("redcap_", LINK_TYPES)
  link_vector <- setdiff(link_vector, "redcap_base")
  #do it!
  project <- update_project_links(project)
  pid_pattern <- paste0("pid=", project$redcap$project_id)
  version_pattern <- paste0("redcap_v", project$redcap$version)
  for (the_link in link_vector) {
    expect_false(is.null(project$links[[the_link]]))
    expect_true(grepl(pid_pattern, project$links[[the_link]]))
    expect_true(grepl(version_pattern, project$links[[the_link]]))
  }
  # version changed!
  version_old <- project$redcap$version
  version_new <- "14.2.3"
  project$redcap$version <- version_new
  project <- update_project_links(project)
  expect_identical(project$redcap$version, version_new)
  pid_pattern <- paste0("pid=", project$redcap$project_id)
  version_pattern <- paste0("redcap_v", project$redcap$version)
  for (the_link in link_vector) {
    expect_false(is.null(project$links[[the_link]]))
    expect_true(grepl(pid_pattern, project$links[[the_link]]))
    expect_true(grepl(version_pattern, project$links[[the_link]]))
  }
})
# has_keyring_token (Internal)
test_that("has_keyring_token returns NULL when no projects exist", {
  local_mocked_bindings(
    get_projects = function(...) BLANK_PROJECT_DETAILS
  )
  expect_null(has_keyring_token())
})
test_that("has_keyring_token returns NULL for missing project names", {
  projects <- data.frame(project_name = "TEST_CLASSIC",
                         stringsAsFactors = FALSE)
  local_mocked_bindings(
    get_projects = function(...) projects
  )
  expect_null(has_keyring_token("MISSING_PROJECT"))
})
test_that("has_keyring_token returns NULL when keyring is invalid", {
  skip_if_not_installed("keyring")
  projects <- data.frame(project_name = "TEST_CLASSIC",
                         stringsAsFactors = FALSE)
  local_mocked_bindings(
    get_projects = function(...) projects,
    is_valid_keyring = function(...) FALSE
  )
  expect_null(has_keyring_token("TEST_CLASSIC"))
})
# has_envvar_token (Internal)
test_that("has_envvar_token works", {
  withr::local_envvar(c(
    REDCAPSYNC_TEST_CLASSIC = NA,
    REDCAPSYNC_TEST_DATA = NA,
    REDCAPSYNC_TEST_LONGITUDINAL = NA
  ))
  a <- mock_test_project("TEST_CLASSIC")$.internal |> extract_project_details()
  b <- mock_test_project("TEST_DATA")$.internal |> extract_project_details()
  c <- mock_test_project("TEST_LONGITUDINAL")$.internal |>
    extract_project_details()
  projects <- a |> bind_rows(b) |> bind_rows(c)
  local_mocked_bindings(
    get_projects = function(...) projects
  )
  withr::local_envvar(c(
    REDCAPSYNC_TEST_DATA = "ABC123"
  ))
  expect_identical(has_envvar_token(), c(FALSE, TRUE, FALSE))
  local_mocked_bindings(
    get_projects = function(...) a
  )
  #missing from projects
  expect_null(has_envvar_token(project_names = "TEST_DATA"))
  # empty projects
  local_mocked_bindings(
    get_projects = function(...) BLANK_PROJECT_DETAILS
  )
  expect_null(has_envvar_token())
})
# is_valid_keyring (Internal)
test_that("is_valid_keyring works!", {
  skip_if_not_installed("keyring")
  # No keyring package
  local_mocked_bindings(
    has_keyring_pkg = function() FALSE
  )
  expect_false(is_valid_keyring())
  # Keyring package available, but no system support
  local_mocked_bindings(
    has_keyring_pkg = function() TRUE,
    has_keyring_support = function() FALSE
  )
  expect_false(is_valid_keyring(keyring = "keyring"))
  expect_true(is_valid_keyring())
  # Keyring specified, but locked
  local_mocked_bindings(
    has_keyring_pkg = function() TRUE,
    has_keyring_support = function() TRUE,
    keyring_is_locked = function(keyring) TRUE
  )
  expect_false(is_valid_keyring(keyring = "test_keyring"))
  # Keyring not locked, but doesn't exist
  local_mocked_bindings(
    has_keyring_pkg = function() TRUE,
    has_keyring_support = function() TRUE,
    keyring_is_locked = function(keyring) FALSE,
    keyring_list = function() {
      data.frame(keyring = "existing_keyring", stringsAsFactors = FALSE)
    }
  )
  expect_false(is_valid_keyring(keyring = "nonexistent_keyring"))
  # All good
  local_mocked_bindings(
    has_keyring_pkg = function() TRUE,
    has_keyring_support = function() TRUE,
    keyring_is_locked = function(keyring) FALSE,
    keyring_list = function() {
      data.frame(keyring = "test_keyring", stringsAsFactors = FALSE)
    }
  )
  expect_true(is_valid_keyring(keyring = "test_keyring"))
})

#' @noRd
get_redcap_metadata <- function(project) {
  assert_setup_project(project)
  result <- get_redcap_rcon(project)
  get_users <- project$settings$get_users
  project$metadata <- list()
  project$redcap$project_info <- result$project_info
  # info ----------
  project$redcap$project_title <- project$redcap$project_info$project_title
  project$redcap$project_id <-
    as.character(project$redcap$project_info$project_id)
  is_longitudinal <- project$redcap$project_info$is_longitudinal == "1"
  project$metadata$is_longitudinal <- is_longitudinal
  missing_data_codes <- NA
  if ("missing_data_codes" %in% colnames(project$redcap$project_info)) {
    missing_data_codes <- project$redcap$project_info$missing_data_codes
    if (!is.na(missing_data_codes)) {
      missing_data_codes <- split_choices(missing_data_codes)
    }
  }
  project$metadata$missing_codes <- missing_data_codes
  project$redcap$has_log_access <- !is.null(result$logging)
  project$redcap$has_dag_access <- !is.null(result$dags)
  project$redcap$has_user_access <- !is.null(result$users)
  project$redcap$has_file_repository_access <- !is.null(result$file_repository)
  # instruments --------
  project$metadata$forms <- rename(result$forms,
                                   form_name  = "instrument_name",
                                   form_label = "instrument_label")
  project$metadata$repeating_forms_events <- result$repeating
  project$metadata$forms$repeating <- FALSE
  project$metadata$has_repeating_forms <- FALSE
  project$metadata$has_repeating_events <- FALSE
  has_repeat_forms_or_events <-
    project$redcap$project_info$has_repeating_instruments_or_events == "1"
  project$metadata$has_repeating_forms_or_events <- has_repeat_forms_or_events
  # if(project$redcap$project_info$has_repeating_instruments_or_events=="1")
  if (is.data.frame(project$metadata$repeating_forms_events)) {
    #  test if you can do this if you dont have designer privileges
    # or would have to use another package
    if (nrow(project$metadata$repeating_forms_events) > 0L) {
      project$metadata$forms$repeating <-
        project$metadata$forms$form_name %in%
        project$metadata$repeating_forms_events$form_name
    }
  }
  if (any(project$metadata$forms$repeating)) {
    project$metadata$has_repeating_forms <- TRUE
  }
  # metadata ----------
  project$metadata$fields <- result$fields
  section_header <- project$metadata$fields$section_header
  project$metadata$fields$section_header <- remove_html_tags(section_header)
  field_label <- project$metadata$fields$field_label
  project$metadata$fields$field_label <- remove_html_tags(field_label)
  id_position <- project$settings$id_position
  id_col <- as.character(project$metadata$fields[id_position, 1L])
  project$metadata$id_col <- id_col
  project$metadata$form_key_cols <- get_key_col_list(data_list = project)
  form_key_cols <- project$metadata$form_key_cols
  project$metadata$raw_structure_cols <- form_key_cols |> unlist() |> unique()
  project$metadata$fields <- add_field_elements(project$metadata$fields)
  project$metadata$choices <- fields_to_choices(project$metadata$fields)
  # add a check for existing conflict possibilities
  project$metadata$has_coding_conflicts <- FALSE
  field_names <- unique(project$metadata$choices$field_name)
  if (length(field_names) > 0L) {
    choices <- project$metadata$choices
    has_conflict <- field_names |>
      lapply(function(field_name) {
        vec_to_check <- choices$name[which(choices$field_name == field_name)]
        anyDuplicated(vec_to_check) > 0L
      }) |>
      unlist()
    project$metadata$has_coding_conflicts <- any(has_conflict)
    if (project$metadata$has_coding_conflicts) {
      conflict_rows <- which(has_conflict)
      project$metadata$coding_conflict_field_names <- field_names[conflict_rows]
    }
  }
  # is longitudinal ------
  if (project$metadata$is_longitudinal) {
    raw_structure_cols_vector <- c(project$metadata$raw_structure_cols,
                                   "arm_number",
                                   "event_name")
    project$metadata$raw_structure_cols <- unique(raw_structure_cols_vector)
    project$metadata$arms <- result$arms
    arm_name_col <- which(colnames(project$metadata$arms) == "arm_num")
    colnames(project$metadata$arms)[arm_name_col] <- "arm_number"
    project$metadata$has_arms <- TRUE
    project$metadata$has_multiple_arms <- nrow(project$metadata$arms) > 1L
    project$metadata$has_arms_that_matter <- project$metadata$has_multiple_arms
    project$metadata$event_mapping <- result$mapping
    project$metadata$events <- result$events
    arm_name_col <- which(colnames(project$metadata$events) == "arm_num")
    colnames(project$metadata$events)[arm_name_col] <- "arm_number"
    project$metadata$events$repeating <- FALSE
    project$metadata$event_mapping$repeating <- FALSE
    unique_event_names <- project$metadata$events$unique_event_name
    unique_event_names_map <- project$metadata$event_mapping$unique_event_name
    if (is.data.frame(project$metadata$repeating_forms_events)) {
      keep <- which(is.na(project$metadata$repeating_forms_events$form_name))
      rep_events <- project$metadata$repeating_forms_events$event_name[keep]
      project$metadata$events$repeating <- unique_event_names %in% rep_events
      is_rep_event <- !is.na(project$metadata$repeating_forms_events$event_name)
      is_rep_form <- !is.na(project$metadata$repeating_forms_events$form_name)
      rep_rows <- which(is_rep_event & is_rep_form)
      rep_form_ind <- project$metadata$repeating_forms_events[rep_rows, ]
      if (nrow(rep_form_ind) > 0L) {
        rows_event_mapping <- seq_len(nrow(rep_form_ind)) |>
          lapply(function(i) {
            criteria_one <- unique_event_names_map == rep_form_ind$event_name[i]
            form_names <- project$metadata$event_mapping$form
            criteria_two <- form_names == rep_form_ind$form_name[i]
            which(criteria_one & criteria_two)
          }) |>
          unlist()
        project$metadata$event_mapping$repeating[rows_event_mapping] <- TRUE
      }
    }
    project$metadata$events$forms <-
      unique_event_names |>
      lapply(function(events) {
        event_rows <- which(unique_event_names_map == events)
        project$metadata$event_mapping$form[event_rows] |>
          unique() |>
          paste0(collapse = " | ")
      }) |>
      unlist()
    if (project$metadata$has_arms_that_matter) {
      arm_number_vector <- project$metadata$arms$arm_number
      project$metadata$has_arms_that_matter <- arm_number_vector |>
        lapply(function(arm) {
          arm_row <- which(project$metadata$event_mapping$arm_number == arm)
          project$metadata$event_mapping$form[arm_row]
        }) |>
        check_match() |>
        xor(TRUE)
    }
    project$metadata$forms$repeating_via_events <- FALSE
    rep_via_event_rows <- project$metadata$forms$form_name |>
      lapply(function(form_name) {
        form_rows <- which(project$metadata$event_mapping$form == form_name)
        anyDuplicated(project$metadata$event_mapping$arm_num[form_rows]) > 0L
      }) |>
      unlist() |>
      which()
    project$metadata$forms$repeating_via_events[rep_via_event_rows] <- TRUE
  } else {
    project$metadata$has_arms <- FALSE
    project$metadata$has_multiple_arms <- FALSE
    project$metadata$has_arms_that_matter <- FALSE
    project$metadata$event_mapping <- NA
    project$metadata$events <- NA
  }
  # other-------
  project$redcap$users <- NA
  has_users <- is_something(result$users)
  if (get_users && project$redcap$has_user_access && has_users) {
    keep_cols <- c("unique_role_name", "role_label")
    project$redcap$users <- result$user_roles[, keep_cols] |>
      merge(result$user_role_assignment, by = "unique_role_name") |>
      merge(result$users, by = "username", all.y = TRUE)
  }
  project$redcap$file_repository <- NA
  get_file_repository <- project$settings$get_file_repository
  if (get_file_repository && project$redcap$has_file_repository_access) {
    project$redcap$file_repository <- result$file_repository
  }
  #dags
  project$settings$last_metadata_update <- now_time()
  invisible(project)
}
#' @noRd
get_redcap_rcon <- function(project) {
  rcon <- redcapConnection(url = project$links$redcap_uri,
                           token = get_project_token(project))
  result <- list(
    project_info = try_else_null(all_character_cols(rcon$projectInformation())),
    arms = try_else_null(all_character_cols(rcon$arms())),
    events = try_else_null(all_character_cols(rcon$events())),
    mapping = try_else_null(all_character_cols(rcon$mapping())),
    forms = try_else_null(all_character_cols(rcon$instruments())),
    repeating = try_else_null(all_character_cols(rcon$repeatInstrumentEvent())),
    fields = try_else_null(all_character_cols(rcon$metadata())),
    logging = try_else_null(all_character_cols(
      exportLogging(rcon = rcon, beginTime = Sys.time() - 100000L)
    )),
    users = try_else_null(all_character_cols(rcon$users())),
    user_roles = try_else_null(all_character_cols(rcon$user_roles())),
    user_role_assignment = try_else_null(all_character_cols(
      rcon$user_role_assignment()
    )),
    dags = try_else_null(all_character_cols(rcon$dags())),
    dag_assignment = try_else_null(all_character_cols(rcon$dag_assignment())),
    file_repository = try_else_null(all_character_cols(rcon$file_repository()))
  )
  result
}
#' @noRd
add_field_elements <- function(fields) {
  #assert_fields should be here
  form_names <- unique(fields$form_name)
  for (form_name in form_names) {
    field_name <- paste0(form_name, "_complete")
    field_label <- field_name |>
      strsplit("_") |>
      unlist() |>
      str_to_title() |>
      paste(collapse = " ")
    new_row <- data.frame(
      field_name = field_name, # check if conflicts,
      form_name = form_name,
      field_type = "radio",
      field_label = field_label,
      select_choices_or_calculations =
        "0, Incomplete | 1, Unverified | 2, Complete",
      stringsAsFactors = FALSE
    )
    last_row <- nrow(fields)
    row_index <- which(fields$form_name == form_name)
    if (length(row_index) == 0L) {
      row_index <- last_row
    }
    x_row <- last(row_index)
    top <- fields[1L:x_row, ]
    bottom <- NULL
    if (last_row > x_row) {
      bottom <- fields[(x_row + 1L):last_row, ]
    }
    fields <- top |>
      bind_rows(new_row) |>
      bind_rows(bottom)
  }
  if (any(fields$field_type == "checkbox")) {
    checkbox_fields <- fields$field_name[which(fields$field_type == "checkbox")]
    for (field_name in checkbox_fields) {
      field_row <- which(fields$field_name == field_name)
      x <- split_choices(fields$select_choices_or_calculations[field_row])
      new_rows <- data.frame(
        field_name = paste0(field_name, "___", x$code),
        form_name = fields$form_name[field_row],
        field_label = x$name,
        field_type = "checkbox_choice",
        select_choices_or_calculations = "1, Checked | 0, Unchecked",
        stringsAsFactors = FALSE
      )
      last_row <- nrow(fields)
      top <- fields[1L:field_row, ]
      bottom <- NULL
      if (last_row > field_row) {
        bottom <- fields[(field_row + 1L):last_row, ]
      }
      fields <- top |>
        bind_rows(new_rows) |>
        bind_rows(bottom)
    }
  }
  if (any(fields$field_type == "yesno")) {
    yn_rows <- which(fields$field_type == "yesno")
    fields$select_choices_or_calculations[yn_rows] <- "1, Yes | 0, No"
  }
  if (any(fields$field_type == "truefalse")) {
    tf_rows <- which(fields$field_type == "truefalse")
    fields$select_choices_or_calculations[tf_rows] <- "1, True | 0, False"
  }
  fields
}
#' @noRd
get_redcap_data <- function(project,
                            labelled = TRUE,
                            records = NULL) {
  form_list <- list()
  denormalized <- get_redcap_denormalized(
    project = project,
    records = records
  ) # add check for dag and api
  form_list <- normalize_redcap(denormalized = denormalized,
                                project = project,
                                labelled = labelled)
  form_list
}
#' @noRd
get_redcap_denormalized <- function(project,
                                    records = NULL) {
  fields <- null_if_na(project$settings$fields)
  if (project$settings$get_type != "identified") {
    fields <- fields |> append(
      get_identifier_fields(
        data_list = project,
        get_type = project$settings$get_type,
        invert = TRUE
      ) |>
        unique()
    )# message if fields clash with get_type
  }
  denormalized <- redcap_read(
    batch_size = project$settings$batch_size_download,
    interbatch_delay = 0.2,
    continue_on_error = FALSE,
    redcap_uri = project$links$redcap_uri,
    token = get_project_token(project),
    records = records,
    fields = fields,
    forms = null_if_na(project$settings$forms),
    events = null_if_na(project$settings$events),
    raw_or_label = "raw",
    filter_logic = quotes_if_na(project$settings$filter_logic),
    verbose = config$verbose(),
    id_position = project$settings$id_position
  )$data
  denormalized <- all_character_cols(denormalized)
  denormalized
}
#' @noRd
get_redcap_files <- function(project,
                             original_file_names = FALSE,
                             overwrite = FALSE) {
  file_rows <- which(project$metadata$fields$field_type == "file")
  project_name <- project$project_name
  out_dir <- file.path(project$dir_path, "REDCap", project_name, "files")
  if (length(file_rows) > 0L) {
    dir.create(out_dir, showWarnings = FALSE, recursive = TRUE)
    for (field_name in project$metadata$fields$field_name[file_rows]) {
      out_dir_folder <- file.path(out_dir, field_name)
      dir.create(out_dir_folder,
                 showWarnings = FALSE,
                 recursive = TRUE)
      field_row <- which(project$metadata$fields$field_name == field_name)
      form_name <- project$metadata$fields$form_name[field_row]
      form_row <- which(project$metadata$forms$form_name == form_name)
      is_repeating <- project$metadata$forms$repeating[form_row]
      form <- project$data[[form_name]]
      rows_to_save <- which(!is.na(form[[field_name]]))
      for (i in rows_to_save) {
        file_name <- form[[field_name]][i]
        record_id <- form[[project$metadata$id_col]][i]
        repeat_instance <- form[["redcap_repeat_instance"]][i]
        repeat_instance <- ifelse(is.null(repeat_instance), "", repeat_instance)
        redcap_event_name <- form[["redcap_event_name"]][i]
        is_null_event <- is.null(redcap_event_name)
        redcap_event_name <- ifelse(is_null_event, "", repeat_instance)
        if (!original_file_names) {
          if (anyDuplicated(file_name) > 0L) {
            warning(
              "You have duplicate file names in ",
              form_name,
              ", ",
              field_name,
              ". Therefore will use system generated names",
              immediate. = TRUE
            )
            original_file_names <- FALSE
          }
        }
        file_name <- ifelse(
          original_file_names,
          file_name,
          paste0(
            form_name,
            "_",
            field_name,
            "_",
            ifelse(is_repeating, "inst_", ""),
            repeat_instance,
            "ID_",
            record_id,
            ".",
            file_ext_alias(file_name)
          )
        )
        out_file_path <- file.path(out_dir_folder, file_name)
        if (!file.exists(out_file_path) || overwrite) {
          #try catch?
          redcap_file_download_oneshot(
            redcap_uri = project$links$redcap_uri,
            token = get_project_token(project),
            field = field_name,
            record = form[[project$metadata$id_col]][i],
            directory = out_dir_folder,
            file_name = file_name,
            event = redcap_event_name,
            repeat_instance = repeat_instance,
            verbose = FALSE
          )
          succeeded <- file.exists(out_file_path)
          if (succeeded) {
            cli_bullets(c(">" = "{.code {file_name}} saved."))
          } else {
            cli_alert_danger("{.code {file_name}} failed to save!")
          }
        }
      }
    }
  }
  cli_alert_success("Checked for files! Stored at ... {.path {out_dir}}")
}
#' @noRd
get_redcap_log <- function(project,
                           log_begin_date = Sys.Date() - 10L) {
  assert_setup_project(project)
  assert_date(log_begin_date)
  log_begin_time <- as.character(log_begin_date) |>
    paste("00:00:00") |>
    as.POSIXct()
  redcap_log <- exportLogging(
    rcon = redcapConnection(url = project$links$redcap_uri,
                            token = get_project_token(project)),
    logtype = character(0L),
    user = character(0L),
    record = character(0L),
    beginTime = log_begin_time,
    endTime = as.POSIXct(character(0L))
  )
  #addtrycatch NULL
  if (is.null(redcap_log)) {
    return(NULL)
  }
  if (is.data.frame(redcap_log)) {
    if (nrow(redcap_log) > 0L) {
      redcap_log <- clean_redcap_log(
        redcap_log = redcap_log,
        drop_exports = project$settings$log_drop_exports
      )
    }
  }
  redcap_log # deal with if NA if user does not have log privileges.
}
#' @noRd
get_redcap_records <- function(project) {
  assert_setup_project(project)
  id_col <- project$metadata$id_col
  records <- suppressMessages({
    redcap_read_oneshot(
      redcap_uri = project$links$redcap_uri,
      token = get_project_token(project),
      fields = id_col,
      raw_or_label = "raw"
    )
  })
  # consider error check
  records <- records$data[[id_col]] |> unique() |> as.character()
  records
}
#' @noRd
upload_form_to_redcap <- function(to_be_uploaded, project, batch_size = 500L) {
  redcap_write(
    ds_to_write = as_tibble(all_character_cols(to_be_uploaded)),
    batch_size = batch_size,
    interbatch_delay = 0.2,
    continue_on_error = FALSE,
    redcap_uri = project$links$redcap_uri,
    token = get_project_token(project),
    overwrite_with_blanks = TRUE
  )
}
#' @noRd
LOG_COLNAMES <- c(
  "timestamp",
  "username",
  "action",
  "details",
  "record",
  "action_type"
)
#' @noRd
USERS_COLNAMES <- c(
  "username",
  "email",
  "firstname",
  "lastname",
  "expiration",
  "data_access_group",
  "data_access_group_id",
  "design",
  "alerts",
  "user_rights",
  "data_access_groups",
  "reports",
  "stats_and_charts",
  "manage_survey_participants",
  "calendar",
  "data_import_tool",
  "data_comparison_tool",
  "logging",
  "file_repository",
  "data_quality_create",
  "data_quality_execute",
  "api_export",
  "api_import",
  "api_modules",
  "mobile_app",
  "mobile_app_download_data",
  "record_create",
  "record_rename",
  "record_delete",
  "lock_records_all_forms",
  "lock_records",
  "lock_records_customization",
  "forms_export"
)
#' @noRd
FIELD_COLNAMES <- c(
  "field_name",
  "form_name",
  "section_header",
  "field_type",
  "field_label",
  "select_choices_or_calculations",
  "field_note",
  "text_validation_type_or_show_slider_number",
  "text_validation_min",
  "text_validation_max",
  "identifier",
  "branching_logic",
  "required_field",
  "custom_alignment",
  "question_number",
  "matrix_group_name",
  "matrix_ranking",
  "field_annotation"
)
#' @noRd
ARMS_COLNAMES <- c(
  "arm_number",
  "arm_name"
)
#' @noRd
EVENTS_COLNAMES <- c(
  "event_name",
  "arm_number",
  "unique_event_name",
  "custom_event_label",
  "event_id"
)
#' @noRd
EVENT_MAPPING_COLNAMES <- c(
  "arm_num",
  "unique_event_name",
  "form"
)
#' @noRd
FORMS_COLNAMES <- c(
  "form_name",
  "form_label"
)
CHOICES_COLNAMES <- c(
  "field_name",
  "code",
  "name"
)
#' @noRd
PROJECT_INFO_COLNAMES <- c(
  "project_id",
  "project_title",
  "creation_time",
  "production_time",
  "in_production",
  "project_language",
  "purpose",
  "purpose_other",
  "project_notes",
  "custom_record_label",
  "secondary_unique_field",
  "is_longitudinal",
  "has_repeating_instruments_or_events",
  "surveys_enabled",
  "scheduling_enabled",
  "record_autonumbering_enabled",
  "randomization_enabled",
  "ddp_enabled",
  "project_irb_number",
  "project_grant_number",
  "project_pi_firstname",
  "project_pi_lastname",
  "display_today_now_button",
  "missing_data_codes",
  "external_modules",
  "bypass_branching_erase_field_prompt"
)
#' @noRd
MISSING_CODES <- data.frame(
  code = c(
    "NI",
    "INV",
    "UNK",
    "NASK",
    "ASKU",
    "NAV",
    "MSK",
    "NA",
    "NAVU",
    "NP",
    "QS",
    "QI",
    "TRC",
    "UNC",
    "DER",
    "PINF",
    "NINF",
    "OTH"
  ),
  name = c(
    "No information",
    "Invalid",
    "Unknown",
    "Not asked",
    "Asked but unknown",
    "Temporarily unavailable",
    "Masked",
    "Not applicable",
    "Not available",
    "Not present",
    "Sufficient quantity",
    "Insufficient quantity",
    "Trace",
    "Unencoded",
    "Derived",
    "Positive infinity",
    "Negative infinity",
    "Other"
  ),
  stringsAsFactors = FALSE
)

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# get_redcap_log (Internal)
test_that("get_redcap_log works!", {
  skip_on_cran()
  skip_if_offline()
  # need dev server with log access?
})
test_that("get_redcap_log works on fixture!", {
  project_name <- "TEST_REPEATING"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  local_mocked_bindings(
    exportLogging = function(...) call_list$logging
  )
  result <- get_redcap_log(project)
  expect_data_frame(result)
})
# get_redcap_metadata (Internal)
test_that("get_redcap_ works on real server, simple!", {
  skip_on_cran()
  skip_if_offline()
  withr::local_options(c(redcapsync.config.allow.test.names = TRUE))
  project_name <- "TEST_REDCAPR_SIMPLE"
  project <- real_test_project(project_name)$.internal
  expect_data_frame(as.data.frame(project$metadata$forms), nrows = 0L)
  expect_data_frame(as.data.frame(project$metadata$fields), nrows = 0L)
  expect_data_frame(as.data.frame(project$metadata$choices), nrows = 0L)
  expect_data_frame(as.data.frame(project$redcap$users), nrows = 0L)
  expect_data_frame(as.data.frame(project$redcap$project_info), nrows = 0L)
  project_with_metadata <- withr::with_envvar(REAL_DEV_TOKENS, {
    get_redcap_metadata(project)
  })
  expect_data_frame(project_with_metadata$metadata$forms, min.rows = 1L)
  expect_data_frame(project_with_metadata$metadata$fields, min.rows = 1L)
  expect_data_frame(project_with_metadata$metadata$choices, min.rows = 1L)
  expect_data_frame(project_with_metadata$redcap$users, min.rows = 1L)
  expect_data_frame(project_with_metadata$redcap$project_info, nrows = 1L)
  # data_test
  expect_length(project_with_metadata$data, 0L)
  data_list <- withr::with_envvar(REAL_DEV_TOKENS, {
    suppressWarnings({
      get_redcap_data(project_with_metadata)
    })
  })
  expect_list(data_list, min.len = 1L)
  expect_data_frame(data_list$demographics, min.rows = 1L, min.cols = 2L)
  project_records <- withr::with_envvar(REAL_DEV_TOKENS, {
    get_redcap_records(project_with_metadata)
  })
  expect_character(project_records, unique = TRUE, min.len = 1L)
  expect_in(project_records, data_list$demographics$record_id)
})
test_that("get_redcap_ works on real server, longitudinal!", {
  skip_on_cran()
  skip_if_offline()
  withr::local_options(c(redcapsync.config.allow.test.names = TRUE))
  project_name <- "TEST_REDCAPR_SIMPLE"
  project <- real_test_project(project_name)$.internal
  expect_data_frame(as.data.frame(project$metadata$forms), nrows = 0L)
  expect_data_frame(as.data.frame(project$metadata$fields), nrows = 0L)
  expect_data_frame(as.data.frame(project$metadata$choices), nrows = 0L)
  expect_data_frame(as.data.frame(project$redcap$users), nrows = 0L)
  expect_data_frame(as.data.frame(project$redcap$project_info), nrows = 0L)
  project_with_metadata <- withr::with_envvar(REAL_DEV_TOKENS, {
    get_redcap_metadata(project)
  })
  expect_data_frame(project_with_metadata$metadata$forms, min.rows = 1L)
  expect_data_frame(project_with_metadata$metadata$fields, min.rows = 1L)
  expect_data_frame(project_with_metadata$metadata$choices, min.rows = 1L)
  expect_data_frame(project_with_metadata$redcap$users, min.rows = 1L)
  expect_data_frame(project_with_metadata$redcap$project_info, nrows = 1L)
  # data_test
  expect_length(project_with_metadata$data, 0L)
  data_list <- withr::with_envvar(REAL_DEV_TOKENS, {
    suppressWarnings({
      get_redcap_data(project_with_metadata)
    })
  })
  expect_list(data_list, min.len = 1L)
  expect_data_frame(data_list$demographics, min.rows = 1L, min.cols = 2L)
  project_records <- withr::with_envvar(REAL_DEV_TOKENS, {
    get_redcap_records(project_with_metadata)
  })
  expect_character(project_records, unique = TRUE, min.len = 1L)
  expect_in(project_records, data_list$demographics$record_id)
})
test_that("get_redcap_metadata works with fixture data (classic)", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  local_mocked_bindings(
    get_redcap_rcon = function(...) call_list
  )
  result <- get_redcap_metadata(project)
  expect_identical(result$project_name, project$project_name)
  expect_identical(call_list$project_info, project$redcap$project_info)
  project$metadata <- BLANK_PROJECT$metadata # clear existing data
  expect_null(project$metadata$fields)
  expect_null(project$metadata$forms)
  expect_list(result)
  expect_false(is.null(result$metadata))
  expect_false(is.null(result$redcap$project_info))
  expect_data_frame(result$metadata$forms, min.rows = 1L)
  expect_data_frame(result$metadata$fields, min.rows = 1L)
  expect_false(is.null(result$redcap$project_title))
  expect_false(is.null(result$redcap$project_id))
  expect_data_frame(result$metadata$forms)
  expect_true("form_name" %in% colnames(result$metadata$forms))
  expect_data_frame(result$metadata$fields)
  expect_true("field_name" %in% colnames(result$metadata$fields))
  expect_true("form_name" %in% colnames(result$metadata$fields))
  expect_data_frame(result$metadata$choices)
  expect_data_frame(result$redcap$users)
  expect_true(result$redcap$has_user_access)
  expect_logical(result$redcap$has_log_access)
  if (is_something(result$metadata$missing_codes)) {
    expect_data_frame(result$metadata$missing_codes)
  }
  expect_identical(result$project_name, project_name)
})
test_that("get_redcap_metadata works with fixture data (longitudinal)", {
  project_name <- "TEST_REDCAPR_LONGITUDINAL"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  local_mocked_bindings(
    get_redcap_rcon = function(...) call_list
  )
  project$metadata <- BLANK_PROJECT$metadata # clear existing data
  expect_null(project$metadata$fields)
  expect_null(project$metadata$forms)
  result <- get_redcap_metadata(project)
  expect_list(result)
  expect_true(result$metadata$is_longitudinal)
  expect_true(result$metadata$has_arms)
  expect_data_frame(result$metadata$arms, min.rows = 1L)
  expect_data_frame(result$metadata$events, min.rows = 1L)
})
test_that("get_redcap_metadata works with fixture data (repeating forms)", {
  project_name <- "TEST_REPEATING"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  local_mocked_bindings(
    get_redcap_rcon = function(...) call_list
  )
  project$metadata <- BLANK_PROJECT$metadata # clear existing data
  result <- get_redcap_metadata(project)
  expect_list(result)
  expect_true(result$metadata$has_repeating_forms_or_events)
})
# get_redcap_records (Internal)
test_that("get_redcap_records works!", {
})
# get_redcap_rcon (Internal)
test_that("get_redcap_rcon works!", {
  skip_on_cran()
  skip_if_offline()
  # compare fixture colnames to real con colnames
})
test_that("get_redcap_rcon returns expected structure without real API calls", {
  project <- mock_test_project()$.internal
  # Create a fake rcon with the methods used by get_redcap_rcon
  fake_rcon <- list(
    projectInformation = function() {
      list(project_id = "9999",
           project_title = "Fake Project",
           has_repeating_instruments_or_events = "0")
    },
    arms = function() {
      data.frame(arm = character(0L), stringsAsFactors = FALSE)
    },
    events = function() {
      data.frame(event = character(0L), stringsAsFactors = FALSE)
    },
    mapping = function() data.frame(),
    instruments = function() {
      data.frame(instrument = "form1", stringsAsFactors = FALSE)
    },
    repeatInstrumentEvent = function() data.frame(),
    metadata = function() {
      data.frame(field_name = character(0L), stringsAsFactors = FALSE)
    },
    users = function() {
      data.frame(username = character(0L), stringsAsFactors = FALSE)
    },
    user_roles = function() data.frame(),
    user_role_assignment = function() data.frame(),
    dags = function() data.frame(),
    dag_assignment = function() data.frame(),
    fileRepository = function() data.frame()
  )
  local_mocked_bindings(
    redcapConnection = function(...) fake_rcon,
    exportLogging = function(...) data.frame()
  )
  out <- get_redcap_rcon(project)
  # replace with real data from fixtures
  expect_type(out, "list")
  # core elements present
  expect_true("project_info" %in% names(out))
  expect_true("arms" %in% names(out))
  expect_true("events" %in% names(out))
  expect_true("mapping" %in% names(out))
  expect_true("forms" %in% names(out))
  expect_true("repeating" %in% names(out))
  expect_true("fields" %in% names(out))
  expect_true("logging" %in% names(out))
  expect_true("users" %in% names(out))
  expect_true("user_roles" %in% names(out))
  expect_true("user_role_assignment" %in% names(out))
  expect_true("dags" %in% names(out))
  expect_true("dag_assignment" %in% names(out))
  expect_true("file_repository" %in% names(out))
  # check a few returned values come from our fake rcon
  expect_identical(out$project_info$project_id, "9999")
  expect_s3_class(out$forms, "data.frame")
})
# upload_form_to_redcap (Internal)
test_that("upload_form_to_redcap works!", {
})
# add_field_elements (Internal)
test_that("add_field_elements works!", {
})
# get_redcap_data (Internal)
test_that("get_redcap_data works with fixture data (classic)", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  local_mocked_bindings(
    get_redcap_denormalized = function(...) call_list$data
  )
  result <- get_redcap_data(project)
  expect_identical(result, project$data) # matches fixture call
  result_not_labelled <- get_redcap_data(project, labelled = FALSE)
  expect_in(c("0", "1"), result_not_labelled$other$var_yesno)
})
test_that("get_redcap_data works with fixture data (longitudinal)", {
  project_name <- "TEST_REDCAPR_LONGITUDINAL"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  local_mocked_bindings(
    get_redcap_rcon = function(...) call_list
  )
  project$metadata <- BLANK_PROJECT$metadata # clear existing data
  expect_null(project$metadata$fields)
  expect_null(project$metadata$forms)
  result <- get_redcap_metadata(project)
  expect_list(result)
  expect_true(result$metadata$is_longitudinal)
  expect_true(result$metadata$has_arms)
  expect_data_frame(result$metadata$arms, min.rows = 1L)
  expect_data_frame(result$metadata$events, min.rows = 1L)
})
test_that("get_redcap_data works with fixture data (repeating forms)", {
  project_name <- "TEST_REPEATING"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  local_mocked_bindings(
    get_redcap_rcon = function(...) call_list
  )
  project$metadata <- BLANK_PROJECT$metadata # clear existing data
  result <- get_redcap_metadata(project)
  expect_list(result)
  expect_true(result$metadata$has_repeating_forms_or_events)
})
# get_redcap_denormalized (Internal)
test_that("get_redcap_denormalized works!", {
  skip_on_cran()
  skip_if_offline()
  withr::local_options(c(redcapsync.config.allow.test.names = TRUE))
  project <- real_test_project("TEST_REDCAPR_LONGITUDINAL")$.internal
  expect_data_frame(as.data.frame(project$data), nrows = 0L)
  project_data <- suppressWarnings({
    withr::with_envvar(REAL_DEV_TOKENS, {
      get_redcap_denormalized(project)
    })
  })
  expect_data_frame(project_data, min.rows = 1L)
})
test_that("get_redcap_denormalized works no API call!", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  local_mocked_bindings(
    redcap_read = function(...) call_list
  )
  result <- get_redcap_denormalized(project)
  expect_data_frame(result, min.rows = 1L)
})
# get_redcap_files (Internal)
test_that("get_redcap_files works!", {
  skip_on_cran()
  skip_if_offline()
})
test_that("get_redcap_files works, no API call!", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  project <- mock_test_project()$.internal
  out_file_path1 <- file.path(
    project$dir_path,
    "REDCap",
    "TEST_CLASSIC",
    "files",
    "var_sig",
    "other_var_sig_ID_1.png"
  )
  out_file_path2 <- file.path(
    project$dir_path,
    "REDCap",
    "TEST_CLASSIC",
    "files",
    "var_file",
    "other_var_file_ID_1.csv"
  )
  out_file_paths <- c(out_file_path1, out_file_path2)
  expect_all_false(file.exists(out_file_paths))
  local_mocked_bindings(
    redcap_file_download_oneshot = function(...) {
      cli_alert_danger("API FAILED!")
    }
  )
  get_redcap_files(project, original_file_names = FALSE, overwrite = FALSE)
  expect_all_false(file.exists(out_file_paths))
  local_mocked_bindings(
    redcap_file_download_oneshot = function(...) {
      file.create(out_file_paths, showWarnings = FALSE)
    }
  )
  get_redcap_files(project, original_file_names = FALSE, overwrite = FALSE)
  expect_all_true(file.exists(out_file_paths))
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  drop_nas(project$data$other$var_sig)
  out_file_path1 <- file.path(
    project$dir_path,
    "REDCap",
    "TEST_CLASSIC",
    "files",
    "var_sig",
    "signature_2025-11-19_1956.png"
  )
  drop_nas(project$data$other$var_file)
  out_file_path2 <- file.path(
    project$dir_path,
    "REDCap",
    "TEST_CLASSIC",
    "files",
    "var_file",
    "Test_DataDictionary_2025-11-19.csv"
  )
  out_file_paths <- c(out_file_path1, out_file_path2)
  expect_all_false(file.exists(out_file_paths))
  get_redcap_files(project, original_file_names = TRUE, overwrite = FALSE)
  expect_all_true(file.exists(out_file_paths))
})

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----eval=FALSE---------------------------------------------------------------
# library(REDCapSync)

## ----eval=FALSE---------------------------------------------------------------
# save_dir <- tempdir() # replace with your intended folder
# 
# project <- load_project("TEST_CLASSIC")
# 
# project$sync() #optional sync
# 
# dataset <- project$load_dataset("REDCapSync")
# 
# # add quick custom variable
# # in future version you can use project$add_field for this
# dataset$data$merged$letter_b <- dataset$data$merged$var_text_letters == "b"
# 
# # send data in global environment
# dataset$to_envir(globalenv())
# 
# # save data in custom location
# dataset$save(dir_other = save_dir)
# 
# # open folder where file was saved
# browseURL(save_dir)

## ----eval=FALSE---------------------------------------------------------------
# save_dir <- tempdir() # replace with your intended folder
# 
# dataset <- projects$load("TEST_CLASSIC")$sync()$load_dataset("REDCapSync")
# # add quick custom variable
# # in future version you can use project$add_field for this
# dataset$data$merged$letter_b <- dataset$data$merged$var_text_letters == "b"
# 
# # send data in global environment
# dataset$to_envir(globalenv())
# 
# # save data in custom location
# dataset$save(dir_other = save_dir) # dir_other not needed for non-test-projects
# 
# # open folder where file was saved
# browseURL(save_dir)

## ----eval=FALSE---------------------------------------------------------------
# projects$load("TEST_CLASSIC")$
#   sync()$
#   load_dataset("REDCapSync", envir = globalenv())

## ----eval=FALSE---------------------------------------------------------------
# save_dir <- tempdir() # replace with your intended folder
# 
# #load and sync (load_project(...) is the same as projects$load(...))
# project <- load_project("TEST_LONGITUDINAL")$sync()
# 
# # see field_names
# project$metadata$fields$field_name
# 
# males_dataset <- project$generate_dataset(dataset_name = "males",
#                                           envir = globalenv(), # global env
#                                           filter_field = "birth_sex",
#                                           filter_choices = "Male")
# # if you check merged and survey it will only contain males
# unique(males_dataset$data$merged$birth_sex) # Male
# unique(males_dataset$data$survey$birth_sex) # Male
# 
# females_dataset <- project$generate_dataset(dataset_name = "females",
#                                             envir = globalenv(), # global env
#                                             filter_field = "birth_sex",
#                                             filter_choices = "Female")
# # if you check merged and survey it will only contain males
# unique(females_dataset$data$merged$birth_sex) # Female
# unique(females_dataset$data$survey$birth_sex) # Female
# 
# # if you want this saved to excel one time you can do
# females_dataset$save(dir_other = save_dir) # dir_other not needed for default
# 
# # To keep this analysis refreshing with each sync use project$add_dataset(..,)
# project$add_dataset(dataset_name = "males",
#                     filter_field = "birth_sex",
#                     filter_choices = "Male")
# # not it can be loaded by name!
# males_dataset <- project$load_dataset("males") # will not send to global
# 
# project$load_dataset("males", envir = globalenv()) # will send to global
# 
# project$save_dataset("males") # will save AND keep track in project object
# 
# project$sync() # once defined will save datasets with sync

## ----eval=FALSE---------------------------------------------------------------
# #load and sync add datasets and save
# project <- load_project("TEST_LONGITUDINAL")$
#   sync()$
#   add_dataset(dataset_name = "males",
#               filter_field = "birth_sex",
#               filter_choices = "Male")$
#   add_dataset(dataset_name = "females",
#               filter_field = "birth_sex",
#               filter_choices = "Female")
# # load males
# project$load_dataset("males", envir = globalenv()) # will send to global
# 
# # load females
# project$load_dataset("females", envir = globalenv()) # will send to global
# 
# # load all default
# project$load_dataset("REDCapSync", envir = globalenv()) # will send to global
# 
# # can trigger saves with `project$save_datasets()` or default `project$sync()`
# project$save_datasets()
# project$sync()

generate_hex <- function(length = 32L) {
  c(0L:9L, letters[1L:6L]) |>
    sample(length, replace = TRUE) |>
    paste(collapse = "") |>
    toupper()
}
generate_comments <- function(action_type = "Add",
                              record = "1",
                              field_name = "var_branching",
                              comment = "some comment") {
  text_string <- paste0(action_type,
                        " field comment (Record: ",
                        record,
                        ", Field: ",
                        field_name)
  if (!is.null(comment)) {
    text_string <- paste0(text_string, ", Comment: \"", comment, "\"")
  }
  text_string <- paste0(text_string, ")")
  text_string
}
real_test_project <- function(project_name = "TEST_REDCAPR_SIMPLE") {
  assert_choice(project_name, TEST_REDCAPR_NAMES)
  redcap_uri <- "https://redcap-dev-2.ouhsc.edu/redcap/api/"
  project <- setup_project(project_name = project_name,
                           dir_path = cache_path(),
                           redcap_uri = redcap_uri,
                           hard_reset = TRUE)
  project
}
mock_test_project <- function(project_name = "TEST_CLASSIC") {
  assert_choice(project_name, TEST_PROJECT_NAMES)
  project <- load_test_project(project_name = project_name,
                               dir_path = cache_path())
  project
}
mock_test_calls <- function(project_name = "TEST_CLASSIC") {
  assert_choice(project_name, TEST_PROJECT_NAMES)
  file_name <- paste0(project_name, "_call_list.rds")
  call_list <- readRDS(test_path("fixtures", file_name))
  call_list
}
REAL_DEV_TOKENS <- c(
  REDCAPSYNC_TEST_REDCAPR_SIMPLE = "9A068C425B1341D69E83064A2D273A70",
  REDCAPSYNC_TEST_REDCAPR_LONGITUDINAL = "DA6F2BB23146BD5A7EA3408C1A44A556",
  REDCAPSYNC_TEST_REDCAPR_CLIN_TRIAL = "F9CBFFF78C3D78F641BAE9623F6B7E6A"
)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----example, eval=FALSE------------------------------------------------------
# library(REDCapSync)
# projects$print()
# # View(projects$df())                   # Vew your previously saved projects

## ----eval=FALSE---------------------------------------------------------------
# # 1.) setting your token -------------------------------------------------------
# Sys.setenv(REDCAPSYNC_FIRST_PROJECT = "YoUrNevErShaReToken")    # in console
# # or WAY BETTER put this in your .Renviron file...
# # REDCAPSYNC_FIRST_PROJECT = 'YoUrNevErShaReToken
# # Then save file, restart R session (`.rs.restartR()`) and library(REDCapSync)
# 
# # 2.) setting up a project -----------------------------------------------------
# project <- setup_project(
#   project_name = "FIRST_PROJECT",
#   redcap_uri = "https://redcap.fake.edu/api/",             # same as REDCapR
#   dir_path = getwd(),                            # choose appropriate folder
#   sync_frequency = "daily",                          # only checks max daily
#   get_entire_log = TRUE                       # for small or medium projects
# )
# 
# # install.packages("keyring")
# project$test_token()                    # will launch keyring if token fails
# 
# # 3.) running project$sync() ---------------------------------------------------
# project$sync()
# 
# project$generate_dataset("custom", envir = globalenv())

## ----eval=FALSE---------------------------------------------------------------
# #1. set in your Renviron file
# usethis::edit_r_environ()                                 #finds your file
# # another way to find same file
# rstudioapi::navigateToFile(file.path(Sys.getenv("HOME")), ".Renviron"))
# # Then you add --> `REDCAPSYNC_FIRST_PROJECT = 'YoUrNevErShaReToken'`
# # then save file and restart R session (`.rs.restartR()`) and reload package
# 
# #2. set to your token with keyring package! (make sure is installed)
# keyring::key_set(service = config$keyring.service(), # for keyring
#                  username = "FIRST_PROJECT", # is project_name
#                  keyring = config$keyring()) # now enter token in pop-up
# # ... or if have project object already
# project$set_keyring_token() # now enter token in pop-up
# 
# #3. set each time in your console or script (not recommended!) RISKY
# Sys.setenv(REDCAPSYNC_FIRST_PROJECT = "YoUrNevErShaReToken")
# 
# # Confirm
# # If it worked you will see your token when you run...
# Sys.getenv("REDCAPSYNC_FIRST_PROJECT") # for R eviron
# # or...
# keyring::key_get(service = config$keyring.service(), # for keyring
#                  username = "FIRST_PROJECT", # is project_name
#                  keyring = config$keyring())

## ----eval=FALSE---------------------------------------------------------------
# # raw REDCap -------------------------------------------------------------------
# forms <- project$metadata$forms
# fields <- project$metadata$fields
# choices <- project$metadata$choices
# log <- project$redcap$log
# users <- project$redcap$users
# form1 <- project$data$form1
# form2 <- project$data$form2
# # project$data |> list2env(globalenv()) # all forms to global
# 
# # summarized REDCap ------------------------------------------------------------
# # has added columns using log to annotate metadata and users etc. customizable
# # help(dataset)
# # dataset <- project$load_dataset("REDCapSync") # de-identified and no free text
# dataset <- project$generate_dataset("custom") # identified (see parameters)
# annotated_forms <- dataset$metadata$forms
# annotated_fields <- dataset$metadata$fields
# annotated_choices <- dataset$metadata$choices
# annotated_users <- dataset$users
# merged <- dataset$data$merged
# # dataset$data |> list2env(globalenv()) # all forms to global
# # dataset$to_envir(globalenv()) # send to R global env
# # can modify if you want before save (the actual dataset object not global)
# # dataset$save() # will save to output folder by default

## ----eval=FALSE---------------------------------------------------------------
# project$load_dataset(summary_name = "REDCapSync", envir = globalenv())

## ----eval=FALSE---------------------------------------------------------------
# project$generate_dataset(
#   dataset_name = "custom_name"
#   filter_field = "has_something"  # Raw REDCap variable name
#   filter_choices = "Yes"          # REDCap labelled choice
#   exclude_identifiers = TRUE,
#   exclude_free_text = TRUE,
#   include_metadata = TRUE,
#   envir = globalenv()
# )

#' @keywords internal
"_PACKAGE"
## usethis namespace: start
#' @importFrom checkmate assert
#' @importFrom checkmate assert_character
#' @importFrom checkmate assert_choice
#' @importFrom checkmate assert_class
#' @importFrom checkmate assert_data_frame
#' @importFrom checkmate assert_date
#' @importFrom checkmate assert_directory
#' @importFrom checkmate assert_directory_exists
#' @importFrom checkmate assert_environment
#' @importFrom checkmate assert_integerish
#' @importFrom checkmate assert_list
#' @importFrom checkmate assert_logical
#' @importFrom checkmate assert_names
#' @importFrom checkmate assert_path_for_output
#' @importFrom checkmate assert_posixct
#' @importFrom checkmate assert_r6
#' @importFrom checkmate assert_set_equal
#' @importFrom checkmate assert_string
#' @importFrom checkmate assert_tibble
#' @importFrom checkmate assert_true
#' @importFrom checkmate check_directory_exists
#' @importFrom checkmate check_set_equal
#' @importFrom checkmate check_true
#' @importFrom checkmate expect_character
#' @importFrom checkmate expect_class
#' @importFrom checkmate expect_data_frame
#' @importFrom checkmate expect_directory_exists
#' @importFrom checkmate expect_file_exists
#' @importFrom checkmate expect_list
#' @importFrom checkmate expect_logical
#' @importFrom checkmate expect_r6
#' @importFrom checkmate expect_scalar_na
#' @importFrom checkmate test_character
#' @importFrom checkmate test_choice
#' @importFrom checkmate test_class
#' @importFrom checkmate test_directory_exists
#' @importFrom checkmate test_file_exists
#' @importFrom checkmate test_integerish
#' @importFrom checkmate test_logical
#' @importFrom checkmate test_scalar_na
#' @importFrom cli cli_abort
#' @importFrom cli cli_alert_danger
#' @importFrom cli cli_alert_info
#' @importFrom cli cli_alert_success
#' @importFrom cli cli_alert_warning
#' @importFrom cli cli_bullets
#' @importFrom cli cli_code
#' @importFrom cli cli_li
#' @importFrom cli cli_progress_bar
#' @importFrom cli cli_progress_update
#' @importFrom cli cli_text
#' @importFrom cli cli_h1
#' @importFrom cli cli_h2
#' @importFrom cli cli_h3
#' @importFrom cli cli_warn
#' @importFrom cli format_message
#' @importFrom dplyr as_tibble
#' @importFrom dplyr bind_rows
#' @importFrom dplyr filter
#' @importFrom dplyr first
#' @importFrom dplyr last
#' @importFrom dplyr pull
#' @importFrom dplyr rename
#' @importFrom dplyr select
#' @importFrom dplyr starts_with
#' @importFrom dplyr tibble
#' @importFrom R6 R6Class
#' @importFrom redcapAPI exportLogging
#' @importFrom redcapAPI exportVersion
#' @importFrom redcapAPI REDCAP_METADATA_FIELDTYPE
#' @importFrom redcapAPI REDCAP_METADATA_VALIDATION_TYPE
#' @importFrom redcapAPI REDCAP_PROJECT_PURPOSE
#' @importFrom redcapAPI REDCAP_SYSTEM_FIELDS
#' @importFrom redcapAPI redcapConnection
#' @importFrom REDCapR redcap_delete
#' @importFrom REDCapR redcap_file_download_oneshot
#' @importFrom REDCapR redcap_read
#' @importFrom REDCapR redcap_read_oneshot
#' @importFrom REDCapR redcap_report
#' @importFrom REDCapR redcap_write
#' @importFrom REDCapR sanitize_token
#' @importFrom stringr str_extract
#' @importFrom stringr str_length
#' @importFrom stringr str_match
#' @importFrom stringr str_remove
#' @importFrom stringr str_replace_all
#' @importFrom stringr str_split_fixed
#' @importFrom stringr str_to_title
#' @importFrom stringr str_trunc
#' @importFrom utils browseURL
#' @importFrom utils head
#' @importFrom utils object.size
#' @importFrom utils write.csv
## usethis namespace: end
NULL

#' @title Synchronize REDCap Data
#' @description
#' Syncs with REDCap via `project` object that user defined with
#' \link{setup_project}
#' @details
#' Syncs all projects by default but can be used to hands-free sync one or
#' defined set projects. This is not intended to return project object. User
#' should use `load_project("project_name")`. However, by default will invisibly
#' return the last project in the set of project_names.
#' @param save_datasets Logical (TRUE/FALSE). If TRUE, saves datasets to
#' directory.
#' @param hard_check Will check REDCap even if not due (see `sync_frequency`
#' parameter from `setup_project()`)
#' @param hard_reset Logical that forces a fresh update if TRUE. Default is
#' `FALSE`.
#' @param project_names character vector of project project_names previously
#' setup. If NULL, will get all from `get_projects()`
#' @examples
#' \dontrun{
#' # if you setup 3 projects "ONE", "TWO", "THREE" using [setup_project()]...
#' sync() # will check all three projects for updates if due for sync
#' sync("TWO") # will only check and sync project TWO
#' project <- sync("TWO") # can also be used for invisible return ...
#' sync(c("ONE","THREE")) # will only check and sync projects ONE and THREE
#' project <- sync(c("ONE","THREE")) # invisible return of THREE ...
#' }
#' @returns Invisibly returns last project object (R6 [REDCapSyncProject] class)
#'
#' @seealso
#' \code{vignette("REDCapSync", package = "REDCapSync")}
#' \code{vignette("Cache", package = "REDCapSync")}
#' [setup_project] for initializing projects
#' [project] for using the project objects
#'
#' @export
sync <- function(project_names = NULL,
                 save_datasets = TRUE,
                 hard_check = FALSE,
                 hard_reset = FALSE) {
  sweep_dirs_for_cache(project_names = project_names)
  if (is.null(project_names)) {
    projects <- get_projects()
    project_names <- projects$project_name
    if (length(project_names) == 0L) {
      cli_alert_info("No projects in cache. Start with `?setup_project()`")
      return(invisible())
    }
  }
  cli_progress_bar("Syncing REDCap Projects", total = length(project_names))
  for (project_name in project_names) {
    project <- try_else_null(load_project(project_name))
    if (is.null(project)) {
      cli_alert_danger("Unable to load {project_name}")
      #add to bad list
    } else {
      project$sync(
        save_to_dir = TRUE,
        save_datasets = save_datasets,
        hard_check = hard_check,
        hard_reset = hard_reset
      )
    }
    cli_progress_update()
  }
  # consider adding message/df
  cli_alert_success("All projects are synced!")
  invisible(project)
}
#' @noRd
sync_project <- function(project,
                         save_datasets = TRUE,
                         save_to_dir = TRUE,
                         hard_check = FALSE,
                         hard_reset = FALSE) {
  assert_setup_project(project)
  assert_logical(hard_reset,
                 any.missing = FALSE,
                 len = 1L)
  assert_logical(save_to_dir,
                 any.missing = FALSE,
                 len = 1L)
  hard_reset <- hard_reset || project$internals$hard_reset
  do_sync <- due_for_sync(project_name = project$project_name) ||
    hard_reset || hard_check
  was_updated <- FALSE
  if (!do_sync) {
    info_message <- paste0("{project$project_name} not due for sync ",
                           "({project$settings$sync_frequency})")
    cli_alert_info(info_message)
  }
  if (do_sync) {
    project <- sync_project_check(project = project, hard_reset = hard_reset)
    was_updated <- project$internals$was_updated
  }
  if (project$settings$add_default_datasets) {
    missing_redcapsync <- !is_something(project$datasets$REDCapSync)
    missing_redcapsync_raw <- !is_something(project$datasets$REDCapSync_raw)
    if (missing_redcapsync || missing_redcapsync_raw) {
      project <- add_default_datasets(project)
    }
  }
  # ?add_default_fields
  if (save_to_dir && !is.null(project$dir_path)) {
    if (is_something(project$data) && project$settings$get_files) {
      get_redcap_files(
        # would want track internally?
        project,
        original_file_names = project$settings$original_file_names,
        overwrite = TRUE
      ) # account for needed overwrites from new records
    }
    if (project$settings$get_file_repository) {
      # get redcap file repo
    }
    if (save_datasets) {
      first_stamp <- project$internals$last_dataset_save
      project <- save_project_datasets(project = project,
                                       hard_reset = hard_reset)
      second_stamp <- project$internals$last_dataset_save
      was_updated <- was_updated ||
        !identical(first_stamp, second_stamp)
    }
    if (was_updated) {
      project <- save_project(project)
    } else {
      project$internals$last_directory_save <- project$internals$last_sync
      project_details <- extract_project_details(project)
      saveRDS(
        object = project_details,
        file = get_project_path2(project, type = "details")
      ) # add error check
      add_project_details_to_cache(project_details)
    }
  }
  invisible(project)
}
#' @noRd
sync_project_hard_reset <- function(project) {
  assert_setup_project(project)
  project <- get_redcap_metadata(project = project)
  project <- update_project_links(project)
  if (project$settings$get_data) {
    project$data <- list()
    project$datasets <- list()
    project$data <- get_redcap_data(
      project = project,
      labelled = project$settings$labelled
    )
    # if error records comma
    if (project$redcap$has_log_access) {
      if (project$settings$get_entire_log) {
        creation_time <- project$redcap$project_info$creation_time
        log_begin_date <- as.Date(as.POSIXct(creation_time))
        log_begin_date <- log_begin_date - 1L
      } else {
        log_begin_date <- Sys.Date() - project$settings$log_days
      }
      # don't need old log in this case
      project$redcap$log <- get_redcap_log(project = project,
                                           log_begin_date = log_begin_date)
    }
    right_now <- now_time()
    project$internals$last_full_update <-
      project$internals$last_metadata_update <-
      project$internals$last_data_update <- right_now
    record_summary <- extract_project_records(project)
    if (is_something(record_summary)) {
      project$record_summary <- record_summary
      project$record_summary$last_api_call <- right_now
    }
  }
  project$internals$hard_reset <- FALSE
  cli_alert_success("Full {project$project_name} update!")
  project
}
#' @noRd
sync_project_check <- function(project, hard_reset = FALSE, records = NULL) {
  project$internals$was_updated <- FALSE
  project <- test_project_token(project)
  connected <- project$internals$last_test_connection_outcome
  if (!connected) {
    cli_alert_danger("Could not connect to REDCap")
    return(invisible(project))
  }
  hard_reset <- hard_reset ||
    !is_something(project$internals$last_metadata_update) ||
    !is_something(project$internals$last_data_update) ||
    !is_something(project$internals$last_full_update)
  if (!hard_reset) {
    interim_log <- get_interim_log(project)
    if (is_something(interim_log)) {
      log_changes <- analyze_log(interim_log, id_col = project$metadata$id_col)
      log_change_messages(log_changes)
      hard_reset <- log_changes$hard_reset
      project$redcap$log <- interim_log |>
        bind_rows(project$redcap$log) |>
        unique()
      project$internals$was_updated <- TRUE
      if (!hard_reset) {
        if (log_changes$refresh_data || !is.null(records)) {
          deleted_records <- log_changes$deleted_records
          updated_records <- log_changes$updated_records
          refresh_records <- c(deleted_records, updated_records, records)
          refresh_records <- unique(refresh_records)
          project <- sync_project_refresh(project = project,
                                          refresh_records = refresh_records)
          if (is.null(project)) {
            # anything wrong with refresh should force hard_reset
            hard_reset <- TRUE
          }
        }
      }
    }
  }
  if (hard_reset) {
    project <- sync_project_hard_reset(project)
    project$internals$was_updated <- TRUE
  }
  project$internals$last_sync <- now_time()
  project
}
#' @noRd
sync_project_refresh <- function(project, refresh_records) {
  assert_setup_project(project)
  id_col <- project$metadata$id_col
  project$data <- all_character_cols_list(project$data)
  all_records <- get_redcap_records(project)
  project <- remove_records_from_project(project = project,
                                         records = refresh_records)
  form_list <- get_redcap_data(
    project = project,
    labelled = project$settings$labelled,
    records = refresh_records # still checking deletes
  )
  records_received <- extract_values_from_form_list(
    form_list = form_list,
    col_name = id_col
  )
  #check diff with refresh?
  if (!all(names(form_list) %in% project$metadata$forms$form_name)) {
    cli_alert_danger("Full update triggered: data mismatch")
    return(NULL)
  }
  if (!all(records_received %in% all_records)) {
    cli_alert_danger("Full update triggered: data mismatch")
    return(NULL)
  }
  for (form_name in names(form_list)) {
    project$data[[form_name]] <- project$data[[form_name]] |>
      all_character_cols() |>
      bind_rows(form_list[[form_name]])
    reorder_records <- match(all_records, project$data[[form_name]][[id_col]])
    project$data[[form_name]] <- project$data[[form_name]][reorder_records, ]
    rownames(project$data[[form_name]]) <- NULL
  }
  records_dataset <- project$record_summary[[id_col]]
  missing_records_i <- which(!records_received %in% records_dataset)
  missing_from_dataset <- records_received[missing_records_i]
  if (length(missing_from_dataset) > 0L) {
    x <- data.frame(
      record = missing_from_dataset,
      last_api_call = NA,
      was_saved = FALSE,
      stringsAsFactors = FALSE
    )
    colnames(x)[1L] <- id_col
    project$record_summary <- bind_rows(project$record_summary, x)
    new_order <- order(project$record_summary[[id_col]], decreasing = TRUE)
    project$record_summary <- project$record_summary[new_order, ]
  }
  project$internals$last_data_update <- now_time()
  row_new <- which(project$record_summary[[id_col]] %in% records_received)
  project$record_summary$last_api_call[row_new] <-
    project$internals$last_data_update
  project$record_summary$was_saved[row_new] <- FALSE
  project$record_summary
  reorder_records <- match(all_records, project$record_summary[[id_col]])
  project$record_summary <- project$record_summary[reorder_records, ]
  rownames(project$record_summary) <- NULL
  invisible(project)
}
#' @noRd
generate_comment_table <- function(redcap_log, only_most_recent = FALSE) {
  assert_data_frame(redcap_log)
  assert_logical(only_most_recent)
  redcap_log <- redcap_log[which(redcap_log$action_type == "Comment"), ]
  redcap_log$comment_type <- redcap_log$details |>
    strsplit(" ") |>
    lapply(first) |>
    unlist()
  inner <- sub("^.*?\\((.*)\\)$", "\\1", redcap_log$details)
  redcap_log$comment_field <- str_match(inner, ", Field:\\s([^,]+)")[, 2L]
  if (only_most_recent && is_something(redcap_log)) {
    redcap_log <- redcap_log |>
      split(redcap_log$record) |>
      lapply(function(record) {
        record |> split(record$comment_field) |> lapply(first) |>  bind_rows()
      }) |>
      bind_rows()
    keep_rows <- which(redcap_log$comment_type != "Delete")
    redcap_log <- redcap_log[keep_rows, ]
    inner <- sub("^.*?\\((.*)\\)$", "\\1", redcap_log$details)
  }
  if (nrow(redcap_log) == 0L) {
    return(NULL)
  }
  redcap_log$comment_details <- inner |>
    str_remove("^.*, Comment:\\s\"") |>
    str_remove("\"$")
  redcap_log$comment_details[which(
    redcap_log$comment_details == paste0(
      "Record: ",
      redcap_log$record,
      ", Field: ",
      redcap_log$comment_field
    )
  )] <- NA
  redcap_log
}
#' @noRd
due_for_sync <- function(project_name) {
  projects <- get_projects()
  if (!project_name %in% projects$project_name) {
    return(TRUE)
  }
  assert_names(projects$project_name, must.include = project_name)
  project_row <- which(projects$project_name == project_name)
  now <- now_time()
  last_sync <- projects$last_sync[project_row]
  sync_frequency <- projects$sync_frequency[project_row]
  if (is.na(last_sync)) {
    return(TRUE)
  }
  interval <- switch(
    sync_frequency,
    always  = return(TRUE),
    never   = return(FALSE),
    once    = return(FALSE),
    hourly  = lubridate::dhours(1L),
    daily   = lubridate::ddays(1L),
    weekly  = lubridate::dweeks(1L),
    monthly = lubridate::dmonths(1L),
    cli_abort("Unknown `sync_frequency`: {sync_frequency}")
  )
  now >= (last_sync + interval - lubridate::dminutes(5L))
}
#' @noRd
remove_from_form_list <- function(form_list, id_col, records = NULL) {
  if (!is_something(form_list)) {
    return(form_list)
  }
  if (!is_df_list(form_list)) {
    cli_abort("form_list is not a list of data.frames as expected.")
  }
  if (is.null(records)) {
    return(form_list)
  }
  rows_for_forms <- which(unlist(lapply(names(form_list), function(form_name) {
    nrow(form_list[[form_name]]) > 0L
  })))
  form_names <- names(form_list)[rows_for_forms]
  for (form_name in form_names) {
    chosen_rows <- which(!form_list[[form_name]][[id_col]] %in% records)
    form_list[[form_name]] <- form_list[[form_name]][chosen_rows, ]
  }
  form_list
}
#' @noRd
remove_records_from_project <- function(project, records) {
  id_col <- project$metadata$id_col
  if (length(records) == 0L) {
    cli_abort(
      paste0(
        "no records supplied to remove_records_from_project.",
        "Try {.code project$sync(hard_reset = TRUE)}"
      )
    )
  }
  project$data <- remove_from_form_list(form_list = project$data,
                                        id_col = id_col,
                                        records = records)
  project$transformation$data <- remove_from_form_list(
    form_list = project$transformation$data,
    id_col = id_col,
    records = records
  )
  keep_rows <- !project$record_summary[[id_col]] %in% records
  project$record_summary <- project$record_summary[which(keep_rows), ]
  invisible(project)
}

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# sync (Exported)
test_that("sync works!", {
  tempdir_test <- sanitize_path(withr::local_tempdir())
  withr::local_envvar(R_USER_CACHE_DIR = tempdir_test)
  local_mocked_bindings(
    sweep_dirs_for_cache = function(...) NULL,
    load_project = function(...) NULL,
    get_projects = function(...) BLANK_PROJECT_DETAILS
  )
  expect_message(sync(), "No projects in cache")
  local_mocked_bindings(
    sweep_dirs_for_cache = function(...) NULL,
    get_projects = function(...) {
      data.frame(project_name = "TEST_CLASSIC", stringsAsFactors = FALSE)
    }
  )
  expect_message(sync(), "Unable to load")
  local_mocked_bindings(
    load_project = function(...) mock_test_project()
  )
  expect_message(sync(), "TEST projects do not communicate with the API")
})
# due_for_sync (Internal)
test_that("due_for_sync works!", {
  # true for nonexistent
  expect_true(due_for_sync("NONEXISTENT_PROJECT"))
  project <- mock_test_project()$.internal
  project_details <- extract_project_details(project)
  project_details$last_sync <- NA
  add_project_details_to_cache(project_details)
  #returns TRUE when last_sync is NA
  expect_true(due_for_sync(project$project_name))
  #returns TRUE when sync_frequency is always
  project_details$sync_frequency <- "always"
  project_details$last_sync <- now_time()
  add_project_details_to_cache(project_details)
  expect_true(due_for_sync(project$project_name))
  #returns FALSE when sync_frequency is never
  project_details$sync_frequency <- "never"
  project_details$last_sync <- now_time() - lubridate::ddays(100L)
  add_project_details_to_cache(project_details)
  expect_false(due_for_sync(project$project_name))
  #returns FALSE for hourly when last_sync < 1 hour ago
  project_details$sync_frequency <- "hourly"
  project_details$last_sync <- now_time() - lubridate::dminutes(30L)
  add_project_details_to_cache(project_details)
  expect_false(due_for_sync(project$project_name))
  #returns TRUE for hourly when last_sync >= 1 hour ago
  project_details$last_sync <- now_time() - lubridate::dminutes(30L)
  add_project_details_to_cache(project_details)
  expect_false(due_for_sync(project$project_name))
  #returns FALSE for daily when last_sync < 1 day ago
  project_details$sync_frequency <- "daily"
  project_details$last_sync <- now_time() - lubridate::dhours(12L)
  add_project_details_to_cache(project_details)
  expect_false(due_for_sync(project$project_name))
  #returns TRUE for daily when last_sync >= 1 day ago
  project_details$last_sync <- now_time() - lubridate::ddays(2L)
  add_project_details_to_cache(project_details)
  expect_true(due_for_sync(project$project_name))
  #returns FALSE for weekly when last_sync < 1 week ago
  project_details$sync_frequency <- "weekly"
  project_details$last_sync <- now_time() - lubridate::ddays(3L)
  add_project_details_to_cache(project_details)
  expect_false(due_for_sync(project$project_name))
  #returns TRUE for weekly when last_sync >= 1 week ago
  project_details$last_sync <- now_time() - lubridate::dweeks(2L)
  add_project_details_to_cache(project_details)
  expect_true(due_for_sync(project$project_name))
  #returns FALSE for monthly when last_sync < 1 month ago
  project_details$sync_frequency <- "monthly"
  project_details$last_sync <- now_time() - lubridate::ddays(15L)
  add_project_details_to_cache(project_details)
  expect_false(due_for_sync(project$project_name))
  #returns TRUE for monthly when last_sync >= 1 month ago
  project_details$last_sync <- now_time() - lubridate::dmonths(2L)
  add_project_details_to_cache(project_details)
  expect_true(due_for_sync(project$project_name))
  #returns FALSE for once when already synced
  project_details$sync_frequency <- "once"
  project_details$last_sync <- now_time()
  add_project_details_to_cache(project_details)
  expect_false(due_for_sync(project$project_name))
  #returns FALSE for once when already synced
  project_details$last_sync <- now_time()
  add_project_details_to_cache(project_details)
  expect_false(due_for_sync(project$project_name))
  #returns TRUE for once when never synced
  project_details$last_sync <- NA
  add_project_details_to_cache(project_details)
  expect_true(due_for_sync(project$project_name))
})
# get_interim_log (Internal)
test_that("get_interim_log works!", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  interim_log <- head(project$redcap$log, n = 10L)
  # Update record 3
  project$redcap$has_log_access <- FALSE
  expect_message(get_interim_log(project), "You do not have logging access")
  project$redcap$has_log_access <- TRUE
  local_mocked_bindings(
    get_redcap_log = function(...) interim_log
  )
  expect_data_frame(get_interim_log(project), nrows = 0L)
  new_interim_log_info <- data.frame(
    timestamp = as.character(Sys.time()),
    username = "u1231",
    action = "Create record (API) 99",
    details = "some details",
    record = "99",
    action_type = "Create",
    stringsAsFactors = FALSE
  )
  interim_log <- bind_rows(new_interim_log_info, interim_log)
  local_mocked_bindings(
    get_redcap_log = function(...) interim_log
  )
  extracted_interim <- get_interim_log(project)
  expect_data_frame(extracted_interim, nrows = 1L)
  expect_identical(new_interim_log_info, extracted_interim)
})
# remove_from_form_list (Internal)
test_that("remove_from_form_list works!", {
  # Setup: list of data.frames with record_id
  df1 <- data.frame(
    record_id = c("1", "2", "3"),
    value = c("A", "B", "C"),
    stringsAsFactors = FALSE
  )
  df2 <- data.frame(
    record_id = c("2", "3", "4"),
    value = c("X", "Y", "Z"),
    stringsAsFactors = FALSE
  )
  form_list <- list(form1 = df1, form2 = df2)
  # Remove records "2" and "3"
  out <- remove_from_form_list(form_list = form_list,
                               id_col = "record_id",
                               records = c("2", "3"))
  expect_identical(out$form1$record_id, "1")
  expect_identical(out$form2$record_id, "4")
  # Remove no records (NULL)
  out2 <- remove_from_form_list(form_list, id_col = "record_id", records = NULL)
  expect_identical(out2$form1$record_id, c("1", "2", "3"))
  expect_identical(out2$form2$record_id, c("2", "3", "4"))
  # Remove all records
  out3 <- remove_from_form_list(form_list = form_list,
                                id_col = "record_id",
                                records = c("1", "2", "3", "4"))
  expect_identical(nrow(out3$form1), 0L)
  expect_identical(nrow(out3$form2), 0L)
  # Empty form_list returns itself
  expect_identical(remove_from_form_list(form_list = list(),
                                         id_col = "record_id",
                                         records = "1"),
                   list())
})
# remove_records_from_project (Internal)
test_that("remove_records_from_project works!", {
  project <- mock_test_project()$.internal
  # Add some records to form_example and transformation$data
  project$data$form_example <- data.frame(
    record_id = c("1", "2", "3"),
    var_text = c("A", "B", "C"),
    stringsAsFactors = FALSE
  )
  project$metadata$id_col <- "record_id"
  project$transformation$data <- list(
    form_example = data.frame(
      record_id = c("1", "2", "3"),
      var_text = c("A", "B", "C"),
      stringsAsFactors = FALSE
    )
  )
  # Remove records "2" and "3"
  project2 <- remove_records_from_project(project, records = c("2", "3"))
  expect_identical(project2$data$form_example$record_id, "1")
  expect_identical(project2$transformation$data$form_example$record_id, "1")
  # Remove no records (empty vector) should error
  expect_error(remove_records_from_project(project, records = character(0L)))
  # Remove all records
  project3 <- remove_records_from_project(project, records = c("1", "2", "3"))
  expect_identical(nrow(project3$data$form_example), 0L)
  expect_identical(nrow(project3$transformation$data$form_example), 0L)
})
# sync_project (Internal)
test_that("sync_project will update if new", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  project$internals$last_test_connection_outcome <- TRUE
  project$internals$last_metadata_update <- now_time() - lubridate::ddays(100L)
  project$internals$last_data_update <- now_time() - lubridate::ddays(100L)
  project$internals$was_updated <- FALSE
  interim_log <- head(project$redcap$log, n = 10L)
  ordered_records <- project$record_summary$record_id
  row_update <- which(project$data$text$record_id == "50")
  row_existing <- which(project$data$text$record_id == "2")
  var_text_integer_update <- project$data$text$var_text_integer[row_update]
  var_text_integer_existing <- project$data$text$var_text_integer[row_existing]
  log_changes <-  analyze_log(interim_log, id_col = project$metadata$id_col)
  updated_records <- log_changes$updated_records
  expect_contains(updated_records, "50")
  expect_false("2" %in% updated_records)
  new_data <- filter_data_list(
    data_list = project,
    filter_field = "record_id",
    filter_choices = updated_records
  )
  row_update <- which(new_data$text$record_id == "50")
  row_existing <- which(new_data$text$record_id == "2")
  expect_length(row_update, 1L)
  expect_length(row_existing, 0L)
  new_var_text_integer <- new_data$text$var_text_integer[row_update]
  expect_identical(new_var_text_integer, var_text_integer_update)
  new_data$text$var_text_integer[row_update] <- "99"
  new_var_text_integer <- new_data$text$var_text_integer[row_update]
  expect_true(new_var_text_integer != var_text_integer_update)
  # Mock the test_project_token to simulate successful connection
  local_mocked_bindings(
    test_project_token =  function(...) project,
    get_interim_log =  function(...) interim_log,
    get_redcap_data =  function(...) new_data,
    get_redcap_records = function(...) ordered_records,
    due_for_sync = function(...) TRUE
  )
  # Call sync_project_check with hard_reset = FALSE
  result <- sync_project(
    project = project,
    save_datasets = TRUE,
    save_to_dir = TRUE,
    hard_reset = FALSE
  )
  row_update <- which(result$data$text$record_id == "50")
  row_existing <- which(result$data$text$record_id == "2")
  var_text_integer_update <- result$data$text$var_text_integer[row_update]
  var_text_integer_existing <- result$data$text$var_text_integer[row_existing]
  expect_identical(new_var_text_integer, var_text_integer_update)
  expect_identical(var_text_integer_existing, var_text_integer_existing)
})
test_that("sync_project will work if not due", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  local_mocked_bindings(
    due_for_sync = function(...) FALSE
  )
  expect_message(sync_project(
    project = project,
    save_datasets = FALSE,
    save_to_dir = TRUE,
    hard_reset = FALSE
  ), "not due for sync")
})
# sync_project_check (Internal)
test_that("sync_project_check cannot connect!", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  project$internals$last_test_connection_outcome <- FALSE
  local_mocked_bindings(
    test_project_token = function(...) project
  )
  expect_message(sync_project_check(project), "Could not connect")
})
test_that("sync_project_check wont update if nothing new", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  # Setup: ensure project is ready for sync_project_check
  project$internals$last_test_connection_outcome <- TRUE
  project$internals$last_metadata_update <- now_time() - lubridate::ddays(100L)
  project$internals$last_data_update <- now_time() - lubridate::ddays(100L)
  project$internals$was_updated <- FALSE
  interim_log <- head(project$redcap$log, n = 0L)
  local_mocked_bindings(
    test_project_token = function(...) project,
    get_interim_log = function(...) interim_log
  )
  # Mock the test_project_token to simulate successful connection
  # Call sync_project_check with hard_reset = FALSE
  result <- sync_project_check(project, hard_reset = FALSE)
  expect_false(result$internals$was_updated)
})
# sync_project_hard_reset (Internal)
test_that("sync_project_hard_reset works!", {
  project_name <- "TEST_CLASSIC"
  project <- mock_test_project(project_name)$.internal
  call_list <- mock_test_calls(project_name)
  local_mocked_bindings(
    get_redcap_metadata = function(...) project,
    update_project_links = function(...) project,
    get_redcap_data = function(...) project$data,
    get_redcap_log = function(...) project$redcap$log
  )
  result <- project
  result$metadata <- BLANK_PROJECT$metadata
  result["data"] <- list(NULL)
  result <- sync_project_hard_reset(result)
  expect_identical(result$metadata, project$metadata)
  expect_identical(result$data, project$data)
})
# sync_project_refresh (Internal)
test_that("sync_project_refresh works!", {
})

#' @noRd
assert_dir <- function(dir_path, silent = FALSE) {
  # param check
  dir_path <- clean_dir_path(dir_path)
  if (!file.exists(dir_path)) {
    cli_abort("dir_path does not exist")
  }
  for (folder in DIR_FOLDERS) {
    if (!file.exists(file.path(dir_path, folder))) {
      cli_abort("'{dir_path}/{folder}' missing! Use `setup_project()`")
    }
  }
  if (!silent) {
    cli_alert_success("Directory is Valid! {.file {dir_path}}")
  }
  dir_path
}
#' @noRd
assert_web_link <- function(link) {
  #change to validate api endpoint
  if (is.null(link))
    cli_abort("link is NULL")
  # Check if the link starts with "https://" or "http://"
  if (!grepl("^https?://", link)) {
    cli_abort("Invalid web link. It must start with 'http://' or 'https://'.")
  }
  # Remove trailing slash if present
  link <- gsub("/$", "", link)
  # Check if the link ends with one of the specified web endings
  if (!grepl("\\.(edu|com|org|net|gov|io|xyz|info|co|uk)$", link)) {
    cli_abort("Invalid web link. It must end with '.edu', '.com', etc.")
  }
  # Add a trailing slash
  link <- paste0(link, "/")
  link
}
#' @noRd
assert_env_name <- function(x, max.chars = 26L, all_caps = FALSE) {
  assert_character(x, len = 1L, min.chars = 1L, any.missing = FALSE)
  # can change to min.length for projects
  assert_integerish(max.chars,
                    len = 1L,
                    lower = 1L,
                    upper = 255L,
                    any.missing = FALSE)
  assert_string(
    x,
    n.chars = NULL,
    min.chars = 1L,
    max.chars = max.chars,
    pattern =  "^[A-Za-z][A-Za-z0-9_]*$",
    fixed = NULL,
    ignore.case = TRUE
  )
  if (all_caps) {
    assert_string(
      x,
      n.chars = NULL,
      min.chars = 1L,
      max.chars = max.chars,
      pattern = "^[A-Z][A-Z0-9_]*$",
      fixed = NULL,
      ignore.case = FALSE
    )
  }
  invisible(x)
}
#' @noRd
assert_blank_project <- function(project) {
  assert_list(project, names = "unique", len = length(BLANK_PROJECT))
  assert_names(names(project),
               type = "unique",
               must.include = names(BLANK_PROJECT))
  invisible(project)
}
#' @noRd
assert_setup_project <- function(project) {
  assert_blank_project(project)
  assert_env_name(project$project_name, max.chars = 31L, all_caps = TRUE)
  # dir_path
  # redcap_uri
  assert_env_name(project$token_name, max.chars = 50L, all_caps = TRUE)
  #redcap --------------
  assert_choice(project$redcap$timezone, OlsonNames())
  #settings --------------
  assert_choice(project$settings$sync_frequency, choices = SYNC_FREQUENCY)
  assert_logical(project$settings$labelled, len = 1L, any.missing = FALSE)
  assert_choice(project$settings$get_type, choices = GET_TYPE)
  assert(
    test_scalar_na(project$settings$records) ||
      test_character(
        project$settings$records, # add exist warning
        min.chars = 1L,
        unique = TRUE,
        min.len = 1L,
        any.missing = FALSE
      )
  )
  assert(
    test_scalar_na(project$settings$fields) ||
      test_character(
        project$settings$fields, # add exist warning
        min.chars = 1L,
        unique = TRUE,
        min.len = 1L,
        any.missing = FALSE
      )
  )
  assert(
    test_scalar_na(project$settings$forms) ||
      test_character(
        project$settings$forms, # add exist warning
        min.chars = 1L,
        unique = TRUE,
        min.len = 1L,
        any.missing = FALSE
      )
  )
  assert(
    test_scalar_na(project$settings$events) ||
      test_character(
        project$settings$events, # add exist warning
        min.chars = 1L,
        unique = TRUE,
        min.len = 1L,
        any.missing = FALSE
      )
  )
  assert(
    test_scalar_na(project$settings$filter_logic) ||
      test_character(
        project$settings$filter_logic, # add exist warning
        min.chars = 1L,
        unique = TRUE,
        min.len = 1L,
        any.missing = FALSE
      )
  )
  assert_integerish(
    project$settings$id_position,
    len = 1L,
    lower = 1L,
    any.missing = FALSE
  )
  assert_logical(project$settings$get_users, len = 1L, any.missing = FALSE)
  assert_logical(project$settings$get_data, len = 1L, any.missing = FALSE)
  assert_integerish(
    project$settings$batch_size_download,
    len = 1L,
    lower = 1L,
    any.missing = FALSE
  )
  assert_integerish(
    project$settings$batch_size_upload,
    len = 1L,
    lower = 1L,
    any.missing = FALSE
  )
  assert_logical(project$settings$get_entire_log, len = 1L, any.missing = FALSE)
  assert_integerish(
    project$settings$log_days,
    len = 1L,
    lower = 1L,
    any.missing = FALSE
  )
  assert_logical(project$settings$log_drop_details,
                 len = 1L,
                 any.missing = FALSE)
  assert_logical(project$settings$log_drop_exports,
                 len = 1L,
                 any.missing = FALSE)
  assert_logical(project$settings$get_files, len = 1L, any.missing = FALSE)
  assert_logical(project$settings$get_file_repository,
                 len = 1L,
                 any.missing = FALSE)
  assert_logical(project$settings$original_file_names,
                 len = 1L,
                 any.missing = FALSE)
  assert_logical(project$settings$add_default_datasets,
                 len = 1L,
                 any.missing = FALSE)
  #internals --------------
  assert_logical(project$internals$hard_reset, len = 1L, any.missing = FALSE)
  assert_logical(project$internals$was_updated,
                 len = 1L,
                 any.missing = FALSE)
  invisible(project)
}
#' @noRd
assert_project_name <- function(project_name,
                                allow_test_names = config$allow.test.names()) {
  if (grepl("[a-z]", project_name)) {
    project_name <- toupper(project_name)
    end_message <- paste0("`project_name` must be all caps!",
                          " For example, try '{project_name}'")
    cli_abort(end_message)
  }
  assert_env_name(project_name, max.chars = 31L, all_caps = TRUE)
  if (project_name %in% c("CONFIG", "TEST")) {
    cli_abort("`project_name` can't be '{project_name}'! Please choose again.")
  }
  if (startsWith(project_name, "CONFIG_")) {
    cli_abort("`project_name` can't start with 'CONFIG_'! Please choose again.")
  }
  if (!allow_test_names && startsWith(project_name, "TEST_")) {
    cli_abort("`project_name` can't start with 'TEST_'! Please choose again.")
  }
  project_name
}
#' @noRd
test_setup_project <- function(project) {
  project <- tryCatch(
    expr = {
      suppressWarnings({
        assert_setup_project(project = project)
      })
    },
    error = function(e) {
      NULL
    }
  )
  !is.null(project)
}
#' @noRd
assert_project_details <- function(project_details, nrows = NULL) {
  assert_data_frame(
    x = project_details,
    nrows = nrows,
    ncols = length(BLANK_PROJECT_COLS)
  )
  assert_names(colnames(project_details), must.include = BLANK_PROJECT_COLS)
  if (nrow(project_details) > 0L) {
    project_details$project_name |>
      lapply(function(project_name) {
        assert_env_name(project_name, max.chars = 31L, all_caps = TRUE)
      })
    assert_names(project_details$project_name, type = "unique")
  }
  project_details
}
#' @noRd
assert_project_path <- function(project_path) {
  assert_path_for_output(x = project_path,
                         overwrite = TRUE,
                         extension = "RData")
  assert_true(endsWith(basename(project_path), PROJECT_PATH_SUFFIX))
}
#' @noRd
test_project_details <- function(project_details, nrows = NULL) {
  project_details <- tryCatch(
    expr = {
      suppressWarnings({
        assert_project_details(project_details = project_details, nrows = nrows)
      })
    },
    error = function(e) {
      NULL
    }
  )
  !is.null(project_details)
}
#' @noRd
test_openxlsx_style <- function(x) {
  test_class(x, "Style") && identical(attr(class(x), "package"), "openxlsx")
}
#' @noRd
test_env_name <- function(x, max.chars = 26L, all_caps = FALSE) {
  x <- try_else_null({
    suppressWarnings({
      assert_env_name(x = x,
                      max.chars = max.chars,
                      all_caps = all_caps)
    })
  })
  !is.null(x)
}
#' @noRd
test_unique_character <- function(x) {
  test_character(
    x,
    min.chars = 1L,
    unique = TRUE,
    min.len = 1L,
    any.missing = FALSE
  )
}

tempdir_file <- sanitize_path(withr::local_tempdir())
withr::local_envvar(R_USER_CACHE_DIR = tempdir_file)
# assert_blank_project (Internal)
test_that("assert_blank_project works!", {
})
# assert_dir (Internal)
test_that("assert_dir works!", {
})
# assert_env_name (Internal)
test_that("assert_env_name works!", {
  expect_no_error(assert_env_name("REDCapSync"))
  expect_no_error(assert_env_name("REDCap_Sync"))
  expect_no_error(assert_env_name("my_form"))
  expect_no_error(assert_env_name("a"))
  expect_no_error(assert_env_name("A"))
  expect_error(assert_env_name("1A"))
  expect_error(assert_env_name("My Form"))
  expect_error(assert_env_name("my-form"))
  expect_error(assert_env_name("varName!"))
  expect_error(assert_env_name("_form"))
  expect_no_error(assert_env_name("ABCD", max.chars = 4L))
  expect_error(assert_env_name("ABCDE", max.chars = 4L))
})
# assert_project_details (Internal)
test_that("assert_project_details works!", {
})
# assert_project_path (Internal)
test_that("assert_project_path works!", {
})
# assert_setup_project (Internal)
test_that("assert_setup_project works!", {
})
# assert_web_link (Internal)
test_that("assert_web_link works!", {
})
# test_env_name (Internal)
test_that("test_env_name works!", {
})
# test_project_details (Internal)
test_that("test_project_details works!", {
})
# test_setup_project (Internal)
test_that("test_setup_project works!", {
})
# test_unique_character (Internal)
test_that("test_unique_character works!", {
})

---
title: "Datasets"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Datasets}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r, eval=FALSE}
library(REDCapSync)
```

Start by loading an existing or test project. There are two named datasets by default: REDCapSync and REDCapSync_raw.

```{r, eval=FALSE}
save_dir <- tempdir() # replace with your intended folder

project <- load_project("TEST_CLASSIC")

project$sync() #optional sync

dataset <- project$load_dataset("REDCapSync")

# add quick custom variable
# in future version you can use project$add_field for this
dataset$data$merged$letter_b <- dataset$data$merged$var_text_letters == "b"

# send data in global environment
dataset$to_envir(globalenv())

# preview file from temporary directory
dataset$preview()
 
# save data in custom location
dataset$save(dir_other = save_dir)

# open folder where file was saved
browseURL(save_dir)
```

You can also chain together certain functions... So this is the same as above.

```{r, eval=FALSE}
save_dir <- tempdir() # replace with your intended folder

dataset <- projects$load("TEST_CLASSIC")$sync()$load_dataset("REDCapSync")
# add quick custom variable
# in future version you can use project$add_field for this
dataset$data$merged$letter_b <- dataset$data$merged$var_text_letters == "b"

# send data in global environment
dataset$to_envir(globalenv())
 
# save data in custom location
dataset$save(dir_other = save_dir) # dir_other not needed for non-test-projects

# open folder where file was saved
browseURL(save_dir)
```

This single line will load, sync, generate dataset, and send to your global environment!

```{r, eval=FALSE}
projects$load("TEST_CLASSIC")$
  sync()$
  load_dataset("REDCapSync", envir = globalenv())
```

You can also generate ad-hoc custom datasets. But if you plan to keep it refreshing to the directory you can add it to your project. The next time you run sync it will be passed to you directory as defined.

```{r, eval=FALSE}
save_dir <- tempdir() # replace with your intended folder

#load and sync (load_project(...) is the same as projects$load(...))
project <- load_project("TEST_LONGITUDINAL")$sync()

# see field_names
project$metadata$fields$field_name

males_dataset <- project$generate_dataset(dataset_name = "males", 
                                          envir = globalenv(), # global env
                                          filter_field = "birth_sex", 
                                          filter_choices = "Male")
# if you check merged and survey it will only contain males
unique(males_dataset$data$merged$birth_sex) # Male
unique(males_dataset$data$survey$birth_sex) # Male

females_dataset <- project$generate_dataset(dataset_name = "females", 
                                            envir = globalenv(), # global env
                                            filter_field = "birth_sex", 
                                            filter_choices = "Female")
# if you check merged and survey it will only contain males
unique(females_dataset$data$merged$birth_sex) # Female
unique(females_dataset$data$survey$birth_sex) # Female

# if you want this saved to excel one time you can do
females_dataset$save(dir_other = save_dir) # dir_other not needed for default

# To keep this analysis refreshing with each sync use project$add_dataset(..,)
project$add_dataset(dataset_name = "males", 
                    filter_field = "birth_sex", 
                    filter_choices = "Male")
# not it can be loaded by name!
males_dataset <- project$load_dataset("males") # will not send to global

project$load_dataset("males", envir = globalenv()) # will send to global

project$save_dataset("males") # will save AND keep track in project object

project$sync() # once defined will save datasets with sync
```


Putting it all together, you should think of meaningful datasets, add them to project with `project$add_dataset`, load them for use in R with `project$load_dataset`, and/or save as excel sheets with `project$save_dataset` or `project$sync`.

```{r, eval=FALSE}
#load and sync add datasets and save
project <- load_project("TEST_LONGITUDINAL")$
  sync()$
  add_dataset(dataset_name = "males", 
              filter_field = "birth_sex", 
              filter_choices = "Male")$
  add_dataset(dataset_name = "females", 
              filter_field = "birth_sex", 
              filter_choices = "Female")
# load males
project$load_dataset("males", envir = globalenv()) # will send to global

# load females
project$load_dataset("females", envir = globalenv()) # will send to global

# load all default
project$load_dataset("REDCapSync", envir = globalenv()) # will send to global

project$load_dataset("males")$preview() # open temporary excel
project$load_dataset("females")$preview() # open temporary excel

# can trigger saves with `project$save_datasets()` or default `project$sync()`
project$save_datasets()
project$sync()
```

---
title: "RosyREDCap"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{RosyREDCap}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

RosyREDCap is an R package with a shiny app and various plotting functions. It is not on CRAN yet. You can find it on github --> [`RosyREDCap`](https://thecodingdocs.github.io/RosyREDCap/ "RosyREDCap R package")

---
title: "Projects"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Projects}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
library(REDCapSync)
```

The goal of REDCapSync is to encapsulate a REDCap project into a standardized object. R6 gives a way of keeping some elements internal to the object. Behind-the-scenes is an R list where all of the metadata and data is stored.

```{r, eval=TRUE}
TEST_CLASSIC <- load_project(project_name = "TEST_CLASSIC")
listviewer::jsonedit(TEST_CLASSIC$.internal)
```

The R6 object contains those things behind-the-scenes but then uses several to build an environment. This allows `project$sync()` to update `project` without having to do `project <- project |> sync_something() |> another_function()`

```{r setup, eval=TRUE}
project <- load_project("TEST_CLASSIC")

# projects have read-only active bindings
names(REDCapSyncProject$active)

# projects have public methods
names(REDCapSyncProject$public_methods) |> setdiff("initialize")
```

---
title: "Tokens"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Tokens}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```


# REDCap API tokens

REDCap API tokens are equivalent to your username and password. They should never be shared with anyone. Ideally, they should never be directly written in an R script, especially if you plan on sharing it in the future. **Never commit a file containing your token on git or GitHub because this will forever be in the history.** If there is ever any doubt in your mind you should quickly regenerate your token on the REDCap website. If you are working on a *real* REDCap project it's good practice to regenerate your token periodically, such as once a week, just to be safe.

Precisely because tokens are sensitive, REDCapSync is designed to only reference the name of your token, such as "REDCAPSYNC_FIRST_PROJECT". If REDCapSync ever wants to use the token to make an API call to REDCap, it will check `Sys.getenv("REDCAPSYNC_FIRST_PROJECT")`. By default are token names start with "REDCAPSYNC_", followed by the `project_name` you chose in `setup_project()`. Below demonstrates how to set and check your tokens...

## 1. Setting Your Token using User Environment Variables (preferred)

You may find want to reuse a token, and you may have several projects, so a convenient way to store the tokens in a separate location. One way to do this is your personal .Renviron file. Again, you should check always confirm the location of this file and make sure it's not a part of any cloud storage or git or GitHub. You can use `usethis::edit_r_environ()` to semi-permanently save your token. Each time you launch an R session this file will run and your token will be visible to you if specifically called with `Sys.getenv("REDCAPSYNC_FIRST_PROJECT")`.

1.  Open the .Renviron file with `usethis::edit_r_environ()`
2.  Add the token like this... `REDCAPSYNC_FIRST_PROJECT = "faKeTokeN"`
3.  Save the file and Close
4.  Restart R Session (Session tab or `.rs.restartR()`).
5.  Confirm with `Sys.getenv("REDCAPSYNC_FIRST_PROJECT")`

```{r, eval=FALSE}
#Install usethis if you don't have it.
#install.packages("usethis") 
usethis::edit_r_environ()
# Now save your token.... (without the comment symbol '#')
# REDCAPSYNC_FIRST_PROJECT = "faKeTokeN"
# Save the file and Close
# Restart R Session (session tab)
# .rs.restartR() # this will also restart R session for you.
Sys.getenv("REDCAPSYNC_FIRST_PROJECT") # now should contain your token
```

The only reference to your token that is ever made or saved in REDCapSync is with its name. For example, any REDCap project object will have the token name at `project$token_name`. In the future, we may set our token in a different way. 

## 2. Setting Your Token using `keyring` package

By default REDCapSync will use keyring = NULL, which is your OS system default and is typically unlocked by default while you are logged in. 

```{r, eval=FALSE}
project <- load_project("TEST_CLASSIC")

project$set_keyring_token() # now enter token in pop-up
# internally the package checks ...
#1. Sys.getenv()
token <- Sys.getenv(project$.internal$token_name)
#2. followed by keyring
token <- keyring::key_get(service = config$keyring.service(),
                          username = project$project_name,
                          keyring = config$keyring())
# you could set with the following; not preferred because token in script
# if you do this set in securely in file that no one can ever see; never github!
keyring::key_set_with_value(service = config$keyring.service(),
                            username = project$project_name,
                            password = "VeryNOTsecureWayToSetYourTOKEN",
                            keyring = config$keyring())
```

You can change default configuration to give you more fine-tuning of keyrings using configuration settings...

```{r, eval=FALSE}
options(redcapsync.config.keyring = "REDCapSync")
# with this kind of method you could define multiple tokens in a private/secret
# .R file which can be sourced to set all tokens in akeyring. Can also be 
# locked and unlocked (see keyring webstite for details)
```

## 3. Setting Your Token for One Session

You can set manually with base R. Unless you specifically set the token `Sys.getenv("REDCAPSYNC_FIRST_PROJECT")` will be blank.

```{r, eval=FALSE}
# Set your token manually
# again having this in a script is not advised but possible
Sys.setenv(REDCAPSYNC_FIRST_PROJECT="a_FaKe_TOkEn") 

# Get your token
Sys.getenv("REDCAPSYNC_FIRST_PROJECT")
#>[1] "a_FaKe_TOkEn"
```

Now the token is set for this R session only. If you restarted R, it would be blank again.

# Testing Your Token (optional)

If you are ever having any issues with your token, you can test your project object with `project$test_token()`.

```{r, eval=FALSE}
project <- load_project("TEST_CLASSIC") # replace project name with yours

project$test_token()
```

# Additional Resources

-   <https://www.dartistics.com/renviron.html>

---
title: "Cache"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Cache}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The REDCapSync package benefits from storing/caching information about your different projects. The two most important pieces of information is the `project_name` and `directory` where it's stored. This is enough information to be able to find where you have chosen to securely store your files, load what has already been collected, and then communicate with R to fetch any new updates. REDCapSync's ability to store your projects in a standardized directory is what allows for powerful pipeline tasks. Importantly, no direct project data or tokens are stored in the cache!

```{r, eval=F}
library(REDCapSync) # don't forget to load the package
```

Using the `hoadr` package, R finds the standard location where R typically stores cached package data. For example, on Mac the location might look like, "/Users/yourmacname/Library/Caches/R/REDCapSync". 

# Location
The exact path can be found with ` tools::R_user_dir`. The only thing that REDCapSync currently stores in this cache is the projects data.frame, which can be loaded with `projects$load("PROJECT")`.

```{r, eval=F}
redcapsync_cache_location <- tools::R_user_dir(package = "REDCapSync", 
                                               which = "cache")
# launch the folder on your computer
utils::browseURL(redcapsync_cache_location)
# can modify with config options
config$cache.dir()
```

# Clearing

You can clear the entire cache or only delete specific projects. This will only delete REDCapSync's "knowledge" of project information and location; it will not delete any files from any project directories. If you want to delete files, please do so using your own methods.

```{r, eval=F}
# clear entire cache (all projects)
cache_clear()
# clear specific project
cache_clear("OLD_PROJECT1")
# clear specific project1
cache_clear(c("OLD_PROJECT2", "OLD_PROJECT3"))
```

# Projects

If you setup and subsequently sync/save a project, it will appear using the code below. If you haven't setup any projects, it will be an empty data.frame.

```{r, eval=F}
print.data.frame(projects$df)
```

---
title: "Cache"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Cache}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The REDCapSync package benefits from storing/caching information about your different projects. The two most important pieces of information is the `project_name` and `directory` where it's stored. This is enough information to be able to find where you have chosen to securely store your files, load what has already been collected, and then communicate with R to fetch any new updates. REDCapSync's ability to store your projects in a standardized directory is what allows for powerful pipeline tasks. Importantly, no direct project data or tokens are stored in the cache!

```{r, eval=F}
library(REDCapSync) # don't forget to load the package
```

Using the `hoadr` package, R finds the standard location where R typically stores cached package data. For example, on Mac the location might look like, "/Users/yourmacname/Library/Caches/R/REDCapSync". 

# Location
Your exact path can be found for any R package with `rappdirs::user_cache_dir("R")`. The only thing that REDCapSync stores in this cache is the projects data.frame, which can be loaded with `projects$load("PROJECT")`.

```{r, eval=F}
user_cache_location <- rappdirs::user_cache_dir("R") 
redcapsync_cache_location <- file.path(user_cache_location, "REDCapSync")
redcapsync_cache_location
# launch the folder on your computer
utils::browseURL(redcapsync_cache_location)
# can modify with config options
config$cache.dir()
```

# Clearing

You can clear the entire cache or only delete specific projects. This will only delete REDCapSync's "knowledge" of project information and location; it will not delete any files from any project directories. If you want to delete files, please do so using your own methods.

```{r, eval=F}
# clear entire cache (all projects)
cache_clear()
# clear specific project
cache_clear("OLD_PROJECT1")
# clear specific project1
cache_clear(c("OLD_PROJECT2", "OLD_PROJECT3"))
```

# Projects

If you setup and subsequently sync/save a project, it will appear using the code below. If you haven't setup any projects, it will be an empty data.frame.

```{r, eval=F}
print.data.frame(projects$df)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setup.R
\name{setup_project}
\alias{setup_project}
\alias{load_project}
\title{Setup or Load REDCapSync Project}
\usage{
setup_project(
  project_name,
  dir_path,
  redcap_uri,
  token_name = paste0("REDCAPSYNC_", project_name),
  sync_frequency = "daily",
  labelled = TRUE,
  get_type = "identified",
  records = NA,
  fields = NA,
  forms = NA,
  events = NA,
  filter_logic = NA,
  id_position = 1L,
  get_users = TRUE,
  get_data = TRUE,
  batch_size_download = 1000L,
  batch_size_upload = 500L,
  get_entire_log = FALSE,
  log_days = 10L,
  log_drop_details = FALSE,
  log_drop_exports = FALSE,
  get_files = FALSE,
  get_file_repository = FALSE,
  original_file_names = FALSE,
  add_default_datasets = TRUE,
  timezone = Sys.timezone(),
  hard_reset = FALSE
)

load_project(project_name)
}
\arguments{
\item{project_name}{Character scalar. Unique uppercase project name with no
spaces or symbols.}

\item{dir_path}{Optional character scalar. Directory where REDCap project
files are stored. If omitted, the project is only available in the current R
session and cannot be persisted to disk.}

\item{redcap_uri}{Character scalar. Base URL of the REDCap API endpoint.}

\item{token_name}{Optional character scalar. Environment variable name used
to store the REDCap token. Defaults to \verb{REDCAPSYNC_<project_name>}.}

\item{sync_frequency}{Character scalar. Allowed values are "always",
"hourly", "daily", "weekly", "monthly", "once", and "never". This
setting controls how often future syncs are considered due.}

\item{labelled}{Logical scalar. If \code{TRUE}, data will be preserved as labelled
values. Default is \code{TRUE}.}

\item{get_type}{Character scalar. REDCap API export type. One of
"identified", "deidentified", "deidentified_strict", or
"deidentified_super_strict". Default is "identified".}

\item{records}{Optional character vector of record IDs to request.}

\item{fields}{Optional character vector of field names to request.}

\item{forms}{Optional character vector of form names to request.}

\item{events}{Optional character vector of event names to request.}

\item{filter_logic}{Optional character scalar. REDCap filter logic used to
limit returned records or record-events.}

\item{id_position}{Integer scalar of the variable that uniquely identifies
the subject (typically record_id). This defaults to the first variable in the
data dictionary.}

\item{get_users}{Logical scalar. If \code{TRUE}, the project will be configured to
retrieve REDCap users during sync.}

\item{get_data}{Logical scalar. If \code{TRUE}, the project will be configured to
retrieve REDCap data during sync.}

\item{batch_size_download}{Integer scalar. Number of records to request per
download batch. Default is \code{1000}.}

\item{batch_size_upload}{Integer scalar. Number of records to process per
upload batch. Default is \code{500}.}

\item{get_entire_log}{Logical scalar. If \code{TRUE}, REDCap activity logs are
retrieved in full. Default is \code{FALSE}.}

\item{log_days}{Integer scalar. Number of days of log history to consider
when a new project is being set up or a hard reset occurs. Default is \code{10}.}

\item{log_drop_details}{Logical scalar. If \code{TRUE}, the log details are
excluded.}

\item{log_drop_exports}{Logical scalar. If \code{TRUE}, the log export events are
excluded.}

\item{get_files}{Logical scalar. If \code{TRUE}, file attachments are configured
to be retrieved from REDCap. Default is \code{FALSE}.}

\item{get_file_repository}{Logical scalar. If \code{TRUE}, the file repository is
configured to be retrieved from REDCap. Default is \code{FALSE}.}

\item{original_file_names}{Logical scalar. If \code{TRUE}, retrieved files keep
their original names. Default is \code{FALSE}.}

\item{add_default_datasets}{Logical scalar. If \code{TRUE}, dataset templates are
added to a new project. Default is \code{TRUE}.}

\item{timezone}{Optional character scalar. Time zone used for REDCap data
timestamps. Defaults to \code{Sys.timezone()}.}

\item{hard_reset}{Logical scalar. If \code{TRUE}, any existing project with the
same \code{project_name} is ignored and a fresh project object is initialized.
Default is \code{FALSE}.}
}
\value{
R6 project object with \link{REDCapSyncProject} class.
}
\description{
Setup or load a REDCapSync \link{project} object.
Prepares a new REDCapSync project by recording the REDCap URI,
token name, sync settings, and optional data selection preferences.
}
\details{
Unless \code{hard_reset = TRUE}, it will first attempt to load an
existing project from cache or the supplied \code{dir_path} before creating a
new project object.
If settings differ from the loaded project, they will be used or it will
trigger a hard reset with a warning. \code{setup_project()} itself does not
perform a full REDCap sync. It configures the project so that subsequent sync
actions may retrieve data, users, files, and logs according to the project
settings.
}
\examples{
# Initialize the project object with the REDCap API token and URL
save_folder <- tempdir() # replace with real folder
project <- setup_project(
  project_name = "FIRST_PROJECT",
  dir_path = save_folder,
  redcap_uri = "https://redcap.yourinstitution.edu/api/"
)

# object can be named whatever you choose to assign
# TEST projects can be loaded in addition to real projects
project_test <- load_project("TEST_CLASSIC")

}
\seealso{
\code{vignette("REDCapSync", package = "REDCapSync")}
\code{vignette("Tokens", package = "REDCapSync")}
\link{projects} for shortcuts of cached setup projects
\link{project} for using the project objects
\code{vignette("Cache", package = "REDCapSync")}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REDCapSyncProject.R
\name{project}
\alias{project}
\alias{REDCapSyncProject}
\title{REDCapSync Project Object}
\value{
An R6 \code{REDCapSyncProject} class generator for internal use. Users interact
with instances created by \link{setup_project} or \link{load_project}.
}
\description{
\link[R6:R6Class]{R6} project object for managing REDCap data access and
synchronization. The project object is your main interface to REDCap data
through the \link{REDCapSync} package. It stores your REDCap configuration, data,
metadata, and provides methods to sync, transform, and export your data.
}
\details{
\subsection{Workflow}{

\strong{Initialize} a project with \code{\link[=setup_project]{setup_project()}}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{project <- setup_project(
  project_name = "FIRST_PROJECT",
  redcap_uri = "https://redcap.yourinstitution.edu/api/",
  dir_path = "~/redcap_projects"
)
}\if{html}{\out{</div>}}

\strong{Synchronize} REDCap data into your project object using the REDCap API
and log-based change detection:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{project$sync()
}\if{html}{\out{</div>}}

\strong{Load} previously saved REDCap project object using cache:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{project <- load_project("FIRST_PROJECT)
projects$names() # should see "FIRST_PROJECT"
}\if{html}{\out{</div>}}

\strong{Access} REDCap data, metadata, and users via read-only fields:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{project$data           # Named list of REDCap forms/instruments
project$metadata       # REDCap field definitions, forms, and choices
project$redcap$users   # REDCap user information
project$redcap$log     # REDCap log information

# you can bypass read-only if needed
users <- project$redcap$users
}\if{html}{\out{</div>}}

\strong{Transform and export} your data into clean, analysis-ready datasets to be
used in R and/or Excel:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{project$add_dataset("analysis_data", ...) # adds to project for re-use
project$save_datasets() # can save to Excel but is also part of future syncs
}\if{html}{\out{</div>}}

\strong{Upload} corrected or new data back to REDCap:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{project$upload(updated_data)
}\if{html}{\out{</div>}}
}

\subsection{Design Features}{

The project object uses \strong{log-based sync}: Since REDCap maintains a detailed
change log, \code{project$sync()} only retrieves and updates records that have
changed since the last sync. This dramatically reduces API calls and improves
performance for large projects.

All methods use \strong{method chaining}: Methods invisibly return the project
object (self), allowing fluent code:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{dataset <- load_project("TEST_CLASSIC")$sync()$load_dataset("REDCapSync")
}\if{html}{\out{</div>}}
}

\subsection{Read-Only Fields}{
\itemize{
\item \code{project$project_name}: Project identifier (set with \link{setup_project})
\item \code{project$dir_path}: Persistent storage directory (set with \link{setup_project})
\item \code{project$data}: Named list of synchronized REDCap forms
\item \code{project$metadata}: REDCap field metadata (forms, fields, choices)
\item \code{project$redcap}: REDCap project info, users, and activity log
\item \code{project$.internal}: Internal project object (for advanced use)
}
}

\subsection{TEST Projects}{

All TEST projects start with "TEST_" and by default the average user cannot
create a project that starts with "TEST_". They are produced from actual
server REDCap projects but never contained real data and were scrubbed of
any "real" user/log data. They are subject to change in future versions as
the package matures. Keep in mind there are many combinations of REDCap
structures to account for.

TEST projects are meant to be used for demonstration and testing
purposes. In general, try to use actual REDCap project(s) to explore the
package.

This is how to load a project with or without specifying a directory.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{# return the list of available test projects
REDCapSync::projects$test_names()

project <- load_project("TEST_CLASSIC")

YOUR_DIRECTORY <- getwd()
project <- setup_project("TEST_CLASSIC", dir_path = YOUR_DIRECTORY)
# now can test object functions (does not use API)
project$sync() # will also save datasets
}\if{html}{\out{</div>}}

The currently available TEST projects are as follows:
\itemize{
\item \code{TEST_CLASSIC}: Classic project (nothing repeats). Has every data type.
Contains data for survival analysis testing. Contains comments.
\item \code{TEST_REPEATING}: Non-longitudinal project with repeating instruments.
\item \code{TEST_LONGITUDINAL}: Longitudinal project (with events).
\item \code{TEST_MULTIARM}: Multi-arm and longitudinal project.
\item \code{TEST_EDGE}: Edge-case project. Meant to test things that may be uncommon
but should be accounted for as package matures.
\item \code{TEST_DATA}: Placeholder project meant to contain more rich test dataset.
\item \code{TEST_CANCER}: Placeholder project meant to contain more rich test dataset.
\item \code{TEST_REDCAPR_SIMPLE}: Borrowed with permission from \link[REDCapR:REDCapR]{REDCapR}.
\item \code{TEST_REDCAPR_LONGITUDINAL}: Borrowed with permission from \link[REDCapR:REDCapR]{REDCapR}.
\item \code{TEST_REDCAPR_CLIN_TRIAL}: Borrowed with permission from \link[REDCapR:REDCapR]{REDCapR}.
}
}
}
\examples{
# Load a test project
projects$test_names() # available test projects
project <- setup_project("TEST_CLASSIC", dir_path = tempdir())

# Sync data from REDCap
project$sync()

# Access data and metadata
head(project$data$text)
project$metadata$fields[1:5, ]

forms <- project$metadata$forms # unchanged metadata
fields <- project$metadata$fields # unchanged metadata
choices <- project$metadata$choices # unchanged metadata

users <- project$redcap$users # unchanged users
log <- project$redcap$log # unchanged log

project$test_token()

dataset <- project$load_dataset("REDCapSync")

}
\seealso{
\code{vignette("REDCapSync", package = "REDCapSync")}
\link{projects} for shortcuts of cached setup projects
\code{vignette("Tokens", package = "REDCapSync")}
\link{setup_project} for initializing projects
\link{dataset} for using the dataset objects
\code{vignette("Datasets", package = "REDCapSync")}
\code{vignette("Uploads", package = "REDCapSync")}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{project_name}}{Read-only character string of project_name as
assigned with \link{setup_project}.}

    \item{\code{dir_path}}{Read-only directory path assigned with \link{setup_project}.}

    \item{\code{data}}{Read-only named list where each name is an instrument name.
See public methods for \link{REDCapSyncProject}.}

    \item{\code{metadata}}{Read-only named list with REDCap metadata. See
public methods for \link{REDCapSyncProject}.}

    \item{\code{redcap}}{Read-only named list with REDCap information including
users and log.}

    \item{\code{.internal}}{Read-only internal project object for custom workflows}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-REDCapSyncProject-initialize}{\code{REDCapSyncProject$new()}}
    \item \href{#method-REDCapSyncProject-print}{\code{REDCapSyncProject$print()}}
    \item \href{#method-REDCapSyncProject-sync}{\code{REDCapSyncProject$sync()}}
    \item \href{#method-REDCapSyncProject-add_dataset}{\code{REDCapSyncProject$add_dataset()}}
    \item \href{#method-REDCapSyncProject-load_dataset}{\code{REDCapSyncProject$load_dataset()}}
    \item \href{#method-REDCapSyncProject-remove_datasets}{\code{REDCapSyncProject$remove_datasets()}}
    \item \href{#method-REDCapSyncProject-generate_dataset}{\code{REDCapSyncProject$generate_dataset()}}
    \item \href{#method-REDCapSyncProject-save_datasets}{\code{REDCapSyncProject$save_datasets()}}
    \item \href{#method-REDCapSyncProject-save_dataset}{\code{REDCapSyncProject$save_dataset()}}
    \item \href{#method-REDCapSyncProject-save}{\code{REDCapSyncProject$save()}}
    \item \href{#method-REDCapSyncProject-set_keyring_token}{\code{REDCapSyncProject$set_keyring_token()}}
    \item \href{#method-REDCapSyncProject-test_token}{\code{REDCapSyncProject$test_token()}}
    \item \href{#method-REDCapSyncProject-url_launch}{\code{REDCapSyncProject$url_launch()}}
    \item \href{#method-REDCapSyncProject-url_record_launch}{\code{REDCapSyncProject$url_record_launch()}}
    \item \href{#method-REDCapSyncProject-upload}{\code{REDCapSyncProject$upload()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-initialize}{}}}
\subsection{\code{REDCapSyncProject$new()}}{
  Active binding are read-only


  The end user will not see \code{project$new()}. This is handled internally.
Users should construct objects using \code{\link[=setup_project]{setup_project()}}. The remain
methods will be accessible to any user.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$new(project)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{project}}{a list object meant to be stored internally within R6}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-print"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-print}{}}}
\subsection{\code{REDCapSyncProject$print()}}{
  Print some key project information
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$print()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-sync"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-sync}{}}}
\subsection{\code{REDCapSyncProject$sync()}}{
  Updates the REDCap data for (\code{project} object) by checking REDCap log for
changes. Sync is performed according to the \code{sync_frequency} set in
\code{\link[=setup_project]{setup_project()}} by default. Use \code{hard_check} to force a check, or \code{ hard_reset} to force a complete refresh. As a default, this object will
be saved to your directory when necessary.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$sync(
  save_datasets = TRUE,
  save_to_dir = TRUE,
  hard_check = FALSE,
  hard_reset = FALSE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{save_datasets}}{Logical. Save dataset outputs during sync. Default is
\code{TRUE}.}
      \item{\code{save_to_dir}}{Logical. Save updated project object to directory. Default
is \code{TRUE}.}
      \item{\code{hard_check}}{Logical. Force REDCap API check regardless of
\code{sync_frequency}. Default is \code{FALSE}.}
      \item{\code{hard_reset}}{Logical. Overwrite existing dataset files. Default is
\code{FALSE}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-add_dataset"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-add_dataset}{}}}
\subsection{\code{REDCapSyncProject$add_dataset()}}{
  Add a new dataset entry
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$add_dataset(
  dataset_name,
  transformation_type = "default",
  merge_form_name = "merged",
  filter_field = NULL,
  filter_choices = NULL,
  filter_list = NULL,
  filter_strict = TRUE,
  field_names = NULL,
  form_names = NULL,
  exclude_identifiers = FALSE,
  exclude_free_text = FALSE,
  date_handling = "none",
  labelled = TRUE,
  clean = TRUE,
  drop_blanks = FALSE,
  drop_missing_codes = FALSE,
  drop_others = NULL,
  include_metadata = TRUE,
  include_records = TRUE,
  include_users = TRUE,
  include_log = FALSE,
  annotate_from_log = TRUE,
  include_comments = TRUE,
  with_links = TRUE,
  separate = FALSE,
  use_csv = FALSE,
  dir_other = NULL,
  file_name = NULL,
  hard_reset = FALSE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{dataset_name}}{Character. Name of the dataset configuration to create,
load, or reference.}
      \item{\code{transformation_type}}{Character. How to transform data: "default"
(merge non-repeating then add to repeating), "none" (no transformation), or
"merge_non_repeating" (merge non-repeating only). Default is "default".}
      \item{\code{merge_form_name}}{Character. Name for the merged non-repeating form.
Default is "merged".}
      \item{\code{filter_field}}{Character. Field name to filter dataset on.}
      \item{\code{filter_choices}}{Vector. Allowed values for \code{filter_field}.}
      \item{\code{filter_list}}{List. Named list where names are field names and values
are allowed value sets. Alternative to \code{filter_field}/\code{filter_choices}.}
      \item{\code{filter_strict}}{Logical. If \code{TRUE}, filter all forms. If \code{FALSE}, apply
filter only to records (ID column). Default is \code{TRUE}.}
      \item{\code{field_names}}{Character vector. Fields to include. Default \code{NULL}
includes all fields.}
      \item{\code{form_names}}{Character vector. Forms to include. Default \code{NULL} includes
all forms.}
      \item{\code{exclude_identifiers}}{Logical. Exclude identifier fields. Default is
\code{TRUE}.}
      \item{\code{exclude_free_text}}{Logical. Exclude free-text fields (for
deidentification). Default is \code{FALSE}.}
      \item{\code{date_handling}}{Character. Date handling strategy: "none",
"exclude_dates", "random_shift_by_record", "random_shift_by_project",
"zero_by_record", or "zero_by_project". Default is "none".}
      \item{\code{labelled}}{Logical. Convert to labelled data if \code{TRUE}. Default is
\code{TRUE}.}
      \item{\code{clean}}{Logical. Clean data (standardize missing values). Default is
\code{TRUE}.}
      \item{\code{drop_blanks}}{Logical. Drop records with blank fields. Default is
\code{TRUE}.}
      \item{\code{drop_missing_codes}}{Logical. Convert REDCap missing codes to \code{NA}.
Default is \code{FALSE}.}
      \item{\code{drop_others}}{Character vector of additional values to drop.}
      \item{\code{include_metadata}}{Logical. Include field metadata in dataset. Default
is \code{TRUE}.}
      \item{\code{include_records}}{Logical. Include record-level information. Default is
\code{TRUE}.}
      \item{\code{include_users}}{Logical. Include user information. Default is \code{TRUE}.}
      \item{\code{include_log}}{Logical. Include REDCap activity log. Default is \code{FALSE}.}
      \item{\code{annotate_from_log}}{Logical. Annotate data using the REDCap log. Default
is \code{TRUE}.}
      \item{\code{include_comments}}{Logical. Include field comments. Default is \code{TRUE}.}
      \item{\code{with_links}}{Logical. Include hyperlinks in Excel exports. Default is
\code{FALSE}.}
      \item{\code{separate}}{Logical. Separate each form into distinct files (vs.
multi-tab Excel). Default is \code{FALSE}.}
      \item{\code{use_csv}}{Logical. Use CSV format instead of Excel. Default is \code{FALSE}.}
      \item{\code{dir_other}}{Character. Output directory (default is project's \code{output}
folder).}
      \item{\code{file_name}}{Character. Dataset file name (default is
\verb{<project_name>_<dataset_name>}).}
      \item{\code{hard_reset}}{Logical. Overwrite existing dataset files. Default is
\code{FALSE}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-load_dataset"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-load_dataset}{}}}
\subsection{\code{REDCapSyncProject$load_dataset()}}{
  Load dataset if previously defined with \code{add_dataset}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$load_dataset(dataset_name, envir = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{dataset_name}}{Character. Name of the dataset configuration to create,
load, or reference.}
      \item{\code{envir}}{Environment to assign dataset objects. Default is \code{NULL}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-remove_datasets"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-remove_datasets}{}}}
\subsection{\code{REDCapSyncProject$remove_datasets()}}{
  Clear all or specified datasets from the \code{project} object.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$remove_datasets(dataset_names = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{dataset_names}}{Character vector. Dataset names to remove/operate on.
Default is \code{NULL}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-generate_dataset"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-generate_dataset}{}}}
\subsection{\code{REDCapSyncProject$generate_dataset()}}{
  Generate \link{dataset} object. This is usually handled
internally for some default behavior but is provided here for ad-hoc
custom datasets.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$generate_dataset(
  dataset_name,
  envir = NULL,
  transformation_type = "default",
  merge_form_name = "merged",
  filter_field = NULL,
  filter_choices = NULL,
  filter_list = NULL,
  filter_strict = TRUE,
  field_names = NULL,
  form_names = NULL,
  exclude_identifiers = FALSE,
  exclude_free_text = FALSE,
  date_handling = "none",
  labelled = TRUE,
  clean = TRUE,
  drop_blanks = FALSE,
  drop_missing_codes = FALSE,
  drop_others = NULL,
  include_metadata = TRUE,
  include_records = TRUE,
  include_users = TRUE,
  include_log = FALSE,
  annotate_from_log = TRUE,
  include_comments = FALSE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{dataset_name}}{Character. Name of the dataset configuration to create,
load, or reference.}
      \item{\code{envir}}{Environment to assign dataset objects. Default is \code{NULL}.}
      \item{\code{transformation_type}}{Character. How to transform data: "default"
(merge non-repeating then add to repeating), "none" (no transformation), or
"merge_non_repeating" (merge non-repeating only). Default is "default".}
      \item{\code{merge_form_name}}{Character. Name for the merged non-repeating form.
Default is "merged".}
      \item{\code{filter_field}}{Character. Field name to filter dataset on.}
      \item{\code{filter_choices}}{Vector. Allowed values for \code{filter_field}.}
      \item{\code{filter_list}}{List. Named list where names are field names and values
are allowed value sets. Alternative to \code{filter_field}/\code{filter_choices}.}
      \item{\code{filter_strict}}{Logical. If \code{TRUE}, filter all forms. If \code{FALSE}, apply
filter only to records (ID column). Default is \code{TRUE}.}
      \item{\code{field_names}}{Character vector. Fields to include. Default \code{NULL}
includes all fields.}
      \item{\code{form_names}}{Character vector. Forms to include. Default \code{NULL} includes
all forms.}
      \item{\code{exclude_identifiers}}{Logical. Exclude identifier fields. Default is
\code{TRUE}.}
      \item{\code{exclude_free_text}}{Logical. Exclude free-text fields (for
deidentification). Default is \code{FALSE}.}
      \item{\code{date_handling}}{Character. Date handling strategy: "none",
"exclude_dates", "random_shift_by_record", "random_shift_by_project",
"zero_by_record", or "zero_by_project". Default is "none".}
      \item{\code{labelled}}{Logical. Convert to labelled data if \code{TRUE}. Default is
\code{TRUE}.}
      \item{\code{clean}}{Logical. Clean data (standardize missing values). Default is
\code{TRUE}.}
      \item{\code{drop_blanks}}{Logical. Drop records with blank fields. Default is
\code{TRUE}.}
      \item{\code{drop_missing_codes}}{Logical. Convert REDCap missing codes to \code{NA}.
Default is \code{FALSE}.}
      \item{\code{drop_others}}{Character vector of additional values to drop.}
      \item{\code{include_metadata}}{Logical. Include field metadata in dataset. Default
is \code{TRUE}.}
      \item{\code{include_records}}{Logical. Include record-level information. Default is
\code{TRUE}.}
      \item{\code{include_users}}{Logical. Include user information. Default is \code{TRUE}.}
      \item{\code{include_log}}{Logical. Include REDCap activity log. Default is \code{FALSE}.}
      \item{\code{annotate_from_log}}{Logical. Annotate data using the REDCap log. Default
is \code{TRUE}.}
      \item{\code{include_comments}}{Logical. Include field comments. Default is \code{TRUE}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-save_datasets"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-save_datasets}{}}}
\subsection{\code{REDCapSyncProject$save_datasets()}}{
  saves datasets to Excel via setting provided to
\code{add_dataset}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$save_datasets(hard_reset = FALSE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{hard_reset}}{Logical. Overwrite existing dataset files. Default is
\code{FALSE}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-save_dataset"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-save_dataset}{}}}
\subsection{\code{REDCapSyncProject$save_dataset()}}{
  saves dataset to Excel via setting provided to
\code{add_dataset}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$save_dataset(dataset_name)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{dataset_name}}{Character. Name of the dataset configuration to create,
load, or reference.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-save"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-save}{}}}
\subsection{\code{REDCapSyncProject$save()}}{
  Save project object to directory chosen with
\link{setup_project}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$save()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-set_keyring_token"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-set_keyring_token}{}}}
\subsection{\code{REDCapSyncProject$set_keyring_token()}}{
  Set keyring token. See vignette and config for detail.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$set_keyring_token()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-test_token"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-test_token}{}}}
\subsection{\code{REDCapSyncProject$test_token()}}{
  test connection via communication with API.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$test_token(keyring_if_error = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{keyring_if_error}}{Logical. Launch pop-up for keyring if token is not
valid or fails to call REDCap. Max 3 attempts. Default is \code{TRUE}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-url_launch"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-url_launch}{}}}
\subsection{\code{REDCapSyncProject$url_launch()}}{
  opens links in browser.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$url_launch(link_type = "home", open_browser = TRUE)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{link_type}}{Character. REDCap link type: "base", "home", "record_home",
"records_dashboard", "api", "api_playground", "codebook", "user_rights",
"setup", "logging", "designer", "dictionary", "data_quality", or
"identifiers".}
      \item{\code{open_browser}}{Logical. Open link in browser. Default is \code{TRUE}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-url_record_launch"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-url_record_launch}{}}}
\subsection{\code{REDCapSyncProject$url_record_launch()}}{
  opens record links in browser.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$url_record_launch(
  record = NULL,
  page = NULL,
  instance = NULL,
  open_browser = TRUE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{record}}{Character. Record ID.}
      \item{\code{page}}{Character. REDCap form/instrument name.}
      \item{\code{instance}}{Character. Repeating instance number.}
      \item{\code{open_browser}}{Logical. Open link in browser. Default is \code{TRUE}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncProject-upload"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncProject-upload}{}}}
\subsection{\code{REDCapSyncProject$upload()}}{
  This will only overwrite and new data. It will not directly delete any
data. Because this is a function that can mess up your data, use it
very carefully. Remember all changes are saved in the REDCap log if
there's an issue. Missing rows and columns are allowed!
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncProject$upload(to_be_uploaded)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{to_be_uploaded}}{Data frame in raw coded format ready to upload to
REDCap.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{config}
\alias{config}
\title{Configuration}
\usage{
config
}
\value{
list of functions that returns config values
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
Internal configuration helpers used to retrieve package configuration
values from options or environment variables.

Configuration is resolved in the following order:
\enumerate{
\item \code{getOption("redcapsync.config.option.name")}
\item \code{Sys.getenv("REDCAPSYNC_CONFIG_OPTION_NAME")} # skipped for functions
\item Default if unable to find and validate from above.
}
}
\details{
The config function is operational but only some methods presently
affect internal code. Most users will not ever need to modify default config.
This is included to improve future versions of the package.
\itemize{
\item Working configs: \code{allow.test.names}, \code{cache.dir}, \code{keyring},
\code{keyring.service}, \code{openxlsx.header.style}, \code{openxlsx.body.style}.
\item Partial coverage: \code{offline}.
\item Placeholder: \code{verbose}, \code{show.api.messages}
}
\subsection{allow.test.names}{

Logical for \code{\link[=setup_project]{setup_project()}} allowing \code{project_name} starting with TEST_.
Default is \code{FALSE}. This is provided for developers and testing environments.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# check current value package is using...
config$allow.test.names()
# set with options (which will be prioritized over envvar)
getOption("redcapsync.config.allow.test.names") # get
options(redcapsync.config.allow.test.names = FALSE) # set

# set with ennvar (which will be prioritized when options not defined)
Sys.getenv("REDCAPSYNC_CONFIG_ALLOW_TEST_NAMES")  # get
Sys.setenv(REDCAPSYNC_CONFIG_ALLOW_TEST_NAMES = FALSE) # or set in .Renviron
}\if{html}{\out{</div>}}
}

\subsection{show.api.messages}{

Logical for showing display API messages from REDCapR and redcapAPI.
Default is \code{FALSE}.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# check current value package is using...
config$show.api.messages()
# set with options (which will be prioritized over envvar)
getOption("redcapsync.config.show.api.messages") # get
options(redcapsync.config.show.api.messages = FALSE) # set

# set with ennvar (which will be prioritized when options not defined)
Sys.getenv("REDCAPSYNC_CONFIG_SHOW_API_MESSAGES") # get
Sys.setenv(REDCAPSYNC_CONFIG_SHOW_API_MESSAGES = FALSE) # or set in .Renviron
}\if{html}{\out{</div>}}
}

\subsection{verbose}{

Logical for showing display API messages from REDCapR and redcapAPI.
Default is \code{FALSE}.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# check current value package is using...
config$verbose()
# set with options (which will be prioritized over envvar)
getOption("redcapsync.config.verbose") # get
options(redcapsync.config.verbose = FALSE) # set

# set with ennvar (which will be prioritized when options not defined)
Sys.getenv("REDCAPSYNC_CONFIG_VERBOSE") # get
Sys.setenv(REDCAPSYNC_CONFIG_VERBOSE = FALSE) # or set in .Renviron
}\if{html}{\out{</div>}}
}

\subsection{offline}{

Logical for offline, which if TRUE will block any API calls.
Default is \code{FALSE}.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# check current value package is using...
config$offline()
# set with options (which will be prioritized over envvar)
getOption("redcapsync.config.offline") # get
options(redcapsync.config.offline = FALSE) # set

# set with ennvar (which will be prioritized when options not defined)
Sys.getenv("REDCAPSYNC_CONFIG_OFFLINE") # get
Sys.setenv(REDCAPSYNC_CONFIG_OFFLINE = FALSE) # or set in .Renviron
}\if{html}{\out{</div>}}
}

\subsection{cache.dir}{

Character file path overriding the default cache directory. Default follow
system standards via rappdir, hoardr, or R_USER_CACHE_DIR

\if{html}{\out{<div class="sourceCode">}}\preformatted{# check current value package is using...
config$cache.dir()
# set with options (which will be prioritized over envvar)
getOption("redcapsync.config.cache.dir") # get
options(redcapsync.config.cache.dir = "file/path/to/keep/cache") # set

# set with ennvar (which will be prioritized when options not defined)
Sys.getenv("REDCAPSYNC_CONFIG_CACHE_DIR") # get
Sys.setenv(REDCAPSYNC_CONFIG_CACHE_DIR = "file/path/to/keep/cache") # set
}\if{html}{\out{</div>}}
}

\subsection{keyring}{

Character keyring name (parameter from \code{\link[keyring]{keyring}}
package). Default is NULL, which is at the system level. For locking use a
keyring like "REDCapSync"

\if{html}{\out{<div class="sourceCode">}}\preformatted{# check current value package is using...
config$keyring()
# set with options (which will be prioritized over envvar)
getOption("redcapsync.config.keyring") # get
options(redcapsync.config.keyring = "REDCapSync") # set

# set with ennvar (which will be prioritized when options not defined)
Sys.getenv("REDCAPSYNC_CONFIG_KEYRING") # get
Sys.setenv(REDCAPSYNC_CONFIG_KEYRING = "REDCapSync") # set in session
}\if{html}{\out{</div>}}
}

\subsection{keyring.service}{

Character keyring service name (parameter from \code{\link[keyring]{keyring}}
package). Default is "R-REDCapSync".

\if{html}{\out{<div class="sourceCode">}}\preformatted{# check current value package is using...
config$keyring.service()
# set with options (which will be prioritized over envvar)
getOption("redcapsync.config.keyring.service") # get
options(redcapsync.config.keyring.service = "REDCapSync") # set

# set with ennvar (which will be prioritized when options not defined)
Sys.getenv("REDCAPSYNC_CONFIG_KEYRING_SERVICE") # get
Sys.setenv(REDCAPSYNC_CONFIG_KEYRING_SERVICE = "REDCapSync") # set in session
}\if{html}{\out{</div>}}
}

\subsection{openxlsx.header.style}{

Excel sheet header using \code{\link[openxlsx:createStyle]{openxlsx::createStyle()}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{# set with options (which will be prioritized over envvar)
getOption("redcapsync.config.openxlsx.header.style") # get
new_style <- openxlsx::createStyle(fontSize = 12L)
options(redcapsync.config.openxlsx.header.style = new_style)
# does not look up envvar due to it's function type
}\if{html}{\out{</div>}}
}

\subsection{openxlsx.body.style}{

Excel sheet body using \code{\link[openxlsx:createStyle]{openxlsx::createStyle()}}

\if{html}{\out{<div class="sourceCode">}}\preformatted{# set with options (which will be prioritized over envvar)
getOption("redcapsync.config.openxlsx.body.style") # get
new_style <- openxlsx::createStyle(fontSize = 12L)
options(redcapsync.config.openxlsx.body.style = new_style)
# does not look up envvar due to it's function type
}\if{html}{\out{</div>}}
}
}
\section{Option Names (searched first)}{
\if{html}{\out{<div class="sourceCode">}}\preformatted{option_list <- list(
  redcapsync.config.allow.test.names = NULL,
  redcapsync.config.show.api.messages = NULL,
  redcapsync.config.verbose = NULL,
  redcapsync.config.offline = NULL,
  redcapsync.config.cache.dir = NULL,
  redcapsync.config.keyring = NULL,
  redcapsync.config.keyring.service = NULL,
  redcapsync.config.openxlsx.header.style = NULL,
  redcapsync.config.openxlsx.body.style = NULL
)
}\if{html}{\out{</div>}}
}

\section{Environment Variable Names (searched second)}{
\if{html}{\out{<div class="sourceCode">}}\preformatted{envvar_list <- list(
  REDCAPSYNC_CONFIG_ALLOW_TEST_NAMES = NA,
  REDCAPSYNC_CONFIG_SHOW_API_MESSAGES = NA,
  REDCAPSYNC_CONFIG_VERBOSE = NA,
  REDCAPSYNC_CONFIG_OFFLINE = NA,
  R_USER_CACHE_DIR = NA, # affects your entire cache for any package
  REDCAPSYNC_CONFIG_CACHE_DIR = NA, # affects only REDCapSync Cache
  REDCAPSYNC_CONFIG_KEYRING = NA,
  REDCAPSYNC_CONFIG_KEYRING_SERVICE = NA
)
}\if{html}{\out{</div>}}
}

\examples{

# disable with environment variable
Sys.setenv(REDCAPSYNC_CONFIG_OFFLINE = FALSE)

config$offline()

# change to offline
options(redcapsync.config.offline = TRUE)

config$offline()
}
\seealso{
\code{vignette("Projects", package = "REDCapSync")}
\link{setup_project} for initializing projects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REDCapSyncDataset.R
\name{dataset}
\alias{dataset}
\alias{REDCapSyncDataset}
\title{REDCapSync Dataset Object}
\value{
An R6 \code{REDCapSyncDataset} object containing dataset output, metadata,
records, user information, and optional REDCap log data.
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#experimental}{\figure{lifecycle-experimental.svg}{options: alt='[Experimental]'}}}{\strong{[Experimental]}}
\link[R6:R6Class]{R6} object representing a standardized dataset generated from
a \link{REDCapSync} project. Use this object to inspect transformed REDCap data,
export datasets, and assign prepared analysis tables into the calling
environment.
}
\details{
A \code{REDCapSyncDataset} can be created ad-hoc from a project with
\code{dataset <- project$generate_dataset()}. For reusability, you can also
define with \code{project$add_dataset()} and then load with
\code{dataset <- project$load_dataset()}.

You can add variables manually if you want them to be passed to Excel. But
\code{add_field} is a feature in development.
\subsection{Typical workflow}{

\if{html}{\out{<div class="sourceCode r">}}\preformatted{setup project from test object
project <- setup_project(
  project_name = "TEST_CLASSIC",
  dir_path = tempdir()
)
# Create and save a filtered dataset
project$add_dataset(
  dataset_name = "analysis_set",
  filter_field = "var_yesno",
  filter_choices = "Yes",
  field_names = c("ecog_at_diagnosis", "stage_at_diagnosis")
)
# generate dataset for R environment
dataset <- project$load_dataset("analysis_set")
# optional send to global environment
dataset$to_envir(globalenv()) # keep in mind potential name conflicts
# Optional modify (final save depends on what is in the dataset object)
dataset$data$merged$stage_2 <- dataset$data$merged$stage_at_diagnosis == "II"
# save to directory
dataset$save() # can specify `dir_other`, by default saves to output folder
}\if{html}{\out{</div>}}

This object is designed for users who want a stable dataset output from
REDCap without modifying the underlying project. This is also used
behind-the-scenes in the RosyREDCap shiny app.
}

\subsection{Key features}{
\itemize{
\item Stores a project-specific dataset definition and resulting data
\item Keeps metadata, record details, user information, and comments in sync
\item Saves datasets in Excel or CSV formats with optional hyperlinks
\item Exports dataset components to a user-specified environment
}
}
}
\examples{
project <- load_project("TEST_CLASSIC")

dataset <- project$generate_dataset(
  dataset_name = "stage_2_patients",
  filter_field = "stage_at_diagnosis",
  filter_choices = "II",
  field_names = c("ecog_at_diagnosis", "stage_at_diagnosis")
 )

dataset$save(dir_other = tempdir())

}
\seealso{
\link{project} for using the project objects
\code{vignette("Datasets", package = "REDCapSync")}
\code{vignette("RosyREDCap", package = "REDCapSync")}
}
\section{Public fields}{
  \if{html}{\out{<div class="r6-fields">}}
  \describe{
    \item{\code{data}}{list of data where names are forms}

    \item{\code{metadata}}{list of metadata}

    \item{\code{records}}{data.frame of records with timestamps}

    \item{\code{users}}{data.frame of users with timestamps}

    \item{\code{log}}{data.frame of log}

    \item{\code{comments}}{data.frame of comments}
  }
  \if{html}{\out{</div>}}
}
\section{Active bindings}{
  \if{html}{\out{<div class="r6-active-bindings">}}
  \describe{
    \item{\code{project_details}}{Read-only list of dataset details from
\code{\link[=setup_project]{setup_project()}}.}

    \item{\code{dataset_details}}{Read-only list of dataset details}

    \item{\code{redcap}}{Read-only list of dataset details from \link{project}}

    \item{\code{links}}{Read-only list of dataset details from \link{project}}
  }
  \if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
  \itemize{
    \item \href{#method-REDCapSyncDataset-initialize}{\code{REDCapSyncDataset$new()}}
    \item \href{#method-REDCapSyncDataset-print}{\code{REDCapSyncDataset$print()}}
    \item \href{#method-REDCapSyncDataset-save}{\code{REDCapSyncDataset$save()}}
    \item \href{#method-REDCapSyncDataset-to_envir}{\code{REDCapSyncDataset$to_envir()}}
  }
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncDataset-initialize"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncDataset-initialize}{}}}
\subsection{\code{REDCapSyncDataset$new()}}{
  The end user will not see \code{dataset$new()}. This is handled internally.
Users should construct objects using \link{REDCapSyncProject}.
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncDataset$new(
  project,
  dataset_name,
  transformation_type = "default",
  merge_form_name = "merged",
  filter_field = NULL,
  filter_choices = NULL,
  filter_list = NULL,
  filter_strict = TRUE,
  field_names = NULL,
  form_names = NULL,
  exclude_identifiers = FALSE,
  exclude_free_text = FALSE,
  date_handling = "none",
  labelled = TRUE,
  clean = TRUE,
  drop_blanks = FALSE,
  drop_missing_codes = FALSE,
  drop_others = NULL,
  include_metadata = TRUE,
  include_users = TRUE,
  include_records = TRUE,
  include_log = FALSE,
  annotate_from_log = TRUE,
  include_comments = FALSE
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{project}}{Project object from \code{\link[=setup_project]{setup_project()}} or \code{\link[=load_project]{load_project()}}.}
      \item{\code{dataset_name}}{Character. Name of the dataset to generate or load.
If the dataset already exists in the project, the existing definition is
reused.}
      \item{\code{transformation_type}}{Character. Data transformation strategy: "default"
(preferred merged output), "none" (raw data structure), or
"merge_non_repeating" (merge only non-repeating forms). Default is
"default".}
      \item{\code{merge_form_name}}{Character. Name used for merged non-repeating records.
Default is "merged".}
      \item{\code{filter_field}}{Character. Field used for filtering the dataset.}
      \item{\code{filter_choices}}{Vector. Allowed values for \code{filter_field}.}
      \item{\code{filter_list}}{List. Named list mapping field names to allowed values.
Use instead of \code{filter_field}/\code{filter_choices} for more complex filters.}
      \item{\code{filter_strict}}{Logical. If \code{TRUE}, filters are applied to every form.
If \code{FALSE}, filters apply only to the record identifier. Default is
\code{TRUE}.}
      \item{\code{field_names}}{Character vector. Variables to include in the dataset.
Default is \code{NULL} (all fields).}
      \item{\code{form_names}}{Character vector. Forms to include in the dataset.
Default is \code{NULL} (all forms).}
      \item{\code{exclude_identifiers}}{Logical. Remove identifier fields. Default is
\code{TRUE}.}
      \item{\code{exclude_free_text}}{Logical. Remove free text fields. Default is
\code{FALSE}.}
      \item{\code{date_handling}}{Character. Date handling method: "none",
"exclude_dates", "random_shift_by_record", "random_shift_by_project",
"zero_by_record", or "zero_by_project". Default is "none".}
      \item{\code{labelled}}{Logical. Convert values to labelled vectors if \code{TRUE}.
Default is \code{TRUE}.}
      \item{\code{clean}}{Logical. Clean the dataset by standardizing missing values and
blanks. Default is \code{TRUE}.}
      \item{\code{drop_blanks}}{Logical. Drop records with blank fields. Default is
\code{FALSE}.}
      \item{\code{drop_missing_codes}}{Logical. Convert REDCap missing codes to \code{NA}.
Default is \code{FALSE}.}
      \item{\code{drop_others}}{Character vector of additional values to remove.}
      \item{\code{include_metadata}}{Logical. Include field metadata in the dataset.
Default is \code{TRUE}.}
      \item{\code{include_users}}{Logical. Include user information in the dataset.
Default is \code{TRUE}.}
      \item{\code{include_records}}{Logical. Include record-level details. Default is
\code{TRUE}.}
      \item{\code{include_log}}{Logical. Include REDCap activity log details. Default is
\code{FALSE}.}
      \item{\code{annotate_from_log}}{Logical. Annotate metadata and records using the
change log. Default is \code{TRUE}.}
      \item{\code{include_comments}}{Logical. Include REDCap comments. Default is
\code{FALSE}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncDataset-print"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncDataset-print}{}}}
\subsection{\code{REDCapSyncDataset$print()}}{
  Print some key dataset information
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncDataset$print()}
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncDataset-save"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncDataset-save}{}}}
\subsection{\code{REDCapSyncDataset$save()}}{
  Return flat list
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncDataset$save(
  with_links = TRUE,
  separate = FALSE,
  use_csv = FALSE,
  dir_other = NULL,
  file_name = NULL
)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{with_links}}{Logical. Include hyperlinks in Excel exports. Default is
\code{TRUE}.}
      \item{\code{separate}}{Logical. Save each form as a separate file instead of a
multi-sheet workbook. Default is \code{FALSE}.}
      \item{\code{use_csv}}{Logical. Write CSV files instead of Excel. Default is
\code{FALSE}.}
      \item{\code{dir_other}}{Character. Directory where the dataset file should be
saved. Defaults to the project's output folder.}
      \item{\code{file_name}}{Character. Base file name for saved datasets. Defaults to
\verb{<project_name>_<dataset_name>}.}
    }
    \if{html}{\out{</div>}}
  }
}

\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-REDCapSyncDataset-to_envir"></a>}}
\if{latex}{\out{\hypertarget{method-REDCapSyncDataset-to_envir}{}}}
\subsection{\code{REDCapSyncDataset$to_envir()}}{
  export dataset to envir of your choosing. Keep in mind
potential name conflicts
  \subsection{Usage}{
    \if{html}{\out{<div class="r">}}
    \preformatted{REDCapSyncDataset$to_envir(envir = NULL)}
    \if{html}{\out{</div>}}
  }
  \subsection{Arguments}{
    \if{html}{\out{<div class="arguments">}}
    \describe{
      \item{\code{envir}}{Environment to assign exported dataset objects. Default is
\code{NULL}.}
    }
    \if{html}{\out{</div>}}
  }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cache.R
\name{cache_clear}
\alias{cache_clear}
\title{Clear your cached projects}
\usage{
cache_clear(project_names = NULL)
}
\arguments{
\item{project_names}{character vector of project project_names previously
setup. If NULL, will get all from \code{get_projects()}}
}
\value{
Message of outcome and invisible NULL.
}
\description{
Finds the location of the cache established by \code{\link[hoardr]{hoard}}
and deletes stored project information (not data)! If you provide
\code{project_names}, it will remove only those projects from the cache. If you
want to truly delete the project files, you must do so at the project
directory you set up.
}
\details{
The cache only stores information like project_name, token_name, directory
location, and other details from \code{\link[=setup_project]{setup_project()}}.The default location of
the cache location is defined by using R_USER_CACHE_DIR if set. Otherwise,
it follows platform conventions via \link[hoardr:hoardr]{hoardr}, saving a file
"R/REDCapSync/projects.rds". No direct project data is stored in the cache.
Notably, tokens and data are not stored here. The key variables stored in
the cache are...
\itemize{
\item \code{project_name} - unique identifier for REDCapSync package
\item \code{redcap_uri} - server location
\item \code{token_name} - where to find token environment with \code{\link[=Sys.getenv]{Sys.getenv()}}
\item \code{dir_path} - where to saved project and associated files locally
\item \code{project_id} - obtained from API call and "locks-in" the connection
\item \code{redcap_version} - obtained from API call and affects links
\item \code{last_sync} and \code{sync_frequency} - informs REDCap sync of when to update
\item other variables from project info and some internal package mechanics
}
}
\examples{
\dontrun{
cache_clear("OLD_PROJECT")
cache_clear() # every project
}

}
\seealso{
\code{vignette("Cache", package = "REDCapSync")}
\link{setup_project} for initializing projects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/projects.R
\name{projects}
\alias{projects}
\title{Manage REDCapSync projects}
\format{
A named list of functions for managing REDCapSync projects.
}
\usage{
projects
}
\description{
\code{projects} is the central interface for managing REDCap projects in
\strong{REDCapSync}. It provides a convenient, persistent registry of projects
that can be accessed across R sessions.

Each time a project is set up or synced, basic metadata is stored locally so
that projects can be reloaded and updated without reconfiguration.
}
\details{
\code{projects} is implemented as a singleton list object and serves as a single
entry point to the REDCapSync workflow. All project-level operations—such as
setup, loading, syncing, and removal—are accessed through this object.

A key advantage of this design is support for \strong{method chaining}, even
without loading the namespace. Because many project methods return project
objects, you can write concise, readable workflows that operate in sequence:

\preformatted{
# load, sync, and generate dataset to global
REDCapSync::projects$load("TEST_CLASSIC")$
  sync()$
  generate_dataset(envir = globalenv(),
                   filter_field = "var_branching",
                   filter_choices = "Yes")
}
}
\section{Methods}{


\describe{

\item{\code{print()}}{
Display a summary of all registered projects.
\itemize{
\item Shows total number of projects
\item Lists project names
\item Indicates how many are due for sync
}

Returns the project data frame invisibly.
}

\item{\code{any()}}{
Check whether any projects are registered.

Returns a logical scalar.
}

\item{\code{n()}}{
Return the number of registered projects.

Returns an integer.
}

\item{\code{names()}}{
Return the names of registered projects.

Returns a character vector.
}

\item{\code{test_names()}}{
Return the names of test projects.

Returns a character vector.
}

\item{\code{df()}}{
Retrieve the project registry as a data frame.

Returns a data frame containing stored project metadata.
}

\item{\code{load(project_name)}}{
Load a project by name. This is a wrapper for \link{load_project}.

\strong{Arguments}
\itemize{
\item \code{project_name}: Character. Name of the project.
}

Returns a \code{REDCapSyncProject} object.
}

}
}

\examples{
project_df <- projects$df()
projects$print()
projects$any()
projects$n()
projects$names()
project <- projects$load("TEST_CLASSIC")
}
\seealso{
\code{vignette("REDCapSync", package = "REDCapSync")}
\code{vignette("Cache", package = "REDCapSync")}
\code{vignette("RosyREDCap", package = "REDCapSync")}
\link{setup_project} for initializing projects
\link{project} for using the project objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sync.R
\name{sync}
\alias{sync}
\title{Synchronize REDCap Data}
\usage{
sync(
  project_names = NULL,
  save_datasets = TRUE,
  hard_check = FALSE,
  hard_reset = FALSE
)
}
\arguments{
\item{project_names}{character vector of project project_names previously
setup. If NULL, will get all from \code{get_projects()}}

\item{save_datasets}{Logical (TRUE/FALSE). If TRUE, saves datasets to
directory.}

\item{hard_check}{Will check REDCap even if not due (see \code{sync_frequency}
parameter from \code{setup_project()})}

\item{hard_reset}{Logical that forces a fresh update if TRUE. Default is
\code{FALSE}.}
}
\value{
Invisibly returns last project object (R6 \link{REDCapSyncProject} class)
}
\description{
Syncs with REDCap via \code{project} object that user defined with
\link{setup_project}
}
\details{
Syncs all projects by default but can be used to hands-free sync one or
defined set projects. This is not intended to return project object. User
should use \code{load_project("project_name")}. However, by default will invisibly
return the last project in the set of project_names.
}
\examples{
\dontrun{
# if you setup 3 projects "ONE", "TWO", "THREE" using [setup_project()]...
sync() # will check all three projects for updates if due for sync
sync("TWO") # will only check and sync project TWO
project <- sync("TWO") # can also be used for invisible return ...
sync(c("ONE","THREE")) # will only check and sync projects ONE and THREE
project <- sync(c("ONE","THREE")) # invisible return of THREE ...
}
}
\seealso{
\code{vignette("REDCapSync", package = "REDCapSync")}
\code{vignette("Cache", package = "REDCapSync")}
\link{setup_project} for initializing projects
\link{project} for using the project objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/REDCapSync-package.R
\docType{package}
\name{REDCapSync-package}
\alias{REDCapSync}
\alias{REDCapSync-package}
\title{REDCapSync: Encapsulated 'REDCap' Projects for Synchronized Data Pipelines}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Wraps dozens of 'REDCap' API endpoints into a standardized R6 object. Research Electronic Data Capture ('REDCap') is a survey and database web application software maintained by Vanderbilt University. It has a robust application programming interface (API) utilized by several R packages. 'REDCapSync' uses 'redcapAPI' and 'REDCapR' behind-the-scenes to retrieve all metadata, data, and log details for a project. To minimize unnecessary server calls, the interim 'REDCap' log is analyzed and used to only update necessary records. Furthermore, the user can define custom datasets that save to a directory. Those datasets continue to refresh when projects are synced. Having a secure, standardized, API-efficient, project-agnostic R object for 'REDCap' projects, streamlines downstream use in scripts, functions, and shiny applications.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/thecodingdocs/REDCapSync}
  \item \url{https://thecodingdocs.github.io/REDCapSync/}
  \item Report bugs at \url{https://github.com/thecodingdocs/REDCapSync/issues}
}

}
\author{
\strong{Maintainer}: Brandon Rose \email{thecodingdocs@gmail.com} (\href{https://orcid.org/0009-0009-7813-1960}{ORCID}) [copyright holder]

Authors:
\itemize{
  \item Brandon Rose \email{thecodingdocs@gmail.com} (\href{https://orcid.org/0009-0009-7813-1960}{ORCID}) [copyright holder]
}

Other contributors:
\itemize{
  \item Natalie Goulett [contributor]
}

}
\keyword{internal}
